/*
 * Adaptor.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.util.*;
import org.apache.xmlrpc.*;
import org.apache.log4j.Logger;

import de.dfki.lt.hog.util.Utility;

/**
 * <code>Adaptor</code> is an abstract class which provides methods for
 * a remote module that functions as an XML-RPC server.
 * It is instantied with a <code>Properties</code> object which is
 * created through a config file. In this file a server port must be
 * specified on which a subclass of <code>Adaptor</code> listens. The
 * <code>addHandler</code> method registers a handler which processes
 * incoming requests. <code>Adpator</code> is a default implementation
 * of <code>XmlRpcServerInterface</code>.
 *
 * @author <a href="mailto:robba@dfki.de">Robert Barbey</a>
 * @version 2003-12-05
 */
public abstract class Adaptor
    implements XmlRpcServerInterface {

    /**
     * Key for port of XML-RPC server in <code>Properties</code>. */
    public static final String WEBSERVER_PORT_PROP
	= "xmlrpc.server.port";

    /**
     * Contains a configuration of this <code>Adpator</code> created
     * through a config file. */
    private Properties config;

    /**
     * An XML-RPC Server which processes incoming requests. */
    private WebServer adaptorServer;

    /**
     * Contains a port on which <code>adaptorServer</code> listens. */
    private int adaptorServerPort;

    /**
     * Contains all handlers added to the <code>WebServer</code>. */
    private Hashtable handlers;

    /**
     * A logger for <code>Adaptor</code>s */
    static Logger adaptorLog = Logger.getLogger(Adaptor.class);

    /**
     * Creates an <code>Adaptor</code> with a given configuration as
     * <code>Properties</code> object. Instantiates a
     * <code>WebServer</code> listening on a port contained in
     * configuration.
     *
     * @param adaptorProps <code>Properties</code> configurating
     *    this <code>Adaptor</code> */
    public Adaptor(Properties adaptorProps) {
	config = adaptorProps;
	handlers = new Hashtable();

        adaptorServerPort =
                Integer.parseInt(getConfig().getProperty(WEBSERVER_PORT_PROP));

        // oeze: due to new apache xmlrpc server (2.0):
        // does not throw ioexception anymore
        //try {
        adaptorServer = new WebServer(adaptorServerPort);
	//} catch (java.io.IOException ioe) {
        //    adaptorLog.fatal(Utility.exception2String(ioe));
	//}
    }

    /**
     * Adds a handler object to the <code>WebServer</code> of this
     * <code>Adaptor</code>.
     * A specific method is called by
     * <code>aHandlerName.aMethodName</code> where
     * <code>aMethodName</code> is implemented by <code>aHandler</code>.
     */
    public void addHandler(String aHandlerName, Object aHandler){
	getServer().addHandler(aHandlerName, aHandler);
	getHandlers().put(aHandlerName, aHandler);
    }

    /**
     * Returns a handler object from <code>handlers</code>
     * hashtable.
     *
     * @param aHandlerName name of handler in hashtable
     * @return corresponding handler object
     */
    public Object getHandler(String aHandlerName) {
	return getHandlers().get(aHandlerName);
    }

    /**
     * Returns configuration of this <code>Adaptor</code>
     * @return configuration of <code>Adaptor</code> as
     *    <code>Properties</code> object */
    public Properties getConfig() {
	return this.config;
    }

    /**
     * Returns hashtable of handlers registeres so far
     *
     * @return hashtable of handler objects
     */
    public Hashtable getHandlers() {
	return this.handlers;
    }

    /**
     * Returns XML-RPC server of this <code>Adaptor</code>
     * @return server of this <code>Adaptor</code> */
    public WebServer getServer() {
	return this.adaptorServer;
    }

    /**
     * Returns port on which the server is listening
     * @return port of XML-RPC server */
    public int getServerPort() {
	return this.adaptorServerPort;
    }

    /**
     * Shuts down an <code>Adaptor</code>.
     * Has to be overriden for shutting down special components of
     * derived subclass.
     */
    public void shutdown()
    {
	getServer().shutdown();
    }
}
