/*
 * Annotation.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import org.w3c.dom.Node;
import org.apache.log4j.Logger;
import de.dfki.lt.hog.util.Utility;

/**
 * Provides methods to handle an annotation during a session.
 *
 *
 * @author <a href="mailto:robba@dfki.de">Robert Barbey</a>
 * @version 2003-11-19
 */
public class Annotation
{
    /**
     * Identifies this <code>Annotation</code> object in a session. */
    private String annotationId;

    /**
     * Identifies collection this <code>Annotation</code> belongs to. */
    private String annotationCollectionId;

    /**
     * Identifies session this <code>Annotation</code> belongs to. */
    private String annotationSessionId;

    /**
     * Annotation as an XML string. */
    private String annotationXml;

    /**
     * Annotation as a DOM tree.	 */
    private Node annotationDOMTree;

    /**
     * Language of this <code>Annotation</code> as an abbreviated
     * string.
     */
    private String annotationLanguage;

    /**
     * The depth of the module that created this annotation.
     */
    private int depth = -1;

    /** A log4j Logger. */
    static Logger log = Logger.getLogger(Annotation.class);

    /**
     * Creates a new <code>Annotation</code> object for given session
     * and language.
     *
     * @param annotation actual annotation
     * @param language language of this <code>Annotation</code>
     */
    public Annotation(String sessId, String annCollId,
		      String annotation, String language, String anId)
    {
	annotationSessionId = sessId;
	annotationCollectionId = annCollId;
	setAnnotationId(anId);
	annotationLanguage = language;
	annotationXml = annotation;
    }

    private void setAnnotationId(String anId)
    {
	this.annotationId = anId;
    }

    /**
     * Returns an identification of this <code>Annotation</code> object
     * @return ID for this <code>Annotation</code> */
    public String getAnnotationId()
    {
	return this.annotationId;
    }

    /**
     * Returns an identification of the collection this
     * <code>Annotation</code>belongs to.
     * @return ID for collection this <code>Annotation</code> belongs to.
     */
    public String getAnnotationCollectionId()
    {
	return this.annotationCollectionId;
    }

    /**
     * Returns a identification of the session this
     * <code>Annotation</code>belongs to.
     * @return ID for collection this <code>Annotation</code> belongs to.
     */
    public String getAnnotationSessionId()
    {
	return this.annotationSessionId;
    }

    /**
     * Returns unprocessed annotation as an XML string
     * @return unprocessed annotation as XML */
    public String getAnnotationXml()
    {
	return annotationXml;
    }

    /**
     * Returns unprocessed annotation as a DOM tree
     * @return unprocessed annotation as DOM */
    public Node getAnnotationDOMTree()
    {
	return annotationDOMTree;
    }

    /**
     * Returns language of this <code>Annotation</code>
     * @return language abbreviation */
    public String getAnnotationLanguage()
    {
	return annotationLanguage;
    }

    /**
     * Returns the depth of the module that created this annotation.
     * @return the depth of the module that created this annotation.
     */
    public int getDepth() {
      // ignore non-xml annotations
      if ( (this.annotationId.indexOf("raw") > -1) ||
           (this.annotationId.indexOf("pic") > -1)) {
        return 0;
      }
      log.debug("Getting depth of "
                + this.annotationSessionId + "."
                + this.annotationCollectionId + "."
                + this.annotationId + "\n");
      log.debug("getDepth: this.annotationXml: " + this.annotationXml);
      // if set, return this.depth directly
      if (this.depth > -1) {
        return this.depth;
      } else {
        // retrieve it from the metadata
        try {
          return Integer.parseInt( (new Metadata(
              this.annotationXml).getConfig().getProperty("module.depth")));
        } catch (NumberFormatException nfe) {
          if (!this.annotationId.equals("xmltext")) {
            log.warn("Could not get depth of annotation "
                     + this.annotationSessionId + "."
                     + this.annotationCollectionId + "."
                     + this.annotationId + " - assuming depth = 0\n"
                     + Utility.exception2String(nfe));
            }
            return 0;
        } // catch
      } // else
    } // getDepth()

    /**
     *
     * @param depth
     */
    public void setDepth(int depth) {
      this.depth = depth;
    }

}
