/*
 * AnnotationCollection.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */


package de.dfki.lt.hog;

import java.util.Hashtable;

/**
 * An <code>AnnotationCollection</code> comprises all
 * <code>Annotation</code> objects created during a session.
 *
 * @author <a href="mailto:robba@dfki.de">Robert Barbey</a>
 * @version 2003-12-12
 */
public class AnnotationCollection
{
    /**
     * Contains all <code>Annotation</code> objects this
     * <code>AnnotationCollection</code> manages.
     */
    private Hashtable annotations;

    /**
     * Returns all <code>Annotation</code> objects as a
     * <code>Hashtable</code>.
     *
     * @return <code>Annotation</code> objects as <code>Hashtable</code>
     */
    public Hashtable getAnnotations()
    {
	return this.annotations;
    }

    /**
     * Sets the private field </code>annotations</code>.
     *
     * @param aHashtable new value for field <code>annotation</code>
     */
    private void setAnnotations(Hashtable aHashtable)
    {
	this.annotations = aHashtable;
    }

    /**
     * Identifies the <code>Session</code> to which this
     * <code>AnnotationCollection</code> belongs to.
     */
    private String annCollSessionId;

    /**
     * Returns identification of the <code>Session</code> to which this
     * <code>AnnotationCollection</code> belongs to.
     * @return identification of <code>Session</code>
     */
    public String getAnnCollSessionId()
    {
	return this.annCollSessionId;
    }

    /**
     * Contains the identification for this
     * <code>AnnotationCollection</code>.
     */
    private String annotationCollectionId;

    /**
     * Returns the identification for this
     * <code>AnnotationCollection</code>.
     *
     * @return identification
     */
    public String getAnnotationCollectionId()
    {
	return this.annotationCollectionId;
    }

    /**
     * Sets a new identification for this
     * <code>AnnotationCollection</code>.
     *
     * @param anId an identification
     */
    private void setAnnotationCollectionId(String anId)
    {
	this.annotationCollectionId = anId;
    }

    /**
     * Creates a new <code>AnnotationCollection</code>.
     * It is initialized with an empty
     * <code>Hashtable</code> of <code>Annotation</code> objects and
     * an identification.
     */
    public AnnotationCollection(String sessId)
    {
	Hashtable hash = new Hashtable();
	annCollSessionId = sessId;
	setAnnotations(hash);
	setAnnotationCollectionId(createCollectionId());
    }

    /**
     * Creates a new <code>AnnotationCollection</code> with an
     * initial <code>Annotation</code> object.
     * This is helpful when a new <code>Annotation</code> is added to
     * session
     * while no <code>AnnotationCollection</code> exists.
     *
     * @param initialAnnotation an initial <code>Annotation</code> object
     */
    public AnnotationCollection(String sessId,
				Annotation initialAnnotation)
    {
	this(sessId);
	getAnnotations().put(initialAnnotation.getAnnotationId(),
			     initialAnnotation);
    }

    /**
     * Creates an identification for this
     * <code>AnnotationCollection</code>.
     */
    private String createCollectionId()
    {
	String sessId = getAnnCollSessionId();
	Session session = (Session) MoCoMan.getSessions().get(sessId);
	String id = "collection"
	    + Integer.toString(session.getAnnotationCollections().size()
			       + 1);

	return id;
    }

    /**
     * Creates an <code>Annotation</code> object for this
     * <code>AnnotationCollection</code>.
     *
     * @param annString annotation string
     * @param annLanguage language of given annotation
     * @return identification
     */
    public String createAnnotation(String annString, String annLanguage,
				   String anId)
    {
	Annotation ann = new Annotation(getAnnCollSessionId(),
					getAnnotationCollectionId(),
					annString, annLanguage, anId);
	String annId = ann.getAnnotationId();
	getAnnotations().put(annId, ann);

	return annId;
    }

    /**
     * Returns an <code>Annotation</code> object specified by a given
     * identification.
     * @param annId identification of <code>Annotation</code> to return
     * @return corresponding <code>Annotation</code>
     */
    public Annotation getAnnotation(String annId)
    {
	return (Annotation) getAnnotations().get(annId);
    }

    /**
     * Removes all <code>Annotation</code> objects from the hashtable.
     */
    public void removeAll()
    {
	if (!getAnnotations().isEmpty())
	    getAnnotations().clear();
    }

}
