/*
 * Application.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.io.IOException;
import java.util.Properties;
import org.apache.log4j.Logger;

/**
 * An abstract class providing some basic methods an objects for an application.
 * Among these is a logger (Log4J) and methods to configure its logging outputs
 * (level, file type, etc.). An application is configured by an external config
 * file (usually <tt>application.cfg</tt>).
 *
 * @version 2003-12-23
 * @author <a href="mailto:robba@dfki.de">Robert Barbey </a>
 */
public abstract class Application {
	/**
	 * A logger for this <code>Application</code>
	 */
	public static Logger applicationLog = Logger.getLogger(Application.class);

	/**
	 * Key for database location property in configuration properties
	 */
	public static final String DATABASE_URI = "database.uri";

	/**
	 * Key for database user property in configuration properties
	 */
	public static final String DATABASE_USER = "database.user";

	/** A configuration for this <code>Application</code> */
	private Properties config;

	/** User name to use in the database */
	private String databaseUser;

	/** URI of the database as <code>String</code> */
	private String databaseUri;

	/**
	 * Initializes field values with values from configuration.
	 *
	 * @param applProps
	 */
	protected void init(Properties applProps) throws IOException {
		setConfig(applProps);
		setDatabaseUri(getConfig().getProperty(DATABASE_URI));
		setDatabaseUser(getConfig().getProperty(DATABASE_USER));
	}

	/**
	 * Returns the configuration of this <code>Application</code>.
	 *
	 * @return configuration as <code>Properties</code>
	 */
	protected Properties getConfig() {
		return this.config;
	}

	/**
	 * Sets a new configuration for this <code>Application</code>.
	 *
	 * @param configProp
	 *            configuration as <code>Properties</code>
	 */
	private void setConfig(Properties configProps) {
		config = configProps;
	}

	/**
	 * Returns current user name for database.
	 *
	 * @return user name
	 */
	protected String getDatabaseUser() {
		return this.databaseUser;
	}

	/**
	 * Sets a new user name for the database.
	 *
	 * @param aUserName
	 *            a user name
	 */
	private void setDatabaseUser(String aUserName) {
		this.databaseUser = aUserName;
	}

	/**
	 * Returns the URI for the database
	 *
	 * @return an URI for the database
	 */
	protected String getDatabaseUri() {
		return this.databaseUri;
	}

	/**
	 * Sets a new URI for the database
	 *
	 * @param anUri
	 *            a new URI
	 */
	private void setDatabaseUri(String anUri) {
		this.databaseUri = anUri;
	}

}

