/*
 * Launcher.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.util.*;
import java.util.regex.Matcher;
import java.rmi.*;
import java.lang.reflect.*;
import org.apache.log4j.Logger;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.exceptions.NoSuchModuleException;
import de.dfki.lt.hog.exceptions.ShutdownException;
import de.dfki.lt.hog.util.Utility;

/**
 * <code>Laucher</code> provides static methods that initialize and
 * register all modules in the local registry and in the rmi
 * registry. It also provides a method that starts the run method of a
 * module in a thread.
 *
 * @author Joerg Steffen, DFKI
 * @version $Id: Launcher.java,v 1.3 2006/11/20 13:20:30 dcontag Exp $ */

public class Launcher {

  /**
   * This contains the logger object for logging. */
  static Logger log = Logger.getLogger(Launcher.class);

  /**
   * This is the property for the rmi module in the module config. */
  protected static final String RMI_MODULE_PROP = "rmi.module";


  /**
   * <code>Launcher</code> can't be initiated. */
  private Launcher(){}


  /**
   * This creates an instance of <code>moduleClass</code> using
   * <code>configProps</code> and registers it in the local registry.
   * We assume that each module has exactly one public constructor
   * that takes a properties object as its only argument. If module
   * registers in the rmi registry, use <code>rmiServer</code> and
   * <code>rmiPort</code>.
   *
   * @param moduleClass a <code>String</code> with the class name of
   * the module
   * @param configProps a <code>Properties</code> object with config
   * settings
   * @param rmiServer a <code>String</code> with the rmi server
   * @param rmiPort a <code>String</code> with the rmi port
   * @return a <code>Module</code> that is the new instance of
   * <code>moduleClass</code>
   * @exception InitializationException if initialization fails */
  public static Module registerModule(String moduleClass,
                                      Properties configProps,
                                      String rmiServer,
                                      String rmiPort)
    throws InitializationException {

    // for initialisation it is necessary to remove
    // the numerical ids from moduleClass which contains the modulename
  
    StringBuilder sb = new StringBuilder(moduleClass);
    sb.reverse();
    
    moduleClass = sb.toString().replaceFirst("\\d", "");

    sb = new StringBuilder(moduleClass);
    sb.reverse();
    moduleClass = sb.toString();
    
    log.info("create instance of module: " + moduleClass + " ...");

    // create instance with paramater
    Object[] parameter = {configProps};
    Module module = (Module)createInstance(moduleClass, parameter);

    // register instance of module
    registerModule(module);

    // check if config specifies rmi module and server
    String rmiModule = configProps.getProperty(RMI_MODULE_PROP);
    if (null != rmiModule &&
        null != rmiServer &&
        null != rmiPort)
      try {
        Launcher.registerRmiModule
          (rmiModule, configProps, rmiServer, rmiPort);
      } catch (InitializationException ie) {
        // if rmi registration fails, just return the module;
        // this means: just report the error here
        log.error(Utility.exception2String(ie));
      }
    return module;
  }


  /**
   * This registers <code>aModule</code> in the local registry.
   *
   * @param aModule the {@link Module} to register */
  public static void registerModule(Module aModule) {

    log.info("register module " + aModule.getModuleName());

    // add module to registry
    Registry.getInstance().registerModule(aModule);
  }


  /**
   * This creates an instance of <code>rmiModuleClass</code> using
   * <code>configProps</code> and registers it in the rmi registry
   * using <code>rmiServer</code> and <code>rmiPort</code>.
   * We assume that each rmi module has exactly one public constructor
   * that takes a {@link Module} as its only argument.
   *
   * @param rmiModuleClass a <code>String</code> with the class name of
   * the module
   * @param configProps a <code>Properties</code> object with config
   * settings
   * @param rmiServer a <code>String</code> with the rmi server
   * @param rmiPort a <code>String</code> with the rmi port
   * @return a <code>Remote</code> that is the new instance of
   * <code>rmiModuleClass</code> or <code>null</code> if no rmi
   * server/port is specified
   * @exception InitializationException if initialization fails */
  public static Remote registerRmiModule(String rmiModuleClass,
                                         Properties configProps,
                                         String rmiServer,
                                         String rmiPort)
    throws InitializationException {

    log.info("bind rmi module " + rmiModuleClass + " ...");

    // get module name from properties; this is the name of the module
    // for which we want to create an rmi wrapper
    String moduleName =
      configProps.getProperty(Module.MODULE_NAME_PROP);

    // get module from local system registry
    Module module;
    try {
      module = (Module)Registry.getInstance()
        .retrieveModuleViaName(moduleName);
    } catch (NoSuchModuleException nme) {
      throw new InitializationException(nme.toString());
    }

    // create remote module as rmi wrapper for module (via parameter)
    Object[] parameter = {module};
    Remote remoteModule =
      (Remote)createInstance(rmiModuleClass, parameter);

    // bind remote module in RMI registry under the same name as in
    // the system local registry
    StringBuffer remoteModuleAdress = new StringBuffer()
      .append("//".intern()).append(rmiServer)
      .append(":".intern()).append(rmiPort)
      .append("/".intern()).append(moduleName);
    try {
      Naming.rebind(remoteModuleAdress.toString(), remoteModule);
    } catch (Exception e) {
      throw new InitializationException
        ("could not bind " + remoteModuleAdress.toString() +
         ": " + e.toString());
    }
    log.info("rmi module bound in rmi registry as " +
             remoteModuleAdress.toString());

    return remoteModule;
  }


  /**
   * This creates an instance of <code>className</code> using
   * <code>parameter</code> as parameter for the constructor. We
   * assume that each class has exactly one public constructor that
   * takes an object as its only argument.
   *
   * @param className a <code>String</code> with the class name
   * @param parameter an <code>Object</code> array used as parameter
   * for the constructor
   * @return an <code>Object</code> that is an instance of
   * <code>className</code>
   * @exception InitializationException if initialization fails */
  private static Object createInstance(String className,
                                       Object[] parameter)
    throws InitializationException {

    // get constructor
    Constructor con;
    try {
      // we assume that the relevant classes only have one public
      // constructor!
      con = Class.forName(className).getConstructors()[0];
    } catch (ClassNotFoundException cnfe) {
      throw new InitializationException
        ("class " + className + " not found : " + cnfe.toString());
    }

    // create instance with paramater
    Object module;
    try {
      module = con.newInstance(parameter);
    } catch (InvocationTargetException ite) {
      throw new InitializationException
        (ite.getTargetException().toString());
    } catch (Exception e) {
      throw new InitializationException
        ("could not instantiate " + className);
    }

    return module;
  }


  /**
   * This starts the run method of a module in a thread. The module is
   * identified via its class name. The thread is started in
   * <code>aThreadGroup</code>.
   *
   * @param moduleClass a <code>String</code> with the class name of
   * the module to run
   * @param tg <code>ThreadGroup</code> where the thread
   * is started
   * @return the thread where the module runs
   * @exception ProcessingException thread could not be started */
  /*public static Thread runModule(String moduleClass, ThreadGroup tg)
    throws ProcessingException {

    log.info("starting thread " + moduleClass);
    try {
      // get module from hash map and start it in a thread
      Module module =
        Registry.getInstance().retrieveModuleViaClass(moduleClass);
      if (null != module) {
        Thread t = new Thread(tg, module);
        t.start();
        return t;
      }
      else {
        throw new NoSuchModuleException
          ("module " + moduleClass + " not found in registry");
      }
    } catch (Exception e) {
      throw new ProcessingException
        ("could not start thread: " + e.toString());
    }
    }*/


  /**
   * This executes the shutdown method of the registered instance of
   * <code>moduleClass </code> and unregisters it in the local
   * registry and the rmi registry if necessary.
   *
   * @param moduleClass a <code>String</code> with the class name of
   * the module to shutdown
   * @param rmiServer a <code>String</code> with the rmi server
   * @param rmiPort a <code>String</code> with the rmi port
   * @exception ShutdownException if shutdown fails */
  public static void unregisterModule(String moduleClass,
                                      String rmiServer,
                                      String rmiPort)
    throws ShutdownException {

    // get module from registry
    Module oneModule = null;
    try {
      oneModule =
        Registry.getInstance().retrieveModuleViaClass(moduleClass);
    } catch (NoSuchModuleException nsme) {
      throw new ShutdownException(nsme.toString());
    }

    // shutdown module;
    log.info("shutdown module " + moduleClass);
    oneModule.shutdown();

    // unregister module in local registry
    log.info("unregister module " + oneModule.getModuleName());
    Registry.getInstance().unregisterModule(oneModule);

    // check if module has an rmi instance running
    String rmiModule = oneModule.getProperty(RMI_MODULE_PROP, null);
    if (null != rmiModule &&
        null != rmiServer &&
        null != rmiPort) {
      // unbind remote module in RMI registry if rmi server/port are
      // specified; use the same name for the rmi module as in the
      // local registry
      StringBuffer remoteModuleAdress = new StringBuffer()
        .append("//").append(rmiServer).append(":").append(rmiPort)
        .append("/").append(oneModule.getModuleName());
      try {
        log.info("unbind rmi module " +
                 remoteModuleAdress.toString());
        Naming.unbind(remoteModuleAdress.toString());
      } catch (Exception e) {
        throw new ShutdownException
          ("could not unbind " + remoteModuleAdress.toString() +
           ": " + e.toString());
      }
    }
  }
}






