/*
 * Metadata.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.apache.log4j.Logger;
import org.w3c.dom.*;
import de.dfki.lt.hog.util.Utility;

/**
 * Represents Metadata objects, holding all relevant metadata information.
 * Moreover, provides methods for generating and parsing METADATA
 * elements.
 * Generation either as XML-<tt>String</tt> or DOM <tt>Element</tt>.
 * For an example of usage see {@link #main(String[])}
 * METADATA elements are structured like this:
 * <p><code>
 * &lt;metadata&gt;<br>
 *     &lt;id&gt;<br>
 *       &lt;entry name="propertyname" value="propertyvalue"/&gt;<br>
 *       &lt;entry name="propertyname" value="propertyvalue"/&gt;<br>
 *       ...<br>
 *     &lt;/id&gt;<br>
 *     &lt;conf&gt;<br>
 *       &lt;entry name="propertyname" value="propertyvalue"/&gt;<br>
 *       &lt;entry name="propertyname" value="propertyvalue"/&gt;<br>
 *       ...<br>
 *     &lt;/conf&gt;<br>
 * &lt;metadata&gt;<br>
 * </code></p>
 *
 * @author <a href=mailto:oeze@dfki.de>Oezguer Demir</a>
 * @version $Revision: 1.2 $
 * $Id: Metadata.java,v 1.2 2005/04/26 17:37:18 oeze Exp $
 */
public class Metadata {

  /** The name of the METADATA element. */
  private static final String METADATA_ELEMENT = "metadata";

  /** The name of the METADATA ID data element. */
  private static final String METADATA_ID_ELEMENT = "id";

  /** The name of the METADATA CONF data element. */
  private static final String METADATA_CONF_ELEMENT = "conf";

  /** A log4j Logger. */
  static Logger log = Logger.getLogger(Metadata.class);

  private Properties confProps = null;
  private Properties idProps = null;

  private String sessionID = "";
  private String acID = "";
  private String created = "";
  private String processingTime = "";
  private String component = "";
  private String diagnosis = "";
  private boolean empty = false;

  private String xmlString = null;
  private boolean strUptodate = false;

  private Element domElement = null;
  private Document domDocument = null;
  private boolean domUptodate = false;

  /**
   * Creates an empty <tt>Metadata</tt> object. Only the creation time
   * stamp will be generated and added to this object. All other attributes
   * can be set using the set*() methods.
   */
  public Metadata() {
    this.init();
  }

  /**
   * Creates a new <tt>Metadata</tt> object setting all values as given.
   */
  public Metadata(String sessionID,
                  String annotationCollectionID,
                  String processingTime,
                  String component,
                  String diagnosis,
                  boolean empty,
                  Properties config) {
    this.init();
    this.sessionID = sessionID;
    this.acID = annotationCollectionID;
    this.processingTime = processingTime;
    this.component = component;
    this.diagnosis = diagnosis;
    this.empty = empty;
    this.confProps = config;
  }

  /**
   * Creates a new <tt>Metadata</tt> object reading all values from the
   * given Metadata XML String.
   * @param xmlString
   */
  public Metadata(String xmlString) {
    this.init();
    if (xmlString.length() > 6) {
      // parse the XML string
      Element md = Utility.string2Element(xmlString);
      // if metadata is not the root, but a child of root, get it
      // (maybe todo: traverse the whole tree to find metadata)
      if (! (md.getTagName().equalsIgnoreCase("metadata"))) {
        md = (Element) md.getElementsByTagName(METADATA_ELEMENT).item(0);
      }
      // process the <id> entries
      Element id = (Element) md.getElementsByTagName(METADATA_ID_ELEMENT).item(
          0);
      NodeList entries = id.getElementsByTagName("entry");
      for (int i = 0; i < entries.getLength(); i++) {
        Element entry = (Element) entries.item(i);
        this.idProps.setProperty(entry.getAttribute("name"),
                                 entry.getAttribute("value"));
      }
      this.sessionID = this.idProps.getProperty("sessionid");
      this.acID = this.idProps.getProperty("acid");
      this.processingTime = this.idProps.getProperty("processingtime");
      this.component = this.idProps.getProperty("component");
      this.diagnosis = this.idProps.getProperty("diagnosis");
      this.empty = this.idProps.getProperty("empty").equalsIgnoreCase("true");
      // process the conf entries
      Element conf =
          (Element) md.getElementsByTagName(METADATA_CONF_ELEMENT).item(0);
      entries = conf.getElementsByTagName("entry");
      for (int i = 0; i < entries.getLength(); i++) {
        Element entry = (Element) entries.item(i);
        this.confProps.setProperty(entry.getAttribute("name"),
                                   entry.getAttribute("value"));
      }
    }
  }

  /**
   * Performs all necessary steps to initilize this object. Extensible.
   */
  private void init() {
    this.created = Utility.formatTime();
    this.idProps = new Properties();
    this.confProps = new Properties();
  }

  /**
   * Returns this <tt>Metadata</tt>'s creation time stamp attribute.
   */
  public String getCreated() {
    return this.created;
  }

  /**
   * Sets this <tt>Metadata</tt>'s creation time stamp attribute.
   */
  public void setCreated(String created) {
    this.created = created;
    this.strUptodate = false;
    this.domUptodate = false;
  }

  /**
   * Returns this <tt>Metadata</tt>'s proccesing time attribute.
   */
  public String getProcTime() {
    return this.processingTime;
  }

  /**
   * Sets this <tt>Metadata</tt>'s proccesing time attribute.
   */
  public void setProcTime(String procTime) {
    this.processingTime = procTime;
    this.strUptodate = false;
    this.domUptodate = false;
  }

  /**
   * Returns this <tt>Metadata</tt>'s sessionID attribute.
   */
  public String getSessionID() {
    return this.sessionID;
  }

  /**
   * Sets this <tt>Metadata</tt>'s sessionID attribute.
   */
  public void setSessionID(String sid) {
    this.sessionID = sid;
    this.strUptodate = false;
    this.domUptodate = false;
  }

  /**
   * Returns this <tt>Metadata</tt>'s annotation collection ID attribute.
   */
  public String getAcID() {
    return this.acID;
  }

  /**
   * Sets this <tt>Metadata</tt>'s annotation collection ID attribute.
   */
  public void setAcID(String acid) {
    this.acID = acid;
    this.strUptodate = false;
    this.domUptodate = false;
  }

  /**
   * Returns this <tt>Metadata</tt>'s component attribute.
   */
  public String getComponent() {
    return this.component;
  }

  /**
   * Sets this <tt>Metadata</tt>'s component attribute.
   */
  public void setComponent(String comp) {
    this.component = comp;
    this.strUptodate = false;
    this.domUptodate = false;
  }

  /**
   * Returns this <tt>Metadata</tt>'s diagnosis attribute.
   */
  public String getDiagnosis() {
    return this.diagnosis;
  }

  /**
   * Sets this <tt>Metadata</tt>'s diagnosis attribute.
   */
  public void setDiagnosis(String diag) {
    this.diagnosis = diag;
    this.strUptodate = false;
    this.domUptodate = false;
  }

   /**
   * Returns this <tt>Metadata</tt>'s empty attribute.
   */
  public boolean getEmpty() {
    return this.empty;
  }

  /**
   * Sets this <tt>Metadata</tt>'s empty attribute.
   */
  public void setEmpty(boolean e) {
    this.empty = e;
    this.strUptodate = false;
    this.domUptodate = false;
  }

  /**
   * Returns this <tt>Metadata</tt>'s config <tt>Properties</tt>.
   */
  public Properties getConfig() {
    return this.confProps;
  }

  /**
   * Sets this <tt>Metadata</tt>'s config <tt>Properties</tt>.
   */
  public void setConfig(Properties conf) {
    this.confProps = conf;
    this.strUptodate = false;
    this.domUptodate = false;
  }

  public void setDocument(Document document) {
    this.domDocument = document;
  }

  /**
   * Returns the name of the metadata element as hardcoded in a final
   * variable. See METADATA_ELEMENT in Metadata.java.
   * @return The name of the metadata element.
   */
  public String getMetadataElementName() {
    return METADATA_ELEMENT;
  }

  /**
   * Returns the XML String representing
   * this metadata object.
   */
  public String getXmlString() {
    if ( (this.xmlString != null) && (this.strUptodate != false)) {
      return this.xmlString;
    }
    else {
      this.xmlString = this.updateXmlString();
      this.strUptodate = true;
      return this.xmlString;
    }
  }

  /**
   * Generates (if not already generated) the XML String representing
   * this metadata object and returns it.
   */
  private String updateXmlString() {
    return generateXmlString(this.sessionID, this.acID, this.processingTime,
                             this.component, this.diagnosis, this.empty, this.confProps);
  }

  /**
   * Creates a METADATA element as an XML string.
   * @param sessionID The current session ID
   * @param annotationCollectionID The annotation collection ID
   * @param processingTime The processing time stamp
   * @param component The name of the component
   * @return New metadata <tt>Element</tt>
   * @todo moduleconf2metadata
   */
  private static String generateXmlString(String sessionID,
                                          String annotationCollectionID,
                                          String processingTime,
                                          String component,
                                          String diagnosis,
                                          boolean empty,
                                          Properties config) {
    // replace <, >, " and & with XML entities
    // and clean non-upper-half ASCII characters --> see Utility.cleanXMLString()
    sessionID = Utility.cleanXMLString(sessionID);
    annotationCollectionID = Utility.cleanXMLString(annotationCollectionID);
    processingTime = Utility.cleanXMLString(processingTime);
    component = Utility.cleanXMLString(component);
    diagnosis = Utility.cleanXMLString(diagnosis);
    String sEmpty = Boolean.toString(empty);
    String e = "        <entry ";
    String n = "name";
    String v = "value";

    // simply concatenate <metadata><conf> *entries w/ attrs* </conf></metadata>
    return "\n<" + METADATA_ELEMENT + ">\n" +
        "    <id>\n" +
        e + n + "=\"created\" " + v + "=\"" + Utility.formatTime() + "\"/>\n" +
        e + n + "=\"processingtime\" " + v + "=\"" + processingTime + "\"/>\n" +
        e + n + "=\"sessionid\" " + v + "=\"" + sessionID + "\"/>\n" +
        e + n + "=\"acid\" " + v + "=\"" + annotationCollectionID + "\"/>\n" +
        e + n + "=\"component\" " + v + "=\"" + component + "\"/>\n" +
        e + n + "=\"diagnosis\" " + v + "=\"" + diagnosis + "\"/>\n" +
        e + n + "=\"empty\" " + v + "=\"" + sEmpty + "\"/>\n" +
        "    </id>\n" +
        Utility.conf2Xml(config) +
        "</" + METADATA_ELEMENT + ">\n";
  } // createMetadataAsString()

  /**
   * Returns the DOM Element representing
   * this metadata object.
   */
  public Element getMetadataElement() {
    if ( (this.domElement != null) && (this.domUptodate != false)) {
      return this.domElement;
    }
    else {
      this.domElement = this.updateDomElement();
      this.domUptodate = true;
      return this.domElement;
    }
  }

  /**
   * Generates (if not already generated) the DOM Element representing
   * this metadata object and returns it.
   */
  private Element updateDomElement() {
    this.domUptodate = true;
    return createDomElement(this.domDocument, this.sessionID, this.acID,
                            this.processingTime, this.component, this.diagnosis,
                            this.empty, this.confProps);
  }

  /**
   * Creates a METADATA DOM <tt>Element</tt> for insertion in DOM tree
   * @param document The DOM <tt>Document</tt> into which to insert metadata
   * @param sessionID The current session ID
   * @param annotationCollectionID The annotation collection ID
   * @param processingTime The processing time stamp
   * @param component The name of the component
   * @return New metadata <tt>Element</tt>
   * @todo moduleconf2metadata
   */
  private static Element createDomElement(Document document,
                                          String sessionID,
                                          String annotationCollectionID,
                                          String processingTime,
                                          String component,
                                          String diagnosis,
                                          boolean empty,
                                          Properties config) {

    String n = "name";
    String v = "value";

    // No need; org.w3c.dom.* take care of this /////////////////////////
    //sessionID = Utility.cleanXMLString(sessionID);
    //annotationCollectionID = Utility.cleanXMLString(annotationCollectionID);
    //processingTime = Utility.cleanXMLString(processingTime);
    //component = Utility.cleanXMLString(component);
    /////////////////////////////////////////////////////////////////////

    // create <metadata> and <id>
    Element metadata = (Element) document.createElement(METADATA_ELEMENT);
    Element id = (Element) document.createElement(METADATA_ID_ELEMENT);
    metadata.appendChild(id);

    // create <entry name="created" value="...">
    Element tmpEntry = (Element) document.createElement("entry");
    tmpEntry.setAttribute(n, "created");
    tmpEntry.setAttribute(v, Utility.formatTime());
    id.appendChild(tmpEntry);

    // create <entry name="processingtime" value="...">
    tmpEntry = (Element) document.createElement("entry");
    tmpEntry.setAttribute(n, "processingtime");
    tmpEntry.setAttribute(v, processingTime);
    id.appendChild(tmpEntry);

    // create <entry name="sessionID" value="...">
    tmpEntry = (Element) document.createElement("entry");
    tmpEntry.setAttribute(n, "sessionid");
    tmpEntry.setAttribute(v, sessionID);
    id.appendChild(tmpEntry);

    // create <entry name="annotationCollectionID" value="...">
    tmpEntry = (Element) document.createElement("entry");
    tmpEntry.setAttribute(n, "acid");
    tmpEntry.setAttribute(v, annotationCollectionID);
    id.appendChild(tmpEntry);

    // create <entry name="component" value="...">
    tmpEntry = (Element) document.createElement("entry");
    tmpEntry.setAttribute(n, "component");
    tmpEntry.setAttribute(v, component);
    id.appendChild(tmpEntry);

    // create <entry name="diagnosis" value="...">
    tmpEntry = (Element) document.createElement("entry");
    tmpEntry.setAttribute(n, "diagnosis");
    tmpEntry.setAttribute(v, diagnosis);
    id.appendChild(tmpEntry);

    // create <entry name="empty" value="...">
    tmpEntry = (Element) document.createElement("entry");
    tmpEntry.setAttribute(n, "empty");
    tmpEntry.setAttribute(v, Boolean.toString(empty));
    id.appendChild(tmpEntry);

    // create <conf>
    Element conf = (Element) document.createElement(METADATA_CONF_ELEMENT);
    metadata.appendChild(conf);

    // create all <entry name="component" value="..."> from confProps
    Enumeration properties = config.keys();
    while (properties.hasMoreElements()) {
      String name = (String) properties.nextElement();
      String value = (String) config.getProperty(name);
      tmpEntry = (Element) document.createElement("entry");
      tmpEntry.setAttribute("name", name);
      tmpEntry.setAttribute("value", value);
      conf.appendChild(tmpEntry);
    }

    return metadata;
  } // createMetadata()

  /**
   * Generates and returns metadata XML statically.
   */
  public static String generateXml(String sessionID,
                                   String annotationCollectionID,
                                   String processingTime,
                                   String component,
                                   String diagnosis,
                                   boolean empty,
                                   Properties config) {
    Metadata md = new Metadata(sessionID, annotationCollectionID,
                               processingTime, component, diagnosis,
                               empty, config);
    return md.getXmlString();
  }

  /**
   * Generates and returns metadata XML statically setting the conf to empty.
   */
  public static String generateXml(String sessionID,
                                   String annotationCollectionID,
                                   String processingTime,
                                   String component,
                                   boolean empty,
                                   String diagnosis) {
    Metadata md = new Metadata(sessionID, annotationCollectionID,
                               processingTime, component, diagnosis,
                               empty, new Properties());
    return md.getXmlString();
  }

  /**
   * Creates and returns a DOM metadata Element statically.
   */
  public static Element createMetadataElement(Document document,
                                              String sessionID,
                                              String annotationCollectionID,
                                              String processingTime,
                                              String component,
                                              String diagnosis,
                                              boolean empty,
                                              Properties config) {
    Metadata md = new Metadata(sessionID, annotationCollectionID,
                               processingTime, component, diagnosis,
                               empty, config);
    md.setDocument(document);
    return md.getMetadataElement();
  }

  /**
   * For testing purposes.
   */
  public static void main(String[] args) {

    Properties props = new Properties();
    String confFile = ("conf/rtagger.cfg");
    try {
      FileInputStream fis = new FileInputStream(confFile);
      props.load(fis);
      fis.close();
    }
    catch (IOException ioex) {
      log.error(Utility.exception2String(ioex));
    }

    // using the static method
    System.out.println("By calling the static method .generateXml():");
    System.out.println(generateXml("sessionXX", "annotationCollectionXX",
                                   "0:00,0", "<compName\"test\">", false, "OK"));

    // create a metadata object from *scratch* and get the XML string
    Metadata fromScratch = new Metadata("sessionXX", "annotationCollectionXX",
                                        "0:00,0", "<compName\"test\">", "OK", false,
                                        props);
    System.out.println("Generated from *scratch*:");
    System.out.println(fromScratch.getXmlString());

    // create another metadata object using the xml string
    System.out.println("Parsing <metadata>:\n");
    Metadata fromXML = new Metadata(fromScratch.getXmlString());
    System.out.println("SessionID: " + fromXML.getSessionID());
    System.out.println("AnnCollID: " + fromXML.getAcID());
    System.out.println("Component: " + fromXML.getComponent());
    System.out.println("Diagnosis: " + fromXML.getDiagnosis());
    System.out.println("Empty:     " + fromXML.getEmpty());
    System.out.println("Created:  "  + fromXML.getCreated());
    System.out.println("Proc.time: " + fromXML.getProcTime());
    System.out.println("Conf: " + Utility.map2String(fromXML.getConfig()));

    // create an empty metadata obj, print it, set some attrs., print it again
    Metadata empty = new Metadata();
    System.out.println("Empty object: ");
    System.out.println(empty.getXmlString());
    empty.setSessionID("sid_0001");
    empty.setAcID("acid_001");
    empty.setComponent("<compName\"test\">");
    empty.setDiagnosis("OK");
    empty.setProcTime("0:04,3");
    empty.setConfig( (Properties) fromXML.getConfig().clone());
    System.out.println("Populated object:");
    System.out.println(empty.getXmlString());

    // get a DOM element
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder builder = null;
    try {
      builder = factory.newDocumentBuilder();
    }
    catch (ParserConfigurationException pce) {
      // Parser with specified options can't be built
      log.error(Utility.exception2String(pce));
    }
    Document document = builder.newDocument();
    System.out.println("Created DOM element:\n");
    System.out.println(Utility.node2String(createDomElement(
        document, "sessionXX",
        "annotationCollectionXX",
        "0:00,0", "<compName\"test\">", "OK", false,
        props)));

    System.out.println("---");
    System.out.println(Utility.conf2Xml(MoCoMan.getInstance().getConfig()));

    System.out.println("---");

    System.out.println("Trying to get sessionID from xml string:\n");
    Metadata md = new Metadata(
        "<?xml version='1.0'?>"+
        "<text> <metadata> <id> "+
        "<entry name=\"created\" value=\"Mon, 9 Feb 2004 22:04:31 +0100\"/> "+
        "<entry name=\"processingtime\" value=\"0.0\"/> "+
        "<entry name=\"sessionid\" value=\"session1\"/> "+
        "entry name=\"acid\" value=\"collection1\"/> "+
        "<entry name=\"component\" value=\"pythonClientApp\"/> "+
        "<entry name=\"diagnosis\" value=\"\"/> "+
        "<entry name=\"empty\" value=\"false\"/> "+
        "</id> <conf> </conf> </metadata> York.</text>");
    System.out.println(md.getSessionID());

  } // main

} // class
