/*
 * MoCoMan.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.io.*;
import java.text.SimpleDateFormat;
import org.apache.log4j.*;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.Transformer;
import org.w3c.dom.Node;
import de.dfki.lt.hog.database.XMLDBAnnotationDatabase;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;

/**
 * The Module Communication Manager for the DeepThought project. It
 * manages sessions opened with several applications along with
 * documents analysed during such a session. For communication with
 * remote modules XML-RPC is used and a config file
 * <code>mocoman.cfg</code> specifies the port of its server.
 * <p>
 * <code>MoCoMan</code> is implemented as a Singleton class which means
 * that not a constructor but the method <code>getInstance</code> has to
 * be called in order to get an instance of <code>MoCoMan</code>.
 *
 * @version $Date 2004/09/13 11:20:22 $ by $Author: uschaefer $
 * @author <a href="mailto:robba@dfki.de">Robert Barbey</a>
 */
public class MoCoMan {

  /**
   * Package prefix used for modules in config-files */
  private static final String MODULE_PREFIX = "de.dfki.lt.hog";

  /**
   * Key for database usage in <code>Properties</code> */
  private static final String DATABASE_USAGE_PROP = "xmldb.usage";

  /**
   * Key in the properties fopr the location of the XML:DB
   * Annotation Database.
   */
  private static final String DATABASE_LOCATION_PROP = "xmldb.location";

  /**
   * This contains the regexp for a valid word graph line in the text file.
   */
  private static final Pattern LINE_PATTERN = Pattern
      .compile("((\\w+\\.)+(\\w)+)=(((\\w+)/)+(\\w+\\.\\w+))");

  private static final Pattern ID_PATTERN = Pattern.compile("([0-9]*)");

  /** 
   * name of protocol file property
   */
  private static final String PROTOCOL_FILE_PROP = "protocol.file";

  /**
   * This contains the only instance of <code>MoCoMan</code>.
   * @see #getInstance()
   */
  private static MoCoMan instance = null;

  /**
   * This is the XML-RPC server that uses this instance as its backend.
   * This variable might also become a list in future - it should be
   * possible to run more than one server using this instance.
   */
  private MoCoManServer server = null;

  /**
   * Contains the location of the XMLDB Annotation Database */
  private String databaseLocation;

  /**
   * XMLDB Annotation Database for <code>MoCoMan</code> */
  private XMLDBAnnotationDatabase mocoBase;

  /**
   * True, iff the XMLDB Annotation Database is used.*/
  private boolean databaseUsage;

  /**
   * All sessions are saved in a <code>Hashtable</code> along with
   * their IDs as key and the corresponding session as value. At the
   * end of a session its entry will be deleted. */
  private static Hashtable sessions;

  /**
   * A list of queries processed so far.
   * Some sort of cache for the last querys, so a new query will be
   * looked up first if it was processed already. */
  //oe: leads to serious time performance problems on a large scale
  // private LinkedList queryQueue;
  //////
  /**
   * A configuration for this <code>MoCoMan</code>.
   * It is derived from a config file <tt>mocoman.cfg</tt> via
   * <code>Properties.load()</code> method. In this configuration
   * all needed attributes are specified, such as
   * which port is used for the XML-RPC server or the
   * format and level of the logger.
   */
  private Properties config;

  /** A log4j Logger. */
  static Logger log = Logger.getLogger(MoCoMan.class);

  /** The date this instance was created (used for uptime calculation). */
  private final long startTime = System.currentTimeMillis();

  /** Some modules require to track the number of processed sentences */
  private int sentenceCounter = 0;

  /** A unique ID based on the system time */
  private long currentUniqueID = System.currentTimeMillis();

  /** Date-Time format for protocol file */
  private SimpleDateFormat protocolDateTimeFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

  /**
   * Creates an instance of <code>MoCoMan</code>.
   * Reads the config file, derives a <code>Properties</code>
   * object from it, sets up a logger, an XML-RPC server and the
   * XML:DB annotation database.
   * gregul: changed from private to protected for tomcat version of hog
   */
   protected MoCoMan() {
    setConfig(new Properties());
    setSessions(new Hashtable());
    //oe: leads to serious time performance problems on a large scale
    // queryQueue = new LinkedList();
    /////

    // read configuration from system properties (was mocoman.cfg)
    getConfig().setProperty(DATABASE_USAGE_PROP,
        System.getProperty(DATABASE_USAGE_PROP, "false"));
    getConfig().setProperty(DATABASE_LOCATION_PROP,
        System.getProperty(DATABASE_LOCATION_PROP, "no database specified"));
    getConfig().setProperty(PROTOCOL_FILE_PROP,
        System.getProperty(PROTOCOL_FILE_PROP, "false"));
    // setup database
    Boolean tempBool = new Boolean(getConfig().getProperty(DATABASE_USAGE_PROP));
    databaseUsage = tempBool.booleanValue();

    if (databaseUsage) {
      databaseLocation = getConfig().getProperty(DATABASE_LOCATION_PROP);
      setDatabase(new XMLDBAnnotationDatabase(databaseLocation));
      log.info("XML:DB Annotation Database opened: " + databaseLocation);
    } else {
      log.info("XML:DB Annotation Database is not used");
    }

    log.info("Instance of MoCoMan created");
  }

  /**
   * Creates an instance of <code>MoCoMan</code> if none exists.
   * This controls access to the instance of <code>MoCoMan</code>.
   * @return the only instance of <code>MoCoMan</code>
   */
  public static MoCoMan getInstance() {
    if (instance == null) {
      setInstance(new MoCoMan());
    }
    log.info("Returning instance of MoCoMan");

    return instance;
  }

  private static void setInstance(MoCoMan theMoCoMan) {
    instance = theMoCoMan;
  }

  /**
   * Returns configuration for this <code>MoCoMan</code>
   * as <code>Properties</code> object.
   *
   * @return configuration
   */
  protected Properties getConfig() {
    return this.config;
  }

  private void setConfig(Properties mocoProps) {
    this.config = mocoProps;
  }

  /**
   * Returns all <code>Session</code> objects stored in a
   * <code>Hashtable</code>.
   *
   * @return all sessions
   */
  public static Hashtable getSessions() {
    return sessions;
  }

  public String hasSession(String sessionId) {
    if (this.sessions.containsKey(sessionId)) {
      return "yes";
    } else {
      return "no";
    }
  }

  private void setSessions(Hashtable mocoSess) {
    MoCoMan.sessions = mocoSess;
  }

  /**
   * Returns the XML:DB annotation database in which
   * all documents and annotations are stored.
   *
   * @return annotation database
   */
  public XMLDBAnnotationDatabase getDatabase() {
    return this.mocoBase;
  }

  private void setDatabase(XMLDBAnnotationDatabase aDatabase) {
    this.mocoBase = aDatabase;
  }

  /**
   * Returns an URL where the XML:DB annotation database is located.
   *
   * @return database URL
   */
  public String getDatabaseLocation() {
    return this.databaseLocation;
  }

  /**
   * Creates a <code>Session</code> configured through
   * a config file from which
   * <code>Properties</code> are derived.
   * A topologically sorted dependency graph is created and stored
   * as an array.
   * @param configFileName a configuration file name
   * @return an ID for this <code>Session</code>
   */
  public String createSession(String configFileName) {
    return this.createSession(configFileName, "");
  }

  /**
   * This loads a configuration file which contains the modules that should be
   * used and their configuration files.
   *
   * This is a replacement for Properties.load() -
   * It allows identical keys and assures that they do not get overwritten.
   * 
   * @param configFileName a configuration file name
   * @return a hashtable containing the modulenames and their conf-path
   */
  private Map loadConfigFile(String configFileName) {

    // stores the configuration; key=modulename, value=conf-path
    Map<String, String> confTable = new Hashtable<String, String>();

    BufferedReader in;
    String line;

    int moduleId = 0;

    try {
      in = new BufferedReader(new FileReader(configFileName));

      while ((line = in.readLine()) != null) {
        Matcher lineMatcher = LINE_PATTERN.matcher(line);

        if (lineMatcher.matches()) {
          // key, modulename
          String moduleName = lineMatcher.group(1);

          // value, configuration file
          String confPath = lineMatcher.group(4);

          // we need unique keys for the modulenames,
          // so that modules can be used with more than one configuration
          // e.g. the SdlModule

          String newModuleName = moduleName + "0";

          // check if the map already contains this key
          if (confTable.containsKey(newModuleName)) {

            // extract the numerical id from the module name
            Matcher idMatcher = ID_PATTERN.matcher(newModuleName);

            // start looking for numbers at the end of the string
            if (idMatcher.find(moduleName.length())) {

              // string->int
              moduleId = Integer.parseInt(idMatcher.group(0)) + 1;

              // the new module name with a unique id at the end
              newModuleName = moduleName + moduleId;
            }

            // reset
            moduleId = 0;

          } // end of containsKey-if

          // save the result in the config Hashtable
          confTable.put(newModuleName, confPath);

        } // end of lineMatcher-if
      } // end of while

    } catch (IOException ioe) {
      log.error(ioe.getMessage());
    }

    return confTable;

  } // end of loadPropertiesFile

  /**
   * Creates a <code>Session</code> configured through a config file from
   * which <code>Properties</code> are derived. A topologically sorted
   * dependency graph is created and stored as an array.
   * 
   * @param configFileName
   *          a configuration file name
   * @param namePrefix
   *          a prefix to the sessionId
   * @return an ID for this <code>Session</code>
   */
  public String createSession(String configFileName, String sessionId) {

    // Properties sessionProps = new Properties();
    Map<String, String> sessionProperties = new Hashtable<String, String>();
    boolean result;

    // read given config file
    sessionProperties = this.loadConfigFile(configFileName);

    log.info("Generate Properties for Modules");

    Set modules = sessionProperties.keySet();
    Vector moduleVector = new Vector();

    log.debug("MoCoMan.java  - createSession - sessionProperties - "
        + sessionProperties);
    Iterator it = modules.iterator();
    while (it.hasNext()) {
      String moduleName = (String) it.next();
      log.debug("Found: " + moduleName);
      if (!moduleName.startsWith(MODULE_PREFIX)) {
        continue;
      }

      // read config file and create Properties
      String moduleConfigFile = sessionProperties.get(moduleName);
      Properties moduleProps = new Properties();
      try {
        FileInputStream fin = new FileInputStream(moduleConfigFile);
        moduleProps.load(fin);
        fin.close();
      } catch (IOException ioe) {
        log.error("createSession: configFileName=" + configFileName
            + " sessionId=" + sessionId + " -- "
            + Utility.exception2String(ioe));
      }

      // create Module and register it in Registry
      // (done by Launcher)
      Module module = null;
      try {

        module = Launcher.registerModule(moduleName, moduleProps, null, null);
      } catch (InitializationException ie) {
        log
            .error("createSession: configFileName=" + configFileName
                + " sessionId=" + sessionId + " -- "
                + Utility.exception2String(ie));
      }

      module.setMoCoMan(getInstance());

      // sort above created Module into Vector
      if (moduleVector.isEmpty()) {
        moduleVector.add(module);
        log.debug("moduleVector: added " + module.getModuleName() + " at 0.");
      } else {
        int mvs = moduleVector.size();
        // mvs added by uschaefer 2003-11-24
        for (int i = 0; i < mvs; ++i) {
          Module m = (Module) moduleVector.get(i);
          if (module.getModuleDepth() <= m.getModuleDepth()) {
            moduleVector.add(i, module);
            log.debug("moduleVector: added " + module.getModuleName() + " at "
                + i + ".");
            break; // added by uschaefer 2004-03-31
          }
        }

        if (!moduleVector.contains(module)) {
          moduleVector.add(module);
          log.debug("moduleVector: added " + module.getModuleName() + " at "
              + (moduleVector.size() - 1) + ".");
        }
      }
    }

    log.info("Create new Session with Properties and Modules");
    Session session;
    if (sessionId.equals("")) {
      session = new Session(sessionProperties, moduleVector);
    } else {
      session = new Session(sessionProperties, moduleVector, sessionId);
    }
    session.setConfigFileName(configFileName);
    getSessions().put(session.getSessionId(), session);

    if (databaseUsage) {
      result = getDatabase().addSession(session.getSessionId());
      if (result) {
        log.info("Session added to database: " + session.getSessionId());
      } else {
        log.error("Session could not be added to database!");
      }
    }

    return session.getSessionId();
  }

  /**
   * Closes a session with MoCoMan. Invokes the <code>close()</code>
   * method of the <code>Session</code> object to be closed. Removes
   * this <code>Session</code> from the <code>sessions Hashtable</code>
   *
   * @param sessionId the Session to be closed */
  public boolean closeSession(String sessionId) {
    boolean result;

    ((Session) getSessions().get(sessionId)).closeSession();
    getSessions().remove(sessionId);

    ///// TO BE TESTED - FOO
    if (databaseUsage) {
      result = getDatabase().deleteSession(sessionId);
      if (result) {
        log.info("Session deleted from database: " + sessionId);
        return true;
      } else {
        log.error("Session couldn't be deleted from database!");
        return false;
      }
    } //if
    //////
    return true;
  }

  /**
   * Creates an annotation collection for a session specified
   * by a given session ID. If database usage is <code>true</code>
   * also a annotation collection is created in the XML:DB annotation
   * database.
   *
   * @param sessionId a session ID to specify a session for which a
   *                  collection shall be created
   * @return an identification for the created annotation collection
   */
  public String createAnnotationCollection(String sessionId) {
    Session session = (Session) getSessions().get(sessionId);
    String annCollId = session.createAnnotationCollection();
    boolean result;

    if (databaseUsage) {
      result = getDatabase().addDocument(sessionId, annCollId);
      if (result) {
        log.info("AnnotationCollection added to database: " + annCollId);
      } else {
        log.error("Couldn't add AnnotationCollection " + "to database");
      }
    }

    return annCollId;
  }

  /**
   * Creates a new annotation for a given session and annotation
   * collection each of them specified by their ID.
   * If database usage is <code>true</code> the in  the XML:DB
   * annotation database is also created an annotation.
   * The given annotation text is stored in an
   * <code>Annotation</code> object along with its language and
   * ID which is usually the processing component or <tt>rawtext</tt>.
   *
   * @param sessionId session ID referring to the session for which
   *                  an annotation shall be created
   * @param annCollId annotation collection ID referring to the session
   *                  for which an annotation shall be created
   * @param ann       annotation text
   * @param language  annotation language
   * @param anId      annotation ID
   * @return an ID for this annotation
   */
  public String createAnnotation(String sessionId, String annCollId,
      String ann, String language, String anId) {
    log.info("Adding annotation " + sessionId + "." + annCollId + "." + anId
        + ": ");
    log.debug(ann);
    Session session = (Session) getSessions().get(sessionId);
    AnnotationCollection annColl = (AnnotationCollection) session
        .getAnnotationCollections().get(annCollId);
    String annId = annColl.createAnnotation(ann, language, anId);

    boolean result;

    if (databaseUsage) {
      StringBuffer stb = new StringBuffer(ann);
      stb.insert(0, "<?xml version=\"1.0\" ?><annotation>");
      stb.append("</annotation>");
      ann = stb.toString();

      result = getDatabase().addAnnotation(sessionId, annCollId, annId, ann)
          && getDatabase().addIndex(sessionId, "ROOT", "text");
      // not ann as a pattern for indexer!!
      if (result) {
        log.info("Annotation added to database: " + annId);
      } else {
        log.error("Annotation couldn't be created: " + annId);
      }
    }
    return annId;
  }

  /**
   * Creates two initial annotations in a given session/collection:<br>
   * 1. The raw text. ID: rawtext<br>
   * 2. The xmlfied text including metadata. ID: xmltext<br>
   *
   * Use this method for initially putting text into a new session.
   *
   * @param sessionId session ID referring to the session for which
   *                  an annotation shall be created
   * @param annCollId annotation collection ID referring to the session
   *                  for which an annotation shall be created
   * @param rawtext   the raw input text
   * @param language  annatiotion language
   * @param component the invoking component ID
   * @return          the string "xmltext" if the operations were successful
   */
  public String createInitialAnnotation(String sessionId, String annCollId,
      String rawtext, String language, String component) {
    MoCoMan mm = MoCoMan.getInstance();
    // create rawtext annotation
    if ((mm.hasSession(sessionId).equalsIgnoreCase("yes"))
        && (((Session) mm.getSessions().get(sessionId))
            .getAnnotationCollections().containsKey(annCollId))) {
      mm.clearAnnotationCollection(sessionId, annCollId);
    }
    mm.createAnnotation(sessionId, annCollId, rawtext, language, "rawtext");
    // prepare metadata
    StringBuffer xmltext = new StringBuffer();
    xmltext.append("<?xml version='1.0'?><text>");
    xmltext.append(Metadata.generateXml(sessionId, annCollId, "0.0", component,
        false, ""));
    xmltext.append(Utility.cleanXMLString(rawtext));
    xmltext.append("</text>");
    log.info("Added xmltext to " + sessionId + ": " + xmltext);
    mm.createAnnotation(sessionId, annCollId, xmltext.toString(), language,
        "xmltext");
    //writeEncodedFile("/tmp/"+System.getProperty("user.name")+"/hog/server-rawtext.txt", rawtext, "utf-8");
    //writeEncodedFile("/tmp/"+System.getProperty("user.name")+"/hog/server-xmltext.xml", xmltext.toString(), "utf-8");
    // for xml-rpc usage we have to return some string
    return "xmltext";
  }

  /**
   * Creates two initial annotations in a given session/collection:<br>
   * 1. The raw text. ID: rawtext<br>
   * 2. The xmlfied text including metadata. ID: xmltext<br>
   *
   * Use this method for initially putting text into a new session.
   *
   * @param sessionId session ID referring to the session for which
   *                  an annotation shall be created
   * @param annCollId annotation collection ID referring to the session
   *                  for which an annotation shall be created
   * @param rawtext   the raw input text (as UTF-8 byte array for XML-RPC)
   * @param language  annatiotion language
   * @param component the invoking component ID
   * @return          the string "xmltext" if the operations were successful
   */
  public String createInitialAnnotation(String sessionId, String annCollId,
      byte[] rawtext, String language, String component) {
    try {
      return createInitialAnnotation(sessionId, annCollId, new String(rawtext,
          "utf-8"), language, component);
    } catch (UnsupportedEncodingException usee) {
      log.error("UnsupportedEncodingException for encoding utf-8");
      return "UnsupportedEncodingException";
    }
  }

  /**
   * Creates two initial annotations in a given session/collection:<br>
   * 1. The raw text. ID: rawtext<br>
   * 2. The xmlfied text including metadata. ID: xmltext<br>
   *
   * Use this method for initially putting text into a new session.
   *
   * @param sessionId session ID referring to the session for which
   *                  an annotation shall be created
   * @param annCollId annotation collection ID referring to the session
   *                  for which an annotation shall be created
   * @param rawtext   the raw input text (byte array for XML-RPC)
   * @param language  annatiotion language
   * @param component the invoking component ID
   * @param encoding  encoding of the String in byte array rawtext
   * @return          the string "xmltext" if the operations were successful
   */
  public String createInitialAnnotation(String sessionId, String annCollId,
      byte[] rawtext, String language, String component, String encoding) {
    try {
        appendProtocolFile(new String(rawtext, encoding), language);
    return createInitialAnnotation(sessionId, annCollId, new String(rawtext,
        encoding), language, component);
    } catch (UnsupportedEncodingException usee) {
      log.error("UnsupportedEncodingException for encoding " + encoding);
      return "UnsupportedEncodingException";
    }
  }

  public void clearAnnotationCollection(String sessionId, String annCollId) {
    try {
      Session session = (Session) getSessions().get(sessionId);
      AnnotationCollection ac = (AnnotationCollection) session
          .getAnnotationCollections().get(annCollId);
      ac.removeAll();
      log.info("ACID " + annCollId + " cleared.");
    } catch (Throwable t) {
      log.error("clearAnnotationCollection: sessionId=" + sessionId
          + " annCollId=" + annCollId + " -- " + t.toString());
    }
  }

  /**
   * Creates two initial annotations in a given session/collection:<br>
   * 1. The raw text. ID: rawtext<br>
   * 2. The xmlfied text including metadata. ID: xmltext<br>
   *
   * Use this method for initially putting text into a new session.
   *
   * @param sessionId session ID referring to the session for which
   *                  an annotation shall be created
   * @param annCollId annotation collection ID referring to the session
   *                  for which an annotation shall be created
   * @param rawtext   the raw input text
   * @param language  annatiotion language
   * @param component the invoking component ID
   * @param encoding  name of the 8 bit encoding rawtext is encoded in
   * @return          the string "xmltext" if the operations were successful
   */
  public String createInitialAnnotation(String sessionId, String annCollId,
      String rawtext, String language, String component, String encoding) {
    log
        .info("Converting pseudo String encoded as " + encoding
            + " to Unicode.");
    log.debug("createInitialAnnotation(): " + "\nsessionId : " + sessionId
        + "\nannCollId: " + annCollId + "\nrawtext: " + rawtext
        + "\nlanguage: " + language + "\ncomponent: " + component
        + "\nencoding: " + encoding);

    try {
      return createInitialAnnotation(sessionId, annCollId, pseudoString2String(
          rawtext, encoding), language, component);
    } catch (UnsupportedEncodingException usee) {
      log.error("UnsupportedEncodingException for encoding " + encoding);
      return "UnsupportedEncodingException";
    }

  }

  /** Utility method that turns an incorrectly encoded (converted)
   String into a correctly encoded Java Unicode String according
   to the specified encoding. This only works correctly for
   eight bit encodings!
   This is used as server-side input frontend for Strings from
   the Apache XML-RPC implementation.
   @param source source string with incorrect encoding
   @param eightbitencoding name of encoding the source originally had
   @return the correctly encoded string according to the specified
   encoding
   */
  public static String pseudoString2String(String source,
      String eightbitencoding) throws UnsupportedEncodingException {
    byte[] b = new byte[source.length()];
    for (int i = 0; i < source.length(); i++) {
      // create byte array without encoding
      b[i] = (byte) (source.charAt(i) & 0xFF);
    }
    // return as String with correct (=specified) encoding
    return new String(b, eightbitencoding);
  }

  /** Utility method that turns a Java Unicode String into
   a pseudo 8 bit encoded String according to the specified
   encoding. This only works for eight bit encodings!
   This is used as server-side output frontend for Strings
   going to the Apache XML-RPC implementation.
   @param source source string with normal Java Unicode encoding
   @param eightbitencoding name of encoding the target should have
   @return the incorrectly encoded string according to the specified
   encoding
   */
  public static String String2pseudoString(String source,
      String eightbitencoding) throws UnsupportedEncodingException {
    // create byte array with specified encoding
    byte[] b = source.getBytes(eightbitencoding);
    StringBuffer sb = new StringBuffer(b.length);
    for (int i = 0; i < b.length; i++) {
      sb.append((char) ((b[i] & 0x7F) + (b[i] < 0 ? 0x80 : 0)));
    }
    // return a String with pseudo encoding
    return sb.toString();
  }

  /**
   * Analyses a document given as XML string via XML-RPC.
   * This method is necessary since the parameters are transmitted
   * as a <code>Vector</code> object which cannot store simple
   * <code>int</code> values. Just calls {@link #analyse(String
   * sessionId, String annCollId, String annId, int startSpan,
   * int endSpan, int depth)} for analysis.
   *
   * @param sessionId session ID of annotation
   * @param annCollId annotation collection ID
   * @param startSpan
   * @param endSpan
   * @param depth
   * @param encoding desired 8 bit encoding on client side
   * @return result of analysis as XML string
   */
  ////public String analyse (String sessionId, String annCollId,
  ////String annId, Integer startSpan,
  ////Integer endSpan, Integer depth,
  ////String encoding) {
  ////try {
  ////return String2pseudoString(analyse(sessionId, annCollId, annId, startSpan.intValue(),
  ////endSpan.intValue(), depth.intValue()), encoding);
  ////} catch (java.io.UnsupportedEncodingException uee) {
  ////log.error("UnsupportedEncodingException for encoding " + encoding);
  ////return "UnsupportedEncodingException";
  ////}
  ////}

  /**
   * Retrieves an annotation belonging to a given session and
   * annotation collection specified by an annotation ID.
   * This ID is usually the module name which created this
   * annotation or <tt>rawtext</tt>.
   *
   * @param sessId       session ID referring to the session for which
   *                     an annotation was created
   * @param annCollId    annotation collection ID referring to the
   *                     annotation collection for which
   *                     an annotation was created
   * @param annId        the ID of the annotation
   * @param encoding     desired 8 bit encoding on client side
   * @return annotation
   */
  //public String getAnnotation(String sessId, String annCollId,
  //String annId, String encoding) {
  //try {
  //return String2pseudoString(getAnnotation(sessId, annCollId, annId), encoding);
  //} catch (java.io.UnsupportedEncodingException uee) {
  //log.error("UnsupportedEncodingException for encoding " + encoding);
  //return "UnsupportedEncodingException";
  //}
  //}
  /**
   * Analyses a document given as XML string via XML-RPC.
   * This method is necessary since the parameters are transmitted
   * as a <code>Vector</code> object which cannot store simple
   * <code>int</code> values. Just calls {@link #analyse(String
   * sessionId, String annCollId, String annId, int startSpan,
   * int endSpan, int depth)} for analysis.
   *
   * @param sessionId session ID of annotation
   * @param annCollId annotation collection ID
   * @param startSpan
   * @param endSpan
   * @param depth
   * @return result of analysis as XML string
   */
  public String analyse(String sessionId, String annCollId, String annId,
      Integer startSpan, Integer endSpan, Integer depth) {
    return analyse(sessionId, annCollId, annId, startSpan.intValue(), endSpan
        .intValue(), depth.intValue());
  }

  /**
   * Analyses a document given as XML string via XML-RPC.
   * This method is necessary since the parameters are transmitted
   * as a <code>Vector</code> object which cannot store simple
   * <code>int</code> values. Just calls {@link #analyse(String
   * sessionId, String annCollId, String annId, int startSpan,
   * int endSpan, int depth)} for analysis.
   *
   * @param sessionId session ID of annotation
   * @param annCollId annotation collection ID
   * @param startSpan
   * @param endSpan
   * @param depth
   * @return result of analysis as UTF-8-encoded byte array
   */
  public byte[] analyseBinary(String sessionId, String annCollId, String annId,
      Integer startSpan, Integer endSpan, Integer depth) {
    try {
      return analyse(sessionId, annCollId, annId, startSpan.intValue(),
          endSpan.intValue(), depth.intValue()).getBytes("utf-8");
    } catch (UnsupportedEncodingException usee) {
      log.error("UnsupportedEncodingException for encoding utf-8");
      return new byte[0];
    }
  }

  /**
   * Analyses a document given as XML string via XML-RPC.
   * This method is necessary since the parameters are transmitted
   * as a <code>Vector</code> object which cannot store simple
   * <code>int</code> values. Just calls {@link #analyse(String
   * sessionId, String annCollId, String annId, int startSpan,
   * int endSpan, int depth)} for analysis.
   *
   * @param sessionId session ID of annotation
   * @param annCollId annotation collection ID
   * @param startSpan
   * @param endSpan
   * @param depth
   * @return result of analysis as UTF-8-encoded byte array
   */
  public byte[] analyseBinary(String sessionId, String annCollId, String annId,
      int startSpan, int endSpan, int depth) {
    try {
      return analyse(sessionId, annCollId, annId, startSpan, endSpan, depth)
          .getBytes("utf-8");
    } catch (UnsupportedEncodingException usee) {
      log.error("UnsupportedEncodingException for encoding utf-8");
      return new byte[0];
    }
  }

  /**
   * Analyses a document given as XML string via XML-RPC.
   * This method is necessary since the parameters are transmitted
   * as a <code>Vector</code> object which cannot store simple
   * <code>int</code> values. Just calls {@link #analyse(String
   * sessionId, String annCollId, String annId, int startSpan,
   * int endSpan, int depth)} for analysis.
   *
   * @param sessionId session ID of annotation
   * @param annCollId annotation collection ID
   * @param startSpan
   * @param endSpan
   * @param depth
   * @param encoding encoding of the returned byte array string
   * @return result of analysis as byte array with given encoding
   */
  public byte[] analyseBinary(String sessionId, String annCollId, String annId,
      Integer startSpan, Integer endSpan, Integer depth, String encoding) {
    try {
      return analyse(sessionId, annCollId, annId, startSpan.intValue(),
          endSpan.intValue(), depth.intValue()).getBytes(encoding);
    } catch (UnsupportedEncodingException usee) {
      log.error("UnsupportedEncodingException for encoding " + encoding);
      return new byte[0];
    }
  }

  /**
   * Analyses a document given as XML string. This document is
   * specified via its session and annotation ID. The parameters for
   * a span are ignored at the moment.
   * @param sessionId session ID of annotation
   * @param annCollId annotation collection ID
   * @param startSpan
   * @param endSpan
   * @param depth
   * @return result (deepest successful analysis) as XML string
   */
  public String analyse(String sessionId, String annCollId, String annId,
      int startSpan, int endSpan, int depth) {
    String deepestResult = "";
    Session session = (Session) getSessions().get(sessionId);
    AnnotationCollection annColl = (AnnotationCollection) session
        .getAnnotationCollections().get(annCollId);
    Annotation annotation = annColl.getAnnotation(annId);
    Query query = null;

    String doc;
    if (databaseUsage) {
      doc = getDatabase().getAnnotationAsString(sessionId, annCollId, annId);
    } else {
      doc = annotation.getAnnotationXml();
    }

    Vector modules = session.getModuleVector();
    String language = annotation.getAnnotationLanguage();

    boolean result;

    if (queryProcessed(sessionId, annId, startSpan, endSpan, depth) == depth) {
      log.info("This query has been processed already");
      return null;
    }

    query = new Query(sessionId, annId, startSpan, endSpan, depth);

    for (int i = 0; i < modules.size(); ++i) {
      Module module = (Module) modules.get(i);

      if (module.suitable(language, depth)) {
        doc = module.process(doc);
        ////log.error(doc);/////REMOVEME FOOBAR
        Metadata tmpMD = new Metadata(doc);

        if (tmpMD.getDiagnosis().equalsIgnoreCase("ok")
            && (tmpMD.getEmpty() == false)) {
          deepestResult = doc;
        }

        String processAnnId = annColl.createAnnotation(doc, language, module
            .getModuleName());

        if (databaseUsage) {
          result = getDatabase().addAnnotation(sessionId, annCollId,
              processAnnId, doc)
              && getDatabase().addIndex(sessionId, module.getModuleName(),
                  "TOKENS");

          if (result) {
            log.info("Annotation added to database: " + processAnnId);
          } else {
            log.error("Couldn't add annotation " + "to database!");
          }
        }
      }
    }

    String processAnnId = annColl.createAnnotation(deepestResult, language,
        "result");
    if (databaseUsage) {
      result = getDatabase().addAnnotation(sessionId, annCollId, "result",
          deepestResult)
          && getDatabase().addIndex(sessionId, "result", "RESULT");

      if (result) {
        log.info("Result annotation added to database: " + processAnnId);
      } else {
        log.error("Couldn't add result annotation to database!");
      }
    }

    //oe: leads to serious time performance problems on a large scale
    // queryQueue.addLast(query);
    ///////

    return deepestResult;
  }

  /**
   * Analyses a document given as DOM tree.
   * This document is specified via its session and annotation ID.
   * The parameters for a span are ignored at the moment.
   *
   * @param sid session ID of annotation
   * @param aid annotation ID
   * @param startSpan
   * @param endSpan
   * @param depth
   * @return result of analysis as DOM tree
   */
  public Node analyseDOMTree(String sid, String aid, int startSpan,
      int endSpan, int depth) {
    Session session = (Session) getSessions().get(sid);
    Annotation annotation = session.getAnnotation(aid);
    Query query = null;
    Vector modules = session.getModuleVector();
    String language = annotation.getAnnotationLanguage();
    boolean result;
    Node doc;

    if (databaseUsage) {
      doc = getDatabase().getAnnotationAsDOM(sid, aid, "rawtext");
    } else {
      doc = annotation.getAnnotationDOMTree();
    }

    if (queryProcessed(sid, aid, startSpan, endSpan, depth) == depth) {
      log.info("This query has been processed already");
    }

    query = new Query(sid, aid, startSpan, endSpan, depth);

    for (int i = 0; i < modules.size(); ++i) {
      Module module = (Module) modules.get(i);

      if (module.suitable(language, depth)) {
        doc = module.processDOMTree(doc);
        if (databaseUsage) {
          getDatabase().addAnnotation(sid, aid, module.getModuleName(), doc);
        }
      }
    }

    if (databaseUsage) {
      result = getDatabase().addAnnotation(sid, aid, "Tokens", doc)
          && getDatabase().addIndex(sid, aid, "annotation");

      if (result) {
        log.info("Annotation added to database");
        log.info("Index created");
      } else {
        log.error("Annotation couldn't be added to database");
        log.error("No index created");
      }
    }

    //oe: leads to serious time performance problems on a large scale
    //queryQueue.addLast(query);
    /////

    return doc;
  }

  /**
   * Retrieves an annotation belonging to a given session and
   * annotation collection specified by an annotation ID.
   * This ID is usually the module name which created this
   * annotation or <tt>rawtext</tt>.
   *
   * @param sessId       session ID referring to the session for which
   *                     an annotation was created
   * @param annCollId    annotation collection ID referring to the
   *                     annotation collection for which
   *                     an annotation was created
   * @param annId        the ID of the annotation
   * @return annotation
   */
  public String getAnnotation(String sessId, String annCollId, String annId) {
    Session session = (Session) getSessions().get(sessId);

    if (session != null) {
      //log.debug("Session still active, retrieve results directly");
      AnnotationCollection annColl = (AnnotationCollection) session
          .getAnnotationCollections().get(annCollId);
      Annotation ann = (Annotation) annColl.getAnnotations().get(annId);
      if (ann==null) {
	  log.error("Annotation " + sessId + "/" + annCollId + "/" + annId + " does not exist.");
	  return null;
      } else
	  return ann.getAnnotationXml();
    } else {
	log.debug("Session closed, retrieve results from database");
	return getDatabase().getAnnotationAsString(sessId, annCollId, annId);
    }
  }

  /**
   * Retrieves an annotation belonging to a given session and
   * annotation collection specified by an annotation ID.
   * This ID is usually the module name which created this
   * annotation or <tt>rawtext</tt>.
   *
   * @param sessId       session ID referring to the session for which
   *                     an annotation was created
   * @param annCollId    annotation collection ID referring to the
   *                     annotation collection for which
   *                     an annotation was created
   * @param annId        the ID of the annotation
   * @return annotation as UTF-8-encoded byte array
   */
  public byte[] getAnnotationBinary(String sessId, String annCollId,
      String annId) {
    try {
      return getAnnotation(sessId, annCollId, annId).getBytes("utf-8");
    } catch (UnsupportedEncodingException usee) {
      log.error("UnsupportedEncodingException for encoding utf-8");
      return new byte[0];
    }
  }

  /**
   * Retrieves an annotation belonging to a given session and
   * annotation collection specified by an annotation ID.
   * This ID is usually the module name which created this
   * annotation or <tt>rawtext</tt>.
   *
   * @param sessId       session ID referring to the session for which
   *                     an annotation was created
   * @param annCollId    annotation collection ID referring to the
   *                     annotation collection for which
   *                     an annotation was created
   * @param annId        the ID of the annotation
   * @param encoding     encoding for byte array to return
   * @return annotation as byte array encoded with encoding
   */
  public byte[] getAnnotationBinary(String sessId, String annCollId,
      String annId, String encoding) {
    try {
      return getAnnotation(sessId, annCollId, annId).getBytes(encoding);
    } catch (UnsupportedEncodingException usee) {
      log.error("UnsupportedEncodingException for encoding " + encoding);
      return new byte[0];
    }

  }

  /**
   * Retrieves an annotation converting it to 7-bit encoded unicode before
   * returning. Useful for XMLRPC-usage.
   */
  public String getAnnotation7bit(String sessId, String annCollId, String annId) {
    return Utility.unicodeTo7bit(this.getAnnotation(sessId, annCollId, annId));
  }

  public int queryProcessed(String sid, String did, int startSpan, int endSpan,
      int depth) {
    Query query = null;
    int processedDepth = -1;
    int i = 0;

    //oe: leads to serious time performance problems on a large scale
    //while (processedDepth <= depth && i != queryQueue.size()) {
    //  query = (Query) queryQueue.get(i++);
    //  processedDepth =
    //      query.processed(sid, did, startSpan, endSpan, depth);
    //}

    return processedDepth;
  }

  /**
   * Returns the instance of the XML-RPC server frontend using this instance
   * as the backend.
   * @return An XML-RPC server using this instance.
   */
  public MoCoManServer getServer() {
    return this.server;
  }

  /**
   * Links the given XML-RPC server to this instance.
   * @param server An XML-RPC server using this instance.
   */
  public void setServer(MoCoManServer server) {
    this.server = server;
  }

  /**
   * Shuts down MoCoMan.
   * This includes closing the XML:DB annotation database, shutting
   * down the XML-RPC server, clearing the query queue and all
   * hashtable for session and handlers.
   * A boolean value is returned in order to make this method available
   * to XML-RPC clients - void would return <code>null</code> which
   * is not supported by XML-RPC.
   * @return <code>true</code> if the shutdown process was successful.
   */
  public boolean shutdown() {
    log.info("Shutting down MoCoMan ...");

    if (databaseUsage) {
      log.info("Closing XMLDBAnnotationDatabase");
      getDatabase().closeDatabase();
    }

    //oe: leads to serious time performance problems on a large scale
    //    log.info("Clearing query cache");
    /////  queryQueue.clear();

    log.info("Closing all sessions");
    Enumeration sessionKeys = getSessions().keys();
    while (sessionKeys.hasMoreElements()) {
      String tempSessionId = (String) sessionKeys.nextElement();
      Session tempSession = (Session) getSessions().get(tempSessionId);
      tempSession.closeSession();
      getSessions().remove(tempSessionId);
    }

    if (this.server != null) {
      this.server.shutdown(); // uschaefer added condition 2004-01-26
    }
    log.info("MoCoMan offline");
    return true;
  }

  /**
   * Return a greeting (<code>String</code>). Use this method to conveniently
   * test a XML-RPC connection.
   * @return A greeting (<code>String</code>).
   */
  public static String sayHello() {
    return "Hello";
  }

  /**
   * Returns the amount of time elapsed since the creation of this instance
   * in millisecond.
   * @return The number of milliseconds this intance has been existing.
   */
  public long getUptime() {
    return System.currentTimeMillis() - startTime;
  }

  /**
   * Increases the sentence counter of this instance by 1.
   */
  public void increaseSentenceCount() {
    this.sentenceCounter++;
  }

  /**
   * Return the number of processed (and counted) sentences.
   * @return int The number of processed (and counted) sentences.
   */
  public int getSentenceCount() {
    return this.sentenceCounter;
  }

  /**
   * Returns the amount of time elapsed since the creation of this instance
   * as a readable String.
   * @return The uptime as a String.
   */
  public String getFormatedUptime() {
    return Utility.formatDuration(getUptime());
  }

  /**
   * Generates and returns a summary String of the current status of this
   * instance.
   * Example:
   * <pre>
   * Uptime: 4 minutes 44 seconds 816 milliseconds
   * logger.host: localhost
   * logger.path: log
   * xmldb.usage: false
   * xmldb.location: xmldb:xindice://clavinova:8080/db
   * logger.port: 4445
   * logger.level: DEBUG
   * logger.file: hoglog.html
   * logger.format: html
   * xmlrpc.server.port: 8411
   * |--session3
   * |--session2
   * |  |--collection2
   * |  |  |--rawtext
   * |  |  |--xmltext
   * |  |--collection1
   * |--session1
   * |  |--collection2
   * |  |--collection1
   * |  |  |--rawtext
   * |  |  |--xmltext
   * </pre>
   * @return A String displaying a summary of the current status of this
   *         instance.
   */
  public String getStatus() {

    StringBuffer sb = new StringBuffer();

    // user + handlers + uptime + config:
    sb.append("  Process owner: \n    ");
    sb.append(System.getProperty("user.name") + "\n");
    sb.append("  Handlers: \n");
    sb.append(Utility.map2String(getServer().getHandlers(), "    "));
    sb.append("  Uptime: \n");
    sb.append("    " + getFormatedUptime() + "\n");
    sb.append("  Configuration:\n");
    sb.append(Utility.map2String(config, "    "));
    sb.append("    logger.filename:  " + getLogFileName() + "\n");
    // data:
    sb.append("  Data:");
    if (getSessions().size() == 0) {
      sb.append("\n      <No data on server>");
      return sb.toString();
    }
    // all sessions
    Hashtable sessions = getSessions();
    Iterator sesIt = sessions.keySet().iterator();
    while (sesIt.hasNext()) {
      String key = sesIt.next().toString();
      Session ses = (Session) sessions.get(key);
      sb.append(Utility.getNewLine());
      sb.append("    |--");
      sb.append(ses.getSessionId() + " [" + ses.getConfigFileName() + "]");

      // all collections
      Hashtable collections = ses.getAnnotationCollections();
      Iterator colIt = collections.keySet().iterator();
      while (colIt.hasNext()) {
        key = colIt.next().toString();
        AnnotationCollection col = (AnnotationCollection) collections.get(key);
        sb.append("\n");
        sb.append("    |  |--");
        sb.append(col.getAnnotationCollectionId());

        // all annotations
        Hashtable annotations = col.getAnnotations();
        Iterator annoIt = annotations.keySet().iterator();
        while (annoIt.hasNext()) {
          key = annoIt.next().toString();
          Annotation anno = (Annotation) annotations.get(key);
          sb.append("\n");
          sb.append("    |  |  |--");
          sb.append(anno.getAnnotationId());
        }
      }
    }
    return sb.toString();
    /*
     Enumeration sessions = sessionsTable.keys();
     while (sessions.hasMoreElements()) {
     Session session = (Session) sessionsTable.get(sessions.nextElement());
     sb.append(session.getSessionId());
     sb.append(":\n");
     Hashtable cTable = session.getAnnotationCollections();
     Enumeration collections = cTable.keys();
     while (collections.hasMoreElements()) {
     AnnotationCollection collection =
     ( (AnnotationCollection) cTable.get(collections.nextElement()));
     sb.append("\t");
     sb.append(collection.getAnnotationCollectionId());
     sb.append("\n");
     sb.append("and so forth, and so forth (to be implemented)");
     }
     }
     sb.append(getSessions().toString());
     return sb.toString();*/
  } // getStatus()

  public String getStatusXml() {

    StringBuffer xml = new StringBuffer("<status>\n");
    String sep = Utility.getNewLine();

    // handlers
    xml.append(Utility.map2String(getServer().getHandlers(), "  <handler>",
        "</handler>")
        + sep);
    // uptime
    xml.append("  <uptime>" + getFormatedUptime() + "</uptime>" + sep);
    // config:
    xml.append(Utility.conf2Xml(config));
    // data:
    xml.append("  <data>" + sep);
    //   all sessions
    Hashtable sessions = getSessions();
    Iterator sesIt = sessions.keySet().iterator();
    while (sesIt.hasNext()) {
      String key = sesIt.next().toString();
      Session ses = (Session) sessions.get(key);
      xml.append("    <session id=\"" + ses.getSessionId() + "\">" + sep);
      // all collections
      Hashtable collections = ses.getAnnotationCollections();
      Iterator colIt = collections.keySet().iterator();
      while (colIt.hasNext()) {
        key = colIt.next().toString();
        AnnotationCollection col = (AnnotationCollection) collections.get(key);
        xml.append("      <collection id=\"" + col.getAnnotationCollectionId()
            + "\">" + sep);
        // all annotations
        Hashtable annotations = col.getAnnotations();
        Iterator annoIt = annotations.keySet().iterator();
        while (annoIt.hasNext()) {
          key = annoIt.next().toString();
          Annotation anno = (Annotation) annotations.get(key);
          xml.append("        <annotation id=\"" + anno.getAnnotationId()
              + "\"/>" + sep);
        } // while anno
        xml.append("      </collection>" + sep);
      } // while collIt
      xml.append("    </session>" + sep);
    } // while sesIt
    xml.append("  </data>" + sep);
    xml.append("</status>");
    return xml.toString();
  } // getStatusXml()

  /**
   * Returns all annotation IDs in a given session.collection as a comma-sep.
   * list sorted by module depth, represented by a String.
   * Main purpose: XMLRPC.
   */
  public String getAnnotationIds(String sessionId, String collectionId) {
    return getAnnotationIds(sessionId, collectionId, false);
  }

  /**
   * Returns those annotation IDs in a given session.collection where
   * the annotation is present in XML format. The returned list is
   * sorted by module depth, represented by a String. Main purpose: XMLRPC.
   */

  public String getXmlAnnotationIds(String sessionId, String collectionId) {
    return getAnnotationIds(sessionId, collectionId, true);
  }

  /**
   * Does the actual work for
   * #getXmlAnnotationIds(String sessionId, String collectionId)
   */
  private String getAnnotationIds(String sessionId, String collectionId,
      boolean xmlOnly) {
    log.debug("getAnnotationIds(): " + "\nsessionId: " + sessionId
        + "\ncollectionId: " + collectionId + "\n xmlOnly: " + xmlOnly);
    ArrayList annotations = new ArrayList();
    // this one to sort the annotations by depth
    Hashtable tmpTable = new Hashtable();
    // get the annotations
    Session session = (Session) getSessions().get(sessionId);
    AnnotationCollection collection = null;
    if (session != null) {
      collection = (AnnotationCollection) session.getAnnotationCollections()
          .get(collectionId);
    }
    if (collection != null) {
      Hashtable annoHT = collection.getAnnotations();
      Enumeration ids = annoHT.keys();
      // put them into a tmp hashtable
      while (ids.hasMoreElements()) {
        String id = "SDL";
        // filter out SDL*
        while (id.startsWith("SDL")) {
          log.debug("getAnnotationIds(): discarding " + id);
          id = (String) ids.nextElement();
        }
        log.debug("getAnnotationIds(): adding " + id + "to temp. list");
        int d = ((Annotation) annoHT.get(id)).getDepth();
        if (id.equalsIgnoreCase("result")) {
          // we want result to be sorted after its corresponding annotation
          d++;
        }
        Float depth = new Float(d);
        if ((!xmlOnly) || (d > 0)) {
          while (tmpTable.containsKey(depth)) {
            depth = new Float(depth.floatValue() + 0.001);
          }
          tmpTable.put(depth, id);
        } // if
      } // while
      // sort them by depth
      Vector v = new Vector(tmpTable.keySet());
      Collections.sort(v);
      // copy the sorted list into an arraylist
      Enumeration e = v.elements();
      while (e.hasMoreElements()) {
        Float key = (Float) e.nextElement();
        String value = (String) tmpTable.get(key);
        annotations.add(value);
      }
    } // if collection != null
    log.debug("getAnnotationIds: length of 'annotations': "
        + annotations.size());
    log.debug("getAnnotationIds(): returning: \n"
        + Utility.abstractList2String(annotations, "", ","));
    return Utility.abstractList2String(annotations, "", ",");
  } // getAnnotationIds()

  /** Transform an annotation to HTML with
   *  special handling for RMRS, where a &lt;text&gt; element is
   *  inserted iff insertText==1
   *  @param sid session ID
   *  @param acid annotation collection ID
   *  @param aid annotation ID
   *  @param styleSheetName name of the stylesheet to apply on the server
   *  @param insertText =="yes" means do insert the  &lt;text&gt; element
   */
  public String xml2htmlTransformer7bit(String sid, String acid, String aid,
      String stylesheetName, String insertText) {
    return Utility.unicodeTo7bit(xml2htmlTransformer(sid, acid, aid,
        stylesheetName, insertText));
  }

  /** Transform an annotation to HTML with
   *  special handling for RMRS, where a &lt;text&gt; element is
   *  inserted iff insertText==1
   *  @param sid session ID
   *  @param acid annotation collection ID
   *  @param aid annotation ID
   *  @param styleSheetName name of the stylesheet to apply on the server
   *  @param insertText =="yes" means do insert the  &lt;text&gt; element
   */
  public byte[] xmlTransformerBinary(String sid, String acid, String aid,
      String stylesheetName, String insertText) {
    try {
      return xml2htmlTransformer(sid, acid, aid, stylesheetName, insertText)
          .getBytes("utf-8");
    } catch (UnsupportedEncodingException usee) {
      log.error("UnsupportedEncodingException for encoding utf-8");
      return new byte[0];
    }
  }

  /** Transform an annotation to HTML with
   *  special handling for RMRS, where a &lt;text&gt; element is
   *  inserted iff insertText==1
   *  @param sid session ID
   *  @param acid annotation collection ID
   *  @param aid annotation ID
   *  @param styleSheetName name of the stylesheet to apply on the server
   *  @param insertText =="yes" means do insert the  &lt;text&gt; element
   *  @param encoding encoding of the returned byte array string
   */
  public byte[] xmlTransformerBinary(String sid, String acid, String aid,
      String stylesheetName, String insertText, String encoding) {
    try {
      return xml2htmlTransformer(sid, acid, aid, stylesheetName, insertText)
          .getBytes(encoding);
    } catch (UnsupportedEncodingException usee) {
      log.error("UnsupportedEncodingException for encoding " + encoding);
      return new byte[0];
    }
  }

  /** Transform an annotation to HTML with
   *  special handling for RMRS, where a &lt;text&gt; element is
   *  inserted iff insertText==1
   *  @param sid session ID
   *  @param acid annotation collection ID
   *  @param aid annotation ID
   *  @param styleSheetName name of the stylesheet to apply on the server
   *  @param insertText =="yes" means do insert the  &lt;text&gt; element
   */
  public String xml2htmlTransformer(String sid, String acid, String aid,
      String stylesheetName, String insertText) {
    String anno = this.getAnnotation(sid, acid, aid);
    //log.debug("insertText=" + insertText + " " + anno);
    //writeEncodedFile("/tmp/"+System.getProperty("user.name")+"/hog/server-"+aid+".xml", anno, "utf-8");
    String text = null;
    if (insertText.equals("yes")) {
      text = Utility.rmrsText(this.getAnnotation(sid, acid, "rawtext"));
    }
    StringBuffer sbRmrs = new StringBuffer();
    int i = 0;
    if (insertText.equals("yes")) {
      int j = anno.indexOf("<rmrs ");
      while (j > -1) {
        j = anno.indexOf(">", j + 4) + 1;
        sbRmrs.append(anno.substring(i, j));
        sbRmrs.append(text);
        i = j;
        j = anno.indexOf("<rmrs ", i);
      }
    }
    sbRmrs.append(anno.substring(i, anno.length()));
    log.debug(sbRmrs.toString());
    TransformationService transformationService = null;
    Transformer transformer = null;
    try {
      transformationService = new TransformationService(this);
      transformer = transformationService.createTransformer(stylesheetName);
      transformer.setParameter("sessionID", sid);
      transformer.setParameter("collectionID", acid);
      transformer.setParameter("annotationID", aid);
    } catch (TransformerConfigurationException tce) {
      log.error("xml2htmlTransformer: " + Utility.exception2String(tce));
      return "Transformation error: " + tce.getMessage();
    }

    try {
      return transformationService.transform(transformer, sbRmrs.toString());
    } catch (TransformerException te) {
      log.error("xml2htmlTransformer: " + Utility.exception2String(te));
      return "Transformation error: " + te.getMessage();
    }
  }

  /** compute configured log4j HTML appender and return its file content as String
   *
   * @return HTML log file content as String
   */
  public String getLogString() {
    Enumeration appenders = log.getRootLogger().getAllAppenders();
    Appender a;
    while (appenders.hasMoreElements()) {
      a = (Appender) appenders.nextElement();
      if ((a instanceof FileAppender) && (a.getLayout() instanceof HTMLLayout)) {
        return Utility
            .readStringFromFile(((FileAppender) a).getFile(), "utf-8");
      }
    }
    return "no HTML file logger found";
    //    Utility.readStringFromFile(this.loggerPath + "/" + this.loggerFile, "utf-8");
  }

  /**
   * Returns the number of messages currently in the html log file.
   */
  public String getNumLogMesgs() {
    return Integer.toString(Utility.findAllrx("<tr>.*?</tr>",
        this.getLogString()).size());
  }

  /**
   * Returns the last n messages as read from the log4j html log file.
   * The returned messages are embedded in an html-table.
   */
  public String getLastLogMessage(int n) {
    ArrayList allMsgs = Utility.findAllrx("<tr>.*?</tr>", this.getLogString());
    //System.err.println("has SIZE: " + allMsgs.size());
    StringBuffer sb = new StringBuffer(
        "<table cellspacing=\"0\" cellpadding=\"4\" border=\"1\" "
            + "bordercolor=\"#224466\" width=\"100%\">");
    if (n > allMsgs.size()) {
      n = allMsgs.size();
    } else {
      sb.append("<tr>" + "<th>Time</th>" + "<th>Level</th>" + "<th>Class</th>"
          + "<th>Message</th>");
    }
    for (int i = allMsgs.size() - n; i <= allMsgs.size() - 1; i++) {
      sb.append((String) allMsgs.get(i));
    }
    sb.append("</table>");
    return sb.toString();
  }

  public String getAllLogMessages() {
    StringBuffer sb = new StringBuffer(
        "<table cellspacing=\"0\" cellpadding=\"4\" border=\"1\" "
            + "bordercolor=\"#224466\" width=\"100%\">");
    sb.append(Utility.extractHtmlElement(this.getLogString(), "tr"));
    sb.append("</table>");
    return sb.toString();
  }

  public String getLastLogMessage() {
    return this.getLastLogMessage(1);
  }

  public String getLast5LogMessages() {
    return this.getLastLogMessage(5);
  }

  public String getLast20LogMessages() {
    return this.getLastLogMessage(20);
  }

  public String getLast100LogMessages() {
    return this.getLastLogMessage(100);
  }

  public String getLast500LogMessages() {
    return this.getLastLogMessage(500);
  }

  public long getUniqueID() {
    return this.currentUniqueID++;
  }

  public void writeEncodedFile(String sFilename, String sContent,
      String encoding) {
    try {
      byte[] b = sContent.getBytes(encoding);
      BufferedOutputStream file = new BufferedOutputStream(
          new FileOutputStream(sFilename));
      file.write(b);
      file.flush();
      file.close();
    } catch (IOException e) {
      System.err.println(e.toString());
    }
  }

  private String getLogFileName() {
    Enumeration appenders = log.getRootLogger().getAllAppenders();
    Appender a;
    while (appenders.hasMoreElements()) {
      a = (Appender) appenders.nextElement();
      if ((a instanceof FileAppender) && (a.getLayout() instanceof HTMLLayout)) {
        return ((FileAppender) a).getFile();
      }
    }
    return "no HTML file logger found";
  }

  private void appendProtocolFile(String text, String lang) {
      String sFilename = getConfig().getProperty(PROTOCOL_FILE_PROP);
      if (sFilename!=null && !sFilename.equals("${protocol.file}")) {
          try {
           byte[] bbuf = new String(protocolDateTimeFormat.format(new Date()) 
                          + " " + lang + " " + text + "\n").getBytes("utf-8");
           BufferedOutputStream file = new BufferedOutputStream(
                                       new FileOutputStream(sFilename, true));
           file.write(bbuf, 0, bbuf.length);
           file.flush();
           file.close();
          } catch (IOException e) {
              log.warn("appendProtocolFile: " + e.toString());
          }
      }
  }


} // class
