/*
 * MoCoManServer.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.util.*;
import java.io.IOException;
import org.apache.log4j.Logger;
import org.apache.xmlrpc.WebServer;
import org.apache.xmlrpc.XmlRpc;
import de.dfki.lt.hog.util.Utility;

/**
 * This is a single threaded server running as an XML-RPC
 * front-end to <tt>MoCoMan</tt>. Start the server by invoking
 * {@link #main(String[] args)}.
 *
 * Sample usage (Python):
 *
 * 1. Start the server by running this class.
 *
 * 2. python:
 * <code>
 * >>> import xmlrpclib
 * >>> srv = xmlrpclib.Server("http://localhost:8411")
 * >>> print srv.mocoman.getServerPort()
 * 8411
 * >>>
 * </code>
 *
 * @author <a href=mailto:oeze@dfki.de>Oezguer Demir</a>
 * @author <a href="mailto:robba@dfki.de">Robert Barbey</a>
 * @version $Id: MoCoManServer.java,v 1.5 2006/06/08 15:44:39 uschaefer Exp $
 */
public class MoCoManServer
    implements XmlRpcServerInterface {

  /**
   * Key for WebServer port in <code>Properties</code> */
  private static final String WEBSERVER_PORT_PROP =
      "xmlrpc.server.port";

  /**
   * Port on which the XML-RPC server is listening */
  private int mocoServerPort;

  /**
   * Contains all handlers for XML-RPC communication registered so far.
   */
  private Hashtable handlers;

  /** A log4j Logger. */
  static Logger log = Logger.getLogger(MoCoManServer.class);

  /**
   * A server for communication with remote modules or parts of the
   * system via XML-RPC. */
  private WebServer mocoServer;

  private MoCoMan mocoman = MoCoMan.getInstance();

  /**
   * Creates a new instance of MoCoManServer using the port number
   * specified in the config. file. Returns a running(!) server.
   */
  public MoCoManServer() throws IOException {
    // get port number from config file and init
    try {
      mocoman.getConfig().setProperty(WEBSERVER_PORT_PROP, System.getProperty(WEBSERVER_PORT_PROP));
      init(Integer.parseInt(mocoman.getConfig().getProperty(WEBSERVER_PORT_PROP)));
    }
    catch (NumberFormatException e) {
      log.error("Could not parse port number from config file" +
                mocoman.getConfig() + "\n Error: " + Utility.exception2String(e));
    }
    catch (IOException e) {
      throw e;
    }
  }

  /**
   * Creates a new instance of MoCoManServer using the port number given.
   */
  public MoCoManServer(int port) throws IOException {
    try {
      init(port);
    } catch (IOException e) {
      throw e;
    }
  }

  private void init(int port) throws IOException {
    // setup XML-RPC server
    this.setServerPort(port);
    XmlRpc.setEncoding("UTF8");
    this.handlers = new Hashtable();
    this.setServer(new WebServer(getServerPort()));
    log.info("XML-RPC WebServer set up on port "
             + mocoServerPort);

    // add a handler for XML-RPC calls
    this.addHandler("mocoman", mocoman);
    this.addHandler("metadata", new Metadata());
    this.addHandler("util", new Utility());

    // link this instance to mocoman
    this.mocoman.setServer(this);

    // run the server of this instance
    //this.getServer().run();
    this.getServer().start();
    log.info("XML-RPC-Server ready at port " + getServerPort());
  } // init()

//  public void run() {
//    this.getServer().start();
//    this.getServer().run();
//    log.info("XML-RPC-Server ready at port " + getServerPort());
//  }

  /**
   * Returns the server for XML-RPC calls.
   *
   * @return an XML-RPC server
   */
  public WebServer getServer() {
    return this.mocoServer;
  }

  private void setServer(WebServer aServer) {
    this.mocoServer = aServer;
  }

  /**
   * Returns the port on which the XML-RPC server
   * is listening.
   *
   * @return port of server
   */
  public int getServerPort() {
    return this.mocoServerPort;
  }

  private void setServerPort(int aPort) {
    this.mocoServerPort = aPort;
  }

  public void addHandler(String aModuleName, Object aHandler) {
    getServer().addHandler(aModuleName, aHandler);
    getHandlers().put(aModuleName, aHandler);
  }

  /**
   * Returns a <code>Hashtable</code> containing
   * all registered handlers for XML-RPC calls.
   *
   * @return all handlers for XML-RPC calls
   */
  public Hashtable getHandlers() {
    return this.handlers;
  }

  /**
   * Retrieves a single handler object corresponding to
   * a given handler name stored
   * in the <code>handlers</code> hashtable.
   *
   * @return a single handler
   */
  public Object getHandler(String aHandlerName) {
    return getHandlers().get(aHandlerName);
  }

  /**
   * Shuts down MoCoMan.
   * This includes closing the XML:DB annotation database, shutting
   * down the XML-RPC server, clearing the query queue and all
   * hashtable for session and handlers.
   */
  public void shutdown() {
    log.info("Shutting down MoCoMan ...");
    log.info("Shutting down XML-RPC server");
    if (getServer() == null) {
      log.fatal("getServer() == null");
    }
    else {
      getServer().shutdown();
    }
    log.info("MoCoManServer offline");
  }

  public static void main(String[] args) {
    MoCoManServer server;
    try {
      // init the server
      server = new MoCoManServer();
      // start running the server
//      server.run();
      System.err.println("XML-RPC-Server ready at port "
                         + server.getServerPort());
    } // try
    catch (IOException ioe) {
      String eMessage = "Couldn't set up XML-RPC WebServer on port "
          + MoCoMan.getInstance().getConfig().getProperty(WEBSERVER_PORT_PROP)
          + "\nError: " + Utility.exception2String(ioe);
      log.error(eMessage);
      System.err.println(eMessage);
      System.exit(-1);
    } // catch
    /* does not work
    if (args.length > 0) {
      int port = Integer.parseInt(args[0]);
      try {
        server = new MoCoManServer(port);
        System.err.println("XML-RPC-Server ready at port "
                           + server.getServerPort());
      } // try
      catch (IOException ioe) {
          String eMessage = "Couldn't set up XML-RPC WebServer on port "
              + port + "\nError: " + Utility.exception2String(ioe);
          log.error(eMessage);
          System.err.println(eMessage);
          System.exit(-1);
      } // catch
    } // if
    */
  } // main
} // class
