/*
 * Module.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.util.*;
import java.io.*;
import org.w3c.dom.*;
import org.apache.log4j.Logger;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.exceptions.NoSuchPropertyException;
import de.dfki.lt.hog.exceptions.ShutdownException;

/**
 * <code>Module</code> is an abstract class that provides methods for
 * a module to process an XML string or a DOM tree, respectively. <br>
 * DeepThought Adaptation by
 * <a href="mailto:robba@dfki.de">Robert Barbey</a>
 *
 * @author Joerg Steffen, DFKI
 * @version $Id: Module.java,v 1.2 2005/04/26 17:37:20 oeze Exp $ */

public abstract class Module {

    /**
     * Key for name of this <code>Module</code> in the config. */
    public static final String MODULE_NAME_PROP = "module.name";

    /**
     * Key for language in configurating <code>Properties</code> */
    public static final String MODULE_LANGUAGE_PROP = "module.language";

    /**
     * Key for depth of this <code>Module</code>in configuration. */
    public static final String MODULE_DEPTH_PROP = "module.depth";

    /**
     * Key for root element name of this <code>Module</code>in configuration. */
    public static final String MODULE_ROOTELEMENT_PROP = "module.rootelement";

    /**
     * This is the path to the summarizer root directory. */
    public static final String ROOT =
	System.getProperty("summarizer.root");

    /**
     * This is the system local newline string. */
    public static final String LINE_SEPARATOR =
	System.getProperty("line.separator");

    /**
     * A logger for <code>Module</code>s */
    static protected Logger moduleLog = Logger.getLogger(Module.class);

    /**
     * This is the value for a flag property that is set. */
    protected static final String YES = "yes";

    /**
     * This is the value for a flag property that is unset. */
    protected static final String NO = "no";

    /**
     * This is the value for a property that is not used. */
    protected static final String NONE = "none";

    /**
     * This is the property for the config files search path in the
     * config. This property is copied to all module properties in case
     * they need to search a config file themself. */
    protected static final String CONFIG_SEARCH_PATH_PROP =
	"config.search.path";

    /**
     * This contains a <code>Properties</code> object with the
     * individual configs of the module. */
    private Properties config;

    /**
     * This returns the <code>Properties</code> object with the
     * individual configs of the module.
     *
     * @return a <code>Properties</code> object */
    protected Properties getConfig() {
	return this.config;
    }

    /**
     * This sets the individual configs of the module.
     *
     * @param aConfig a <code>Properties</code> object */
    protected void setConfig(Properties aConfig) {
	this.config = aConfig;
    }

    /**
     * This contains the key under which the module is stored in the
     * registry. */
    private String moduleName;


    /** Contains a reference back to the current MoCoMan instance
        (uschaefer).
    */
    private MoCoMan mocoman;

    /** Return a reference back to the current MoCoMan instance
        (uschaefer).
        @return instance of MoCoMan (set by MoCoMan module initialization loop)
    */
    public MoCoMan getMoCoMan() {
        return this.mocoman;
    }

    /** Store a reference back to the MoCoMan instance
        (uschaefer).
        @param mcm instance of MoCoMan (given MoCoMan module initialization loop)
    */
    protected void setMoCoMan(MoCoMan mcm) {
        this.mocoman = mcm;
    }


    /**
     * This returns the key under which the module is stored in the
     * registry.
     *
     * @return a <code>String</code> */
    protected String getModuleName() {
	return this.moduleName;
    }

    /**
     * This sets the module name.
     *
     * @param aName a <code>String</code> */
    protected void setModuleName(String aName) {
	this.moduleName = aName;
    }

    /**
     * This contains an abbreviated string indicating the language
     * of this <code>Module</code>.
     */
    private String moduleLanguage;

    /**
     * Returns language of this <code>Module</code> as an abbreviated string.
     * @return language
     */
    protected String getModuleLanguage()
    {
	return this.moduleLanguage;
    }

    /**
     * This sets the module language.
     * @param aLanguage
     */
    protected void setModuleLanguage(String aLanguage)
    {
	this.moduleLanguage = aLanguage;
    }

    /**
     * Specifies the depth this <code>Module</code> is ment for.
     */
    private int moduleDepth;

    /**
     * Returns depth of this <code>Module</code>
     * @return depth
     */
    protected int getModuleDepth()
    {
	return this.moduleDepth;
    }

    protected void setModuleDepth(int aDepth)
    {
	this.moduleDepth = aDepth;
    }


    /**
     * Root element name for this Module (default set to module name in init()) */
    private String moduleRootElement;

    /**
     * .
     * Return root element name for this Module (default set to module name in init())
     * @return a <code>String</code> */
    protected String getModuleRootElement() {
	return this.moduleRootElement;
    }

    /**
     * Set root element name for this Module
     * @param sElement a <code>String</code> */
    protected void setModuleRootElement(String sElement) {
	this.moduleRootElement = sElement;
    }


    /**
     * This is the default shutdown method. If any special action is
     * needed for shutdown, overide this method in subclass.
     *
     * @exception ShutdownException if shutdown fails */
    public void shutdown()
	throws ShutdownException{};

    /**
     * This makes the basic initialization of the module for a
     * <code>Properties</code> object <code>configProps</code>.
     *
     * @param configProps a <code>Properties</code> object that contains
     * at least the module name
     * @exception InitializationException if initialization fails */
    protected void init(Properties configProps)
	throws InitializationException {

	setConfig(configProps);

	// set module name, language and depth;
	// each module must set these properties
	setModuleName(getConfig().getProperty(MODULE_NAME_PROP, "unknown"));
	setModuleLanguage(getConfig().getProperty(MODULE_LANGUAGE_PROP, "en"));
	setModuleDepth(Integer.parseInt(getConfig().getProperty(MODULE_DEPTH_PROP, "0")));
	setModuleRootElement(getConfig().getProperty(MODULE_ROOTELEMENT_PROP, getModuleName()));
    }

    /**
     * Determines whether this <code>Module</code> is suitable for analysis
     * of a document with given language and depth.
     * @param language
     * @param depth
     * @return true iff this <code>Module</code> is suitable for analysis
     */
    protected boolean suitable(String language, int depth) {
	String lang =  getModuleLanguage(); // was: getConfig().getProperty(MODULE_LANGUAGE_PROP, "de");
	int d = getModuleDepth(); // was: Integer.parseInt(getConfig().getProperty(MODULE_DEPTH_PROP, "1"));

	return (language.equalsIgnoreCase(lang) && d <= depth);
    }

    public abstract String process(String processString);

    public abstract Node processDOMTree(Node domTree);




    // ========================================================================
    // ==   unused methods       ==============================================
    // ========================================================================

    /**
     * This returns a <code>File</code> build from the value of a
     * property as found in the properties object of the module.
     * If the pathname is relative, the summarizer root is added as a
     * prefix.
     *
     * @param aProperty a <code>String<code> with the property
     * @return a <code>File</code> build from the value of the property
     * @exception NoSuchPropertyException if property is not found */
    protected File getFileProperty(String aProperty)
	throws NoSuchPropertyException {

	// get value
	String val = this.getConfig().getProperty(aProperty);
	if (null == val)
	    throw new NoSuchPropertyException
		("no entry for " + aProperty + " found in config properties");
	// create file
	File aFile = new File(val);
	if (aFile.isAbsolute())
	    return aFile;
	else
	    return new File(ROOT + File.separator + val);
    }

    /**
     * This takes a file path and returns it if it's an absolut path or
     * appends it to the summarizer root if it's a relative path.
     *
     * @param path a <code>String</code> with a path
     * @return a possibly modified path */
    protected String getAbsolutOrRelativePath(String path) {
	File testFile = new File(path);
	if (testFile.isAbsolute())
	    return path;
	else
	    return ROOT + File.separator + path;
    }

    /**
     * This searchs a config file in the config file search path and
     * returns it's first occurrence. The config file search path is
     * retrieved from the <code>CONFIG_SEARCH_PATH_PROP</code> property
     * of the config file.
     *
     * @param configFile a <code>String</code> with the config filename
     * @return a <code>File</code> with the first occurrence of the
     * config file in the search path
     * @exception InitializationException when config file is not found */
    protected File searchConfigFile(String configFile)
	throws InitializationException {
	StringTokenizer st =
	    new StringTokenizer
	    (this.getProperty(CONFIG_SEARCH_PATH_PROP), ":".intern());
	List configDirs = new ArrayList();
	while (st.hasMoreTokens()) {
	    configDirs.add
		(this.getAbsolutOrRelativePath(st.nextToken().trim()));
	}
	return this.searchConfigFile(configFile, configDirs);
    }

    /**
     * This returns the value of a property as found in the properties
     * object of the module.
     *
     * @param aProperty a <code>String<code> with the property
     * @return a <code>String</code> with the value of the property
     * @exception NoSuchPropertyException if property is not found */
    protected String getProperty(String aProperty)
	throws NoSuchPropertyException {

	// get value
	String val = this.getConfig().getProperty(aProperty);
	if (null == val)
	    throw new NoSuchPropertyException
		("no entry for " + aProperty + " found in config properties");
	return val;
    }

    /**
     * This returns the value of a property as found in the properties
     * object of the module. If property is not found, it returns the
     * default value.
     *
     * @param aProperty a <code>String<code> with the property
     * @param aDefault a <code>String</code> with the default value
     * @return a <code>String</code> with the value of the property or
     * the default value if property is not found in config */
    protected String getProperty(String aProperty, String aDefault) {

	// get value
	String val = this.getConfig().getProperty(aProperty);
	if (null == val)
	    return aDefault;
	else
	    return val;
    }

    /**
     * This sets the value of a property in the properties object of
     * the module.
     *
     * @param aProperty a <code>String<code> with the property
     * @param aValue a <code>String</code> with the value */
    protected void setProperty(String aProperty,
			       String aValue) {
	this.getConfig().setProperty(aProperty, aValue);
    }

    /**
     * This searchs a config file in the config file search path and
     * returns it's first occurrence.
     *
     * @param configFile a <code>String</code> with the config file's name
     * @param dirList a <code>List</code> of <code>String</code>s with
     * the config file directories
     * @return a <code>File</code> with the first occurrence of the
     * config file in the search path
     * @exception InitializationException when config file is not found */
    protected File searchConfigFile (String configFile,
				     List dirList)
	throws InitializationException {

	// iterate over directory list
	Iterator it = dirList.iterator();
	while (it.hasNext()) {
	    File testFile = new File
		((String)it.next() + File.separator  + configFile);
	    if (testFile.exists()) {
		return testFile;
	    }
	}
	throw new InitializationException
	    ("no config file " + configFile + " found in search path " +
	     // the directory list is build form the property
	     // CONFIG_SEARCH_PATH_PROP, so we can use it here
	     this.getProperty(CONFIG_SEARCH_PATH_PROP));
    }

}
