/*
 * Registry.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.util.*;
import de.dfki.lt.hog.exceptions.NoSuchModuleException;

/**
 * <code>Registry</code> provides methods to store and retrieve {@link
 * Module}s in a central hashmap. This class is realised using a
 * singelton pattern since there can only be exactly one instance of
 * <code>Registry</code>. This class is the local registry in contrast
 * to the (public) rmi registry.
 *
 * @author Joerg Steffen, DFKI
 * @version $Id: Registry.java,v 1.2 2005/04/26 17:37:21 oeze Exp $ */

public class Registry {

  /**
   * This contains the only instance of <code>Registry</code>. */
  private static Registry instance = null;

  /**
   * This contains a hash map that maps module names to the
   * modules. Modules communicate with each other using these module
   * names. */
  private HashMap modules;

  /**
   * This contains a hash map that maps class names to module
   * names. This is needed because the {@link Launcher} communicates
   * with the <code>Registry</code> via class names. */
  private HashMap moduleNames;


  /**
   * This standard constructor creates a new instance of
   * <code>Registry</code>. Not to be used outside this class. */
  private Registry() {
    this.setModules(new HashMap());
    this.setModuleNames(new HashMap());
  }


  /**
   * This controlls the access to the instance of
   * <code>Registry</code>.
   *
   * @return the only existing instance of <code>Registry</code> */
  public static Registry getInstance() {
    if (null == instance)
      instance = new Registry();
    return instance;
  }


  /**
   * This returns the hash map that maps module names to the
   * modules. Modules communicate with each other using these module
   * names.
   *
   * @return the <code>HashMap</code> with modules */
  public HashMap getModules() {
    return this.modules;
  }

  /**
   * This sets the field {@link #modules} to <code>aHashMap</code>.
   *
   * @param aHashMap the <code>HashMap</code> with the modules */
  private void setModules(HashMap aHashMap) {
    this.modules = aHashMap;
  }


  /**
   * This returns the hash map that maps class names to module
   * names. This is needed because the {@link Launcher} communicates
   * with the <code>Registry</code> via class names.
   *
   * @return a <code>HashMap</code> with the module names */
  public HashMap getModuleNames() {
    return this.moduleNames;
  }

  /**
   * This sets the module names.
   *
   * @param aHashMap a <code>HashMap</code> with the module names */
  public void setModuleNames(HashMap aHashMap){
    this.moduleNames = aHashMap;
  }


  /**
   * This registers <code>aModule</code> using the name of the module as
   * defined in the module as key.
   *
   * @param aModule a {@link Module} */
  public void registerModule(Module aModule) {
    this.getModules().put(aModule.getModuleName(), aModule);
    this.getModuleNames()
      .put(aModule.getClass().getName(), aModule.getModuleName());
  }


  /**
   * This unregisters <code>aModule</code>.
   *
   * @param aModule a {@link Module} */
  public void unregisterModule(Module aModule) {
    this.getModules().remove(aModule.getModuleName());
    this.getModuleNames().remove(aModule.getClass().getName());
  }


  /**
   * This retrieves the {@link Module} stored under the module name
   * <code>aModName</code>. This method is used by
   * the Agents to retrieve modules from the registry.
   *
   * @param aModName a <code>String</code> that is used as key
   * @return the {@link Module} stored under that name
   * @exception NoSuchModuleException if requested module doesn't
   * exist in the registry
   */
  public Module retrieveModuleViaName(String aModName)
    throws NoSuchModuleException {
    Object module = this.getModules().get(aModName);
    if (null == module)
      throw new NoSuchModuleException
        ("no module " + aModName + " found in registry");
    else
      return (Module)module;
  }


  /**
   * This retrieves the {@link Module} that is an instance of the
   * class with name <code>aModClass</code>. This method is used by
   * the {@link Launcher} to retrieve modules from the registry.
   *
   * @param aModClass a <code>String</code> with the class name
   * @return the {link Module} that is an instance of this class.
   * @exception NoSuchModuleException if requested module doesn't
   * exist in the registry */
  public Module retrieveModuleViaClass(String aModClass)
    throws NoSuchModuleException {

    String moduleName = (String)this.getModuleNames().get(aModClass);
    if (null == moduleName)
      throw new NoSuchModuleException
        ("no module class " + aModClass + " found in registry");
    else
      return retrieveModuleViaName(moduleName);
  }
}
