/*
 * TransformationService.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.io.File;
import java.io.StringReader;
import java.io.StringWriter;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.URIResolver;
import org.apache.log4j.Logger;
import org.w3c.dom.Node;


/**
 * Service class mainly for module implementations that make use of
 * XSLT transformations. This class also implements a URI resolver
 * that allows to access HoG annotations from within stylesheets
 * using the document() function of XSLT. The URI syntax for HoG
 * annotations is
 * <code>hog://sessionID/collectionID/annotationID</code>.
 *
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @version $Revision: 1.5 $
 * $Id: TransformationService.java,v 1.5 2006/05/04 19:34:45 uschaefer Exp $
 */
public class TransformationService implements URIResolver {

	private TransformerFactory transformerFactory;
	private static Logger log = Logger.getLogger(TransformationService.class);
	private static final String hogURIprefix = "hog://";
	private MoCoMan mocoman;

	/**
         The <code>TransformationService</code> class supports XSL
         transformations plus a
         special handling for access to HoG-internal annotations for the XSLT
         <code>document()</code> function. The syntax in XPath expressions
         is:<br>
         <code>document('hog://sid/acid/aid')</code><br>
         where sid is a sessionID, acid is a collection ID and aid is an
         annotation identifier such as Chunkie or JTok.<br>
         To use the TransformationService in a Module subclass, create a
         field variable for a TransformationService object, say
         transformationService:<br><br>
         <code>
         private TransformationService transformationService = null;
         </code><br>
         ...<br>
         in <code>process()</code>:<br><br>
         <code>
         if (this.transformationService==null) {<br>
         this.transformationService = new TransformationService(getMoCoMan());<br>
         myTransformer = this.transformationService.createTransformer(xslfile);<br>
         }<br>
         <br>
         myTransformer.setParameter("sessionID", sid);<br>
         myTransformer.setParameter("collectionID", acid);<br>
         myTransformer.setParameter("annotationID", aid);<br>
         String result = transformationService.transform(myTransformer, xmlinput);<br>
         </code>
         <br>
         Access to multiple input annotations can be obtained by adding further
         parameters with unique names for annotationIDs (e.g., annotationID2
         etc.).<br>
         There are also DOM-based variants of <code>createTransformer()</code>
         and <code>transform()</code>.<br>
         The method <code>getTransformerFactory()</code> gives access to the
         underlying TransformerFactory.
	 */
        public TransformationService(MoCoMan mocoman) {
            this.transformerFactory = TransformerFactory.newInstance();
            this.mocoman=mocoman;
	}

        /** Alternative constructor (not recommended). setMoCoMan() must be used
         *  afterwards!
         */
        public TransformationService() {
            this.transformerFactory = TransformerFactory.newInstance();
	}

        /**
         * Set MoCoMan object.
         */
        public void setMoCoMan(MoCoMan mcm) {
            this.mocoman = mcm;
        }

	/* (non-Javadoc)
	 * @see javax.xml.transform.URIResolver#resolve(java.lang.String, java.lang.String)
	 */
	public Source resolve(String href, String base) throws TransformerException {
		if (href.startsWith(hogURIprefix)) {
			String[] s = new String[5];
			s = href.split("/", 5);
			if (s.length<5) {
				throw new TransformerException("URI " + href +
					    " not wellformed. Use syntax: " + hogURIprefix + "sid/acid/aid");
			}
			if (s[2].equals("") || s[3].equals("") || s[4].equals("")) {
				throw new TransformerException("URI " + href +
				    " not wellformed. Parsed sid=" + s[2] +
					", acid=" + s[3] + ", aid=" + s[4] + ".");
			}
                        log.debug("URI resolver: getAnnotation(hog://" + s[2] + "/" +  s[3] +"/"+ s[4] +")");
			String sourceString = this.mocoman.getAnnotation(s[2].trim(), s[3].trim(), s[4].trim());
			if (sourceString==null) {
				throw new TransformerException("URI " + href +
					    " not wellformed. Parsed sid=" + s[2] +
						", acid=" + s[3] + ", aid=" + s[4] + ".");
			}

			return new StreamSource(new StringReader(sourceString));

		} else
			return null;
	}


        /** Return the TransformerFactory of this instance
         * @return the TransformerFactory
         */
        public TransformerFactory getTransformerFactory() {
            return this.transformerFactory;
        }

	/**
	 * Create a Transformer object from a stylesheet specified as Source.
	 * @param sheet a stylesheet specified as Source object
	 * @return a new Transformer object
	 */
	public Transformer createTransformer(Source sheet)
               throws TransformerConfigurationException {
		Transformer transformer = null;
                transformer = this.transformerFactory.newTransformer(sheet);
                transformer.setURIResolver(this);
		return transformer;
	}

	/**
	 * Create an empty Transformer object (e.g. for ID transformation DOM to String etc.)
	 * @return a new Transformer object
	 */
	public Transformer createTransformer()
               throws TransformerConfigurationException {
		Transformer transformer = null;
                transformer = this.transformerFactory.newTransformer();
		return transformer;
	}

	/**
	 * Create a Transformer object from a stylesheet specified as filename String.
	 * @param stylesheetfilename a stylesheet specified as filename String
	 * @return a new Transformer object
	 */
	public Transformer createTransformer(String stylesheetfilename)
               throws TransformerConfigurationException{
		StreamSource sheet = new StreamSource(new File(stylesheetfilename));
		return this.createTransformer(sheet);
	}

	/**
	 * Create a Transformer object from a stylesheet specified as filename String.
	 * @param stylesheet a stylesheet specified as XML String
	 * @return a new Transformer object
	 */
	public Transformer createTransformerFromString(String stylesheet)
               throws TransformerConfigurationException {
		StreamSource sheet = new StreamSource(new StringReader(stylesheet));
		return this.createTransformer(sheet);
	}

	/**
	 * Create a Transformer object from a stylesheet specified as File.
	 * @param stylesheetfile a stylesheet specified as File
	 * @return a new Transformer object
	 */
	public Transformer createTransformer(File stylesheetfile)
               throws TransformerConfigurationException {
		StreamSource sheet = new StreamSource(stylesheetfile);
		return this.createTransformer(sheet);
	}

	/**
	 * Transform the XML Source using the specified Transformer.
	 * @param transformer the Transformer object
	 * @param source the XML Source object to be transformed
	 * @return the transformation result as String
	 */
	public String transform(Transformer transformer, Source source) throws TransformerException {
	    StringWriter resultStringWriter = new StringWriter();
	    StreamResult result = new StreamResult(resultStringWriter);
            transformer.transform(source, result);
            return resultStringWriter.toString();
	}

	/**
	 * Transform the XML String using the specified Transformer.
	 * @param transformer the Transformer object
	 * @param sourceString the XML Source as String
	 * @return the transformation result as String
	 */
	public String transform(Transformer transformer, String sourceString) throws TransformerException {
            StreamSource source = new StreamSource(new StringReader(sourceString));
            return this.transform(transformer, source);
	}

	/**
	 * Transform the DOM node using the specified Transformer.
	 * @param transformer the Transformer object
	 * @param node the XML document to be transformed
	 * @return the transformation result as DOM node
	 */
	public Node transform(Transformer transformer, Node node) throws TransformerException {
            DOMSource source = new DOMSource(node);
	    DOMResult result = new DOMResult();
            transformer.transform(source, result);
            return result.getNode();
	}

	/**
	 * Transform the XML Source using the specified Transformer.
	 * @param transformer the Transformer object
	 * @param source the XML Source object to be transformed
	 * @return the transformation result as String
	 */
	public String transformToString(Transformer transformer, Node node) throws TransformerException {
            DOMSource source = new DOMSource(node);
	    StringWriter resultStringWriter = new StringWriter();
	    StreamResult result = new StreamResult(resultStringWriter);
            transformer.transform(source, result);
            return resultStringWriter.toString();
	}

	/**
	 * Transform the XML Source using the specified Transformer.
	 * @param transformer the Transformer object
	 * @param source the XML Source object to be transformed
	 * @return the transformation result as DOM node
	 */
	public Node transformToDom(Transformer transformer, String sourceString) throws TransformerException {
            StreamSource source = new StreamSource(new StringReader(sourceString));
	    DOMResult result = new DOMResult();
            transformer.transform(source, result);
            return result.getNode();
	}

    

	/* test method
	public static void main(String [] args) {
		log.addAppender(new ConsoleAppender(new SimpleLayout() ));
		TransformationService ts = new TransformationService(null);
		Transformer mytrafo = ts.createTransformer("/home/cl-home/uschaefer/src/xsl/insertdoc.xsl");
		String xml = "<a><b>bla</b></a>";
	    System.out.println(ts.transform(mytrafo, xml));
	}
	*/
}

