/*
 * XmlRpcApplication.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.util.Properties;
import java.util.Vector;
import java.io.IOException;
import java.net.MalformedURLException;
import org.apache.xmlrpc.XmlRpcClient;
import de.dfki.lt.hog.util.Utility;

/**
 * An <code>XmlRpcApplication</code> provides methods and fields for
 * an application which uses the XML-RPC protocol to communicate with
 * MoCoMan.
 * These methods and fields are, among others, an XmlRpcClient, an URL
 * to MoCoMan's XmlRpcServer, a handler name and handler method names.
 * For all these fields, there are <code>get</code> and <code>set</code>
 * methods to change and retrieve each of them.
 *
 * @version 2003-12-03
 * @author <a href="mailto:robba@dfki.de">Robert Barbey</a>
 */

public abstract class XmlRpcApplication
    extends Application
    implements XmlRpcClientInterface
{
    /**
     * Key for server URL in the properties of this
     * <code>XmlRpcApplications</code>. */
    public static final String XMLRPC_SERVER_URL = "xmlrpc.server.url";

    /**
     * Key for server port in the properties of this
     * <code>XmlRpcApplications</code>. */
    public static final String XMLRPC_SERVER_PORT = "xmlrpc.server.port";

    /**
     * Key for a handler for XML-RPC calls in the properties of this
     * <code>XmlRpcApplications</code>. */
    public static final String XMLRPC_HANDLER_NAME =
	"xmlrpc.handler.name";

    /**
     * Key for a handler method for XML-RPC calls in the properties of
     * this <code>XmlRpcApplications</code>. */
    public static final String XMLRPC_HANDLER_METHOD =
	"xmlrpc.handler.method";

    /**
     * Client for XML-RPC calls. */
    private XmlRpcClient xmlRpcApplicationClient;

    /**
     * Contains server URL as <code>String</code> */
    private String xmlRpcServerUrl;

    /**
     * Contains server port as <code>int</code> */
    private int xmlRpcServerPort;

    /**
     * Contains name of the handler for XML-RPC calls as
     * <code>String</code> */
    private String xmlRpcHandlerName;

    /**
     * Contains all method names for possible XML-RPC calls. */
    private String[] xmlRpcHandlerMethods;

    /**
     * */
    /*public XmlRpcApplication(Properties xmlRpcApplProps)
    {
	super(xmlRpcApplProps);
	//setConfig(xmlRpcApplProps);
	init(xmlRpcApplProps);
	}*/

    public void init(Properties xmlRpcApplProps)
    {
	try {
	    super.init(xmlRpcApplProps);
	} catch (IOException ioe) {
	    applicationLog.error(Utility.exception2String(ioe));
	}

	setXmlRpcServerUrl(getConfig().getProperty(XMLRPC_SERVER_URL));

	setXmlRpcHandlerName(getConfig().
			     getProperty(XMLRPC_HANDLER_NAME));

	try {
	    setXmlRpcApplicationClient(new XmlRpcClient(getXmlRpcServerUrl()));
	} catch (MalformedURLException mue) {
	    applicationLog.error(Utility.exception2String(mue));
	}
    }

    /**
     * Returns the XML-RPC client of this
     * <code>XmlRpcApplications</code>.
     *
     * @return an XML-RPC client
     */
    public XmlRpcClient getXmlRpcApplicationClient()
    {
	return this.xmlRpcApplicationClient;
    }

    /**
     * Sets an XML-RPC client for this
     * <code>XmlRpcApplications</code>.
     *
     * @param aClient a new XML-RPC client for this
     *   <code>XmlRpcApplications</code>.
     */
    public void setXmlRpcApplicationClient(XmlRpcClient aClient)
    {
	this.xmlRpcApplicationClient = aClient;
    }

    /**
     * Returns the URL of the server which answers XML-RPC calls.
     *
     * @return server URL
     */
    protected String getXmlRpcServerUrl()
    {
	return this.xmlRpcServerUrl;
    }

    /**
     * Sets an URL to a server for XML-RPC calls
     *
     * @param aUrl a new URL as <code>String</code>
     */
    private void setXmlRpcServerUrl(String aUrl)
    {
	this.xmlRpcServerUrl = aUrl;
    }

    /**
     * Returns the port of the server which answers XML-RPC calls.
     *
     * @return server port
     */
    public int getXmlRpcServerPort()
    {
	return this.xmlRpcServerPort;
    }

    /**
     * Sets a port to a server for XML-RPC calls
     *
     * @param aPort a new port
     */
    public void setXmlRpcServerPort(int aPort)
    {
	this.xmlRpcServerPort = aPort;
    }

    /**
     * Returns the name of the handler of XML-RPC calls.
     *
     * @return a handler name for XML-RPC
     */
    public String getXmlRpcHandlerName()
    {
	return this.xmlRpcHandlerName;
    }

    /**
     * Sets a new handler name for XML-RPC calls.
     *
     * @param aHandlerName a new handler name
     */
    private void setXmlRpcHandlerName(String aHandlerName)
    {
	this.xmlRpcHandlerName = aHandlerName;
    }

    /**
     * Returns an array of names of possible methods for
     * XML-RPC calls.
     *
     * @return names of methods for XML-RPC
     */
    public String[] getXmlRpcHandlerMethods()
    {
	return this.xmlRpcHandlerMethods;
    }

    public void shutdown(String aModuleName)
    {
    }

    public String process(String aMethodName,
			  Vector params)
    {
	return null;
    }

    public void start(String aModuleName)
    {
    }
}
