/*
 * XmlRpcModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog;

import java.util.*;
import java.net.*;
import java.io.*;
import org.w3c.dom.*;
import org.apache.xmlrpc.*;
import org.apache.log4j.Logger;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;

/**
 * Encapsulates all methods needed to communicate via XML-RPC
 * with a remote module. This module is called a <tt>Handler</tt> in
 * the <a href="http://ws.apache.org./xmlrpc">XML-RPC API</a> used,
 * so, this convention is applied here. A handler must have a method
 * which actually processes a request and a server for communication.
 *
 * @version 2003-11-04
 * @author <a href="mailto:robba@dfki.de">Robert Barbey</a>
 */
public abstract class XmlRpcModule
    extends Module
    implements XmlRpcClientInterface {

    /**
     * Key for XML-RPC server URL in configuration file. */
    public static final String XMLRPC_SERVER_URL = "xmlrpc.server.url";

    /**
     * Key for an object which handles an XmlRpc request */
    public static final String XMLRPC_HANDLER_NAME =
	"xmlrpc.handler.name";

    /**
     * Key for a method name of a handler. */
    public static final String XMLRPC_HANDLER_METHOD =
	"xmlrpc.handler.method";

    /**
     * A logger for <code>XmlRpcModule</code>s */
    static protected Logger xmlRpcModuleLog =
	Logger.getLogger(XmlRpcModule.class);

    /**
     * A XmlRpcClient for communication with a XmlRpcServer. */
    private XmlRpcClient xmlRpcClient;

    /**
     * Returns the <code>XmlRpcClient</code> of this
     * <code>XmlRpcModule</code>.
     *
     * @return an XmlRpc client
     */
    protected XmlRpcClient getXmlRpcClient()
    {
	return this.xmlRpcClient;
    }

    /**
     * Sets the client for this <code>XmlRpcModule</code>
     * @param aClient <code>XmlRpcClient</code> for this module
     */
    protected void setXmlRpcClient(XmlRpcClient aClient)
    {
	this.xmlRpcClient = aClient;
    }

    /**
     * URL of XML-RPC server as <code>String</code>. */
    private String xmlRpcServerUrl;

    /**
     * Returns the URL of an XmlRpc server this <code>XmlRpcModule</code>
     * communicates with.
     * @return an URL as <code>String</code>
     */
    public String getXmlRpcServerUrl()
    {
	return this.xmlRpcServerUrl;
    }

    /**
     * Sets the URL of an XmlRpc server this <code>XmlRpcModule</code>
     * communicates with.
     * @param aServerUrl an URL as <code>String</code>
     */
    protected void setXmlRpcServerUrl(String aServerUrl)
    {
	this.xmlRpcServerUrl = aServerUrl;
    }

    /**
     * Name of a class which can handle an XmlRpc request. */
    private String xmlRpcHandlerName;

    /**
     * Returns name of a class which can handle an XmlRpc request.
     * @return class name of handler
     */
    public String getXmlRpcHandlerName()
    {
	return this.xmlRpcHandlerName;
    }

    /**
     * Sets name name of a class which can handle an XmlRpc request.
     * @param aHandlerName class name
     */
    protected void setXmlRpcHandlerName(String aHandlerName)
    {
	this.xmlRpcHandlerName = aHandlerName;
    }

    /**
     * Method name to which an XmlRpc request is past by a handler. */
    private String xmlRpcHandlerMethod;

    /**
     * Returns method name to which an XmlRpc request is past by a
     * handler.
     *
     * @return method name for XmlRpc request
     */
    public String getXmlRpcHandlerMethod()
    {
	return this.xmlRpcHandlerMethod;
    }

    /**
     * Sets a method name to which an XmlRpc request is past by a handler
     * @param aHandlerMethod method name for a handler
     */
    protected void setXmlRpcHandlerMethod(String aHandlerMethod)
    {
	this.xmlRpcHandlerMethod = aHandlerMethod;
    }

    /**
     * Initializes basic field values like name, server URL and
     * XMLRPC client, by a <code>Properties</code> object.
     * First, <code>init()</code> of <code>Module</code> is called;
     * then all other field values are set by their corresponding
     * <code>set</code> methods.
     *
     * @param xmlRpcModuleConfig
     * @throws InitializationException
     * @see Module#init
     */
    protected void init(Properties xmlRpcModuleConfig)
	throws InitializationException
    {
	super.init(xmlRpcModuleConfig);

	setXmlRpcServerUrl(getConfig().getProperty(XMLRPC_SERVER_URL));
	setXmlRpcHandlerName(getConfig().
			     getProperty(XMLRPC_HANDLER_NAME));
	setXmlRpcHandlerMethod(getConfig().
			       getProperty(XMLRPC_HANDLER_METHOD));

	try {
	    setXmlRpcClient(new XmlRpcClient(getXmlRpcServerUrl()));
	} catch (MalformedURLException mue) {
	    xmlRpcModuleLog.error(Utility.exception2String(mue));
	}

    }

    public void start(String aModuleName)
    {
    }

    public String process(String aHandlerCall, Vector params)
    {
	String output = null;

	try {
	    output = (String) getXmlRpcClient().
		execute(aHandlerCall, params);
	} catch (IOException ioe) {
	    xmlRpcModuleLog.error(Utility.exception2String(ioe));
	} catch (XmlRpcException xre) {
	    xmlRpcModuleLog.error(Utility.exception2String(xre));
	}

	return output;
    }

    public void shutdown(String aModuleName)
    {
    }

    public abstract String process(String processString);

    public abstract Node processDOMTree(Node domTree);

}
