/*
 * Chunk.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.container;

import java.util.*;
import org.apache.log4j.Logger;
import de.dfki.lt.hog.util.Utility;

/**
 * Represents a chunk as segmented by Chunkie. Each chunk has daughters
 * which can be chunks themselves.
 *
 *
 * @author <a href=mailto:oeze@dfki.de>Oezguer Demir</a>
 * @version $Revision: 1.2 $
 * $Id: Chunk.java,v 1.2 2005/04/26 17:37:26 oeze Exp $
 */
public class Chunk {

  /** A logger for <code>Module</code>s */
  static protected Logger log = Logger.getLogger(Chunk.class);

  /** The type of this chunk: s(endtence)|c(constituent)|w(ord) */
  private String type;
  /** The POS or CAT of this chunk */
  private String poscat;
  /** The daughters of this chunk */
  private ArrayList daughters;
  /** The sentence ID of this chunk (only for chunks of type S) */
  private String sid;
  /** The token of this chunk (for chunks of type c) */
  private JToken token;
  /** All terminals (words) underneath this chunk */
  private ArrayList words;
  /** The XML representation of this chunk */
  private StringBuffer xmlBuffer;

  /**
   * Creates an empty chunk.
   */
  public Chunk() {
    init();
  } // constructor()

  /**
   * Creates a chunk of type "w" using the given JToken.
   * @param token
   */
  public Chunk(JToken token) {
    init();
    this.type = "w";
    this.token = token;
  }

  /**
   * Initializes all member variables.
   */
  private void init() {
    this.type = "";
    this.poscat = "";
    this.daughters = new ArrayList();
    this.sid = "";
    this.token = new JToken();
    this.words = null;
    this.xmlBuffer = null;
  } // init()

  public void setType(String type) {
    this.type = type;
  }

  public String getType() {
    return this.type;
  }

  public void setToken(JToken token) {
    this.token = token;
  }

  public JToken getToken() {
    return this.token;
  }

  public void setPoscat(String poscat) {
    this.poscat = poscat;
  }

  public String getPoscat() {
    return this.poscat;
  }

  public void setSid(String sid) {
    this.sid = sid;
  }

  public String getSid() {
    return this.sid;
  }

  public void addDaughter(Chunk daughter) {
    this.daughters.add(daughter);
  }

  /**
   * Returns a copy of the daughters of this chunk.
   * @return A copy of the daughters of this chunk.
   */
  public ArrayList getDaughters() {
    return ( (ArrayList)this.daughters.clone());
  }

  public boolean isTerminal() {
    if (this.type.equals("w")) {
      return true;
    }
    else {
      return false;
    }
  }

  public int getStart() {
    if (this.isTerminal()) {
      return this.token.getOffset();
    }
    else {
      return ( (Chunk)this.getWords().get(0)).getToken().getOffset();
    }
  }

  public int getEnd() {
    if (this.isTerminal()) {
      return (this.token.getOffset() + this.token.getLength() - 1);
    }
    else {
      ArrayList terminals = this.getWords();
      int lastIndex = terminals.size() - 1;
      Chunk lastTerminal = (Chunk) terminals.get(lastIndex);
      return lastTerminal.getEnd();
    }
  }

  public ArrayList getWords() {
    // If not already generated, generate the terminal nodes list
    if (this.words == null) {
      this.words = new ArrayList();
      this.words.add(this);
      expandChunks(this.words);
    } // if
    return this.words;
  }

  /**
   * Helps getWords() to expand this instance in order to return an
   * ArrayList of terminal node. Method runs recursively.
   *
   * @param chunks An ArrayList containing possibly only one node.
   * @return The expanded ArrayList containing terminal nodes only.
   */
  private static ArrayList expandChunks(ArrayList chunks) {
    // is any of the chunks is an NT?
    boolean stillToDo = false;
    for (int i = 0; i < chunks.size(); i++) {
      if ( ( (Chunk) chunks.get(i)).isTerminal() == false) {
        stillToDo = true;
      } // if
    } // for i

    // if not, abort the recursion
    if (stillToDo == false) {
      return chunks;
    }
    // if yes, continue recursively
    else {
      // search through the chunks
      for (int i = 0; i < chunks.size(); i++) {
        // expand the first NT encounterd by replacing it by its daughters
        if ( ( (Chunk) chunks.get(i)).isTerminal() == false) {
          Chunk chunk2expand = (Chunk) chunks.get(i);
          chunks.remove(i);
          ArrayList daughters2add = chunk2expand.getDaughters();
          for (int j = daughters2add.size() - 1; j >= 0; j--) {
            chunks.add(i, daughters2add.get(j));
          } // for j
        } // if
      } // for i
      return expandChunks(chunks);
    } // else
  } // expandChunks()

  /**
   * Returns the xml structure of this node and its daughters as a string.
   * This method is primarily for class-internal usage.
   * @param depth The current level of indendation (recursion).
   * @param xml The xml string built so far (recursion).
   */
  private void buildXMLStringBuffer(int depth, StringBuffer xml) {
    String endElement;
    StringBuffer indent = new StringBuffer();
    String newLine = System.getProperties().getProperty("line.separator", "\n");
    for (int i = 0; i < depth; i++) {
      indent.append("  ");
    }
    xml.append(indent);
    if (this.isTerminal()) {
      xml.append("<w pos=\"" + this.getPoscat() + "\"" +
                 " cstart=\"" + this.getStart() +
                 "\" cend=\"" + this.getEnd() +
                 "\">" +
                 Utility.cleanXMLString(this.token.getToken()));
      endElement = "</w>";
      indent = new StringBuffer();
    }
    else {
      if (this.type.equals("s")) {
        endElement = "</s>";
        xml.append("<s id=\"" + this.sid + "\"" + " cstart=\"" +
                   this.getStart() + "\" cend=\"" +
                   this.getEnd() +
                   "\">" + newLine);
      }
      else {
        endElement = "</chunk>";
        xml.append("<chunk cat=\"" + ( (Chunk)this).getPoscat() + "\"" +
                   " cstart=\"" + this.getStart() +
                   "\" cend=\"" + this.getEnd() +
                   "\">" +
                   newLine);
      } // if
      ArrayList daughters = ( (Chunk)this).getDaughters();
      for (int i = 0; i < daughters.size(); i++) {
        Chunk currentDaughter = (Chunk) daughters.get(i);
        currentDaughter.buildXMLStringBuffer(depth + 1, xml);
      } // for i
    } // else
    xml.append(indent + endElement + newLine);
  } // buildXMLString()

  public String toXML() {
    this.xmlBuffer = new StringBuffer();
    this.buildXMLStringBuffer(0, this.xmlBuffer);
    return this.xmlBuffer.toString();
  }

  public String toString() {
    StringBuffer out = new StringBuffer();
    out.append("Type: " + this.type + "\n");
    out.append("Poscat: " + this.poscat + "\n");
    out.append("Sid: " + this.sid + "\n");
    out.append("Token:\n" + this.token.toString());
    out.append("Daughters: ");
    for (int i = 0; i < this.daughters.size(); i++) {
      out.append( ( (Chunk) daughters.get(i)).getType() + " ");
    }
    return out.toString();
  }
} // class
