/*
 * GraphNode.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.container;

import java.util.*;
import org.apache.log4j.Logger;
import de.dfki.lt.hog.util.Utility;

/**
 * Represents a graph node as in syntactic parse trees. Each graph node has
 * daughters which can be graph nodes themselves.
 *
 * @author <a href=mailto:oeze@dfki.de>Oezguer Demir</a>
 * @version $Revision: 1.1 $
 * $Id: GraphNode.java,v 1.1 2005/09/27 17:09:35 oeze Exp $
 */
public class GraphNode {

  /** A logger for <code>Module</code>s */
  static protected Logger log = Logger.getLogger(GraphNode.class);

  /** The type of this node: s(endtence)|nt(non-terminal)|t(terminal) */
  private String type;
  /** The POS or CAT of this node */
  private String poscat;
  /** The label of the edge leading to the mother of this node */
  private String label;
  /** The daughters of this node */
  private ArrayList daughters;
  /** The ID of this node */
  private String id;
  /** The token of this node (for nodes of type nt) */
  private JToken jtoken;
  /** All terminals (words) underneath this node */
  private ArrayList words;
  /** The XML representation of this node */
  private StringBuffer xmlBuffer;
  /** The buffer used to build an ascii representation of the syntax tree */
  private StringBuffer treeStringBuffer;

  /**
   * Creates an empty node.
   */
  public GraphNode() {
    init();
  } // constructor()

  /**
   * Creates a node of type "w" using the given JToken.
   * @param token
   */
  public GraphNode(JToken token) {
    init();
    this.type = "t";
    this.jtoken = token;
  }

  /**
   * Initializes all member variables.
   */
  private void init() {
    this.type = "";
    this.poscat = "";
    this.daughters = new ArrayList();
    this.id = "";
    this.label = "";
    this.jtoken = new JToken();
    this.words = null;
    this.xmlBuffer = null;
  } // init()

  public void setType(String type) {
    this.type = type;
  }

  public String getType() {
    return this.type;
  }

  public void setJToken(JToken token) {
    this.jtoken = token;
  }

  public JToken getJToken() {
    return this.jtoken;
  }

  public void setPoscat(String poscat) {
    this.poscat = poscat;
  }

  public String getPoscat() {
    return this.poscat;
  }

  public void setLabel(String label) {
    this.label = label;
  }

  public String getLabel() {
    return this.label;
  }

  public void setId(String id) {
    this.id = id;
  }

  public String getId() {
    return this.id;
  }

  public void addDaughter(GraphNode daughter) {
    this.daughters.add(daughter);
  }

  /**
   * Returns a copy of the daughters of this node.
   * @return A copy of the daughters of this node.
   */
  public ArrayList getDaughters() {
    return ((ArrayList)this.daughters.clone());
  }

  public boolean isTerminal() {
    if (this.type.equals("t")) {
      return true;
    } else {
      return false;
    }
  }

  public int getStart() {
    if (this.isTerminal()) {
      return this.jtoken.getOffset();
    } else {
      return ((GraphNode)this.getWords().get(0)).getJToken().getOffset();
    }
  }

  public int getEnd() {
    if (this.isTerminal()) {
      return (this.jtoken.getOffset() + this.jtoken.getLength() - 1);
    } else {
      ArrayList terminals = this.getWords();
      int lastIndex = terminals.size() - 1;
      GraphNode lastTerminal = (GraphNode) terminals.get(lastIndex);
      return lastTerminal.getEnd();
    }
  }

  public ArrayList getWords() {
    // If not already generated, generate the terminal nodes list
    if (this.words == null) {
      this.words = new ArrayList();
      this.words.add(this);
      expandGraphNodes(this.words);
    } // if
    return this.words;
  }

  /**
   * Helps getWords() to expand this instance in order to return an
   * ArrayList of terminal node. Method runs recursively.
   *
   * @param nodes An ArrayList containing possibly only one node.
   * @return The expanded ArrayList containing terminal nodes only.
   */
  private static ArrayList expandGraphNodes(ArrayList nodes) {
    // is any of the nodes is an NT?
    boolean stillToDo = false;
    for (int i = 0; i < nodes.size(); i++) {
      if (((GraphNode) nodes.get(i)).isTerminal() == false) {
        stillToDo = true;
      } // if
    } // for i

    // if not, abort the recursion
    if (stillToDo == false) {
      return nodes;
    }
    // if yes, continue recursively
    else {
      // search through the nodes
      for (int i = 0; i < nodes.size(); i++) {
        // expand the first NT encounterd by replacing it by its daughters
        if (((GraphNode) nodes.get(i)).isTerminal() == false) {
          GraphNode node2expand = (GraphNode) nodes.get(i);
          nodes.remove(i);
          ArrayList daughters2add = node2expand.getDaughters();
          for (int j = daughters2add.size() - 1; j >= 0; j--) {
            nodes.add(i, daughters2add.get(j));
          } // for j
        } // if
      } // for i
      return expandGraphNodes(nodes);
    } // else
  } // expandGraphNodes()

  /**
   * Returns the xml structure of this node and its daughters as a string.
   * This method is primarily for class-internal usage.
   * @param depth The current level of indendation (recursion).
   * @param xml The xml string built so far (recursion).
   */
  private void buildXMLStringBuffer(int depth, StringBuffer xml) {
    String endElement;
    StringBuffer indent = new StringBuffer();
    String lbElement = "";
    String newLine = System.getProperties().getProperty("line.separator", "\n");
    for (int i = 0; i < depth; i++) {
      indent.append("  ");
    }
    xml.append(indent);
    if (!this.getLabel().equals("")) {
      lbElement = " lb=\"" + this.getLabel() + "\"";
    } else {
      lbElement = "";
    }
    if (this.isTerminal()) {
      xml.append("<node cat=\"" + this.getPoscat() + "\"" +
                 lbElement +
                 " id=\"" + this.getId() + "\"" +
                 " cstart=\"" + this.getStart() +
                 "\" cend=\"" + this.getEnd() +
                 "\">" +
                 Utility.cleanXMLString(this.jtoken.getToken()));
      endElement = "</node>";
      indent = new StringBuffer();
    } else {
      // if (this.type.equals("s")) {
      //   endElement = "</s>";
      //   xml.append("<s id=\"" + this.sid + "\"" + " cstart=\"" +
      //              this.getStart() + "\" cend=\"" +
      //              this.getEnd() +
      //              "\">" + newLine);
      // }
      // else {
      endElement = "</node>";
      xml.append("<node cat=\"" + ((GraphNode)this).getPoscat() + "\"" +
                 lbElement +
                 " id=\"" + this.getId() + "\"" +
                 " cstart=\"" + this.getStart() +
                 "\" cend=\"" + this.getEnd() +
                 "\">" +
                 newLine);
      // } // else
      ArrayList daughters = ((GraphNode)this).getDaughters();
      for (int i = 0; i < daughters.size(); i++) {
        GraphNode currentDaughter = (GraphNode) daughters.get(i);
        currentDaughter.buildXMLStringBuffer(depth + 1, xml);
      } // for i
    } // else
    xml.append(indent + endElement + newLine);
  } // buildXMLString()

  public String toXML() {
    this.xmlBuffer = new StringBuffer();
    this.buildXMLStringBuffer(0, this.xmlBuffer);
    return this.xmlBuffer.toString();
  }

  public String toString() {
    StringBuffer out = new StringBuffer();
    out.append("Type: " + this.type + "\n");
    out.append("Poscat: " + this.poscat + "\n");
    out.append("Id: " + this.id + "\n");
    out.append("Token:\n" + this.jtoken.toString());
    out.append("\nDaughters: ");
    for (int i = 0; i < this.daughters.size(); i++) {
      out.append(((GraphNode) daughters.get(i)).getType() + " ");
    }
    return out.toString();
  }

  /**
   * Returns the tree structure of this GraphNode as a string.
   * @return The tree structure of this GraphNode as a string.
   */
  public String toTreeString() {
    this.treeStringBuffer = new StringBuffer();
    this.buildTreeStringBuffer(0, this.treeStringBuffer);
    return this.treeStringBuffer.toString();
  }

  private void buildTreeStringBuffer(int depth, StringBuffer tree) {
    for (int i = 0; i < depth; i++) {
      tree.append("|   ");
    }
    tree.append("|--" + "[edge]" + "--");
    if (this.type.equals("t")) {
      tree.append("(" + this.getPoscat() + ") "
                  + "\"" + this.getJToken().getToken() + "\""
                  + System.getProperties().getProperty("line.separator", "\n"));
    } else {
      tree.append("[" + this.getPoscat() + "] "
                  + System.getProperties().getProperty("line.separator", "\n"));
      ArrayList daughters = this.getDaughters();
      for (int i = 0; i < daughters.size(); i++) {
        GraphNode currentDaughter = (GraphNode) daughters.get(i);
        currentDaughter.buildTreeStringBuffer(depth + 1, tree);
      } // for i
    } // else
  } // buildTreeString()

} // class
