/*
 * XmlStringBuffer.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.container;

/**
 * <code>XmlStringBuffer</code> creates an XML string with element and
 * attribute tags. Internally a <code>StringBuffer</code> is used for
 * connecting elements and attributes. It only creates XML strings of
 * the kind illustrated by the example.
 * <p> Example: <br>
 * <code>XmlStringBuffer xml = new XmlStringBuffer(); <br>
 * xml.addElement("METADATA/");<br>
 * xml.addAttribute("created", "2003-09-23");<br>
 * xml.addAttribute("user", "robba");<br>
 * xml.addComment("this is a comment");<br>
 * xml.addElement("W", "Fahrrad");<br>
 * xml.addAttribute("id", "W5029");<br>
 * xml.addAttribute("pos", "NOUN");<br>
 * System.out.println(xml.toString());</code>
 * <p>produces <br>
 * <code>&lt;TOKENS&gt;&lt;METADATA created="2003-09-23" user="robba"/&gt;&lt;!-- this is a comment --&gt;&lt;W id="W5029" pos="NOUN"&gt;Fahhrad&lt;/W&gt;&lt;/TOKENS&gt;</code>;
 *
 * @author <a href="mailto:robba@dfki.de">Robert Barbey</a>
 * @version 2003-09-25
 * @see java.lang.StringBuffer
 */

public class XmlStringBuffer {

    /**
     * Contains XML string. */
    private StringBuffer xmlBuffer;

    /**
     * Contains current XML element tag. */
    private String currentElement;

    /**
     * Contains last inserted attribute for current element tag.*/
    private String currentValue;

    /**
     * <code>true</code> if current XML tag is of kind <tt><META/></tt>. */
    private boolean isEmptyElement;

    /**
     * Contains current position for inserting into <code>xmlBuffer</code>. */
    private int insertPosition;

    /**
     * Creates <code>XmlStringBuffer</code> with default XML version
     * 1.0 and encoding UTF-8. */
    public XmlStringBuffer() {
			this("1.0", "utf-8");
    }

    /**
     * Creates <code>XmlStringBuffer</code> with a given
     * version and enconding. */
    public XmlStringBuffer(String version, String encoding) {
			String header = "<?xml version=\"" + version + "\" encoding=\"" + encoding + "\"?>";
			xmlBuffer = new StringBuffer(header);
			currentElement = "";
			currentValue = "";
			isEmptyElement = false;
			insertPosition = header.length();
			this.addElement("TOKENS");
		}

    /**
     * Adds a new XML element tag without a value. An empty element is created, if
     * <code>tag</code> ends with '/'. */
    public XmlStringBuffer addElement(String tag) {
			return this.addElement(tag, null);
    }

    /**
     * Adds new XML element tag. An empty element is created, if
     * <code>tag</code> ends with '/'. */
    public XmlStringBuffer addElement(String tag, String value) {
			if (value == null) value = "";

			if (!currentElement.equals("") && !currentElement.equals("!COMMENT!") && !currentElement.equals("TOKENS") && !isEmptyElement)
			    insertPosition += currentValue.length() + currentElement.length()+3;

			if (tag.charAt(tag.length()-1) == '/') {
	    	String temp = "<" + tag + ">";
	    	xmlBuffer.insert(insertPosition, temp);
	    	currentElement = tag;
				currentValue = value;
	    	insertPosition += tag.length()+2;
	    	isEmptyElement = true;

	    	return this;
			}

			String temp = "<" + tag + ">" + value + "</" + tag + ">";
			xmlBuffer.insert(insertPosition, temp);
			currentElement = tag;
			currentValue = value;
			insertPosition += tag.length() + 2;
			isEmptyElement = false;

			return this;
    }

    /**
     * Adds attributes for current XML element tag. */
    public XmlStringBuffer addAttribute(String attr, String value) {
			if (currentElement.equals("!COMMENT!")) {
					System.err.println("No element tag for adding this attribute!");
					System.err.println("Add element tag first!");

					return this;
			}

			if (isEmptyElement) insertPosition -= 2;
			else --insertPosition;

			String temp = " " + attr + "=\"" + value + "\"";
			xmlBuffer.insert(insertPosition, temp);

			if (isEmptyElement) insertPosition += temp.length()+2;
			else insertPosition += temp.length()+1;

			return this;
    }

    public XmlStringBuffer addAttribute(String tag, int value) {
			return this.addAttribute(tag, String.valueOf(value));
    }

		/**
			* Adds an XML Comment */
		public XmlStringBuffer addComment(String comment) {
			String temp = "<!-- " + comment + " -->";
			xmlBuffer.insert(insertPosition, temp);
			insertPosition += temp.length();
			currentElement = "!COMMENT!";
			currentValue = "";

			return this;
		}

    /**
     * Returns this <code>XmlStringBuffer</code> as a <code>String</code>.
     * @return <code>XmlStringBuffer</code> as a <code>String</code>
     */
    public String toString() {
			return xmlBuffer.toString();
    }
}
