/*
 * AnnotationDatabase.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.database;
import org.w3c.dom.Node;
//import org.xmldb.api.base.ResourceIterator;

/**
 * <p>This is an abstract class for a Deepthought annotation database
 * independent of a specific underlying database system.
 * This class uses concepts like Session and Document (identified via Strings)
 * that are used to structure the Deepthought annotation database.
 * In this Database, a session comprises a collection of a documents.
 * A document itself is a collection (!) of XML annotations that are
 * produced from different modules and can be accessed via annotation
 * layer names.</p>
 *
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @version 2003-09-18
 */
public abstract class AnnotationDatabase {


    private String sDatabaseLocation = null;

	/** Default Constructor:
	 *  Creates an Annotation database object. Normally, the database is initialized
	 *  using the setDatabaseLocation() method or the alternative constructor
	 *  with the same signature.
	 */
	public AnnotationDatabase() {
		this.setDatabaseLocation(null);
	}

	/** Constructor with database location prefix as parameter.
	 * @param databaseLocation a String identifying the location of the database, e.g. URI
	 */
	public AnnotationDatabase(String databaseLocation) {
		this.setDatabaseLocation (databaseLocation);
	}

	/** Sets the database location prefix.
	 * @param databaseLocation a String identifying the location of the database, e.g. URI.
	 */
	public void setDatabaseLocation(String databaseLocation) {
		this.sDatabaseLocation = databaseLocation;
	}

	/** Returns the database location prefix.
	 * @return a String identifying the location of the database, e.g. URI, as set by setDatabaseLocation().
	 */
	protected String getDatabaseLocation() {
		return this.sDatabaseLocation;
	}

	/** Opens the annotation database.
	 * @param connectString optional connection String for the underlying database
	 * @return true iff successful
	 */

	public abstract boolean openDatabase(String connectString) throws Exception;

	/** Closes the annotation database .
	 */
	public abstract void closeDatabase() throws Exception;

	/** Adds a new session to the annotation database.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @return true iff successful
	 */
	public abstract boolean addSession(String sessionID) throws Exception;

	/** Adds a new session to the annotation database.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @param metaInfo metainformation associated with the session (XML String)
	 * @return true iff successful
	 */
	public abstract boolean addSession(String sessionID, String metaInfo) throws Exception;

	/** Adds a new session to the annotation database.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @param metaInfo metainformation associated with the session (DOM object)
	 * @return true iff successful
	 */
	public abstract boolean addSession(String sessionID, Node metaInfo);

	/** Deletes a session from the annotation database.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @return true iff successful
	 */
	public abstract boolean deleteSession(String sessionID);

	/** Adds a new, empty annotation collection for a document to the annotation database.
       	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
       	 * @param docName identifier of the added document (must be unique within the same session collection)
       	 * @return true iff successful
       	 */
	public abstract boolean addDocument(String sessionID, String docName);

	/** Deletes an annotation collection for a document from the annotation database.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @param docName identifier of the document to be deleted
	 * @return true iff successful
	 */
	public abstract boolean deleteDocument(String sessionID, String docName);

	/** Adds XML annotation (as String) for the specified document.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @param docName identifier of the document for which annotation is to be added
	 * @param layerName name of the layer type (such as Tokens, NamedEntities etc.) of the annotation
	 * @param annotation the XML annotation as String
	 * @return true iff successful
	 */
	public abstract boolean addAnnotation (String sessionID, String docName, String layerName, String annotation);

	/** Adds XML annotation (as DOM tree) for specified document.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @param docName identifier of the document for which annotation is to be added
	 * @param layerName name of the layer type (such as Tokens, NamedEntities etc.) of the annotation
	 * @param domTree the XML annotation as DOM object
	 * @return true iff successful
	 */
	public abstract boolean addAnnotation (String sessionID, String docName, String layerName, Node domTree);

	/** Returns the annotation for the specified document and layer as String.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @param docName identifier of the document for which annotation is to be added
	 * @param Layer name of the layer type (such as Tokens, NamedEntities etc.) of the annotation
	 * @return annotation as string
	 */
	public abstract String getAnnotationAsString (String sessionID, String docName, String Layer);

	/** Returns the annotation for the specified document and layer as DOM tree.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @param docName  identifier of the document for which annotation is to be retrieved
	 * @param Layer name of the layer type (such as Tokens, NamedEntities etc.) of the annotation
	 * @return annotation as DOM object
	 */
	public abstract Node getAnnotationAsDOM (String sessionID, String docName, String Layer);

	/** Returns the result set (Iterator) of an XPath query. For details and syntax of
	 * XPath see <a href="http://w3c.org/TR/xpath.html">http://w3c.org/TR/xpath.html</a>
	 * and <a href="http://xml.apache.org/xindice">http://xml.apache.org/xindice</a>.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @param docName  identifier of the document for which annotation is to be retrieved
	 * @param query XPath query string
	 * @param namespaces space-separated list of namespaces and their URIs
	 * @return org.xmldb.api.base.ResourceIterator over the results
	 */
//	public abstract ResourceIterator XPathQuery (String sessionID, String docName, String query, String namespaces);

	/**  Executes an XUpdate `query'. For details and syntax of
	 * XUpdate see <a href="http://xmldb.org/xupdate">http://xmldb.org/xupdate</a>
	 * and <a href="http://xml.apache.org/xindice">http://xml.apache.org/xindice</a>.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @param query the XUpdate query (XML syntax) as String
	 * @return number of updated entries
	 */
	public abstract long XUpdateQuery (String sessionID, String query);

	/** Adds an index to the annotation database.
	 * @param sessionID the session Identifier (single token starting with a non-numerical character)
	 * @param indexID name of the index (must be unique within a collection)
	 * @param indexPattern pattern for the indexer to be generated
	 * @return true iff successful
	 */
	protected abstract boolean addIndex(String sessionID, String indexID, String indexPattern);
}
