/*
 * XMLDBAccess.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.database;

/**
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @version 2003-09-18
 */

import org.xmldb.api.base.Collection;
import org.xmldb.api.base.Database;
import org.xmldb.api.base.XMLDBException;
import org.xmldb.api.base.Resource;
import org.xmldb.api.base.ResourceIterator;
import org.xmldb.api.base.ResourceSet;
import org.xmldb.api.modules.XMLResource;
import org.xmldb.api.DatabaseManager;
import org.xmldb.api.modules.XUpdateQueryService;
import org.xmldb.api.modules.XPathQueryService;
import org.apache.xindice.util.XindiceException;
import org.apache.xindice.client.xmldb.services.CollectionManager;
import org.apache.xindice.xml.dom.DOMParser;
import org.apache.xindice.xml.TextWriter;
import org.apache.xindice.xml.dom.DocumentImpl;
import org.w3c.dom.*;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.StringTokenizer;
import org.apache.log4j.*;

/** The `low-level' XML:DB interface for XML annotation.
 *   <p>This database interface should not be directly accessed
 *   from DeepThought architecture and applications.
 *   XMLDBAnnotationDatabase should be used for this
 *   purpose and is itself based on XMLDBAccess.
 *   Only a few classes mentioned in this file necessarily are
 *   Xindice-specific. The rest is XML:DB standard.</p>
 *
 *   <p>Tested with Xindice 1.1b2rc as of 2003-07-25 (CVS checkout)
 *   and tomcat-4.1.24-LE-jdk14 on SUNJDK 1.4.2 linux.</p>
 *
 *   <p>Annotation is grouped in collections that can consist of
 *   (hierarchically structured) collections or XML documents.
 *   XML documents can be stored or retrieved as XMLResource s.
 *   and from/to there as texts (String/Files), DOM tree or SAX events.</p>
 *
 *   <p>XPath queries return a ResourceIterator which can be used to
 *   iterate over a result list of XMLResource objects (either XML
 *   as text Strings or converted to DOM trees using the
 *   org.xmldb.api.modules.XMLResource.getContentAsDOM() method).</p>
 *
 *  <p> In general, conversion between XML Strings and DOM (SAX) can be done
 *   using the org.apache.xindice.xml.dom.DOMParser.toDocument(String)
 *   and org.xmldb.api.modules.XMLResource.{g,s}etContentAs{DOM,SAX} methods.</p>
 *
 *   <p>This code is partially based on Xindice command and examples classes sources
 *   included in the XIndice source distribution.</p>
 *
 *  @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 *  @see <a href="http://www.xmldb.org">http://www.xmldb.org</a>
 *  and <a href="http://xml.apache.org/xindice">http://xml.apache.org/xindice</a>.
 *
 *
 */

public class XMLDBAccess {

	 static Logger logger = Logger.getLogger(XMLDBAccess.class);

    // in order to avoid reinitialization of the Database manager each time a command is executed...
	 private static boolean alreadyInitialized = false;

	// Class name of the Standard Xindice Indexer
	 public static final String XINDICE_NAME_INDEXER="org.apache.xindice.core.indexer.NameIndexer";

	 // Class name of the Standard Xindice Indexer
	 public static final String XINDICE_VAL_INDEXER="org.apache.xindice.core.indexer.ValueIndexer";

	 // Version of the XML:DB API that we are using
	 public static final String XMLDBAPIVERSION="1.0";

    /**
	 * Adds a new collection as child to an existing collection
	 * @param collectionURI The URI of the parent collection as String
	 * @param newCollection The name of the new collection as String
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws XMLDBException
	 * @throws XindiceException
    */
     public static void addCollection (String collectionURI, String newCollection)
		throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			XMLDBException, XindiceException {
				Collection collection = null;
				try {
					collection = getCollection(collectionURI);
					CollectionManager service = (CollectionManager) collection.getService("CollectionManager", XMLDBAPIVERSION);
					try {
						service.dropCollection(newCollection);
						logger.info("Dropped existing collection with name: " + newCollection);
					}
					catch (Exception e) {
						; // nothing, this may be the first pass.
					}
					// Build up the Collection XML configuration.
					String collectionConfig =
						"<collection compressed=\"true\" name=\""
							+ newCollection
							+ "\">"
							+ "   <filer class=\"org.apache.xindice.core.filer.BTreeFiler\" gzip=\"true\"/>"
							+ "</collection>";
					service.createCollection(newCollection, DOMParser.toDocument(collectionConfig));
					logger.info("Collection " + newCollection + " created.");
				}
				catch (XMLDBException e) {
					logger.error("XML:DB Exception occured " + e.errorCode + " " + e.getMessage());
				}
				finally {
					if (collection != null) {
						collection.close();
					}
				}
	}

	/**
		* Adds an XML document specified through a filename as child to an existing collection
		* @param collectionURI The URI of the parent collection as String
		* @param filename The name of the document to be added as String
		* @return String the Identifier (name) of the added document
		* @throws ClassNotFoundException
 		* @throws InstantiationException
		* @throws IllegalAccessException
 		* @throws XMLDBException
 		* @throws XindiceException
 		* @throws IOException
	   */
    public static String addDocumentFromFile (String collectionURI, String filename)
		throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			XMLDBException, XindiceException, IOException {
	  			Collection collection = null;
	  			String id = null;
				try {
					collection = getCollection(collectionURI); // e.g. "xmldb:xindice://localhost:8080/db/addressbook"
					File file = new File(filename);
					String data = readFileFromDisk(file);
					XMLResource document = (XMLResource) collection.createResource(file.getName(), "XMLResource");
					document.setContent(data);  // setContentAsDOM?
					collection.storeResource(document);
					id = document.getId();
					logger.info("Document " + filename + " inserted as " + id);
				}
				catch (XMLDBException e) {
					logger.error( "XML:DB Exception occured " + e.errorCode + " " + e.getMessage());
				}
				finally {
					if (collection != null) {
						collection.close();
					}
				}
				return id;
    }

	/**
			* Adds an XML document specified as String with an additional document
			*  identifier (name) as child to an existing collection
			* @param collectionURI The URI of the parent collection as String
			* @param XMLData The XML data of the document as String
			* @return the Identifier (name) of the added document
			* @throws ClassNotFoundException
     		* @throws InstantiationException
    		* @throws IllegalAccessException
     		* @throws XMLDBException
     		* @throws XindiceException
     		* @throws IOException
		   */
	public static String addDocumentFromString (String collectionURI, String XMLData, String id)
		  throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			  XMLDBException, XindiceException, IOException {
				  Collection collection = null;
				  try {
					  collection = getCollection(collectionURI); // e.g. "xmldb:xindice://localhost:8080/db/addressbook"
					  XMLResource document = (XMLResource) collection.createResource(id, "XMLResource");
					  document.setContent(XMLData);
					  collection.storeResource(document);
					  id = document.getId();
					  logger.info("Document inserted as " + id);
				  }
				  catch (XMLDBException e) {
					  logger.error( "XML:DB Exception occured " + e.errorCode + " " + e.getMessage());
				  }
				  finally {
					  if (collection != null) {
						  collection.close();
					  }
				  }
				  return id;
	  }


	/**
				* Adds an XML document specified as DOM with an additional document
				*  identifier (name) as child to an existing collection
				* @param collectionURI The URI of the parent collection as String
				* @param XMLData The XML data of the document as DOM
				* @return the Identifier (name) of the added document
				* @throws ClassNotFoundException
				* @throws InstantiationException
				* @throws IllegalAccessException
				* @throws XMLDBException
				* @throws XindiceException
				* @throws IOException
			   */
		public static String addDocumentFromDOM (String collectionURI, Node XMLData, String id)
			  throws ClassNotFoundException, InstantiationException, IllegalAccessException,
				  XMLDBException, XindiceException, IOException {
					  Collection collection = null;
					  try {
						  collection = getCollection(collectionURI); // e.g. "xmldb:xindice://localhost:8080/db/addressbook"
						  XMLResource document = (XMLResource) collection.createResource(id, "XMLResource");
						  document.setContentAsDOM((Document)XMLData);
						  collection.storeResource(document);
						  id = document.getId();
						  logger.info("Document inserted as " + id);
					  }
					  catch (XMLDBException e) {
						  logger.error( "XML:DB Exception occured " + e.errorCode + " " + e.getMessage());
					  }
					  finally {
						  if (collection != null) {
							  collection.close();
						  }
					  }
					  return id;
		  }

	/** Auxiliary method for XPathQuery. Adds a list of namespace prefixes and URIs to the
	 * query service object.
	 * @param service XPathQueryService object
	 * @param namespacesString AV list (String) of namespace prefixes and URIs
	 * @throws XMLDBException
	 */
	protected static void addNamespaces(XPathQueryService service, String namespacesString)
		throws XMLDBException {

			if ((namespacesString != "") && (namespacesString != null)) {
				StringTokenizer st = new StringTokenizer(namespacesString, "=;");
				if (st.countTokens() % 2 != 0) {
					logger.error( "mismatched namespace prefixes and uris");
					return;
				}
				while (st.hasMoreTokens()) {
					service.setNamespace(st.nextToken(), st.nextToken());
				}
			}
	}

	/**
	 * Deletes a Collection.
	 * @param collectionURI The collection to be deleted as URI String
	 * @return true iff collection could be deleted
	 */
	public static boolean deleteCollection (String collectionURI)
		throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			XMLDBException, XindiceException {
				Collection collection = null;
				String delCollection = "";
				StringTokenizer st = new StringTokenizer (collectionURI, "/");
				while (st.hasMoreTokens()) delCollection = st.nextToken();
				boolean ret = false;
				try {
					collection = getCollection(collectionURI);
					CollectionManager service = (CollectionManager) collection.getService("CollectionManager", XMLDBAPIVERSION);
					service.dropCollection(delCollection);
					ret = true;
				}
				catch (XMLDBException e) {
					logger.error( "XML:DB Exception occured " + e.errorCode + " " + e.getMessage());
				}
				finally {
					if (collection != null) {
						collection.close();
					}
				}
				return ret;
	  }

    /**
     * Deletes a document identified by name from the specified database collection.
     * @param collectionURI The URI of the parent collection as String
     * @param docname the document identifier (name)
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws IllegalAccessException
     * @throws XMLDBException
     * @throws XindiceException
     */
    public static void deleteDocument (String collectionURI, String docname)
		throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			XMLDBException, XindiceException {
				Collection collection = null;
				try {
					collection = getCollection(collectionURI);
					Resource document = collection.getResource(docname);
					collection.removeResource(document);
					logger.info("Document " + docname + " removed");
				}
				catch (XMLDBException e) {
					logger.error( "XML:DB Exception occured " + e.errorCode + " " + e.getMessage());
				}
				finally {
					if (collection != null) {
						collection.close();
					}
				}
    }

/**
 * Auxiliary method to avoid re-initialization of the Database manager
 */
	protected static void ensureInitialized()
		throws ClassNotFoundException, InstantiationException,
			IllegalAccessException, XMLDBException {
				if (alreadyInitialized) return;
				String driver = "org.apache.xindice.client.xmldb.DatabaseImpl";
				Class driverClass = Class.forName(driver);
				Database database = (Database) driverClass.newInstance();
				DatabaseManager.registerDatabase(database);
				alreadyInitialized = true;
	}

/**
 * Exports a database (collection) to an XML file directory tree.
 * @param collectionURI
 * @param directory the root directory to export the collection to
 */
	public static void exportDB (String collectionURI, String  directory) {
		}

/**
 * Returns a collection object specified by its name.
 * @param collectionUri The URI of the collection as String
 * @return Collection The requested collection
 * @throws ClassNotFoundException
 * @throws InstantiationException
 * @throws IllegalAccessException
 * @throws XMLDBException
 * @throws XindiceException
 */
	protected static Collection getCollection(String collectionUri)
		throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			XMLDBException, XindiceException {
			ensureInitialized();
			Collection collection = DatabaseManager.getCollection(collectionUri);
			if (collection == null) {
				throw new XindiceException("DatabaseManager.getCollection(" + collectionUri + ") returned null.");
			}
			return collection;
	}

	/**
	 * Import a Database collection from an XML file directory tree.
	 * @param collectionURI The URI of the root collection as String
	 * @param directory The path of the source directory as String
	 */
	public static void importDB (String collectionURI, String directory) {
	  }

    /**
     * List all child documents of the specified collection.
     * @param collectionURI The URI of the parent collection as String
     * @return a String array containing the names of the child documents
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws IllegalAccessException
     * @throws XMLDBException
     * @throws XindiceException
     */
    public static String[] listDocuments(String collectionURI)
    	throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			XMLDBException, XindiceException {
				Collection collection = null;
				String[] documentarray = null;
				try {
						collection = getCollection( collectionURI );
						if ( collection == null ) {
							logger.error( "Collection not found!");
						} else documentarray = collection.listResources();
				} finally {
					// Release the collection object
					if ( collection != null ) {
						collection.close();
					}
				}
			    return documentarray;
   }


	/**
	 * Auxiliary method that reads an XML file into a String object.
	 * Encoding ????
	 * @param file The file to be opened
	 * @return String XML String with default encoding assumed (!)
	 * @throws IOException
	 */
	public static String readFileFromDisk(File file)
		throws IOException {
			FileInputStream insr = new FileInputStream(file);
			byte[] fileBuffer = new byte[(int) file.length()];
			insr.read(fileBuffer);
			insr.close();
			return new String(fileBuffer);
	}

    /**
     * Retrieves a document from the specified collection as String.
     * @param collectionURI The URI of the parent collection as String
     * @param docname the name of the document to be returned
     * @return document as String
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws IllegalAccessException
     * @throws XMLDBException
     * @throws XindiceException
     */
    public static String retrieveDocumentAsString (String collectionURI, String docname)
		throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			XMLDBException, XindiceException {
    			String res = null;
				Collection collection = null;
				try {
					collection = getCollection(collectionURI);
					XMLResource xmlResource = (XMLResource) collection.getResource(docname);
					if (xmlResource != null) {
						res = (String) xmlResource.getContent();
					}
					else {
						logger.error( "Document not found");
					}
				}
				catch (XMLDBException e) {
					logger.error( "XML:DB Exception occured " + e.errorCode + " " + e.getMessage());
				}
				finally {
					if (collection != null) {
						collection.close();
					}
				}
				return res;
	}

	/**
	  * Retrieves a document from the specified collection as DOM object.
	  * @param collectionURI The URI of the parent collection as String
	  * @param docname the name of the document to be returned
	  * @return Node DOM node
	  * @throws ClassNotFoundException
	  * @throws InstantiationException
	  * @throws IllegalAccessException
	  * @throws XMLDBException
	  * @throws XindiceException
	  */
	 public static Node retrieveDocumentAsDOM (String collectionURI, String docname)
		 throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			 XMLDBException, XindiceException {
				 Node res = null;
				 Collection collection = null;
				 try {
					 collection = getCollection(collectionURI);
					 XMLResource xmlResource = (XMLResource) collection.getResource(docname);
					 if (xmlResource != null) {
						 res = xmlResource.getContentAsDOM();
					 }
					 else {
						 logger.error( "Document not found");
					 }
				 }
				 catch (XMLDBException e) {
					 logger.error( "XML:DB Exception occured " + e.errorCode + " " + e.getMessage());
				 }
				 finally {
					 if (collection != null) {
						 collection.close();
					 }
				 }
				 return res;
	 }
	/**
	 * Shutdown the database. Data will be lost unless exportDB() is called before shutdown.
	 * @param collectionURI The URI of the parent collection as String
	 */
    public static  void shutdown(String collectionURI) {
      }

    /**
     * Apply an update query (XUpdate standard) to the specified collection.
     * @param collectionURI The URI of the parent collection as String
     * @param updateString the XUpdate expression (XML syntax!)
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws IllegalAccessException
     * @throws XMLDBException
     * @throws XindiceException
     * @return count of updated entries
     */
    public static long update(String collectionURI, String updateString)
		throws ClassNotFoundException, InstantiationException, IllegalAccessException,
			XMLDBException, XindiceException {
				Collection collection = null;
				long count = -1;
				try {
						collection = getCollection(collectionURI);
							/*  e.g.
								    <xu:modifications version="1.0"
									      xmlns:xu="http://www.xmldb.org/xupdate">
									   <xu:remove select="/person/phone[@type = 'home']"/>
									   <xu:update select="/person/phone[@type = 'work']">
									       480-300-3003
									   </xu:update>
									</xu:modifications>
		                    */
						XUpdateQueryService service = (XUpdateQueryService) collection.getService("XUpdateQueryService", XMLDBAPIVERSION);
						count = service.update(updateString);
						logger.info(count + " entries updated.");
				}
				catch (XMLDBException e) {
					logger.error( "XML:DB Exception occured " + e.errorCode + " " + e.getMessage());
				}
				finally {
					if (collection != null) {
						collection.close();
					}
				}
				return count;
	}

    /**
     * Apply an XPath query to the specified collection.
     * @param collectionURI The URI of the parent collection as String
     * @param queryString The XPath query as String
     * @param namespaces a String containing a AV list of namespace prefixes and URIs
     * @return ResourceIterator an Iterator containing XMLResource objects the represent
     * (multiple) query results
     * @throws XMLDBException
     */
    public static ResourceIterator XPathQuery(String collectionURI, String queryString, String namespaces)
    	throws XMLDBException {
			Collection collection = null;
			XPathQueryService service = null;
			ResourceIterator results = null;
			try {
				collection = getCollection(collectionURI);
				if (collection == null) {
					logger.error( "Collection not found!");
					return null;
				}
				service = (XPathQueryService) collection.getService("XPathQueryService", "1.0");
				if (namespaces!=null) addNamespaces(service, namespaces);
				ResourceSet resultSet = service.query(queryString);
				results = resultSet.getIterator();
			/*	while (results.hasMoreResources()) {
					XMLResource resource = (XMLResource) results.nextResource();
					String documentstr = (String) resource.getContent();
					System.out.println(documentstr);
				} */
			}
			catch (Exception e) {
				logger.error( e.getMessage());
			}
			finally {
				if (collection != null) {
					collection.close();
				}
			}
		return results;
	}

	/**
	 * Adds a Indexer to the specified collection .
	 * @param collectionURI The URI of the collection to be indexed as String
	 * @param indexerName The name of the indexer (must be unique within a collection)
	 * @param indexerPattern A Pattern String with shape elem, elem@attr, *, *@attr, elem@* or *@*
	 * @param type the type of the index (String or null)
	 * @param pagesize page size (String of null)
	 * @param maxkeysize maximum key size (String or null)
	 * @return boolean for success or not
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws XMLDBException
	 * @throws XindiceException
	 */
	public static boolean addIndexer(String collectionURI, String indexerName, String indexerPattern,
	                                                         String type, String pagesize, String maxkeysize)
	    throws ClassNotFoundException, InstantiationException, IllegalAccessException,
	    	XMLDBException, XindiceException {
				Collection collection = null;
				try {
					collection = getCollection(collectionURI); // e.g. "xmldb:xindice://localhost:8080/db/addressbook"
					if ( collection == null ) {
					  logger.error( "Collection not found!");
					  return false;
					}
					// Create a collection manager instance for the collection
					CollectionManager colman = (CollectionManager)collection.getService("CollectionManager",XMLDBAPIVERSION);
				    Document doc = new DocumentImpl();
					// Create the index element to hold attributes
					Element idxEle = doc.createElement("index");
					idxEle.setAttribute("class", XINDICE_VAL_INDEXER);
					idxEle.setAttribute("name", indexerName);
					idxEle.setAttribute("pattern", indexerPattern);
					// Setup optional index attributes
					if (type != null) {
						if ( type.equalsIgnoreCase("name") )
						   idxEle.setAttribute("class", XINDICE_NAME_INDEXER);
						else
						   idxEle.setAttribute("type", type);
					}
					if (pagesize != null) {
						idxEle.setAttribute("pagesize", pagesize);
					}
					if (maxkeysize != null) {
						idxEle.setAttribute("maxkeysize", maxkeysize );
					}
					// Add Element to the document
					doc.appendChild(idxEle);
	                // show....
					String indexstr = TextWriter.toString(doc);
					System.out.println("Index node element = ");
					System.out.println("\t" + indexstr + "\n" );
					// Create the indexer for this collection manager
					colman.createIndexer(doc);
					System.out.println("CREATED : " + indexerName);
				} finally {
					if ( collection != null ) {
						collection.close();
					}
				}
				return true;
	}

}


