/*
 * ChasenModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.*;
import java.util.*;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Logger;
import org.w3c.dom.*;
import de.dfki.lt.hog.Metadata;
import de.dfki.lt.hog.MoCoMan;
import de.dfki.lt.hog.Module;
import de.dfki.lt.hog.TransformationService;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;
import de.dfki.lt.hog.util.CorrectCharPosHandler;

/**
 * Module definition for ChaSen
 * Created 2004-06-04
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @version  $Id: ChasenModule.java,v 1.6 2007/09/10 16:48:09 dcontag Exp $
 */

/* 2004-06-08 : uschaefer removed chasen server mode (no longer supported by ChaSen)
 * 2004-07-22 : uschaefer added empty result handling (metadata)
 * 2004-11-25 : uschaefer moved to pet XML input chart
 */

public class ChasenModule extends Module {
    //static final String EOS = "EOS";
    static final String EOL = System.getProperty("line.separator");
    private static final String CHASEN_SHUTDOWN_COMMAND = "\4";
    static Logger log = Logger.getLogger(ChasenModule.class);
    private String xmlHeader = null;
    private Process process = null;
    private String sChasenRootElement = null;
    private String sChasenInputAnnotationName = null;
    private String sChasenOutputEncoding = null;
    public BufferedWriter chasenIn;
    public BufferedReader chasenOut;
    private String chasenSmafOutput = null;
    private String chasenSmafStylesheet = null;
    private TransformationService transformationService = null;
    private Transformer transformer = null;
    
    /**  initialize ChaSen
     */
    public ChasenModule(Properties chasenModuleProps) {

                Runtime runtime;

                try {
                  super.init(chasenModuleProps);
                } catch (InitializationException ie) {
                  log.fatal(Utility.exception2String(ie));
                }
                // read configuration properties
                // global ones:
                sChasenRootElement = getConfig().getProperty("module.rootelement", "chasen");
                sChasenInputAnnotationName = getConfig().getProperty("chasen.inputannotation", "rawtext");
                //sChasenOutputAnnotationName = getConfig().getProperty("chasen.outputannotation", "ChaSenpic");
                // for init only:
                String sChasenOptions = getConfig().getProperty("chasen.options", "-r conf/.chasenrc");
                String sChasenBinary =  getConfig().getProperty("chasen.binary", "components/chasen/bin/runchasen");
                String sChasenLibs = getConfig().getProperty("chasen.libs", "components/chasen/lib");
                String sChasenInputEncoding = getConfig().getProperty("chasen.inputencoding", "EUC-JP");
                sChasenOutputEncoding = getConfig().getProperty("chasen.outputencoding", "EUC-JP");
                xmlHeader = "<?xml version='1.0'?>";
                if (!sChasenBinary.startsWith(System.getProperty("file.separator")))
                    sChasenBinary = System.getProperty("root.dir") + System.getProperty("file.separator") + sChasenBinary;
                if (!sChasenLibs.startsWith(System.getProperty("file.separator")))
                    sChasenLibs = System.getProperty("root.dir") + System.getProperty("file.separator") + sChasenLibs;
                String[] libs = { "LD_LIBRARY_PATH=" + sChasenLibs, // + ":/lib:/usr/lib:/usr/local/lib"
                                  "PATH=." };  // + ":/bin:/usr/bin:/usr/local/bin"
                
                // defaults to no, if not set in chasen.cfg
                chasenSmafOutput = getConfig().getProperty("chasen.smafOutput", "0");
                chasenSmafStylesheet = getConfig().getProperty("chasen.smafStylesheet");
                
                log.info("Initializing ChaSen ...");
                runtime = Runtime.getRuntime();
                try {
                    process = runtime.exec(sChasenBinary + " " + sChasenOptions, libs);
                    String lineIn;
                    chasenOut = new BufferedReader(
                                  new InputStreamReader(process.getInputStream(),
                                                        sChasenOutputEncoding));
                    chasenIn = new BufferedWriter(
                                 new OutputStreamWriter(process.getOutputStream(),
                                                        sChasenInputEncoding));
                    log.debug("Invoked: " + sChasenBinary + " " + sChasenOptions);
                    //                    chasenIn.write(EOL);
                    //                    chasenIn.flush();
                    //                    while (!(lineIn=chasenOut.readLine()).equals(EOS)) {
                    //                        log.debug("ChaSen: " + lineIn);
                    //                    }
                    log.info("ChaSen initialized.");
                } catch (IOException ioe) {
                  log.fatal(Utility.exception2String(ioe));
                }

        }

        public void shutdown() {
            String lineIn;
            log.info("Shutting down ChaSen.");
            try {
                if (process!=null) {
                    chasenIn.write(CHASEN_SHUTDOWN_COMMAND);
                    chasenIn.flush();
                    chasenIn.close();
                    chasenOut.close();
                    process=null;
                }
            } catch (IOException ioe) {
              log.error(Utility.exception2String(ioe));
            }
        }

        /* (non-Javadoc)
         * @see de.dfki.lt.hog.Module#process(java.lang.String)
         */
        public String process(String processString) {
        	String duration = "00:00,000";
        	long t1 = System.currentTimeMillis();
            MoCoMan mocoman = getMoCoMan();
            String inputText = Utility.getComputedAnnotation(mocoman, processString, sChasenInputAnnotationName);

            log.debug("input for CHASEN: " + inputText);
            Metadata md = new Metadata(processString);
            String sessionID = md.getSessionID();
            String acID = md.getAcID();

            // process text with CHASEN, output is PIC format -> ChasenPic
            // maybe we should also return a proper XML format

            String result = "";
            String lineIn;

            try {
                chasenIn.write(inputText + EOL);  // send raw text to ChaSen
                chasenIn.flush();
                result = insertIDs(chasenOut.readLine());
                duration = Utility.formatDuration(System.currentTimeMillis() - t1);
                log.info("Processing time of ChaSen: " + duration);
                log.debug("result: " + result);

                //mocoman.createAnnotation(sessionID, acID, result, getModuleLanguage(), sChasenOutputAnnotationName);

                result = xmlHeader + "<" + sChasenRootElement + ">" +
                    Metadata.generateXml(sessionID, acID, duration,
                                         getModuleName(),
                                         "OK", (result.trim().equals("")),
                                         getConfig())
                    + result + "</" + sChasenRootElement + ">";

                String pixmlResult = CorrectCharPosHandler.correctCharPos(result, inputText, sChasenOutputEncoding, "cstart", "cend"); 
                String smafResult = null;
                
                if (chasenSmafOutput.equals("1")) {
                    // smaf is enabled
                    
                    if (this.transformationService == null) {
                        this.transformationService = new TransformationService(
                                getMoCoMan());
                    }

                    try {
                        transformer = this.transformationService
                                .createTransformer(chasenSmafStylesheet);
                    } catch (TransformerConfigurationException tce) {
                        log.error(Utility.exception2String(tce));
                    }

                    try {
                        smafResult = this.transformationService.transform(
                                transformer, pixmlResult);
                    } catch (TransformerException te) {
                        log.error(Utility.exception2String(te));
                    }
                    return smafResult;
                }
                else {
                    // smaf not enabled
                    return pixmlResult;
                }
                 

            } catch (Exception ioe) {
                log.fatal(Utility.exception2String(ioe));
                return xmlHeader + "<" + sChasenRootElement + ">" +
                    Metadata.generateXml(sessionID, acID, duration,
                                         getModuleName(),
                                         "Error: " + ioe.getMessage(), true,
                                         getConfig()) +
                    "</" + sChasenRootElement + ">";
            }
        }


        /** insert unique ID attributes for each w element
         */

        public static String insertIDs(String input) {
            StringBuffer sb = new StringBuffer();
            int i = 0;
            int j = 0;
            int k = 0;
            while ((i=input.indexOf("<w ", j))>-1) {
                sb.append(input.substring(j, i+3));
                sb.append("id=\"CHA" + (k++) + "\" ");
                j = i+3;
            }
            sb.append(input.substring(j));
            return sb.toString();
        }

        /* (non-Javadoc)
         * @see de.dfki.lt.hog.Module#processDOMTree(org.w3c.dom.Node)
         */
        public Node processDOMTree(Node domTree) {
                // TODO Auto-generated method stub
                // use a simple SAX parser to extract text from domTree
                // return a DOM from transformer DOMResult (i.e., copy most of process() except final DOM output)
                return null;
        }

}
