/*
 * ChunkieModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.*;
import java.util.*;
import java.util.regex.*;
import org.apache.log4j.Logger;
import org.w3c.dom.*;
import de.dfki.lt.hog.Metadata;
import de.dfki.lt.hog.MoCoMan;
import de.dfki.lt.hog.Module;
import de.dfki.lt.hog.container.Chunk;
import de.dfki.lt.hog.container.JToken;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;

/**
 * Module definition for Chunkie commandline interface
 *
 * @author <a href=mailto:oeze@dfki.de>Oezguer Demir</a>
 * @version $Revision: 1.9 $
 * $Id: ChunkieModule.java,v 1.9 2005/09/29 12:59:31 uschaefer Exp $
 */

/* Changes:
 * 2004-07-22 : US added empty result handling (metadata)
 * 2004-09-09 : US input + outputencoding properly passed to process streams
 * 2004-09-13 : US added additional configuration for regexp of chunk types
 */

public class ChunkieModule extends Module {

  static Logger log = Logger.getLogger(ChunkieModule.class);
  private BufferedReader procOut;
  private BufferedReader procErr;
  private BufferedWriter procIn;
  private Runtime runtime;
  private Process process;
  private String chunkieOptions = "-v0";
  private String chunkieScript = "";
  private String outputEncoding;
  private String inputEncoding;
  private Pattern chunkPattern;
  private static final String NEWLINE = System.getProperty("line.separator");
  private static final String DATA_ROOT_ELEMENT = "chunks";
  private static int lifeCycle = 0;

  public ChunkieModule(Properties chunkieModuleProps) throws
    InitializationException {
    try {
      super.init(chunkieModuleProps);
    } catch (InitializationException ie) {
      log.fatal(Utility.exception2String(ie));
    }
    chunkieScript = getConfig().getProperty("chunkie.script");
    chunkieOptions =
      chunkieOptions + " " + getConfig().getProperty("chunkie.options");
    outputEncoding =
      getConfig().getProperty("chunkie.outputencoding", "ISO-8859-1");
    inputEncoding =
      getConfig().getProperty("chunkie.inputencoding", "ISO-8859-1");
    String sPattern = getConfig().getProperty("chunkie.phrasalcats", null);
    try {
      if (sPattern == null) {
        log.error("Pattern for chunk categories not specified in ChunkieModule configuration (chunkie.phrasalcats)");
        chunkPattern = Pattern.compile("NP|PP");
      } else {
        chunkPattern = Pattern.compile(sPattern);
      }
    } catch (PatternSyntaxException pse) {
      log.fatal(Utility.exception2String(pse));
    }

    start();
  } // ChunkieModule()

  public void start() {
    log.info("Initializing Chunkie ...");
    runtime = Runtime.getRuntime();
    try {
      log.debug("Invoking " + chunkieScript + " " + chunkieOptions);
      process = runtime.exec(chunkieScript + " " + chunkieOptions);
      procOut =
        new BufferedReader(new InputStreamReader(process.getInputStream(),
                                                 outputEncoding));
      procIn =
        new BufferedWriter(new OutputStreamWriter(process.getOutputStream(),
                                                  inputEncoding));
      procErr =
        new BufferedReader(new InputStreamReader(process.getErrorStream(),
                                                 outputEncoding));
    } catch (IOException ioe) {
      log.fatal(Utility.exception2String(ioe));
    }

//try {
//process = runtime.exec(tntScript + " " + tntOptions);
//log.debug("Invoked: " + tntScript + " " + tntOptions);
//String lineIn;
//procOut = new BufferedReader(
//    new InputStreamReader(process.getInputStream()));
//procIn = new BufferedWriter(
//    new OutputStreamWriter(process.getOutputStream()));
// test
//if (procOut.ready()) {
//  log.info("TnT initialized (tested).");
//  procIn.close();
//  procOut.close();
//}
//else {
//  throw new InitializationException("TnT did not start");
//} // else
//} // try
//catch (IOException ioe) {
//  log.fatal(Utility.exception2String(ioe));
//}
  } // start()

  public void shutdown() {
    log.info("Shutting down Chunkie.");
    try {
      procIn.close();
      procOut.close();
      procErr.close();
      process.destroy();
    } catch (IOException ioe) {
      log.fatal(Utility.exception2String(ioe));
    }
  }

  private static String preProcess(String inString) {
    String outString =
      inString.replace('(', '[');
    outString =
      outString.replace(')', ']');
    return outString;
  }

  private static String postProcess(String dirtyString) {
    String cleanString =
      dirtyString.replace('[', '(');
    cleanString =
      cleanString.replace(']', ')');
    return cleanString;
  }


  /* (non-Javadoc)
   * @see de.dfki.lt.hog.Module#process(java.lang.String)
   */
  public String process(String processString) {
    //lifeCycle++;
    //if (lifeCycle >= 1000) {
    //  shutdown();
    //  start();
    //  lifeCycle = 0;
    //}
    MoCoMan mocoman = getMoCoMan();
    Metadata md = new Metadata(processString);
    String sessionID = md.getSessionID();
    String acID = md.getAcID();
    String root = getModuleRootElement();
    boolean warned = false;
    // process text
    String lineIn = "";
    StringBuffer errBuffer = new StringBuffer();
    String metadata;
    String duration = "00:00,000";
    StringBuffer sbResult = new StringBuffer();
    String diagnosis;
    boolean ok = false;
    boolean inResult = false;
    long t1 = System.currentTimeMillis();
    // tokenized text
    StringBuffer tokenBuffer = new StringBuffer();
    ArrayList jtoks = JToken.parseJTokAnnotation(
      Utility.getComputedAnnotation(mocoman, processString, "JTok"));
    // split sentences according to JTok
    String lineSep;
    for (int i = 0; i < jtoks.size(); i++) {
      JToken tok = (JToken) jtoks.get(i);
      if (tok.getType().startsWith("PTERM")) {
        lineSep = NEWLINE + NEWLINE;
      } else {
        lineSep = NEWLINE;
      }
      tokenBuffer.append(tok.getToken() + lineSep);
    }

    String tokenizedStr = tokenBuffer.toString();
    tokenizedStr = preProcess(tokenizedStr);
    int numJTokens = jtoks.size();
    log.debug(tokenizedStr);
    // process
    try {
      // feed the text to Chunkie
      procIn.write(tokenizedStr + NEWLINE);
      procIn.flush();
      // get results from Chunkie
      int jtokenCounter = 0;
      int structureDepth = 0;
      boolean insideSentence = false;
      // StringBuffer debugChunkieBuffer = new StringBuffer();

      // go go
      boolean firstLine = true;
      while (jtokenCounter != numJTokens) {
          //        log.debug("Chunkie's original output so far (0): " + NEWLINE +
          //        debugChunkieBuffer);

        if (procOut.ready()) {
          lineIn = procOut.readLine();
        } else { // else 1
          for (int i = 0; i <= 50; i++) {
            try {
              Thread.sleep(20);
            } catch (InterruptedException ex) {
              log.error(ex);
            }
            if (procOut.ready()) {
              break;
            }
          } // for
          if (procOut.ready()) {
            lineIn = procOut.readLine();
          } else { // else 2
            lineIn = "FOOBAR";
            jtokenCounter++;
            if (!warned) {
              log.warn("Chunkie's output ended prematurely: " + NEWLINE +
                       sbResult);
              warned = true;
            }
          } // else 2
        } // else 1

        // log.debug("Buffer ready? : " + procOut.ready());
        // debugChunkieBuffer.append(lineIn + NEWLINE);
        // log.debug("Chunkie's original output so far (1): " + NEWLINE +
        //          debugChunkieBuffer);
        // log.debug("chunkie said: " + lineIn);
        sbResult.append(lineIn + NEWLINE);
        if (lineIn.replaceAll("\t*", "").startsWith("(")) {
          structureDepth++;
          insideSentence = true;
        }
        if (lineIn.replaceAll("\t*", "").endsWith(")")) {
          structureDepth--;
        }
        // fix outputs starting with closing parantheses
        if (structureDepth < 0) {
          structureDepth = 0;
          lineIn = "";
          log.warn("Received too many closing parantheses." + NEWLINE +
                   sbResult);
        }
        if ((structureDepth == 0) && insideSentence) {
          insideSentence = false;
        }
        if (((lineIn.replaceAll("\t*", "")).length() > 1)
            && lineIn.endsWith(")")) {
            //  log.debug("lineIn after tabs-removal: " +
            //            lineIn.replaceAll("\t*", ""));
          jtokenCounter++;
        }
        // log.debug("numJTokens = " + numJTokens);
        // log.debug("jtokenCounter = " + jtokenCounter);
        // log.debug("insideSentence = " + insideSentence);
        // log.debug("structureDepth = " + structureDepth);
      } // while

      for (int i = 0; i < structureDepth; i++) {
          //        log.debug("structureDepth: " + structureDepth);
          //        log.debug("reading one line in order to get closing paranthesis");
        lineIn = procOut.readLine();
        //        sbResult.append(lineIn + NEWLINE);
        //        log.debug("read this: " + lineIn);
      }

      StringBuffer errIn = new StringBuffer();
      while (procErr.ready()) {
        char c = (char) procErr.read();
        if (c != '.') {
          errIn.append(c);
        }
      }
      //      log.debug("Chunkie said this on stderr: " + errIn);
      errBuffer.append(errIn);
      if (errIn.toString().equals("")) {
        ok = true;
      }
      //////
      // convert chunkie's parentheses structure to xml
      StringBuffer xmlBuffer = new StringBuffer();

      // check whether result is empty
      String chunkieresult = sbResult.toString();
      log.debug(chunkieresult);
      boolean empty = false;
      if (chunkieresult.trim().equals("")) {
        empty = true;
        log.warn(errBuffer.toString());
      }
      // generate header
      if (ok) {
        diagnosis = "OK";
      } else {
        diagnosis = errBuffer.toString();
        if (diagnosis.startsWith("TnT-Token -")) {
          diagnosis = "OK";
        } else {
          log.error(errBuffer);
        }
      }
      duration = Utility.formatDuration(System.currentTimeMillis() - t1);
      StringBuffer header = new StringBuffer();
      header.append("<?xml version='1.0'?>");
      header.append("<" + root + ">");
      header.append(Metadata.generateXml(sessionID,
                                         acID, duration,
                                         getModuleName(),
                                         diagnosis, empty,
                                         getConfig()));
      header.append("<" + DATA_ROOT_ELEMENT + ">" + NEWLINE);
      xmlBuffer.append(chunkie2xml(chunkieresult, processString, mocoman,
                                   chunkPattern.matcher("dummy")));
      xmlBuffer.append("</" + DATA_ROOT_ELEMENT + ">" + NEWLINE);
      xmlBuffer.append("</" + root + ">");

      String sResult = header.toString() + xmlBuffer.toString();
      sResult = postProcess(sResult);
      log.debug("result: " + sResult);
      log.info("Processing time of ChunkieModule: " + duration);
      //if (lifeCycle == 500) { log.error("spotCheck-500: " + sResult); }
      return sResult;
    } catch (IOException ioe) {
      log.fatal(Utility.exception2String(ioe));
      return "<" + root + ">" +
        Metadata.generateXml(sessionID, acID, duration,
                             getModuleName(),
                             ioe.getMessage(), true,
                             getConfig()) +
        "</" + root + ">";
    }
  }

  /**
   * Converts Chunkie's parentheses structure to XML in two steps:
   * 1. Building a data structure representing the sentences;
   * 2. Generating an XML-String using the data strucure.
   *
   * @param str Chunkie's output as a String
   * @return Chunkie's output converted to XML
   */
  private static String chunkie2xml(String str, String processString,
                                    MoCoMan mocoman, Matcher chunkMatcher) {
    //log.debug("chunkie2xml: got this input." + str);
    ArrayList sentences = new ArrayList();
    Stack stack = new Stack();
    ArrayList jtoks = JToken.parseJTokAnnotation(
      Utility.getComputedAnnotation(mocoman, processString, "JTok"));
    String[] tokens;
    String token;
    int sid = 0;
    int wordPointer = 0;
    // clean the input
    str = Utility.cleanXMLString(str);
    // remove chunkie's comments
    str = str.replaceAll("%%.*\n", "");
    // hack the problem with parentheses in input
    str = str.replaceAll("\\(\\$PUNCT\\s+\\(", "\\(\\$PUNCT \\&#40;");
    str = str.replaceAll("\\(\\$PUNCT\\s+\\)", "\\(\\$PUNCT \\&#41;");
//    String dbs = "\\";
//    str = str.replaceAll("\\(\\$PUNCT\\s+" + dbs + "\"",
//                         "\\(\\$PUNCT \\&#34;");
    // tokenize chunkie's input
    StringTokenizer st = new StringTokenizer(str);
    tokens = new String[st.countTokens()];
    for (int i = 0; i < tokens.length; i++) {
      tokens[i] = st.nextToken();
    }
    for (int i = 0; i < tokens.length; i++) {
      token = (String) tokens[i];
      // build the data structure
      if (token.equals("(")) {
        token = ("(S");
      } // if "("
      if (token.length() != 0) {
        if (token.startsWith("(")) {
          token = token.substring(1, token.length());
          if (token.equals("S")) {
            Chunk s = new Chunk();
            s.setType("s");
            s.setSid("S" + sid++);
            s.setPoscat("S");
            sentences.add(s);
            stack.push(s);
          } else if ((chunkMatcher.reset(token)).matches()) {
            Chunk c = new Chunk();
            c.setType("c");
            c.setPoscat(token);
            ((Chunk) stack.get(stack.size() - 1)).addDaughter(c);
            stack.push(c);
          } else {
            Chunk w = new Chunk();
            w.setType("w");
            w.setPoscat(token);
            if (jtoks.size() <= wordPointer) {
              JToken tok = (JToken) jtoks.get(wordPointer - 1);
              w.setToken(new JToken(token, token.length(),
                                    tok.getOffset() + tok.getLength()));
            } else {
              w.setToken((JToken) jtoks.get(wordPointer));
              wordPointer++;
            }
            ((Chunk) stack.get(stack.size() - 1)).addDaughter(w);
            stack.push(w);
          } // if NP, else
        } // if "(.*"
        if (token.equals(")")) {
          stack.pop();
        } // if ")"
      } // if length != 0
    } // for i
    // convert to XML
    StringBuffer out = new StringBuffer();
    for (int i = 0; i < sentences.size(); i++) {
      out.append(((Chunk) sentences.get(i)).toXML());
    }
    return out.toString();
  } // chunkie2xml()

  /* (non-Javadoc)
   * @see de.dfki.lt.hog.Module#processDOMTree(org.w3c.dom.Node)
   */
  public Node processDOMTree(Node domTree) {
    // TODO Auto-generated method stub
    // use a simple SAX parser to extract text from domTree
    // return a DOM from transformer DOMResult (i.e., copy most of process() except final DOM output)
    return null;
  }
} // class
