/*
 * FreLingModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.*;
import java.util.*;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import org.apache.log4j.Logger;
import org.w3c.dom.*;
import de.dfki.lt.hog.Metadata;
import de.dfki.lt.hog.MoCoMan;
import de.dfki.lt.hog.Module;
import de.dfki.lt.hog.TransformationService;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;

/**
 * Module definition for FreeLing (LKBanalyzer commandline)
 * Created 2006-06-17
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @version  $Id: FreeLingModule.java,v 1.2 2006/06/21 22:43:31 uschaefer Exp $
 */

/* Changes:
   
   */

public class FreeLingModule extends Module {

    static Logger log = Logger.getLogger(FreeLingModule.class);
    private BufferedReader procOut;
    private BufferedWriter procIn;
    private Runtime runtime;  // need not be a class variable
    private Process process;  // need not be a class variable
    private static final String SHUTDOWN_COMMAND = "\u0004";
    private static final String FORMFEED = "\u000c"; 
    private static final String LINEFEED = System.getProperty("line.separator");
    private static final String xmlHeader = "<?xml version='1.0'?>";
    private static final String xmlHeader4LkbWrapper = "<?xml version='1.0' encoding='utf-8'?>";
    private String sRootElement = null;
    private String sInputAnnotationName = null;
    private String sOptions = null;
    private TransformationService transformationService = null;
    private Transformer piXmlTransformer = null;
    
    String sBinary ;
    String sLibs;
    String sInputEncoding;
    String sOutputEncoding;
    
    /**
     */
    public FreeLingModule(Properties moduleProps) {
                try {
                  super.init(moduleProps);
                } catch (InitializationException ie) {
                  log.error(Utility.exception2String(ie));
                }
                // read configuration properties
                // global ones:
                sRootElement = getConfig().getProperty("module.rootelement", "");
                sInputAnnotationName = getConfig().getProperty("freeling.inputannotation", "xmltext");
                sOptions = getConfig().getProperty("freeling.options", "-f lkb.cfg");
                // for init only:
                sBinary =  getConfig().getProperty("freeling.binary", "LKBanalyzer");
                sLibs = getConfig().getProperty("freeling.libs", "components/freeling/lib");
                sInputEncoding = getConfig().getProperty("freeling.inputencoding", "ISO-8859-1");
                sOutputEncoding = getConfig().getProperty("freeling.outputencoding", "ISO-8859-1");
                if (!sBinary.startsWith(System.getProperty("file.separator")))
                    sBinary = System.getProperty("root.dir") + System.getProperty("file.separator") + sBinary;
                if (!sLibs.startsWith(System.getProperty("file.separator")))
                    sLibs = System.getProperty("root.dir") + System.getProperty("file.separator") + sLibs;
        }

        public void shutdown() {
            log.info("Shutting down FreeLing.");
            try {
                procIn.write(SHUTDOWN_COMMAND);
                procIn.flush();
                process = null;
            } catch (IOException ioe) {
                log.fatal(Utility.exception2String(ioe));
            }
        }

        /* (non-Javadoc)
         * @see de.dfki.lt.hog.Module#process(java.lang.String)
         */
        public String process(String processString) {
            MoCoMan mocoman = getMoCoMan();
            Metadata md = new Metadata(processString);
            String sessionID = md.getSessionID();
            String acID = md.getAcID();
            String inputAnnotation;
            String duration = "00:00,000";

            log.info("Initializing FreeLing...");
            runtime = Runtime.getRuntime();
            String[] libs = { "LD_LIBRARY_PATH=" + sLibs, // + ":/lib:/usr/lib:/usr/local/lib"
                              "PATH=." };  // + ":/bin:/usr/bin:/usr/local/bin"

            try {
                process = runtime.exec(sBinary + " " + sOptions, libs);
                log.debug("Invoked: " + sBinary + " " + sOptions);
                String lineIn;
                procOut = new BufferedReader(
                                             new InputStreamReader(process.getInputStream(), // use ErrorStream for debugging
                                                                   sOutputEncoding));
                procIn = new BufferedWriter(
                                            new OutputStreamWriter(process.getOutputStream(),
                                                                   sInputEncoding));
                log.info("FreeLing initialized.");
                
            } catch (IOException ioe) {
                log.fatal(Utility.exception2String(ioe));
            }

            if (this.transformationService==null) {
            	this.transformationService = new TransformationService(getMoCoMan());
            	// build PET XML input chart generating stylesheet 
            	try {
		    piXmlTransformer = this.transformationService.createTransformer(
		      getConfig().getProperty("freeling.picstylesheet", "xsl/pic/sppp2pic.xsl"));
            	} catch (TransformerConfigurationException tce) {
            		log.error(Utility.exception2String(tce));
            	}
            }
            long t1 = System.currentTimeMillis();
                // Add DTD and remove empty lines, add empty line at end

	    inputAnnotation = mocoman.getAnnotation(sessionID, acID, sInputAnnotationName);
            inputAnnotation = xmlHeader4LkbWrapper + LINEFEED + "<text>" + inputAnnotation.substring(inputAnnotation.indexOf("</metadata>")+12);
            log.debug("input for FreeLing: " + inputAnnotation);

            // process text with FreeLing, output is RMRS (XML)
            String lineIn;
            String metadata;
            StringBuffer sbResult = new StringBuffer();
            boolean inResult = false;
            try {
                procIn.write(inputAnnotation + LINEFEED);
                procIn.flush();
                while (!((lineIn=procOut.readLine()).equals("</segment>"))) {
                    log.debug(lineIn);
		    sbResult.append(lineIn);
		}
                sbResult.append("</segment>");
                shutdown();
                
                duration = Utility.formatDuration(System.currentTimeMillis()-t1);
		metadata = Metadata.generateXml(sessionID, acID, duration, getModuleName(), "OK", false, getConfig());

                String sResult = sbResult.toString();
                log.debug(sResult);
		if (piXmlTransformer!=null) {
		    try {
			piXmlTransformer.setParameter("metadata", metadata);
			sResult = this.transformationService.transform(piXmlTransformer, sResult);
		    } catch (TransformerException te) {
			log.error(Utility.exception2String(te));
			return generateErrorResult(sessionID, acID, duration, getModuleName(), "Transformation error: " + te.getMessage());
		    }
		}
	    
                log.debug("result: " + sResult);
                return sResult;
                } catch (IOException ioe) {
                    log.fatal(Utility.exception2String(ioe));
                    return generateErrorResult(sessionID, acID, duration, getModuleName(), 
                                               "Error: " + ioe.getMessage());
                }
        }

        /* (non-Javadoc)
         * @see de.dfki.lt.hog.Module#processDOMTree(org.w3c.dom.Node)
         */
        public Node processDOMTree(Node domTree) {
                // TODO Auto-generated method stub
                // use a simple SAX parser to extract text from domTree
                // return a DOM from transformer DOMResult (i.e., copy most of process() except final DOM output)
                return null;
        }

        public String generateErrorResult( String sessionID, String acID, String duration, String module, String diagnosis) {
            return xmlHeader + "<" + getModuleRootElement() + " readings=\"0\">" +
                   Metadata.generateXml(sessionID, acID, duration, module,
                                        diagnosis, true, getConfig()) +
                "</" + getModuleRootElement() + ">";
    }
}
