/*
 * LingPipe2Module.java
 *
 * (C) 2006 DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.*;
import java.util.*;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.log4j.*;
import org.w3c.dom.*;
import org.xml.sax.*;

import com.aliasi.chunk.*;
import com.aliasi.util.AbstractExternalizable;
import com.aliasi.xml.SAXWriter;
import de.dfki.lt.hog.*;
import de.dfki.lt.hog.exceptions.*;
import de.dfki.lt.hog.util.*;

/**
 * Module definition for LingPipe (Named-Entity Extraction) ({@link http://www.alias-i.com/lingpipe})
 * 
 * based on LingPipeModule.java by Oezguer Demir
 * 
 * @author Daniel Contag
 * @version
 */
public class LingPipe2Module extends Module {

    static Logger log = Logger.getLogger(LingPipe2Module.class);

    private TransformationService transformationService = null;

    private String sTylesheet4rmrs = null;

    private Transformer lingpipe2rmrsTransformer = null;

    private String sTylesheet4pic = null;

    private Transformer lingpipe2picTransformer = null;

    private String sPicOutputAnnotationName = null;

    private String sRawOutputAnnotationName = null;

    public LingPipe2Module(Properties moduleProps)
            throws InitializationException {
        log.info("Initializing LingPipe...");
        try {
            log.debug(Utility.map2String(moduleProps));
            super.init(moduleProps);
        } catch (InitializationException ie) {
            log.fatal(Utility.exception2String(ie));
        }
        sPicOutputAnnotationName = getConfig().getProperty(
                "lingpipe2.pixmlannotation", null);
        sRawOutputAnnotationName = getConfig().getProperty(
                "lingpipe2.rawannotation", "SproutFS");
        sTylesheet4pic = getConfig().getProperty("lingpipe2.stylesheet4pic");
        sTylesheet4rmrs = getConfig().getProperty("lingpipe2.stylesheet4rmrs");
        log.info("LingPipe initialized.");
    }

    public void shutdown() {
        log.info("shutting down LingPipe.");
        // nothing to do here
    }

    /*
     * (non-Javadoc)
     * 
     * @see de.usaar.coli.deepthought.mocoman.Module#process(java.lang.String)
     */
    public String process(String processString) {
        MoCoMan mocoman = getMoCoMan();
        String rawtext = Utility.getComputedAnnotation(mocoman, processString,
                "rawtext");

        log.debug("processString " + processString + " rawtext " + rawtext);

        Metadata md = new Metadata(processString);
        String sessionID = md.getSessionID();
        String acID = md.getAcID();
        log.info("input text: \"" + rawtext + "\"");
        long t1 = System.currentTimeMillis();
        String as = runLingPipeNE(rawtext);
        String duration = Utility.formatDuration(System.currentTimeMillis()
                - t1);
        String root = getModuleRootElement();
        log.info("processing time of LingPipe: " + duration);
        log.info("creating LingPipe XML output");

        String msg = "OK";
        md = new Metadata(sessionID, acID, duration, getModuleName(), msg,
                false, getConfig());
        StringBuffer sbResult = new StringBuffer();
        sbResult.append("<?xml version='1.0'?>");
        sbResult.append("<" + root + ">");
        sbResult.append(md.getXmlString());
        sbResult.append(as);
        sbResult.append("</" + root + ">");
        String sXML = sbResult.toString();
        log.debug(sXML);
        // convert to RMRS and PET XML input chart format, save raw and piXML as
        // additional, separate annotation, return RMRS.
        String rmrs = transformLP(sXML, md);
        // sXML = cleanXMLString(sbResult);
        return rmrs;
    }

    private final String runLingPipeNE(String inputString) {
        log.debug("runLingPipeNE");
        File modelDirectory = new File(getConfig().getProperty(
                "lingpipe2.modelsdir"));
        File newsModelFile = new File(modelDirectory, getConfig().getProperty(
                "lingpipe2.modelfile"));

        log.debug("newsModelFile " + newsModelFile);

        Xml2StringHandler handler = new Xml2StringHandler();
        Chunker neChunker;

        try {
            neChunker = readChunker(newsModelFile);
            log.debug("loaded neChunker");
            Chunking mentionChunking = neChunker.chunk(inputString);

            writeMentionChunking(handler, mentionChunking);
            handler.endDocument();

        } catch (Exception e) {
            log.error(e.getMessage());
        }

        return handler.getString();

    }

    private Chunker readChunker(File modelFile) throws IOException,
            ClassNotFoundException {

        Chunker chunker = (Chunker) AbstractExternalizable
                .readObject(modelFile);
        return chunker;
    }

    private void writeMentionChunking(Xml2StringHandler writer,
            Chunking mentionChunking) throws SAXException {

        Set chunkSet = new TreeSet(Chunk.TEXT_ORDER_COMPARATOR);
        chunkSet.addAll(mentionChunking.chunkSet());
        Iterator it = chunkSet.iterator();
        String text = mentionChunking.charSequence().toString();
        int pos = 0;
        Integer id = 0;

        while (it.hasNext()) {
            Chunk neChunk = (Chunk) it.next();
            int start = neChunk.start();
            int end = neChunk.end();
            String type = neChunk.type();
            String chunkText = text.substring(start, end);
            String whitespace = text.substring(pos, start);
            writer.characters(whitespace);
            writer.startElement("ENAMEX", (id++).toString(), type);
            writer.characters(chunkText);
            writer.endElement("ENAMEX");
            pos = end;
        }
        String whitespace = text.substring(pos);
        writer.characters(whitespace);
    }

    String transformLP(String output, Metadata metadata) {
        String rmrs = null;
        String piXml = null;
        if (this.transformationService == null) {
            this.transformationService = new TransformationService(getMoCoMan());
            try {
                lingpipe2rmrsTransformer = this.transformationService
                        .createTransformer(sTylesheet4rmrs);
                lingpipe2rmrsTransformer.setParameter("encoding", "UTF-8");
                if (sPicOutputAnnotationName != null)
                    lingpipe2picTransformer = this.transformationService
                            .createTransformer(sTylesheet4pic);
            } catch (TransformerConfigurationException tce) {
                log.error(Utility.exception2String(tce));
            }
        }
        // saving raw lingpipe output as separate annotation
        getMoCoMan().createAnnotation(metadata.getSessionID(),
                metadata.getAcID(), output, getModuleLanguage(),
                sRawOutputAnnotationName);
        try {
            log.info("transforming LingPipe output to RMRS");
            rmrs = this.transformationService.transform(
                    lingpipe2rmrsTransformer, output);
            if (sPicOutputAnnotationName != null) {
                log
                        .info("transforming LingPipe output to Pet XML Input Chart Format");
                piXml = this.transformationService.transform(
                        lingpipe2picTransformer, output);
                getMoCoMan().createAnnotation(metadata.getSessionID(),
                        metadata.getAcID(), piXml, getModuleLanguage(),
                        sPicOutputAnnotationName);
            }
        } catch (TransformerException te) {
            log.error(Utility.exception2String(te));
            return generateErrorResult(metadata.getSessionID(), metadata
                    .getAcID(), metadata.getProcTime(), getModuleName(),
                    "Error: " + te.getMessage());
        }
        return rmrs;
    }

    public String generateErrorResult(String sessionID, String acID,
            String duration, String module, String diagnosis) {
        return "<"
                + getModuleRootElement()
                + ">"
                + Metadata.generateXml(sessionID, acID, duration, module,
                        diagnosis, true, getConfig()) + "</"
                + getModuleRootElement() + ">";
    }

    private static class Xml2StringHandler extends SAXWriter {

        private final StringBuffer outBuffer = new StringBuffer();

        private String typeID;

        private int offsetCount = 0;

        private boolean inNE = false;

        public Xml2StringHandler() {
        }

        public void startElement(String qName, String id, String type) {

            if (qName.equals("ENAMEX")) {
                String entityType = type;
                String entityID = id;

                typeID = entityType + "[ID=" + entityID + "]";
                outBuffer.append("<NE id=\"" + entityID + "\" " + "type=\""
                        + entityType + "\" " + "cstart=\"" + offsetCount
                        + "\" " + "string=\"");
                inNE = true;
            }
        }

        public void endElement(String qName) {
            if (qName.equals("ENAMEX")) {
                outBuffer.append("\" cend=\"" + (offsetCount - 1) + "\"/>"
                        + System.getProperty("line.separator"));
                inNE = false;
            }
        }

        public void characters(char[] cs, int start, int length) {
            offsetCount = offsetCount + length;
            if (inNE) {
                outBuffer.append(cs, start, length);
            }
        }

        public String getString() {
            return outBuffer.toString();
        }

        // to avoid problems with the OutputStream of the lingpipe SaxWriter
        public void endDocument() {
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see de.usaar.coli.deepthought.mocoman.Module#processDOMTree(org.w3c.dom.Node)
     */
    public Node processDOMTree(Node domTree) {
        // TODO Auto-generated method stub
        // tokenize as in process(), but use XMLOutputter.createXMLDocument()
        // for output
        // the bad thing: JTok then outputs JDOM, but (directly) no JAXP DOM
        // so, use the jdomDOMoutputter:
        // org.jdom.output.DOMOutputter.output(org.jdom.Document) -> w3c-DOM
        // i.e., copy the whole process() and modifiy the two lines that
        // generate XML Strings
        return null;
    }

}
