/*
 * LingPipeModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.*;
import java.util.*;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.log4j.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import com.aliasi.coref.*;
import com.aliasi.ne.*;
import com.aliasi.sentences.*;
import com.aliasi.tokenizer.*;
import com.aliasi.xml.*;
import de.dfki.lt.hog.*;
import de.dfki.lt.hog.exceptions.*;
import de.dfki.lt.hog.util.*;

/**
 * Module definition for LingPipe (Named-Entity Extraction)
 * ({@link http://www.alias-i.com/lingpipe})
 *
 * @author <a href=mailto:oeze@dfki.de>Oezguer Demir</a>
 * @version $Revision: 1.3 $
 * $Id: LingPipeModule.java,v 1.3 2005/04/26 17:37:36 oeze Exp $
 */
public class LingPipeModule
    extends Module {

  static Logger log = Logger.getLogger(LingPipeModule.class);
  private TransformationService transformationService = null;
  private String sTylesheet4rmrs = null;
  private Transformer lingpipe2rmrsTransformer = null;
  private String sTylesheet4pic = null;
  private Transformer lingpipe2picTransformer = null;
  private String sPicOutputAnnotationName = null;
  private String sRawOutputAnnotationName = null;

  public LingPipeModule(Properties moduleProps) throws
      InitializationException {
    log.info("Initializing LingPipe...");
    try {
      log.debug(Utility.map2String(moduleProps));
      super.init(moduleProps);
    }
    catch (InitializationException ie) {
      log.fatal(Utility.exception2String(ie));
    }
    sPicOutputAnnotationName = getConfig().getProperty("lingpipe.pixmlannotation", null);
    sRawOutputAnnotationName = getConfig().getProperty("lingpipe.rawannotation", "SproutFS");
    sTylesheet4pic = getConfig().getProperty("lingpipe.stylesheet4pic");
    sTylesheet4rmrs = getConfig().getProperty("lingpipe.stylesheet4rmrs");
    log.info("LingPipe initialized.");
  }

  public void shutdown() {
    log.info("shutting down LingPipe.");
    // nothing to do here
  }

  /* (non-Javadoc)
   * @see de.usaar.coli.deepthought.mocoman.Module#process(java.lang.String)
   */
  public String process(String processString) {
    MoCoMan mocoman = getMoCoMan();
    String rawtext =
        Utility.getComputedAnnotation(mocoman, processString, "rawtext");
    Metadata md = new Metadata(processString);
    String sessionID = md.getSessionID();
    String acID = md.getAcID();
    log.info("input text: \"" + rawtext + "\"");
    long t1 = System.currentTimeMillis();
    String as = runLingPipeNE(rawtext);
    String duration = Utility.formatDuration(System.currentTimeMillis() - t1);
    String root = getModuleRootElement();
    log.info("processing time of LingPipe: " + duration);
    log.info("creating LingPipe XML output");

    String msg = "OK";
    md = new Metadata(sessionID, acID, duration,
            getModuleName(), msg, false, getConfig());
    StringBuffer sbResult = new StringBuffer();
    sbResult.append("<?xml version='1.0'?>");
    sbResult.append("<" + root + ">");
    sbResult.append(md.getXmlString());
    sbResult.append(as);
    sbResult.append("</" + root + ">");
    String sXML = sbResult.toString();
    log.debug(sXML);
    // convert to RMRS and PET XML input chart format, save raw and piXML as
    // additional, separate annotation, return RMRS.
    String rmrs = transformLP(sXML, md);
    //sXML = cleanXMLString(sbResult);
    return rmrs;
  }

  private final String runLingPipeNE(String input) {
    final double PRUNING_THRESHOLD = 16.0;
    final String PLAIN_TEXT_TOP_LEVEL_ELEMENT = "DOCUMENT";
    String output = "";
    Decoder newsDecoder = null;
    MentionFactory mentionFactory = new EnglishMentionFactory();
    File modelDirectory =
        new File(getConfig().getProperty("lingpipe.modelsdir"));
    File newsModelFile =
        new File(modelDirectory, getConfig().getProperty("lingpipe.modelfile"));
    TokenCategorizer tokenCategorizer = new IndoEuropeanTokenCategorizer();
    TokenizerFactory tokenizerFactory = new IndoEuropeanTokenizerFactory();
    NEDictionary pronounDictionary = new EnglishPronounDictionary();
    SentenceModel sentenceModel = new IndoEuropeanSentenceModel();
    String[] elementsToAnnotate = new String[] {
        PLAIN_TEXT_TOP_LEVEL_ELEMENT
    };
    SAXFilterHandler firstFilter = null;
    SAXFilterHandler lastFilter = null;

    try {
      newsDecoder =
          new Decoder(newsModelFile, tokenCategorizer, PRUNING_THRESHOLD);
    }
    catch (IOException e) {
      log.error("Could not find news model file = " + newsModelFile);
    }
    try {

      WithinDocCorefAnnotateFilter corefFilter
          = new WithinDocCorefAnnotateFilter(mentionFactory);

      NEAnnotateFilter neAnnotateFilter
          = new NEAnnotateFilter(newsDecoder,
                                 tokenizerFactory,
                                 tokenCategorizer);

      neAnnotateFilter.setPronounDictionary(pronounDictionary);
      neAnnotateFilter.setHandler(corefFilter);

      SentenceAnnotateFilter sentenceFilter
          = new SentenceAnnotateFilter(sentenceModel,
                                       tokenizerFactory,
                                       elementsToAnnotate);

      sentenceFilter.setHandler(new GroupCharactersFilter(
          neAnnotateFilter));

      lastFilter = corefFilter;
      firstFilter = sentenceFilter;

      Xml2StringHandler htmlHandler = new Xml2StringHandler();
      lastFilter.setHandler(htmlHandler);

      char[] inputChars = input.toCharArray();

      firstFilter.startDocument();
      firstFilter.startSimpleElement(PLAIN_TEXT_TOP_LEVEL_ELEMENT);
      firstFilter.characters(inputChars, 0, inputChars.length);
      firstFilter.endSimpleElement(PLAIN_TEXT_TOP_LEVEL_ELEMENT);
      firstFilter.endDocument();

      output = htmlHandler.getString();

    }
    catch (SAXException e) {
      log.debug("SAX Exception during processing=" + e);
    }
    log.debug(output);
    return output;
  }

  String transformLP (String output, Metadata metadata) {
  	String rmrs = null;
  	String piXml = null;
  	if (this.transformationService==null) {
  		this.transformationService = new TransformationService(getMoCoMan());
  		try {
  			lingpipe2rmrsTransformer = this.transformationService.createTransformer(sTylesheet4rmrs);
  			lingpipe2rmrsTransformer.setParameter("encoding", "UTF-8"); // we stay with Java Unicode
  			if (sPicOutputAnnotationName!=null)
  				lingpipe2picTransformer = this.transformationService.createTransformer(sTylesheet4pic);
  		} catch (TransformerConfigurationException tce) {
  			log.error(Utility.exception2String(tce));
  		}
  	}
  	// saving raw lingpipe output as separate annotation
  	getMoCoMan().createAnnotation(metadata.getSessionID(), metadata.getAcID(), output, getModuleLanguage(), sRawOutputAnnotationName);
    try {
    	log.info("transforming LingPipe output to RMRS");
        rmrs = this.transformationService.transform(lingpipe2rmrsTransformer, output);
        if (sPicOutputAnnotationName!=null) {
            log.info("transforming LingPipe output to Pet XML Input Chart Format");
            piXml = this.transformationService.transform(lingpipe2picTransformer, output);
            getMoCoMan().createAnnotation(metadata.getSessionID(), metadata.getAcID(), piXml, getModuleLanguage(), sPicOutputAnnotationName);
        }
    } catch (TransformerException te) {
    	log.error(Utility.exception2String(te));
    	return generateErrorResult(metadata.getSessionID(), metadata.getAcID(), metadata.getProcTime(), getModuleName(), "Error: " + te.getMessage());
    }
    return rmrs;
  }

  public String generateErrorResult( String sessionID, String acID, String duration, String module, String diagnosis) {
    return "<" + getModuleRootElement() + ">" +
           Metadata.generateXml(sessionID, acID, duration, module,
                                diagnosis, true, getConfig()) +
        "</" + getModuleRootElement() + ">";
}

  private static class Xml2StringHandler
      extends SimpleElementHandler {

    private final StringBuffer outBuffer = new StringBuffer();
    private String typeID;
    private int offsetCount = 0;
    private boolean inNE = false;

    public Xml2StringHandler() {}

    public void startElement(String domain, String localName,
                             String qName, Attributes atts) {

      if (qName.equals(NEAnnotateFilter.DEFAULT_ENTITY_ELEMENT)) {
        String entityType =
            atts.getValue(NEAnnotateFilter.DEFAULT_ENTITY_TYPE_ATTRIBUTE);
        String entityID =
            atts.getValue(WithinDocCorefAnnotateFilter.ID_ATTRIBUTE);
        typeID = entityType + "[ID=" + entityID + "]";
        outBuffer.append("<NE id=\"" + entityID + "\" "
                         + "type=\"" + entityType + "\" "
                         + "cstart=\"" + offsetCount + "\" "
                         + "string=\"");
        inNE = true;
        //tmpLength = 0;
      }
    }

    public void endElement(String domain, String localName,
                           String qName) {
      if (qName.equals("ENAMEX")) {
        outBuffer.append("\" cend=\"" + (offsetCount - 1) + "\"/>" +
                         System.getProperty("line.separator"));
        inNE = false;
      }
    }

    public void characters(char[] cs, int start, int length) {
      offsetCount = offsetCount + length;
      if (inNE) {
        outBuffer.append(cs, start, length);
      }
    }

    public String getString() {
      return outBuffer.toString();
    }
  }

  /* (non-Javadoc)
   * @see de.usaar.coli.deepthought.mocoman.Module#processDOMTree(org.w3c.dom.Node)
   */
  public Node processDOMTree(Node domTree) {
    // TODO Auto-generated method stub
    // tokenize as in process(), but use XMLOutputter.createXMLDocument() for output
    // the bad thing: JTok then outputs JDOM, but (directly) no JAXP DOM
    // so, use the jdomDOMoutputter:
    // org.jdom.output.DOMOutputter.output(org.jdom.Document) -> w3c-DOM
    // i.e., copy the whole process() and modifiy the two lines that generate XML Strings
    return null;
  }

}
