/*
 * LoParModule.java
 *
 * (C) 2007 DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.LinkedList;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Logger;
import org.w3c.dom.Node;

import de.dfki.lt.hog.Metadata;
import de.dfki.lt.hog.MoCoMan;
import de.dfki.lt.hog.Module;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;
import de.dfki.lt.hog.TransformationService;

/**
 * Module definition for LoPar ({@link http://www.ims.uni-stuttgart.de/projekte/gramotron/SOFTWARE/LoPar.html})
 * 
 * @author <a href=mailto:daniel.contag@dfki.de>Daniel Contag</a>
 * @version $ $Id: LoParModule.java,v 1.7 2007/07/16 13:48:27 dcontag Exp $
 */
public class LoParModule extends Module {

    static Logger log = Logger.getLogger(LoParModule.class);

    private TransformationService transformationService = null;

    private Transformer transformer = null;

    private Transformer tnt2Lopar = null;

    private Transformer attr2Element = null;

    private Transformer flattenTree = null;

    private Transformer topo2Brackets = null;

    private Transformer prettyPrint = null;

    private Transformer extractFields = null;

    private Transformer extractText4Chunkie = null;

    private Transformer extractChunks = null;

    private Transformer deletePunctuation = null;

    private Transformer topo2Brackets_chunks = null;

    private String styleSheet = null;

    private Pattern p = Pattern.compile("[\\W&&[^	öäüÄÖÜß-]]");

    private String sId = "(<W id=\"[A-Z]\">)";

    private Pattern id = Pattern.compile(sId);

    private String ncfieldPattern = "(<NCFIELD id=\"\\d\" delim=\"start\"/>){1}?.*?(<NCFIELD delim=\"end\"/>){1}?";

    // 32 is flag DOTALL - the dot matches also newlines
    private Pattern ncPattern = Pattern.compile(ncfieldPattern, 32);

    private String linePatternStr = "<w (.*?)>(.*?)</w>";

    private Pattern linePattern = Pattern.compile(linePatternStr);

    private Runtime runtime;

    private String loparBinary = null;

    private String loparOptions = null;

    private String loparModel = null;

    private String tmpDir = null;

    private String annoNameTopoTree = null;

    private String annoNameTopoFlat = null;

    private String annoNameTopoChunks = null;

    private Process process;

    private BufferedReader procOut;

    private BufferedWriter procIn;

    private BufferedReader procErr;

    private String outputEncoding = null;

    private String inputEncoding = null;

    private String loparXslDir = null;

    private String rootDir = null;

    private int chunkieModuleDepth = 0;

    private static final String NEWLINE = System.getProperty("line.separator");

    private String loparResult = null;

    private String sPostprocStylesheet = null;

    public LoParModule(Properties loparProps) {
        try {
            super.init(loparProps);
        } catch (InitializationException ie) {
            log.error(Utility.exception2String(ie));
        }

        rootDir = System.getProperty("root.dir");

        // read configuration properties
        loparBinary = getConfig().getProperty("lopar.binary");
        loparOptions = getConfig().getProperty("lopar.options");
        loparModel = getConfig().getProperty("lopar.model");
        loparXslDir = getConfig().getProperty("lopar.xsldir");
        outputEncoding = getConfig().getProperty("lopar.outputencoding");
        inputEncoding = getConfig().getProperty("lopar.inputencoding");
        tmpDir = getConfig().getProperty("lopar.tmpDir");
        annoNameTopoTree = getConfig().getProperty(
                "lopar.topotree.annotationname");
        annoNameTopoFlat = getConfig().getProperty(
                "lopar.topoflat.annotationname");
        annoNameTopoChunks = getConfig().getProperty(
                "lopar.topochunks.annotationname");
        chunkieModuleDepth = Integer.parseInt(getConfig().getProperty(
                "chunkie.moduleDepth"));
        sPostprocStylesheet=getConfig().getProperty("lopar.postprocstylesheet");
        // start lopar-module-infrastructure
        start();

    }

    /**
     * Starts the lopar binary and takes care of creating input and output,
     * compiles the transformers
     */
    public void start() {

        log.info("Initializing LoPar ...");
        runtime = Runtime.getRuntime();

        String sname = loparBinary + " " + loparModel + " " + loparOptions;
        log.debug("Invoking " + sname);

        try {
            process = runtime.exec(sname);
            procOut = new BufferedReader(new InputStreamReader(process
                    .getInputStream(), outputEncoding));
            procIn = new BufferedWriter(new OutputStreamWriter(process
                    .getOutputStream(), inputEncoding));
            procErr = new BufferedReader(new InputStreamReader(process
                    .getErrorStream(), outputEncoding));
        } catch (IOException ioe) {
            log.fatal(Utility.exception2String(ioe));
        }

        styleSheet = "xsl/preproc/tnt2lopar.xsl";
        tnt2Lopar = compileTransformer(styleSheet);

        styleSheet = loparXslDir + "attr2element.xsl";
        attr2Element = compileTransformer(styleSheet);

        styleSheet = loparXslDir + "flatten_tree.xsl";
        flattenTree = compileTransformer(styleSheet);

        styleSheet = loparXslDir + "topo2brackets.xsl";
        topo2Brackets = compileTransformer(styleSheet);

        styleSheet = loparXslDir + "prettyprint.xsl";
        prettyPrint = compileTransformer(styleSheet);

        styleSheet = loparXslDir + "extractfields.xsl";
        extractFields = compileTransformer(styleSheet);

        styleSheet = loparXslDir + "extracttext4chunkie.xsl";
        extractText4Chunkie = compileTransformer(styleSheet);

        styleSheet = loparXslDir + "extractChunks.xsl";
        extractChunks = compileTransformer(styleSheet);

        styleSheet = loparXslDir + "delete-punctuation.xsl";
        deletePunctuation = compileTransformer(styleSheet);

        styleSheet = sPostprocStylesheet;
        if (styleSheet != null) 
	    topo2Brackets_chunks = compileTransformer(styleSheet);
      
    }


    /*
     * (non-Javadoc)
     * 
     * @see de.dfki.lt.hog.Module#process(java.lang.String)
     */
    public String process(String processString) {
        MoCoMan mocoman = getMoCoMan();
        Metadata md = new Metadata(processString);
        String sessionID = md.getSessionID();
        String acID = md.getAcID();
        long t1 = System.currentTimeMillis();

        // get tnt result
        String tntResult = mocoman.getAnnotation(sessionID, acID, "TnT-ToPo");
        log.debug("TnT-Annotation: " + tntResult);

        // transform tnt-xml to text-input format for lopar
        String loparInput = transform(tnt2Lopar, tntResult);

        log.debug("lopar-input:" + NEWLINE + loparInput);

        // remove punctuation
        StringBuilder sb = new StringBuilder();
        Matcher m;

        for (String line : loparInput.split(NEWLINE)) {
            m = p.matcher(line);
            if (!m.find()) {
                sb.append(line);
                sb.append(NEWLINE);
            }
        }
        loparInput = sb.toString();
        log.debug("cleaned lopar-input:" + NEWLINE + loparInput);

        // feed the text to LoPar
        try {
            procIn.write(loparInput);
            procIn.flush();
            procIn.newLine();
            procIn.flush();

            // get result from LoPar
            loparResult = procOut.readLine();

            log.debug("lopar-result: " + loparResult);

        } catch (IOException ioe) {
            log.fatal(Utility.exception2String(ioe));
        }

        // convert lopar output to raw xml
        String rawXml = lopar2RawXml(loparResult);
        log.debug("topo2xml-result:" + NEWLINE + rawXml);

        // convert raw xml to xml by numbering the indexwords
        String xml = raw2Xml(rawXml);
        log.debug("topo-raw to topo-xml:" + NEWLINE + xml);

        // rename NODE to CL etc.
        String xmlTopoTree = transform(attr2Element, xml);
        log.debug("xmlTopoTree:" + NEWLINE + xmlTopoTree);

        // create annotation for debugging
        String duration = Utility.formatDuration(System.currentTimeMillis()
                - t1);
        // create metadata-object to be reused
        Metadata topoMd = new Metadata(sessionID, acID, duration,
                annoNameTopoTree, "OK", false, getConfig());
        String strMd = createMetaDataXmlString(sessionID, acID, xmlTopoTree, topoMd);
        if (annoNameTopoTree != null) {
            mocoman.createAnnotation(sessionID, acID, strMd,
                    getModuleLanguage(), annoNameTopoTree);
        }

        // flatten binary topo tree
        String xmlTopoFlat = transform(flattenTree, xmlTopoTree);
        log.debug("xmlTopoFlat:" + NEWLINE + xmlTopoFlat);

        // create annotation for debugging
        duration = Utility.formatDuration(System.currentTimeMillis() - t1);
        topoMd.setProcTime(duration);
        topoMd.setComponent(annoNameTopoFlat);
        strMd = createMetaDataXmlString(sessionID, acID, xmlTopoFlat, topoMd);
        if (annoNameTopoFlat != null) {
            mocoman.createAnnotation(sessionID, acID, strMd,
                    getModuleLanguage(), annoNameTopoFlat);
        }

        // write bracket spans
        String tmp = transform(topo2Brackets, xmlTopoFlat);
        String bracketsTopoXml = transform(prettyPrint, tmp);

        log.debug("bracket spans:" + NEWLINE + bracketsTopoXml);

        // extract fields
        String skeleton = transform(extractFields, xmlTopoFlat);

        log.debug("extracted fields skeleton:" + NEWLINE + skeleton);

        // split ncfields needed for chunkie
        String text4Chunkie = transform(extractText4Chunkie, skeleton);

        // text4Chunkie contains contents of NCFIELDS separated by newline
        log.debug("text for chunkie:" + NEWLINE + text4Chunkie);

        // let chunkie analyse the seperate chunks
        String topoChunkie = doChunkieAnalysis(text4Chunkie, skeleton, mocoman,
                sessionID);
	deletePunctuation.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,
						   "yes");

        topoChunkie = transform(deletePunctuation, topoChunkie);
        log.debug("topoChunkie:" + NEWLINE + topoChunkie);

        // renumber tokens
        String xmlTopoChunks = renumberTokens(topoChunkie);
        log.debug("xmlTopoChunks:" + NEWLINE + xmlTopoChunks);

        // create annotation for debugging
        duration = Utility.formatDuration(System.currentTimeMillis() - t1);
        topoMd.setProcTime(duration);
        topoMd.setComponent(annoNameTopoChunks);
        strMd = createMetaDataXmlString(sessionID, acID, xmlTopoChunks, topoMd);
        if (annoNameTopoChunks != null) {
            mocoman.createAnnotation(sessionID, acID, strMd,
                    getModuleLanguage(), annoNameTopoChunks);
        }

	String xmlTopoBrackets = "";

	if (topo2Brackets_chunks!=null) {
	    // extract brackets
	    tmp = transform(topo2Brackets_chunks, xmlTopoChunks);
	    log.debug("topo2brackets_chunks:" + NEWLINE + tmp);
	    
	    String bracketsChunks = transform(prettyPrint, tmp);
	    log.debug("brackets_chunks:" + NEWLINE + bracketsChunks);
	    
	    // workaround using temp-file
	    // write contents of bracketsTopoXml to a tempfile
	    String path = rootDir + "/" + tmpDir + "/temp.xml";
	    createTempFile(path, bracketsTopoXml);
	    
	    // merge brackets
	    styleSheet = loparXslDir + "mergebrackets.xsl";
	    
	    try {
		transformer = this.transformationService
                    .createTransformer(styleSheet);
		this.transformer.setParameter("URI", path);
	    } catch (TransformerConfigurationException tce) {
		log.error(Utility.exception2String(tce));
	    }
	    
	    try {
		tmp = this.transformationService.transform(transformer,
							   bracketsChunks);
	    } catch (TransformerException te) {
		log.error(Utility.exception2String(te));
	    }
	    log.debug("tmp:" + NEWLINE + tmp);
	    
	    // renumber brackets
	    styleSheet = loparXslDir + "renumberbrackets.xsl";

	    try {
		transformer = this.transformationService
                    .createTransformer(styleSheet);
		// see topogui.TopoLogic: script 'topomerge' is always started with
		// this parameter set to 0
		this.transformer.setParameter("sentenceno", "0");
		this.transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,
						   "yes");
	    } catch (TransformerConfigurationException tce) {
		log.error(Utility.exception2String(tce));
	    }
	    try {
		xmlTopoBrackets = this.transformationService.transform(transformer,
								       tmp);
	    } catch (TransformerException te) {
		log.error(Utility.exception2String(te));
	    }
	    
	    log.debug("xmlTopoBrackets:" + NEWLINE + xmlTopoBrackets);
	}
	    
        duration = Utility.formatDuration(System.currentTimeMillis() - t1);
        String root = getModuleRootElement();
        log.info("processing time of LoParModule: " + duration);
        log.info("creating ToPo-LoPar XML output");

        String msg = "OK";
        md = new Metadata(sessionID, acID, duration, getModuleName(), msg,
                false, getConfig());
        StringBuffer sbResult = new StringBuffer();
        sbResult.append("<?xml version='1.0'?>");
        sbResult.append("<" + root + ">");
        sbResult.append(md.getXmlString());

	if (topo2Brackets_chunks==null)
	    sbResult.append(xmlTopoChunks);
	else
	    sbResult.append(xmlTopoBrackets);

        sbResult.append("</" + root + ">");
        String output = sbResult.toString();

        log.debug(output);
        return output;

    }

    /**
     * This renumbers the lowercase <w> tags and gives them an id.
     * 
     * @param topoChunkie -
     *            xml input string with lowercase <w>-tags
     * @return - xml-string with renumbered and uppercase <W>-tags
     */
    private String renumberTokens(String topoChunkie) {
        // in order to be able to extract the brackets we need to number
        // <w id="Wn">word</w> again - n runs from 0 to number of tokens
        Matcher lineMatcher = linePattern.matcher(topoChunkie);
        int i = 0;

        while (lineMatcher.find()) {
            topoChunkie = lineMatcher.replaceFirst("<W id=\"W" + i
                    + "\" $1\\>$2</W>");
            lineMatcher = linePattern.matcher(topoChunkie);
            i++;
        }
        return topoChunkie;
    }

    /**
     * This does chunkie analysis per chunk and puts the results in the NCFIELDS
     * of the extracted fields from topoFlat.
     * 
     * @param text4Chunkie -
     *            chunkie input text
     * @param skeleton -
     *            extracted fields from topoFlat
     * @param mocoman -
     *            mocomanserver
     * @param sessionID -
     *            id of the session we are using
     * @return String skeleton with all NCFIELDS contents replaced by chunkie
     *         analysis
     */
    private String doChunkieAnalysis(String text4Chunkie, String skeleton,
            MoCoMan mocoman, String sessionID) {
        // get chunkie inputs separated in a list
        LinkedList<String> chunkieInputList = new LinkedList<String>();
        for (String line : text4Chunkie.split(NEWLINE)) {
            chunkieInputList.add(line);
        }

        // get the analyses from chunkie
        LinkedList<String> chunkieOutputList = new LinkedList<String>();
        
        for (String line : chunkieInputList) {
            
            String annCollId = mocoman.createAnnotationCollection(sessionID);
            log.debug("annotation collection id: " + annCollId);
            
            String annotationId = mocoman.createInitialAnnotation(sessionID,
                    annCollId, line, getModuleLanguage(), getModuleName());
            log.debug("annotation id: " + annotationId);
            
            String chunkResult = mocoman.analyse(sessionID, annCollId,
                    "xmltext", 0, 0, chunkieModuleDepth);
            
            chunkieOutputList.add(chunkResult);
        }
        log.debug("chunkie-analysis-results:" + NEWLINE + chunkieOutputList);

        // now remove all unneeded xml from the chunks (metadata from hog etc.)
        LinkedList<String> chunksList = new LinkedList<String>();

        for (String line : chunkieOutputList) {
            String outChunk = transform(extractChunks, line);

            // remove xml-header
            outChunk = outChunk.replaceAll(".*xml.*", "");

            // the word tags which are not in ncfields are lowercase,
            // this is problematic for renumbering later,
            // so we put them in capital letters
            outChunk = outChunk
                    .replaceAll("<W (.*?)>(.*?)</W>", "<w $1>$2</w>");
            chunksList.add(outChunk + NEWLINE);
            log.debug("single extracted chunk:" + NEWLINE + outChunk + NEWLINE);
        }

        // now replace all NCFIELDS in skeleton with the chunks from chunkie
        Matcher ncMatcher = ncPattern.matcher(skeleton);
        String topoChunkie = "";

        while (ncMatcher.find()) {
            // insert the chunks into the ncfields and delete them by polling
            // from the list
            topoChunkie = ncMatcher.replaceFirst(chunksList.poll());
            ncMatcher = ncPattern.matcher(topoChunkie);
        }

        return topoChunkie;
    }

    /**
     * Creates a tempfile for merging via stylesheets, workaround
     * 
     * @param path -
     *            input file path
     * @param input -
     *            input xml-string
     */
    private void createTempFile(String path, String input) {
        /*
         * does not work at the moment, workaround with tempfile // needed for
         * building the $URI parameter for the stylesheet's // "Document($URI)"
         * mocoman.createAnnotation(sessionID, acID, bracketsTopoXml,
         * getModuleLanguage(), "topo_brackets"); String aid = "topo_brackets";
         * String uri = "hog://" + sessionID + "/" + acID + "/" + aid;
         */

        PrintWriter pw;
        try {
            File file = new File(path);
            file.createNewFile();

            pw = new PrintWriter(file);
            pw.println(input);
            pw.close();

        } catch (IOException e) {
        }

    }

    /**
     * Converts the topo-raw.xml to topo.xml by numbering the indexwords
     * 
     * @param input -
     *            an input string
     * @return the correctly numbered output string
     */
    private String raw2Xml(String input) {
        String output = "";

        Matcher m = id.matcher(input);
        int i = 0;

        while (m.find()) {
            output = m.replaceFirst("<W id=\"W" + i + "\"\\>");
            m = id.matcher(output);
            i++;
        }

        return output;
    }

    /**
     * modifies the processing time of the metadata object and returns a
     * suitable xml-string for createannotation
     * 
     * @param sessionId -
     *            Session ID
     * @param annCollId -
     *            Annotation collection id
     * @param rawtext -
     *            result
     * @param md -
     *            metadata-object
     * @return - suitable string for createannotation
     */
    private String createMetaDataXmlString(String sessionId, String annCollId,
            String rawtext, Metadata md) {
        // prepare metadata
        StringBuffer xmltext = new StringBuffer();
        xmltext.append("<?xml version='1.0'?>");
        xmltext.append("<" + getModuleName() + ">");

        xmltext.append(md.getXmlString());

        // remove xml-header
        rawtext = rawtext.replaceFirst(".*xml.*", "");
        xmltext.append(rawtext);

        xmltext.append("</" + getModuleName() + ">");

        return xmltext.toString();
    }

    /**
     * Converts lopar output to a 'raw xml'-format
     * 
     * @param input -
     *            the lopar output string in bracket form
     * @return - a 'xmlified' string
     */
    private String lopar2RawXml(String input) {
        StringBuilder rawXml = new StringBuilder();

        String xmlHeader = "<?xml version=\"1.0\"?>\n";
        rawXml.append(xmlHeader);

        String embedCatNode = "\\(([^\\ ]+)\\ (?=\\()";
        String newNode = "<NODE cat=$1>\n";
        String tmp = input.replaceAll(embedCatNode, newNode);

        String lexicalCatNode = "\\(([^\\)]+)\\ \\\"([^\\)]+)\\\"\\)";
        String embeddedCatNode = "<NODE\\ cat=$1\\>\\ \\<W id=\\\"W\\\"\\>$2\\<\\/W\\>\n\\<\\/NODE\\>\n";
        tmp = tmp.replaceAll(lexicalCatNode, embeddedCatNode);

        String closingBracket = "\\)";
        String closingNode = "</NODE>\n";
        tmp = tmp.replaceAll(closingBracket, closingNode);

        rawXml.append(tmp);
        rawXml.append(NEWLINE);

        return rawXml.toString();
    }

    /**
     * Compiles a transformer with a given stylesheet
     * 
     * @param aStyleSheet -
     *            an input stylesheet
     * @return - returns the compiled transformer
     */
    private Transformer compileTransformer(String aStyleSheet) {
        if (this.transformationService == null) {
            this.transformationService = new TransformationService(getMoCoMan());
        }

        try {
            transformer = this.transformationService
                    .createTransformer(aStyleSheet);
        } catch (TransformerConfigurationException tce) {
            log.error(Utility.exception2String(tce));
        }

        return transformer;
    }

    /**
     * Takes a compiled transformer and an input string and returns the
     * transformed string
     * 
     * @param aTransformer -
     *            compiled transformer
     * @param anInputString -
     *            xml input string
     * @return - transformed output string
     */
    private String transform(Transformer aTransformer, String anInputString) {
        String output = "";

        if (this.transformationService == null) {
            this.transformationService = new TransformationService(getMoCoMan());
        }

        try {
            output = this.transformationService.transform(aTransformer,
                    anInputString);
        } catch (TransformerException te) {
            log.error(Utility.exception2String(te));
        }
        return output;
    }

    public Node processDOMTree(Node domTree) {
        // Auto-generated method stub
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see de.dfki.lt.hog.Module#shutdown()
     */
    public void shutdown() {
        log.info("Shutting down the LoParModule.");
        try {
            procIn.close();
            procOut.close();
            procErr.close();
            process = null;

        } catch (IOException ioe) {
            log.fatal(Utility.exception2String(ioe));
        }
    }
}
