/*
 * PetModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import org.apache.log4j.Logger;
import org.w3c.dom.*;
import de.dfki.lt.hog.Metadata;
import de.dfki.lt.hog.MoCoMan;
import de.dfki.lt.hog.Module;
import de.dfki.lt.hog.TransformationService;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;

/**
 * Module definition for PET (cheap commandline).
 * Created 2003-11-25
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @version  $Id: PetModule.java,v 1.9 2007/11/20 16:54:14 uschaefer Exp $
 */

/* Changes:
   2004-01-29 confidence added, diagnosis in metadata
   2004-03-31 hack for preliminary PET input chart
   2004-04-30 generalised input annotation via cfg file (property pet.inputannotation)
   2004-05-18 xml header with encoding (property pet.outputencoding)
   2004-07-06 new cheap message unpack edge limit exhausted
   2004-07-21 handling of cheap -partial option (0 readings, rmrses in <rmrs-list>)
   2004-07-22 added empty result handling (metadata)
   2004-11-05 handling of cheap -results=n option (set readings to min(n, actual readings).
   2004-11-26 PET XML input chart mode added
   2007-05-24 combinepixml stylesheet defined means pixml mode
   2007-09-18 added combinesmaf support
   2007-11-20 added support for -mrs=mrx 
   */

public class PetModule extends Module {

    static Logger log = Logger.getLogger(PetModule.class);
    private boolean yymode = false;
    private boolean pixmlmode = false;
    private boolean smafmode = false;
    private BufferedReader procOut;
    private BufferedWriter procIn;
    private Runtime runtime;  // need not be a class variable
    private Process process;  // need not be a class variable
    // private static final String PET_SHUTDOWN_COMMAND = "\u0004";  // cheap seems to ignore this
    private static final String PET_MRS_MODE = "-mrs=mrx";
    private static final String PET_SHUTDOWN_COMMAND = System.getProperty("line.separator");
    private static final String PET_INIT_TERMINATOR =  "types in";
    private static final String PET_NO_LEX = "no lexicon entries for";
    private static final String PET_NO_ORTH = "no orth path";
    private static final String PET_YY_TOK = "yy_tokenizer:";
    private static final String PET_NO_SAX = "SAX:";
    private static final String PET_EDGE_LIMIT = "edge limit exhausted";
    private static final String PET_UNPACK_EDGE_LIMIT = "unpack edge limit exhausted";
    private static final String PET_NEWLINE = System.getProperty("line.separator");
    private static final String PET_READINGS_PATTERN = "\\([0-9]+?\\) `.+?' \\[[0-9]+?\\] --- [0-9]+? \\(.+?";
    private static final String PET_DERIVATION_LINE="derivation[";
    private static final String PET_YY_MODE1 = "-yy";
    private static final String PET_YY_MODE2 = "-tok=yy_counts";
    private static final String PET_PARTIAL_MODE = "-partial";
    private static final String PET_RESULTS_OPTION = "-results=";

    private String PET_RESULT_STARTER = "<rmrs";
    private String PET_RESULT_TERMINATOR = "</rmrs>";
    private String PET_RMRSLIST_STARTER = "<rmrs-list";
    private String PET_RMRSLIST_TERMINATOR = "</rmrs-list>";
    private int PET_RMRS_OFFSET = 1;

    private String xmlHeader;
    private String sPetRootElement = null;
    private String sPetInputAnnotationNames = null;
    private String sPetSmafPrimaryAnnotationName = null;
    private String sPetOptions = null;
    private int iPetPrintResults = 0;
    private TransformationService transformationService = null;
    private Transformer combineInputTransformer = null;
    private Transformer preprocTransformer = null;
    private String sPetInputDTD;
    private Pattern emptylinesPattern;
    private Matcher emptylinesMatcher; 
    private Transformer postprocTransformer;
    /**
     */
    public PetModule(Properties petModuleProps) {
                try {
                  super.init(petModuleProps);
                } catch (InitializationException ie) {
                  log.error(Utility.exception2String(ie));
                }
                // read configuration properties
                // global ones:
                sPetRootElement = getConfig().getProperty("module.rootelement", "pet");
                sPetInputAnnotationNames = getConfig().getProperty("pet.inputannotation", "rawtext");
		sPetSmafPrimaryAnnotationName = getConfig().getProperty("pet.primarysmafinputannotation", "");
                sPetOptions = getConfig().getProperty("pet.options", "-mrs=rmrx");
		if (sPetOptions.indexOf(PET_MRS_MODE)>-1) {
		    PET_RESULT_STARTER = "<mrs";
		    PET_RESULT_TERMINATOR = "</mrs>";
		    PET_RMRSLIST_STARTER = "<mrs-list";
		    PET_RMRSLIST_TERMINATOR = "</mrs-list>";
                    PET_RMRS_OFFSET=0;
		}
                // for init only:
                String sPetGrammarPrefix = getConfig().getProperty("pet.grammarprefix");
                String sPetGrammarDir = getConfig().getProperty("pet.grammardir");
                String sPetBinary =  getConfig().getProperty("pet.binary", "cheap");
                String sPetLibs = getConfig().getProperty("pet.libs");
                String sPetInputEncoding = getConfig().getProperty("pet.inputencoding", "ISO-8859-1");
                String sPetOutputEncoding = getConfig().getProperty("pet.outputencoding", "ISO-8859-1");
		String dtdfile = null;
                xmlHeader = "<?xml version='1.0'?>";
                if (!getConfig().getProperty("pet.combinestylesheet","").equals("")) {
		    if (sPetSmafPrimaryAnnotationName.equals("")) {
			pixmlmode = true;
			dtdfile = getConfig().getProperty("pet.pixmldtdfile", "xsl/pic/pet-input-chart.dtd");
             		sPetInputDTD = xmlHeader + "\n<!DOCTYPE pet-input-chart [\n" +
						Utility.readStringFromFile(dtdfile, "iso-8859-1") + " ]>\n";
		    }
		    else { 
			smafmode = true;
			dtdfile = getConfig().getProperty("pet.smafdtdfile", "xsl/smaf/smaf.dtd");
             		sPetInputDTD = xmlHeader + "\n<!DOCTYPE smaf [\n" +
						Utility.readStringFromFile(dtdfile, "iso-8859-1") + " ]>\n";

		    }
		    emptylinesPattern = Pattern.compile("(\r?\n)+");
                }
                if (!sPetBinary.startsWith(System.getProperty("file.separator")))
                    sPetBinary = System.getProperty("root.dir") + System.getProperty("file.separator") + sPetBinary;
                if ((sPetOptions.indexOf(PET_YY_MODE1)>-1)||(sPetOptions.indexOf(PET_YY_MODE2)>-1)) yymode = true;
                if (!sPetLibs.startsWith(System.getProperty("file.separator")))
                    sPetLibs = System.getProperty("root.dir") + System.getProperty("file.separator") + sPetLibs;
                String[] libs = { "LD_LIBRARY_PATH=" + sPetLibs, // + ":/lib:/usr/lib:/usr/local/lib"
                                  "PATH=." };  // + ":/bin:/usr/bin:/usr/local/bin"
                File fPetGrammarDir = new File(sPetGrammarDir);
                if (!fPetGrammarDir.exists()||!fPetGrammarDir.isDirectory()) {
                    log.fatal("PET grammar dir " + sPetGrammarDir + " does not exist!");
                }
                int proPos = sPetOptions.indexOf(PET_RESULTS_OPTION);
                if (proPos>-1) {
                	StringTokenizer tk = new StringTokenizer(sPetOptions.substring(proPos + PET_RESULTS_OPTION.length()));
                	if (tk.hasMoreTokens()) {
                		 String sPetPrintResults = tk.nextToken();
                		 try {
                		 	iPetPrintResults = Integer.parseInt(sPetPrintResults);
                		 } catch (NumberFormatException nfe) {
                		 	log.error("Value of cheap option " + PET_RESULTS_OPTION + sPetPrintResults + " could not be interpreted.");
                		 }
                	}
                }
                log.info("Initializing PET...");
                runtime = Runtime.getRuntime();
                try {
                    process = runtime.exec(sPetBinary + " " + sPetOptions + " " + sPetGrammarPrefix,
                                           libs,
                                           fPetGrammarDir
                                           );
                    log.debug("Invoked: " + sPetBinary + " " + sPetOptions + " " + sPetGrammarPrefix);
                    String lineIn;
                    procOut = new BufferedReader(
                                  new InputStreamReader(process.getErrorStream(),
                                                        sPetOutputEncoding));
                    procIn = new BufferedWriter(
                                 new OutputStreamWriter(process.getOutputStream(),
                                                        sPetInputEncoding));
                    while ((lineIn=procOut.readLine())!=null) {
                        log.debug("PET: " + lineIn);
                        if (lineIn.indexOf(PET_NO_ORTH)!=-1) throw new IOException(lineIn);
                        if (lineIn.indexOf(PET_INIT_TERMINATOR)!=-1) break;
                    }
                    log.info("PET initialized.");

                } catch (IOException ioe) {
                    log.fatal(Utility.exception2String(ioe));
                }

        }

        public void shutdown() {
            String lineIn;
            log.info("Shutting down PET.");
            try {
                procIn.write(PET_SHUTDOWN_COMMAND + PET_NEWLINE);
                procIn.flush();
                process = null;
            } catch (IOException ioe) {
                log.fatal(Utility.exception2String(ioe));
            }
        }

        /* (non-Javadoc)
         * @see de.dfki.lt.hog.Module#process(java.lang.String)
         */
        public String process(String processString) {
            MoCoMan mocoman = getMoCoMan();
            Metadata md = new Metadata(processString);
            String sessionID = md.getSessionID();
            String acID = md.getAcID();
            String inputAnnotation;
            String duration = "00:00,000";
            if (this.transformationService==null) {
            	this.transformationService = new TransformationService(getMoCoMan());
            	// build PET XML input chart combining stylesheet 
            	try {
            		combineInputTransformer = this.transformationService.createTransformer(getConfig().getProperty("pet.combinestylesheet", "xsl/pic/combinepixml.xsl"));
            	} catch (TransformerConfigurationException tce) {
            		log.error(Utility.exception2String(tce));
            	}
            	// build pre-processor stylesheet
            	String sPreProcStylesheet = getConfig().getProperty("pet.preprocstylesheet");
            	if (sPreProcStylesheet != null) {
          			try {
        				preprocTransformer = this.transformationService.createTransformer(sPreProcStylesheet);
                  	} catch (TransformerConfigurationException tce) {
                  		log.error(Utility.exception2String(tce));
                  	}
            	}
            	// build partial mode post-processor stylesheet
            	if ((sPetOptions.indexOf(PET_PARTIAL_MODE)!=-1) && (sPetOptions.indexOf(PET_MRS_MODE)==-1)) {
            		String sTylesheet = getConfig().getProperty("pet.postprocstylesheet");
            		String sParam = getConfig().getProperty("pet.postprocfragments");
            		if ((sTylesheet != null) && (sParam != null)) {
            			try {
            				postprocTransformer = this.transformationService.createTransformer(sTylesheet);
            				postprocTransformer.setParameter("fragments", sParam);
                      	} catch (TransformerConfigurationException tce) {
                      		log.error(Utility.exception2String(tce));
                      	}
            		}
            	}
            }
            long t1 = System.currentTimeMillis();
            if (pixmlmode) {
                // PET XML input chart mode: collate input annotations in pet input chart format
                StringBuffer sbUriList = new StringBuffer();
                String sHogUriPrefix = "hog://" + sessionID + "/" + acID + "/";
                StringTokenizer stan = new StringTokenizer(sPetInputAnnotationNames, " ,:\t+");
                while (stan.hasMoreTokens()) { // more than one input annotation
                		String anno = stan.nextToken();
                		if (mocoman.getAnnotation(sessionID, acID, anno)!=null) {
                			sbUriList.append(sHogUriPrefix).append(anno);
                			if (stan.hasMoreTokens()) sbUriList.append(",");
                		}
                }
                log.debug("URI list for combine stylesheet: " + sbUriList.toString());
                combineInputTransformer.setParameter("urilist", sbUriList.toString());
                try {
                	inputAnnotation = 
						this.transformationService.transform(combineInputTransformer,
                									         "<pet-input-chart/>");
                	
                } catch (TransformerException te) {
                    log.error(Utility.exception2String(te));
                    return generateErrorResult(sessionID, acID, duration, getModuleName(), "Error: " + te.getMessage());
                }
                // run preprocessor transformer if defined
                //log.debug("annotation before preproc stylesheet: " + inputAnnotation);
                if (preprocTransformer!=null) {
                	try {
                		inputAnnotation = this.transformationService.transform(preprocTransformer, inputAnnotation);
                	} catch (TransformerException te) {
                        log.error(Utility.exception2String(te));
                        return generateErrorResult(sessionID, acID, duration, getModuleName(), "Error: " + te.getMessage());
                    }
                }
                // Add DTD and remove empty lines, add empty line at end
                inputAnnotation = sPetInputDTD + inputAnnotation;
                inputAnnotation = emptylinesPattern.matcher(inputAnnotation).replaceAll("\n");

            } else if (smafmode) {

       // SMAF input mode: collate input annotations in SMAF format
		// 1 primary + n secondary annotations
                StringBuffer sbUriList = new StringBuffer();
                String sHogUriPrefix = "hog://" + sessionID + "/" + acID + "/";
                StringTokenizer stan = new StringTokenizer(sPetInputAnnotationNames, " ,:\t+");
                while (stan.hasMoreTokens()) { // more than one input annotation
                		String anno = stan.nextToken();
                		if (mocoman.getAnnotation(sessionID, acID, anno)!=null) {
                			sbUriList.append(sHogUriPrefix).append(anno);
                			if (stan.hasMoreTokens()) sbUriList.append(",");
                		}
                }
                log.debug("URI list for combine stylesheet: " + sbUriList.toString());
                combineInputTransformer.setParameter("urilist", sbUriList.toString());
		String primarySmafAnnotation = mocoman.getAnnotation(sessionID, acID, sPetSmafPrimaryAnnotationName);
                try {
                	inputAnnotation = 
			    this.transformationService.transform(combineInputTransformer, primarySmafAnnotation);
                } catch (TransformerException te) {
                    log.error(Utility.exception2String(te));
                    return generateErrorResult(sessionID, acID, duration, getModuleName(), "Error: " + te.getMessage());
                }
                // run preprocessor transformer if defined
                //log.debug("annotation before preproc stylesheet: " + inputAnnotation);
                if (preprocTransformer!=null) {
                	try {
                		inputAnnotation = this.transformationService.transform(preprocTransformer, inputAnnotation);
                	} catch (TransformerException te) {
                        log.error(Utility.exception2String(te));
                        return generateErrorResult(sessionID, acID, duration, getModuleName(), "Error: " + te.getMessage());
                    }
                }
                // Add DTD and remove empty lines, add empty line at end
                inputAnnotation = sPetInputDTD + inputAnnotation;
                inputAnnotation = emptylinesPattern.matcher(inputAnnotation).replaceAll(" ") +  "\n";
	    }

	    else

	    {  // this should now be obsolete, XML input chart mode is strongly recommended
            	StringTokenizer stan = new StringTokenizer(sPetInputAnnotationNames, " ,:\t+");
                // default input is processString
            	if (!stan.hasMoreTokens())
            		inputAnnotation = processString;
            	else {
            		// otherwise concatenate configured input annotation strings
            		StringBuffer stbuf = new StringBuffer(mocoman.getAnnotation(sessionID, acID, stan.nextToken()));
            		while (stan.hasMoreTokens()) { // more than one input annotation
            			stbuf.append(" ").append(mocoman.getAnnotation(sessionID, acID, stan.nextToken()));
            		}
            		if (yymode) // trim is necessary for yy_tokenizer!!
            			inputAnnotation = stbuf.toString().trim();
            		else
            			inputAnnotation = stbuf.toString();
            	}
            }

            log.debug("input for PET: " + inputAnnotation);


            // process text with PET, output is RMRS (XML)
            String lineIn;
            String metadata;
            StringBuffer sbResult = new StringBuffer();
            sbResult.append(xmlHeader);

            boolean inResult = false;
            try {
                procIn.write(inputAnnotation + PET_NEWLINE);
                procIn.flush();
                int reading = 0;
                int readings = 0;
                String confidence = null;
                while ((lineIn=procOut.readLine())!=null) {
                    
                    if (lineIn.startsWith(PET_NO_LEX)
                        || lineIn.startsWith(PET_YY_TOK)
                        || lineIn.startsWith(PET_EDGE_LIMIT)
						|| lineIn.startsWith(PET_NO_SAX)
                        || lineIn.startsWith(PET_UNPACK_EDGE_LIMIT)) {
                        if (lineIn.startsWith(PET_NO_LEX)) {
                            lineIn += procOut.readLine();
                        }
                        log.warn("PET: " + lineIn);
                        return generateErrorResult(sessionID, acID, duration, getModuleName(),
                                "Error: " + lineIn);
                    }

                    if (!inResult && lineIn.matches(PET_READINGS_PATTERN)) {
                        int start = lineIn.indexOf(" --- ")+5;
                        readings = Integer.parseInt(lineIn.substring(start, lineIn.indexOf(" ",start)));
                        log.info("Number of readings found by PET: " + readings + ".");
                        duration = Utility.formatDuration(System.currentTimeMillis()-t1);
                        sbResult.append("<" + sPetRootElement + " readings=\"" + readings +"\">");

                        if (readings>0) {
                            sbResult.append(Metadata.generateXml(sessionID, acID, duration, getModuleName(), "OK", false, getConfig()));
                            if (iPetPrintResults>0) {
                            	readings = Math.min(readings,iPetPrintResults);
                            	log.info("Setting expected number of readings to " + readings + 
                            			 " because " + PET_RESULTS_OPTION + iPetPrintResults + 
										 " was specified.");
                            }
                        }
                        else {
                            // if -partial is specified, a <rmrslist> is returned containing <rmrs>es
                            StringBuffer presult = new StringBuffer();
                            boolean empty = true;
                            if (sPetOptions.indexOf(PET_PARTIAL_MODE)!=-1) {
                                if (sPetOptions.indexOf(PET_MRS_MODE)==-1) {
                                    while (((lineIn=procOut.readLine()).indexOf(PET_RMRSLIST_TERMINATOR))==-1) {
                                        if (lineIn.indexOf(PET_RMRSLIST_STARTER)==-1) {
                                            presult.append(lineIn);
                                            if (lineIn.indexOf(PET_RESULT_TERMINATOR)>-1) empty = false;
                                        }
                                    }
                                } else {
                                    while (((lineIn=procOut.readLine()).indexOf("EOM"))==-1) {
                                        presult.append(lineIn);
                                        if (lineIn.trim().length() > 0) {
                                            empty = false;
                                        }
                                    }                          
                                }
                            }
                            sbResult.append(Metadata.generateXml(sessionID, acID, duration, getModuleName(), "OK", empty, getConfig()));
                            sbResult.append(presult + "</" + sPetRootElement + ">");
                            break;
                        }
                        reading = 0;
                    }

                    if (lineIn.startsWith(PET_DERIVATION_LINE)) {
                        inResult=false;
                        // get confidence value
                        confidence = lineIn.substring(lineIn.indexOf("(")+1, lineIn.indexOf(")"));
                    }

                    if (lineIn.indexOf(PET_RESULT_STARTER)!=-1) {
                        inResult=true;
                        lineIn = PET_RESULT_STARTER + " reading=\"" + reading + "\" " +
                                 ( confidence==null ? "" : "confidence=\"" + confidence + "\" " ) +
                                 lineIn.substring(lineIn.indexOf(PET_RESULT_STARTER) +
                                                  PET_RESULT_STARTER.length() + PET_RMRS_OFFSET );
                        confidence = null;
                    }

                    if (inResult) sbResult.append(lineIn);
                    if (lineIn.indexOf(PET_RESULT_TERMINATOR)!=-1) {
                        reading++;
                        if (reading>=readings) {
                            sbResult.append("</" + sPetRootElement + ">");
                            break;
                        }
                    }
                }
                String sResult = sbResult.toString();
                // post-processing transformer for partial results:
                // return only the n longest partial RMRSes
                if ((readings==0) && (postprocTransformer!=null)) {
                	try {
                		long tpt = System.currentTimeMillis();
                		sResult = this.transformationService.transform(postprocTransformer,sResult);
                		log.info("Processing time of fragment extraction transformation: " +
                				Utility.formatDuration(System.currentTimeMillis()-tpt));
                } catch (TransformerException te) {
                    log.error(Utility.exception2String(te));
                    return generateErrorResult(sessionID, acID, duration, getModuleName(), "Error: " + te.getMessage());
                	}
                }
                
                log.debug("result: " + sResult);
                duration = Utility.formatDuration(System.currentTimeMillis()-t1);
                log.info("Processing time of PET parser: " + duration);
                return sResult;
                } catch (IOException ioe) {
                    log.fatal(Utility.exception2String(ioe));
                    return generateErrorResult(sessionID, acID, duration, getModuleName(), 
                                               "Error: " + ioe.getMessage());
                }
        }

        /* (non-Javadoc)
         * @see de.dfki.lt.hog.Module#processDOMTree(org.w3c.dom.Node)
         */
        public Node processDOMTree(Node domTree) {
                // TODO Auto-generated method stub
                // use a simple SAX parser to extract text from domTree
                // return a DOM from transformer DOMResult (i.e., copy most of process() except final DOM output)
                return null;
        }

        public String generateErrorResult( String sessionID, String acID, String duration, String module, String diagnosis) {
            return xmlHeader + "<" + getModuleRootElement() + " readings=\"0\">" +
                   Metadata.generateXml(sessionID, acID, duration, module,
                                        diagnosis, true, getConfig()) +
                "</" + getModuleRootElement() + ">";
    }
}
