/*
 * RaspModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Properties;
import java.util.Timer;
import java.util.TimerTask;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Logger;
import org.w3c.dom.Node;

import de.dfki.lt.hog.Metadata;
import de.dfki.lt.hog.MoCoMan;
import de.dfki.lt.hog.Module;
import de.dfki.lt.hog.TransformationService;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;

/**
 * Module definition for RASP2 CLI and RMRS converter server Created 2004-01-23
 * 
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @author Torsten Marek
 * @version $Id: RaspModule2.java,v 1.5 2005/09/26 11:10:25 uschaefer Exp $
 */

/*
 * Changes: 2004-07-22 : US added empty result handling (metadata) 2004-09-09 :
 * US input + outputencoding properly passed to process streams
 * 2008-05-15: ported to RASP2, is using the socket interface for RMRS conversion
 */

public class Rasp2Module extends Module {

    private static final int MAX_RASP2_RUNTIME = 30000;

    public static class WatchedProcess extends Process {
        private final Process wrappedProcess;
        private Timer timer; 
        
        public WatchedProcess(Process pWrappedProcess) {
            super();
            wrappedProcess = pWrappedProcess;
        }

        public void startGuard(int millis) {
            if (timer != null) {
                throw new IllegalStateException("Guard already in process.");
            }
            timer = new Timer();
            timer.schedule(new TimerTask() {
                @Override
                public void run() {
                    log.error("RASP2 process killed");
                    wrappedProcess.destroy();
                }
                
            }, millis);
        }
        
        public boolean stopGuard() {
            timer.cancel();
            timer = null;
            try {
                wrappedProcess.exitValue();
                return false;
            } catch (IllegalThreadStateException e) {
                return true;
            }
        }

        public boolean isGuarding() {
            return timer != null;
        }
        
        /**
         * {@inheritDoc}
         */
        @Override
        public void destroy() {
            this.wrappedProcess.destroy();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public int exitValue() {
            return wrappedProcess.exitValue();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public InputStream getErrorStream() {
            return wrappedProcess.getErrorStream();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public InputStream getInputStream() {
            return wrappedProcess.getInputStream();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public OutputStream getOutputStream() {
            return wrappedProcess.getOutputStream();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public int waitFor() throws InterruptedException {
            return wrappedProcess.waitFor();
        }
        
    }
  
    static Logger log = Logger.getLogger(Rasp2Module.class);
    private BufferedReader procOut;
    private OutputStreamWriter procIn;
    private WatchedProcess process;
    private String sRaspScript;
    private String outputEncoding;
    private String inputEncoding;
    private static final String NEWLINE = System.getProperty("line.separator");
    private static final String RASP_SHUTDOWN_COMMAND = NEWLINE + NEWLINE;
    private Socket converterSocket = null;
    private Transformer postprocTransformer = null;
    private static final Pattern OFFSET_MATCHER = Pattern.compile("[se]='(\\d+)'");
    private TransformationService transformationService;

    public Rasp2Module(Properties raspModuleProps)
            throws InitializationException {
        try {
            super.init(raspModuleProps);
        } catch (InitializationException ie) {
            log.error(Utility.exception2String(ie));
        }

        sRaspScript = getConfig().getProperty("rasp2.script");
        outputEncoding = getConfig().getProperty("rasp2.outputencoding",
                "ISO-8859-1");
        inputEncoding = getConfig().getProperty("rasp2.inputencoding",
                "ISO-8859-1");

        log.debug("sRaspScript: " + sRaspScript);
        if (!sRaspScript.startsWith(System.getProperty("file.separator")))
            sRaspScript = System.getProperty("root.dir")
                    + System.getProperty("file.separator") + sRaspScript;
        log.info("Initializing RASP...");
        createConverterConnection();
        setupRaspProcess();
        try {
            initPostTransformer();
        } catch (TransformerConfigurationException e) {
            throw new InitializationException(e.getMessage());
        }
    }

    private void initPostTransformer() throws TransformerConfigurationException {
       String sTylesheet = getConfig().getProperty("rasp2.postprocstylesheet");
        if (sTylesheet != null) {
	    try {
		transformationService = new TransformationService(MoCoMan.getInstance());
		postprocTransformer = transformationService.createTransformer(sTylesheet);
	    } catch (TransformerConfigurationException tce) {
		log.error(Utility.exception2String(tce));
		throw tce;
	    }
	}
    }

    private void setupRaspProcess() {
        try {
            log.debug("Invoking: " + sRaspScript);
            process = new WatchedProcess(Runtime.getRuntime().exec(sRaspScript));
            procOut = new BufferedReader(new InputStreamReader(process.getInputStream(), 
                    outputEncoding));
            procIn = new OutputStreamWriter(process.getOutputStream(),
                    inputEncoding);
            log.info("RASP initialized.");
        } catch (IOException ioe) {
            log.fatal(Utility.exception2String(ioe));
        }
    }

    private void createConverterConnection() {
        String converterHost = getConfig().getProperty("rasp2.converter_host");
        int converterPort = Integer.valueOf(getConfig().getProperty("rasp2.converter_port"));
        try {
            converterSocket = new Socket(converterHost, converterPort);
            log.info(String.format("Connected to converter server on %s:%d", 
                    converterHost, converterPort));
        } catch (UnknownHostException uhe) {
            log.error(Utility.exception2String(uhe));
        } catch (IOException ie) {
            log.error(Utility.exception2String(ie));
        }
    }

    /**
     * {@inheritDoc}
     */
    public void shutdown() {
        log.info("Shutting down RASP.");
        try {
            procIn.write(RASP_SHUTDOWN_COMMAND);
            procIn.flush();
            process = null;
            if (converterSocket != null) {
                converterSocket.close();
            }
        } catch (IOException ioe) {
            log.error(Utility.exception2String(ioe));
        }
    }

    /**
     * {@inheritDoc}
     */
    public String process(String processString) {
        String duration = "00:00,000";
        long t1 = System.currentTimeMillis();
        MoCoMan mocoman = getMoCoMan();
        String rawText = Utility.getComputedAnnotation(mocoman, processString, "rawtext");
        Metadata md = new Metadata(processString);
        String sessionID = md.getSessionID();
        String acID = md.getAcID();
        String lineIn;
        StringBuffer sbResult = new StringBuffer();
        try {
            rawText = rawText.replaceAll("\n", " ").replaceAll("\r", " ");
            String rt = rawText.trim();
            if (rt.equals(".") || rt.equals("")) {
                log.debug("empty input");
                return String.format("<%s>", getModuleRootElement())
                        + Metadata.generateXml(sessionID, acID, duration,
                                getModuleName(), "OK", true, getConfig())
                        + String.format("</%s>", getModuleRootElement());
            }
        
            process.startGuard(MAX_RASP2_RUNTIME);
            
            String firstLine = null;
            try {
                procIn.write(rawText + NEWLINE);
                procIn.flush();
                firstLine = procOut.readLine();
            } finally {
                if (process.isGuarding()) {
                    process.stopGuard();
                }
                if (firstLine == null) {
                    log.error(String.format("RASP2 timed out after %d seconds", 
                            MAX_RASP2_RUNTIME / 1000));
                    // in this case, restart RASP2 process
                    setupRaspProcess();
                    return String.format("<%s>", getModuleRootElement())
                           + Metadata.generateXml(sessionID, acID, duration,
                                   getModuleName(), "timeout", true, getConfig())
                           + String.format("</%s>", getModuleRootElement());
                }
            }

            log.debug("first line: " + firstLine);
            procOut.readLine(); // consume one empty line
            
            int numParses = getNumParses(firstLine);
            if (numParses == 0) {
                numParses = 1;
            }
            firstLine = fixCharacterOffsets(firstLine);
            duration = Utility.formatDuration(System.currentTimeMillis() - t1);
            sbResult.append(String.format("<%s readings=\"%d\">", 
                    getModuleRootElement(), numParses));
            sbResult.append(Metadata.generateXml(sessionID, acID,
                    duration, getModuleName(), "OK", numParses == 0, getConfig()));
            if (numParses == 0) {
                while (true) {
                    lineIn = procOut.readLine().trim();
                    if (lineIn.length() == 0) {
                        break;
                    }
                }
            } else {
                log.debug(String.format("%d parses", numParses));
                StringBuffer resBuffer = new StringBuffer(firstLine + "\n\n");
                while ((lineIn = procOut.readLine()) != null) {
                    lineIn = lineIn.trim();
                    if (lineIn.length() == 0) {
                        sbResult.append(convertRaspOutput(resBuffer.toString()));
                        
                        resBuffer = new StringBuffer(firstLine + "\n\n");
                        numParses -= 1;
                    } else {
                        resBuffer.append(fixCharacterOffsets(lineIn));
                        resBuffer.append('\n');
                    }
                    if (numParses == 0) {
                        break;
                    }
                }
            }
            // check whether result is empty
            sbResult.append("</" + getModuleRootElement() + ">");
            String sResult = sbResult.toString();
            duration = Utility.formatDuration(System.currentTimeMillis() - t1);
            log.info("Processing time of RASP parser: " + duration);
	    if (postprocTransformer==null)
		return sResult;
	    else
		return transformationService.transform(postprocTransformer, sResult);
        } catch (IOException ioe) {
            log.fatal(Utility.exception2String(ioe));
            return returnErrorDocument(duration, sessionID, acID, ioe);
        } catch (TransformerException te) {
            log.fatal(te.getMessage(), te);
            return returnErrorDocument(duration, sessionID, acID, te);
        }
    }

    private String returnErrorDocument(String duration, String sessionID,
            String acID, Exception ioe) {
        return "<"
                + getModuleRootElement()
                + ">"
                + Metadata.generateXml(sessionID, acID, duration,
                        getModuleName(), ioe.getMessage(), true,
                        getConfig()) + "</" + getModuleRootElement() + ">";
    }
    
    private String fixCharacterOffsets(String firstLine) {
        StringBuilder output = new StringBuilder();
        Matcher matcher = OFFSET_MATCHER.matcher(firstLine);
        
        int last = 0;
        while (matcher.find()) {
            int current = matcher.start(1);
            output.append(firstLine.substring(last, current));
            output.append(Integer.valueOf(matcher.group(1)) - 4);
            last = matcher.end(1);
        }
        output.append(firstLine.substring(last));
        return output.toString();
    }

    private String convertRaspOutput(String raspParse) {
        StringBuilder s = new StringBuilder();
        log.debug(String.format("Converting RASP parse '%s'", raspParse));
        try {
            OutputStream out = converterSocket.getOutputStream();
            Writer w = new OutputStreamWriter(out, "UTF-8");
            w.write(raspParse);
            w.write(17);
            w.write(10);
            w.flush();
            InputStream in = converterSocket.getInputStream();
            BufferedReader isr = new BufferedReader(new InputStreamReader(in, "UTF-8"));
            while (true) {
                String line = isr.readLine();
                if (line.length() > 0 && line.charAt(0) == 17) {
                    break;
                } else {
                    s.append(line);
                    s.append('\n');
                }
            }
        } catch (IOException e) {
            log.error(e);
        }
        log.debug(String.format("%d chars returned", s.toString().length()));
        return s.toString();
    }

    private int getNumParses(String lineIn) {
        String[] parts = lineIn.split(" ; ");
        String parseProbs = parts[parts.length - 1];
        if (parseProbs.equals("()")) {
            return 0;
        } else {
            return parseProbs.split(" ").length;
        }
    }

    /**
     * {@inheritDoc}
     */
    public Node processDOMTree(Node domTree) {
        return null;
    }
}
