/*
 * RaspModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.*;
import java.util.Properties;
import org.apache.log4j.Logger;
import org.w3c.dom.Node;
import de.dfki.lt.hog.Metadata;
import de.dfki.lt.hog.MoCoMan;
import de.dfki.lt.hog.Module;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;


/**
 * Module definition for RASP commandline interface
 * Created 2004-01-23
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @version $Id: RaspModule.java,v 1.5 2005/09/26 11:10:25 uschaefer Exp $
 */

/* Changes:
   2004-07-22 : US added empty result handling (metadata)
 * 2004-09-09 : US input + outputencoding properly passed to process streams
*/

public class RaspModule extends Module {

    static Logger log = Logger.getLogger(RaspModule.class);
    private BufferedReader procOut;
    private OutputStreamWriter procIn;
    private Runtime runtime;
    private Process process;
    private String sRaspOptions;
    private String sRaspScript;
    private String outputEncoding;
    private String inputEncoding;
    private static final String RASP_INIT_TERMINATOR = "ready.";
    private static final String RASP_RESULT_STARTER = "<rmrs";
    private static final String RASP_RESULT_END = "</rmrs>";
    private static final String RASP_RESULT_EMPTY = "<rmrs/>";
    private static final String NEWLINE = System.getProperty("line.separator");
    private static final String RASP_SHUTDOWN_COMMAND = NEWLINE + NEWLINE;

    public RaspModule(Properties raspModuleProps)
        throws InitializationException {
      try {
        super.init(raspModuleProps);
      } catch (InitializationException ie) {
        log.error(Utility.exception2String(ie));
      }

      sRaspOptions = getConfig().getProperty("rasp.options");
      sRaspScript =  getConfig().getProperty("rasp.script");
      outputEncoding = getConfig().getProperty("rasp.outputencoding", "ISO-8859-1");
      inputEncoding = getConfig().getProperty("rasp.inputencoding", "ISO-8859-1");

      log.debug("sRaspScript: " + sRaspScript);
      if (!sRaspScript.startsWith(System.getProperty("file.separator")))
        sRaspScript = System.getProperty("root.dir") + System.getProperty("file.separator") + sRaspScript;
      log.info("Initializing RASP...");
      runtime = Runtime.getRuntime();

      try {
        log.debug("Invoking: " + sRaspScript + " " + sRaspOptions);
        process = runtime.exec(sRaspScript + " " + sRaspOptions);
        String lineIn;
        procOut = new BufferedReader(new InputStreamReader(process.getInputStream(), outputEncoding));
        ////log.debug("line after procOut = ...");
        procIn = new OutputStreamWriter(process.getOutputStream(), inputEncoding);
        ////log.debug("line after procIn = ...");
        //RASP doesn't talk during initialization in server mode
        //so keeping these lines runs the loop forever
        //while ((lineIn=procOut.readLine())!=null) {
        //    log.debug("RASP: " + lineIn);
        //    if (lineIn.indexOf(RASP_INIT_TERMINATOR)!=-1) break;
        //}
        // test

        //if (procOut.ready()) {
        log.info("RASP initialized.");
        //} else {
        //  throw new InitializationException("RASP was not initialized");
        //}
      } catch (IOException ioe) {
        log.fatal(Utility.exception2String(ioe));
      }
    }

        public void shutdown() {
            log.info("Shutting down RASP.");
            String lineIn;
            try {
                procIn.write(RASP_SHUTDOWN_COMMAND);
                procIn.flush();
                process = null;
            } catch (IOException ioe) {
                log.error(Utility.exception2String(ioe));
            }
        }

        /* (non-Javadoc)
         * @see de.dfki.lt.hog.Module#process(java.lang.String)
         */
        public String process(String processString) {
        	String duration = "00:00,000";
        	long t1 = System.currentTimeMillis();
            MoCoMan mocoman = getMoCoMan();
            String rawText = Utility.getComputedAnnotation(mocoman, processString, "rawtext");
            ////String sessionID = Utility.getSessionIdFromMetadata(mocoman, processString);
            Metadata md = new Metadata(processString);
            String sessionID = md.getSessionID();
            ////String acID = Utility.getAnnotationCollectionIdFromMetadata(mocoman, processString);
            String acID = md.getAcID();
            String root =  getModuleRootElement();
            // process text with PET, output is RMRS (XML)
            String lineIn;
            String metadata;
            StringBuffer sbResult = new StringBuffer();
            boolean inResult = false;
            try {

                /* the following three lines should be removed (uncomment code in Constructor)
                   if server mode for rasp exists */
                //process = runtime.exec(sRaspScript + " " + sRaspOptions);
                //procOut = new BufferedReader(new InputStreamReader(process.getInputStream()));
                //procIn = new BufferedWriter(new OutputStreamWriter(process.getOutputStream()));

                // we need to check whether rawText already contains a terminal
                // symbol that would accidently combine to a SHUTDOWNCOMMAND
                //if (rawText.equals("\n")) {
                //	log.debug("rawtext:" + rawText);
                //}
                rawText = rawText.replaceAll("\n", " ").replaceAll("\r", " ");
                // problematic? at least neccessary.
                // exception in SAX otherwise
                rawText = replaceXMLcharsPartially(rawText);
                String rt = rawText.trim();
                if (rt.equals(".")||rt.equals("")) {
                    log.debug("empty input");
                    return "<" + root + ">" +
                        Metadata.generateXml(sessionID, acID, duration,
                                             getModuleName(),
                                             "OK", true,
                                             getConfig())
                        + "</" + root + ">";
                }
                procIn.write(rawText + NEWLINE);
                procIn.flush();//////
                int reading = 0;
                // int readings = 0;
                //	String confidence = null;
                while ((lineIn=procOut.readLine())!=null) {
                    if (lineIn.startsWith(RASP_RESULT_EMPTY) || lineIn.startsWith(RASP_RESULT_STARTER)) {
                        if (reading==0) {
                            duration = Utility.formatDuration(System.currentTimeMillis()-t1);
                            sbResult.append("<" + root + ">"); // readings=\"" + readings +"\">");
                            sbResult.append(Metadata.generateXml(sessionID,
                                                                 acID, duration,
                                                                 getModuleName(),
                                                                 "OK", lineIn.startsWith(RASP_RESULT_EMPTY),
                                                                 getConfig()));
                            //if (readings==0) {
                            //sbResult.append("</" + root + ">");
                            //break;
                            reading = 1;
                            inResult=true;
                        }
                    }

                    if (inResult) sbResult.append(lineIn);

                    // else if (lineIn.startsWith("(|")) sbResult.append("<!-- ").append(lineIn).append(" -->");
                    if (lineIn.endsWith(RASP_RESULT_END) || lineIn.startsWith(RASP_RESULT_EMPTY)) {
                        inResult=false;
                        break;
                    }
                }
                // check whether result is empty
                if (!sbResult.toString().equals("")) {
                  sbResult.append("</" + root + ">");
                }
                else {
                  sbResult.append("<" + root + ">" +
                                  Metadata.generateXml(sessionID, acID, duration,
                                                       getModuleName(),
                                                       "OK", true,
                                                       getConfig())
                                  + "</" + root + ">");
                }
                String sResult = sbResult.toString();
                log.debug("result: " + sResult);
                duration = Utility.formatDuration(System.currentTimeMillis()-t1);
                log.info("Processing time of RASP parser: " + duration);
                return sResult;
                } catch (IOException ioe) {
                    log.fatal(Utility.exception2String(ioe));
                    return "<" + root + ">" +
                           Metadata.generateXml(sessionID, acID, duration,
                                                getModuleName(),
                                                ioe.getMessage(), true,
                                                getConfig()) +
                           "</" + root + ">";
                }
        }

        /* (non-Javadoc)
         * @see de.dfki.lt.hog.Module#processDOMTree(org.w3c.dom.Node)
         */
        public Node processDOMTree(Node domTree) {
                // TODO Auto-generated method stub
                // use a simple SAX parser to extract text from domTree
                // return a DOM from transformer DOMResult (i.e., copy most of process() except final DOM output)
                return null;
        }


  /**
   * Cleans an XML <tt>String</tt> by replacing '&amp', '&lt', '&gt',
   * by spaces respectively.
   * This is a hack for RASP versions before october 2005 that do
   * not quote characters properly in RMRS output mode (non-wellformed XML).
   * @param dirtyXMLString The XML <tt>String</tt> to be cleaned
   * @return The cleaned XML <tt>String</tt>
   */
  public final static String replaceXMLcharsPartially(String dirtyXMLString) {

    StringBuffer cleanStrBuffer = new StringBuffer();

    for (int i = 0; i < dirtyXMLString.length(); i++) {
      char c = dirtyXMLString.charAt(i);
      if (c == '&' || c == '<' || c=='>') {
        cleanStrBuffer.append(" ");
      }
      else {
        cleanStrBuffer.append(c);
      }
    } // for
    return cleanStrBuffer.toString();
  } // cleanXMLString()

}
