/*
 * SleepyModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.*;
import java.util.ArrayList;
import java.util.Properties;
import java.util.Stack;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import de.dfki.lt.hog.*;
import de.dfki.lt.hog.container.JToken;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;
import de.dfki.lt.hog.container.GraphNode;
import de.dfki.lt.hog.TransformationService;
import org.apache.log4j.Logger;
import org.w3c.dom.Node;

/* CHANGES
   2005-09-29: uschaefer corrected problem with % (comment only at beginning of line)
   2005-09-29: uschaefer added postprocessing transformation for lemma inclusion
                         e.g. from TreeTaggerModule
*/

/**
 * Module definition for Sleepy commandline interface
 *
 * @author <a href=mailto:oeze@dfki.de>Oezguer Demir</a>
 * @version $Revision: 1.2 $
 * $Id: SleepyModule.java,v 1.2 2005/09/29 12:59:30 uschaefer Exp $
 */
public class SleepyModule extends Module {

  static Logger log = Logger.getLogger(SleepyModule.class);
  private BufferedReader procOut;
  private BufferedReader procErr;
  private BufferedWriter procIn;
  private Runtime runtime;
  private Process process;
  private String sleepyOptions = "";
  private String sleepyScript = "";
  private String outputEncoding;
  private String inputEncoding;
  private String tokenizationAnnotation;
  private TransformationService transformationService = null;
  private Transformer postprocTransformer;
  private String postprocInputAnnotation;
  private static final String NEWLINE = System.getProperty("line.separator");
  private static final String DATA_ROOT_ELEMENT = "parse";

  public SleepyModule(Properties sleepyModuleProps) throws
      InitializationException {
    try {
      super.init(sleepyModuleProps);
    } catch (InitializationException ie) {
      log.fatal(Utility.exception2String(ie));
    }
    sleepyScript = getConfig().getProperty("sleepy.script");
    sleepyOptions = sleepyOptions +
                    " " + getConfig().getProperty("sleepy.options");
    outputEncoding =
        getConfig().getProperty("sleepy.outputencoding", "ISO-8859-1");
    inputEncoding =
        getConfig().getProperty("sleepy.inputencoding", "ISO-8859-1");

    tokenizationAnnotation =
        getConfig().getProperty("sleepy.tokenization", "JTok");

    postprocInputAnnotation =
        getConfig().getProperty("sleepy.postprocinputannotation");

    runtime = Runtime.getRuntime();

  } // SleepyModule()

  public void shutdown() {
    log.info("Shutting down Sleepy.");
    // nothing to do here
  }

  public String process(String processString) {
    String duration = "00:00,000";
    long t1 = System.currentTimeMillis();
    Metadata md = new Metadata(processString);
    String sessionID = md.getSessionID();
    String acID = md.getAcID();
    String root = getModuleRootElement();
    String diagnosis = "OK";

    // tokenize text
    StringBuffer tokenBuffer = new StringBuffer();
    MoCoMan mocoman = MoCoMan.getInstance();

    if (this.transformationService==null) {
        this.transformationService = new TransformationService(mocoman);
        String sTylesheet = getConfig().getProperty("sleepy.postprocstylesheet");
        if (sTylesheet != null) {
            try {
                postprocTransformer = this.transformationService.createTransformer(sTylesheet);
            } catch (TransformerConfigurationException tce) {
      		log.error(Utility.exception2String(tce));
            }
        }
    }

    ArrayList jtoks = JToken.parseJTokAnnotation(
        Utility.getComputedAnnotation(mocoman, processString, tokenizationAnnotation));
    for (int i = 0; i < jtoks.size(); i++) {
      JToken tok = (JToken) jtoks.get(i);
      tokenBuffer.append(tok.getToken() + " ");
    }
    String tokenizedStr = tokenBuffer.toString();
    log.debug("input for Sleepy: " + tokenizedStr);
    // process text
    try {
      process = runtime.exec(sleepyScript + " " + sleepyOptions);
      procOut =
          new BufferedReader(new InputStreamReader(process.getInputStream(),
          outputEncoding));
      procIn =
          new BufferedWriter(new OutputStreamWriter(process.getOutputStream(),
          inputEncoding));
      // feed the text to Sleepy
      procIn.write(tokenizedStr + NEWLINE);
      procIn.flush();
      procIn.close();
      StringBuffer sleepyResultBuffer = new StringBuffer();
      String lineIn;
      StringBuffer sbResult = new StringBuffer();
      while ((lineIn = procOut.readLine()) != null) {
        sleepyResultBuffer.append(lineIn + NEWLINE);
      } // while
      // check whether result is empty
      String sleepyResult = sleepyResultBuffer.toString();
      log.debug("Sleepy output:\n" + sleepyResult);
      boolean empty = false;
      if (sleepyResult.length() < 3) {
        empty = true;
        diagnosis = "Sleepy's output was empty";
        log.warn(diagnosis);
      }
      // convert to sleepy's output to xml
      if (!empty) {
        GraphNode top = parseSleepy(sleepyResult, jtoks);
        sleepyResult = top.toXML();
        int jtoksize = jtoks.size();
        int sleepysize = top.getWords().size();
        if (sleepysize != jtoksize) {
          diagnosis = "Error Token count mismatch between JTok (" + jtoksize + ") and Sleepy (" + sleepysize + ").";
          log.error(diagnosis);
          return "<" + root + ">" +
                    Metadata.generateXml(sessionID, acID, duration, getModuleName(),
                                         diagnosis, true, getConfig()) +
                 "</" + root + ">";
        }
      }

      // generate metadata
      duration = Utility.formatDuration(System.currentTimeMillis() - t1);
      sbResult.append("<" + root + ">");
      sbResult.append(Metadata.generateXml(sessionID,
                                           acID, duration,
                                           getModuleName(),
                                           diagnosis, empty,
                                           getConfig()));
      sbResult.append("<" + DATA_ROOT_ELEMENT + ">" + NEWLINE);

      // paste in results from sleepy
      sbResult.append(sleepyResult);
      sbResult.append("</" + DATA_ROOT_ELEMENT + ">" + NEWLINE);
      sbResult.append("</" + root + ">");
      String sResult = sbResult.toString();
      log.debug("Sleepy result:\n" + sResult);
      log.info("Processing time of SleepyModule: " + duration);

      if (postprocTransformer!=null) {
          if (postprocInputAnnotation!=null) {
              if (mocoman.getAnnotation(sessionID, acID, postprocInputAnnotation)!=null) {
                  postprocTransformer.setParameter("uri", "hog://" + sessionID + "/" + acID + "/" + postprocInputAnnotation);
                  try {
                      sResult = this.transformationService.transform(postprocTransformer, sResult);
                  } catch (TransformerException te) {
                      log.error(Utility.exception2String(te));
                      return "<" + root + ">" +
                          Metadata.generateXml(sessionID, acID, duration, getModuleName(),
                                               "Error: " + te.getMessage(), true, getConfig()) +
                          "</" + root + ">";
                  }
              }
          }
      }
      return sResult;
    } catch (IOException ioe) {
      log.fatal(Utility.exception2String(ioe));
      return "<" + root + ">" +
          Metadata.generateXml(sessionID, acID, duration,
                               getModuleName(),
                               ioe.getMessage(),
                               true,
                               getConfig()) +
          "</" + root + ">";
    }
  }

  private static GraphNode parseSleepy(String inString, ArrayList jtoks) {
    MoCoMan mcm = MoCoMan.getInstance();
    Stack nodeStack = new Stack();
    StringBuffer tag = new StringBuffer();
    String tagStr = "";
    StringBuffer word = new StringBuffer();
    GraphNode currentNode = new GraphNode();
    char mode = 'x'; // t: tag, w: word, i: ignore, p: parameter, x: undefined
    char event = 'X'; // O: open; C: close, X: undefined
    char c;
    int tCounter = 1;
    int ntCounter = 501;
    int sentenceCounter = mcm.getSentenceCount() + 1;
    int jtokCounter = 0;

    // preprocess the input string
    inString = inString.replaceAll("\\(\\)", "0pP4R4N)");
    inString = inString.replaceAll("\\)\\)", "C1P4R4N)");
    inString = inString.replaceAll("\\$\\(", "P4R4NT49");
//    log.debug("preprocessed instring: " + inString);

    // process the input string char by char
    for (int i = 0; i < inString.length(); i++) {
      c = inString.charAt(i);

      // set mode according to input char
      if (c == '%') {
          // if added by uschaefer 2005-09-29
          // (% is start of comment only at beginning of a line!)
        if (mode=='x') mode = 'i'; // mode: cursor over comment line
      }
      if (c == '\n') {
        mode = 'x'; // mode: cursor over empty area
      }
      if (mode != 'i') {
        if (c == '(') {
          event = 'O'; // event: opening bracket
          mode = 't'; // mode: cursor over tag
        }
        if (c == ')') {
          event = 'C'; // event: closing bracket
        }
        if (mode == 't' && c == ' ') {
          event = 'T'; // event: tag completely read
          mode = 'p'; // mode: cursor over parameter area
        }
        if (mode == 'p' && c != ' ' && c != '\n' && c != '(') {
          mode = 'w'; // mode: cursor over word
        }
      } // if mode != i

      // act according to mode
      if (event == 'O') { // if event = opening bracket
        currentNode = new GraphNode();
        if (!nodeStack.empty()) {
          ((GraphNode) nodeStack.peek()).addDaughter(currentNode);
        }
        nodeStack.push(currentNode);
        event = 'X';
      }
      if (event == 'C') { // if event = closing bracket
        if (mode == 'w') { // if cursor over word
          JToken currentJToken = (JToken) jtoks.get(jtokCounter++);
          String wordStr = word.toString()
                           .replaceAll("0pP4R4N", "(")
                           .replaceAll("C1P4R4N", ")")
                           .trim();
//          log.debug("wordstr: " + wordStr);
          if (!wordStr.equals(currentJToken.getToken().trim())) {
            log.error("Mapped JToken does not match sleepy token:\n" +
                      "JToken: " + currentJToken.getToken() + "\n" +
                      "Sleepy token: " + word);
          }
          currentNode.setJToken(currentJToken);
          currentNode.setType("t");
          currentNode.setId("" + sentenceCounter + "_" + tCounter++);
          word = new StringBuffer();
          mode = 'x';
        }
        GraphNode popped = ((GraphNode) nodeStack.pop());
        if (popped.getPoscat().equals("TOP")) {
          mcm.increaseSentenceCount();
          return popped;
        }
        event = 'X'; // reset event
      }
      if (event == 'T') {
        GraphNode nt = ((GraphNode) nodeStack.peek());
        tagStr = tag.toString();
        tagStr = tagStr.replaceAll("P4R4NT49", "\\$\\(");
        if (tag.indexOf("-") > -1) {
          nt.setLabel(tag.substring(tag.indexOf("-") + 1));
          tagStr = tagStr.substring(1, tag.indexOf("-"));
        } else {
          tagStr = tagStr.toString().substring(1);
        }
        nt.setPoscat(tagStr);
        nt.setType("nt");
        nt.setId("" + sentenceCounter + "_" + ntCounter++);
        tag = new StringBuffer(); // reset tag buffer
        event = 'X'; // reset event
      }
      if (mode == 't') {
        tag.append(c);
      }
      if (mode == 'w') {
        word.append(c);
      }
    } // for i
    mcm.increaseSentenceCount();
    return (GraphNode) nodeStack.pop();
  }

  /* (non-Javadoc)
   * @see de.dfki.lt.hog.Module#processDOMTree(org.w3c.dom.Node)
   */
  public Node processDOMTree(Node domTree) {
    // TODO Auto-generated method stub
    // use a simple SAX parser to extract text from domTree
    // return a DOM from transformer DOMResult (i.e., copy most of process() except final DOM output)
    return null;
  }

} // class
