/*
 * SproutModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.util.Properties;
import java.util.ArrayList;
import org.apache.log4j.Logger;
import de.dfki.lt.tfs.FeatureStructure;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Node;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.dom.DOMSource;
import de.dfki.lt.hog.Metadata;
import de.dfki.lt.hog.Module;
import de.dfki.lt.hog.TransformationService;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;
import de.dfki.lt.sprout.runtime.interpreter.SproutInterpreter;
import de.dfki.lt.sprout.runtime.MatchInfo;


/**
 * Module definition for SProUT runtime API.
 * Created 2003-11-17
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @version  $Id: SproutModule.java,v 1.5 2006/06/10 20:30:47 uschaefer Exp $
 */

 /* Changes:
 *  2004-03-31  hack for preliminary PET input chart<br>
 *  2004-04-02  count '('s instead of '\n's <br>
 *  2004-04-30  configuration of annotation names for PET input chart<br>
 *  2004-06-14  added TransformationService
 *  2004-07-22  added empty result handling (metadata)
 *  2004-09-13  XML output encoding fixed (Unicode)
 *  2004-10-12  SproutInterpreter (new sprout version)
 *  2006-05-18  added XML input mode and true optional output transformation
 */

public class SproutModule extends Module {

	static Logger log = Logger.getLogger(SproutModule.class);
    private TransformationService transformationService = null;
    private String sTylesheet = null;
	private Transformer sprout2rmrsTransformer = null;
    private String sTylesheet4pic = null;
	private Transformer sprout2picTransformer = null;
    private Transformer stringTransformer = null;
	private SproutInterpreter sprout = null;
	private String sOutputPath = null;
    private String sPicOutputAnnotationName = null;
    private String sOutputFSAnnotationName = null;
    //        private String sTntInputAnnotationName = null;
    private String sRawTextAnnotationName = null;

	public SproutModule(Properties sproutModuleProps) {
		String sProutConfigFile = null;
		try {
                  super.init(sproutModuleProps);
		} catch (InitializationException ie) {
                  log.fatal(Utility.exception2String(ie));
		}
                // settings for PET input chart mode
                sPicOutputAnnotationName = getConfig().getProperty("sprout.output4pic", null);
                if (sPicOutputAnnotationName!=null) {
                    sTylesheet4pic = getConfig().getProperty("sprout.stylesheet4pic", null);
                    //sTntInputAnnotationName = getConfig().getProperty("sprout.tntinput4pic");
                }

                // global configurations as imposed by cfg file

                sRawTextAnnotationName = getConfig().getProperty("sprout.inputtextannotation", "rawtext");
                sOutputFSAnnotationName = getConfig().getProperty("sprout.outputfsannotation", null);

		sTylesheet = getConfig().getProperty("sprout.stylesheet", null);

		// output feature path (default null means start from root FS including IN and OUT features)
		sOutputPath = getConfig().getProperty("sprout.outputpath", null);

		// start SProUT API with configured cfg file
		sProutConfigFile = getConfig().getProperty("sprout.configfile");

		log.info("Initializing SProUT interpreter...");
		sprout = new SproutInterpreter();

		boolean status = sprout.initialize(sProutConfigFile, null, null) ;

		if (!status)
			log.error("Could not initialize SProUT interpreter");
		log.info("SProUT interpreter initialized.");
	}

	public void shutdown() {
		//  no corresponding method found in SproutInterpreter (Sprout API)
		log.info("shutting down SProUT.");
	}

	/* (non-Javadoc)
	 * @see de.dfki.lt.hog.Module#process(java.lang.String)
	 */
	public String process(String processString) {
            Metadata md = new Metadata(processString);
            String sessionID = md.getSessionID();
            String acID = md.getAcID();
            String rawText = getMoCoMan().getAnnotation(sessionID, acID, sRawTextAnnotationName);
            // create XSL transformers with configured stylesheets
            if (this.transformationService==null) {
                this.transformationService = new TransformationService(getMoCoMan());
                try {
                    stringTransformer = this.transformationService.createTransformer();
                    if (sTylesheet!=null) {
                        sprout2rmrsTransformer = this.transformationService.createTransformer(sTylesheet);
                        sprout2rmrsTransformer.setParameter("encoding", "UTF-8"); // we stay with Java Unicode
                        //sprout2rmrsTransformer.setParameter("rawtext", rawText);
                    }
                    if (sPicOutputAnnotationName!=null)
                        sprout2picTransformer = this.transformationService.createTransformer(sTylesheet4pic);
                } catch (TransformerConfigurationException tce) {
                    log.error(Utility.exception2String(tce));
                }
            }


            // process text with SProUT API, output is Match structures containing
            // disjunctions of typed feature structures
            long t1 = System.currentTimeMillis();

            if ("rawtext".equals(sRawTextAnnotationName))
                sprout.setNewInputTextfromString(rawText);
            else
                sprout.setNewXMLInputTextfromString(rawText);

            log.info("running SProUT interpreter.");
            log.debug("input text: " + rawText);
            ArrayList output = null;
            try {
                output = sprout.RunInterpreter();
            } catch (Throwable t) {
                log.error(t.toString());
                t.printStackTrace(System.err);
            }
            String duration = Utility.formatDuration(System.currentTimeMillis() - t1);
            String result = null;
            log.info("processing time of SProUT interpreter: " + duration);
            log.info("creating SProUT XML output");

            // generate XML structure from TFS and transform XML using XSLT

            if (output != null) {
                log.info("Sprout result size=" + output.size());
                t1 = System.currentTimeMillis();

                DOMSource source = new DOMSource(buildFSDOM(output, sessionID, acID, duration));
                // write FS xml file
                if (sOutputFSAnnotationName!=null) {
                    try {
                        getMoCoMan().createAnnotation(sessionID, acID,
                                                      this.transformationService.transform(stringTransformer, source),
                                                      getModuleLanguage(), sOutputFSAnnotationName);
                    } catch (TransformerException te) {
                        log.error(Utility.exception2String(te));
                    }
                }
                String duration2 = Utility.formatDuration(System.currentTimeMillis() - t1);
                log.info("processing time for DOM generation: " + duration2);

                try {
                    if (sTylesheet!=null) {
                        log.info("transforming SProUT XML output to RMRS");
                        t1 = System.currentTimeMillis();
                        sprout2rmrsTransformer.setParameter("sessionID", sessionID);
                        sprout2rmrsTransformer.setParameter("collectionID", acID);
                        sprout2rmrsTransformer.setParameter("annotationID", getModuleName());
                        result = this.transformationService.transform(sprout2rmrsTransformer, source);
                        String duration3 = Utility.formatDuration(System.currentTimeMillis() - t1);
                        log.info("processing time for DOM transformation: " + duration3);
                    }
                    else
                        result =  this.transformationService.transform(stringTransformer, source);

                    if (sPicOutputAnnotationName!=null) {
                        //String tntpic = Utility.getComputedAnnotation(getMoCoMan(), processString, sTntInputAnnotationName);  //pic-hack
                        //if (tntpic == null) tntpic="";
                        //sprout2picTransformer.setParameter("offset", Integer.toString(countLines(tntpic)));
                    	sprout2picTransformer.setParameter("sessionID", sessionID);
                    	sprout2picTransformer.setParameter("collectionID", acID);
                    	sprout2picTransformer.setParameter("annotationID", sPicOutputAnnotationName);
                        log.info("transforming SProUT XML output to Pet Input Chart Format");
                        String sResult = this.transformationService.transform(sprout2picTransformer, source);
                        getMoCoMan().createAnnotation(sessionID, acID, sResult, getModuleLanguage(), sPicOutputAnnotationName);
                    }
                    } catch (TransformerException te) {
                        log.error(Utility.exception2String(te));
                        return generateErrorResult(sessionID, acID, duration, getModuleName(), "Error: " + te.getMessage());
                    }
            } else {
                log.debug ("empty result");
                return generateErrorResult(sessionID, acID, duration, getModuleName(), "OK (empty result)");
            }
            log.debug(result);
            return result;
        }

    /*
    public int countLines(String str) {  // for pic-hack
        // we only need at least the number of tokens
        // no problem if there are more '('s than tokens
        int i = 0;
        int j = 0;
        while ((j=str.indexOf('(',j))!=-1) {
            i++;
            j++;
        }
        return i;
    }
    */

    public String generateErrorResult( String sessionID, String acID, String duration, String module, String diagnosis) {
            return "<" + getModuleRootElement() + ">" +
                   Metadata.generateXml(sessionID, acID, duration, module,
                                        diagnosis, true, getConfig()) +
                "</" + getModuleRootElement() + ">";
    }

        /* (non-Javadoc)
         * @see de.dfki.lt.hog.Module#processDOMTree(org.w3c.dom.Node)
	 */
	public Node processDOMTree(Node domTree) {
		// TODO Auto-generated method stub
		// use a simple SAX parser to extract text from domTree
		// return a DOM from transformer DOMResult (i.e., copy most of process() except final DOM output)
		return null;
	}

	/** build DOM of Match structure (disjunctions of typed feature structures)
	    * @param Match structure
	    * @return DOM node
	    */
	private Document buildFSDOM(
		ArrayList sproutResults,
                String sessionID,
                String acID,
		String duration) {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = null;
		try {
                  builder = factory.newDocumentBuilder();
		} catch (ParserConfigurationException pce) {
                  // Parser with specified options can't be built
                  log.error(Utility.exception2String(pce));
                  return null;
		}
		Document document = builder.newDocument();

		Element root = (Element) document.createElement(getModuleRootElement());
		document.appendChild(root);
		root.appendChild(
			Metadata.createMetadataElement(document, sessionID,
                                                       acID, duration,
                                                       getModuleName(), "OK",
                                                       (sproutResults.size()==0),
                                                       getConfig()));

		Element disjElement;
		Element infoElement;
		MatchInfo matchInfo;
		FeatureStructure matchFS;

		int mino = 0;
		int dino = 0;
		for (int match = 0; match < sproutResults.size(); match++) {

			ArrayList vMI = (ArrayList)sproutResults.get(match);
			disjElement = (Element) document.createElement("DISJ");
			disjElement.setAttribute("id", "DI" + Integer.toString(dino++));
			root.appendChild(disjElement);

			for (int disj = 0; disj < vMI.size(); disj++) {
				matchInfo = (MatchInfo) vMI.get(disj);

				infoElement = (Element) document.createElement("MATCHINFO");
				infoElement.setAttribute("id", "MI" + Integer.toString(mino++));
				infoElement.setAttribute("rule", matchInfo.getRuleName());
				infoElement.setAttribute(
					"cstart",
					Integer.toString(matchInfo.getCStart()));
				infoElement.setAttribute(
					"cend",
					Integer.toString(matchInfo.getCEnd()));
				infoElement.setAttribute(
					"start",
					Integer.toString(matchInfo.getStart()));
				infoElement.setAttribute(
					"end",
					Integer.toString(matchInfo.getEnd()));
				disjElement.appendChild(infoElement);
				if (sOutputPath == null)
					matchFS = matchInfo.getFs();
				else
					matchFS = (matchInfo.getFs()).getPathValue(sOutputPath);
				matchFS.xmlFS(infoElement, true, document);
			}
			vMI = null;  // hack 2004-12-01 for memory hole
		}
		return document;
	}


}
