/*
 * TreeTaggerModule.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.modules;

import java.io.*;
import java.util.*;
import org.apache.log4j.Logger;
import org.w3c.dom.*;
import de.dfki.lt.hog.Metadata;
import de.dfki.lt.hog.MoCoMan;
import de.dfki.lt.hog.Module;
import de.dfki.lt.hog.container.JToken;
import de.dfki.lt.hog.container.TreeTaggerToken;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;

/**
 * Module definition for TreeTagger commandline interface
 *
 * @author <a href=mailto:oeze@dfki.de>Oezguer Demir</a>
 * @version $Revision: 1.3 $
 * $Id: TreeTaggerModule.java,v 1.3 2007/01/19 14:11:39 gregul Exp $
 */
public class TreeTaggerModule extends Module {

  static Logger log = Logger.getLogger(TreeTaggerModule.class);
  private BufferedReader procOut;
  private BufferedWriter procIn;
  private BufferedReader procErr;
  private Runtime runtime;
  private Process process;
  private String treetaggerOptions = "";
  private String treetaggerScript = "";
  private String outputEncoding;
  private String inputEncoding;
  private String tokenizationAnnotation;
  private int lifeCycle = 0;
  private final int lifeSpan = 1;
  private final int treeTaggerInitTime = 2000;
  private static final String NEWLINE = System.getProperty("line.separator");
  private static final String DATA_ROOT_ELEMENT = "tokens";
  private static final String decoy =
      "d3lt43ch0ch4rl130sc4ry4nk33" + NEWLINE + ".";

  public TreeTaggerModule(Properties treetaggerModuleProps) throws
      InitializationException {
    try {
      super.init(treetaggerModuleProps);
    } catch (InitializationException ie) {
      log.error(Utility.exception2String(ie));
    }
    treetaggerScript =
        getConfig().getProperty("treetagger.script");
    treetaggerOptions =
        getConfig().getProperty("treetagger.options");
    outputEncoding =
        getConfig().getProperty("treetagger.outputencoding", "ISO-8859-1");
    inputEncoding =
        getConfig().getProperty("treetagger.inputencoding", "ISO-8859-1");
    tokenizationAnnotation =
        getConfig().getProperty("treetagger.tokenization", "JTok");

    start();
  } // TreeTaggerModule()

  public void start() throws InitializationException {
    log.info("Initializing TreeTagger ...");
    runtime = Runtime.getRuntime();
    String sname = treetaggerScript + " " + treetaggerOptions;
    log.debug("Invoking " + sname);
    StringBuffer errBuffer = new StringBuffer();
    try {
      process = runtime.exec(sname);
      procOut = new BufferedReader(
          new InputStreamReader(process.getInputStream(), outputEncoding));
      procIn = new BufferedWriter(
          new OutputStreamWriter(process.getOutputStream(), inputEncoding));
      procErr = new BufferedReader(
          new InputStreamReader(process.getErrorStream(), outputEncoding));
      Thread.sleep(treeTaggerInitTime);

			if (readWait("") == false)
			{
  			throw new InitializationException( "Initialization of TreeTagger failed.");
			}
			
/*
      while (!procErr.ready()) {
        Thread.sleep(30);
      }


      while (procErr.ready()) {
        errBuffer.append((char) procErr.read());
      } // while
      //log.debug("ERRBUFFER: " + errBuffer);
      if (!errBuffer.toString().trim()
          .replaceAll("\n", "").replaceAll("\t", "").replaceAll(" ", "")
          .equals("readingparameters...tagging...")) {
        log.error("Initialization of TreeTagger failed.");
        log.error("Read from TreeTagger stderr: " + errBuffer);
        throw new InitializationException(
            "Initialization of TreeTagger failed." +
            "Read from TreeTagger stderr: " + errBuffer);
      }
      log.debug("ERRBUFFER 2: " + errBuffer.toString().trim()
                .replaceAll("\n", "").replaceAll("\t", "").replaceAll(" ", ""));
*/
    } catch (IOException ioe) {
      log.fatal(Utility.exception2String(ioe));
    } catch (InterruptedException ex) {
      log.error(ex);
    }
  }

  public void shutdown() {
    log.info("Shutting down TreeTagger.");
    try {
      procIn.close();
      procOut.close();
      procErr.close();
    } catch (IOException ioe) {
      log.fatal(Utility.exception2String(ioe));
    }
  }

  /* (non-Javadoc)
   * @see de.dfki.lt.hog.Module#process(java.lang.String)
   */
  public String process(String processString) {
    String duration = "00:00,000";
    long t1 = System.currentTimeMillis();
    MoCoMan mocoman = getMoCoMan();
    Metadata md = new Metadata(processString);
    String sessionID = md.getSessionID();
    String acID = md.getAcID();
    String root = getModuleRootElement();
    this.lifeCycle++;
    if (this.lifeCycle >= this.lifeSpan) {
      this.shutdown();
      try {
        this.start();
      } catch (InitializationException ex) {
        log.error(Utility.exception2String(ex));
      }
      this.lifeCycle = 0;
      //(Family)this.getFamily()).sendCondolence();
    }
    // get tokenized text
    ArrayList jtoks = JToken.parseJTokAnnotation(
        Utility.getComputedAnnotation(mocoman, processString, tokenizationAnnotation));
    // split sentences according to JTok
    StringBuffer tokenBuffer = new StringBuffer();
    String lineSep;
    int numberOfTokens = jtoks.size();
    for (int i = 0; i < jtoks.size(); i++) {
      JToken tok = (JToken) jtoks.get(i);
      if (tok.getType().startsWith("PTERM")) {
        lineSep = NEWLINE + NEWLINE;
      } else {
        lineSep = NEWLINE;
      }
      tokenBuffer.append(tok.getToken() + lineSep);
    }
    String tokenizedStr = tokenBuffer.toString();
    // process text
    String lineIn;
    StringBuffer sbResult = new StringBuffer();
    String diagnosis = "";
    try {
      // feed the text to TreeTagger
      procIn.write(tokenizedStr + NEWLINE);
      procIn.flush();
//      log.error("flushed into tt: " + tokenizedStr);
      sbResult.append("<" + DATA_ROOT_ELEMENT + ">" + NEWLINE);
//      log.debug("debug now what 0");
      // get results from TreeTagger
      StringBuffer resultBuffer = new StringBuffer();
      lineIn = "";
      int tokenCounter = 0;
//      log.debug("debug now what 1");

			// this routine is very broken,
			// for some texts i send it it gets caught in an endless loop
			// the diecounter will discontinue the loop IF:
			///  diecounter == 0	
			int diecounter = numberOfTokens * 1000;

      while (tokenCounter != numberOfTokens) {
				if ( diecounter <= 0)
				{
					System.out.println("TREETAGGERMODULE:DIELOOP HIT");
					break;
				}
				diecounter--;
//        log.debug("debug now what 2");
        while (!procOut.ready()) {
//          log.debug("feeding dummy");
          procIn.write(decoy + NEWLINE);
          procIn.flush();
        }
//        log.debug("debug now what 3");
        StringBuffer procOutBuffer = new StringBuffer();
        if (procOut.ready()) {
//          log.debug("procout is ready");
          char c;
          while (procOut.ready()) {
            c = ((char) procOut.read());
//            log.debug("now what 11");
            procOutBuffer.append(c);
            if (c == '\n') {
              break;
            }
//            log.debug("char: " + c);
          }
          // log.debug("foo: " + foo);
//          lineIn = foo.toString();
//          log.debug("linein: " + lineIn);
//          log.debug("foo:" + foo);
        } else {
          while (!procOut.ready()) {
//            log.debug("feeding dummy 2");
            procIn.write(decoy + NEWLINE);
            procIn.flush();
          }
        }
//        log.debug("debug now what 4");
        lineIn = procOutBuffer.toString();
        if (lineIn.startsWith(((JToken) jtoks.get(tokenCounter)).getToken())) {
//          log.debug("lineIn matched: " + lineIn);
          resultBuffer.append(lineIn + NEWLINE);
          tokenCounter++;
        }
//        log.debug("debug now what 5");
      } // while tokenCounter...
      StringBuffer errBuffer = new StringBuffer();
//      log.debug("debug now what 6");
      while (procErr.ready()) {
//        log.debug("debug now what 7");
//        lineIn = procErr.readLine();
        while (procErr.ready()) {
          errBuffer.append((char) procErr.read());
        }
//        log.debug("debug now what 9");
      } // while
      if (errBuffer.toString().equals("") ||
          errBuffer.toString().endsWith("000")) {
        diagnosis = "OK";
      } else {
        diagnosis = errBuffer.toString();
        log.error(diagnosis);
      }
      // convert to XML
      String treetaggerresult = resultBuffer.toString();
      // check whether result is empty
      boolean empty = false;
      if (treetaggerresult.trim().equals("")) {
        empty = true;
      }
      log.debug("TreeTagger output:\n" + treetaggerresult);
			try {
      if (diagnosis.equals("OK")) {
        sbResult.append(
            treetaggertokens2XML(parseTreeTagger(treetaggerresult, jtoks)));
      }
			} catch (Exception e) {
				System.out.println(e.getMessage());	
			}
      duration = Utility.formatDuration(System.currentTimeMillis() - t1);
      // generate the header
      StringBuffer header = new StringBuffer();
      header.append("<?xml version='1.0'?>");
      header.append("<" + root + ">");
      header.append(Metadata.generateXml(sessionID,
                                         acID, duration,
                                         getModuleName(),
                                         diagnosis, empty,
                                         getConfig()));
      sbResult.append("</" + DATA_ROOT_ELEMENT + ">" + NEWLINE);
      sbResult.append("</" + root + ">");
      String sResult = header.toString().concat(sbResult.toString());
      log.info("Processing time of TreeTaggerModule: " + duration);
      log.debug("result: " + sResult);
      return sResult;
    } catch (IOException ioe) {
      log.fatal(Utility.exception2String(ioe));
      return "<" + root + ">" +
          Metadata.generateXml(sessionID, acID, duration,
                               getModuleName(),
                               ioe.getMessage(), true,
                               getConfig()) +
          "</" + root + ">";
    }
  }

  /* (non-Javadoc)
   * @see de.dfki.lt.hog.Module#processDOMTree(org.w3c.dom.Node)
   */
  public Node processDOMTree(Node domTree) {
    // TODO Auto-generated method stub
    // use a simple SAX parser to extract text from domTree
    // return a DOM from transformer DOMResult (i.e., copy most of process() except final DOM output)
    return null;
  }

  /**
   * Parses the output of TreeTagger and returns an ArrayList of
   * {@link TreeTaggerToken} instances.
   * @param treetaggerOutput The output of TreeTagger with switch -z0
   * @param jtokens An ArrayList of {@link JToken} instances
   * @return
   */
  private static ArrayList parseTreeTagger(String treetaggerOutput,
                                           ArrayList jtokens) {
    ArrayList ttTokens = new ArrayList();
    ArrayList lines = Utility.splitLines(treetaggerOutput);
    if (lines.size() == 0) {
      log.error("Token list received by TreeTagger is empty");
      ttTokens.add(new TreeTaggerToken(new JToken(" ", 1, 0)));
      return ttTokens;
    }
    for (int i = 0; i < jtokens.size(); i++) {
      TreeTaggerToken ttToken =
          new TreeTaggerToken((JToken) jtokens.get(i));
      ArrayList line = Utility.splitTabs((String) lines.get(i));
      if (line.size() < 3) {
        log.error("TreeTagger output line incomplete ...");
        log.error("... Input: \n" + treetaggerOutput);
        for (int j = 0; j < line.size(); j++) {
          log.error("line[" + j + "]: " + line.get(j));
        }
      } else {
        ttToken.setToken((String) line.get(0));
        ttToken.setPos((String) line.get(1));
        String lemma = (String) line.get(2);
        if (!lemma.equals("<unknown>")) {
          ttToken.setLemma(lemma);
        }
        ttTokens.add(ttToken);
      }
    }
    return ttTokens;
  }

  private static String treetaggertokens2XML(ArrayList treetaggertokens) {
    StringBuffer out = new StringBuffer();
    for (int i = 0; i < treetaggertokens.size(); i++) {
      out.append(((TreeTaggerToken) treetaggertokens.get(i)).toXML());
    }
    return out.toString();
  }


	/*
  * wait until we see the initstring from the process
  * this method could block indefinitely but most likely wont
  * this complete module should be redone using NIO
	*/

	public boolean readWait(String wait) throws IOException
	{
		boolean cont = true;	
		StringBuffer sb = new StringBuffer();
		String rl = "";
		while ( (rl =  procErr.readLine() ) != null)
		{
			sb.append(rl.replaceAll("\n", "").replaceAll("\t", "").replaceAll(" ", ""));
			if (sb.toString().trim().equals("readingparameters...tagging..."))
			{
				return true;
			}
		}
		if (sb.toString().trim().equals("readingparameters...tagging..."))
		{
			return true;
		}
		return false;
	}


	
	
} // class
