/*
 * SampleTokenizer.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */


package de.dfki.lt.hog.modules.test;

import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

import de.dfki.lt.hog.Module;
import de.dfki.lt.hog.container.XmlStringBuffer;
import de.dfki.lt.hog.exceptions.InitializationException;
import de.dfki.lt.hog.util.Utility;

/**
 * <code>Tokenizer</code> is a <code>StringTokenizer</code> embedded
 * in a Memphis <code>Module</code>. It is configured through a
 * <code>Properties</code> object which sets its language,
 * depth, delimiters and whether this delimiters are also tokens.
 * It tokenizes through the {@link #process} and
 * {@link #processDOMTree} methods.
 * @version 2003
 * @author <a href="mailto:robba@dfki.de">Robert Barbey</a>
 * @see java.Utility.StringTokenizer */

public class SampleTokenizer
    extends Module {

    /**
     * Key for delimiters in configuration.
     */
    public static final String TOKENIZER_DELIMITERS_PROP = "tokenizer.delimiters";

    /**
     * Key for return delimiters property in configuration.
     */
    public static final String TOKENIZER_RETURN_DELIMITERS_PROP = "tokenizer.return";

    /**
     * Creates a <code>Tokenizer</code> module with given
     * <code>Properties</code>
     * @param props properties for this <code>Tokenizer</code> */
    public SampleTokenizer(Properties props) {
	try {
	    this.init(props);
	} catch (InitializationException e) {
	    moduleLog.error(Utility.exception2String(e));
	}
    }

    /**
     * Tokenizes a given <code>String</code> depending on configuration.
     * @param processString <code>String</code> to tokenize
     * @return XML string containing tokenized document
     */
    public String process(String processString) {
	String delimiters = getConfig().getProperty(TOKENIZER_DELIMITERS_PROP);
	Boolean tempBool = new Boolean(getConfig().getProperty(TOKENIZER_RETURN_DELIMITERS_PROP, null));
	boolean returnDelimiters = tempBool.booleanValue();

	XmlStringBuffer xsb = new XmlStringBuffer();

	// if String is from database it start with XMl stuff
	// cut this off
	if (processString.startsWith("<?xml version=\"1.0\"?>")) {
	    StringBuffer stb = new StringBuffer(processString);
	    int delpos = stb.indexOf("<annotation>");
	    delpos = stb.indexOf(">", delpos);
	    stb.delete(0, delpos+1);
	    delpos = stb.indexOf("<");
	    stb.delete(delpos, stb.length());
	    processString = stb.toString();
	}

	StringTokenizer st = new StringTokenizer(processString, delimiters, returnDelimiters);

	int tokenId = 0;
	int charPosition = -1;

	xsb.addElement("METADATA/");
	xsb.addAttribute("component", getModuleName());
	xsb.addAttribute("user", System.getProperty("user.name"));

	while (st.hasMoreTokens()) {
	    charPosition++;
	    String token = st.nextToken();
	    xsb.addElement("W", token);
	    xsb.addAttribute("id", tokenId);
	    xsb.addAttribute("cstart", charPosition);
	    xsb.addAttribute("cend", charPosition + token.length());
	    charPosition += token.length();
	}

	return xsb.toString();
    }

    public Node processDOMTree(Node domTree)
    {
	// parse domTree
	// create new Document
	// create rootelement, elements and attributes
	// return Document
	DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
	org.w3c.dom.Document resultTree = null;
	SAXParserFactory saxpFactory = SAXParserFactory.newInstance();
	saxpFactory.setValidating(false);


	try {
	    DocumentBuilder builder = dbFactory.newDocumentBuilder();
	    resultTree = builder.newDocument();
	} catch (ParserConfigurationException pce) {
	    moduleLog.error(pce.getMessage(), pce);
	}

	Element root = (Element) resultTree.createElement("TOKENS");

	return resultTree;
    }

}
