/*
 * CorrectCharPosHandler .java
 *
 * (C) 2006  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.util;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import java.io.StringReader;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;

/**
 * Correct byte-offset character position to Unicode char-offset character position 
 * (a fix to EUC-JP-only output of ChaSen for use in Heart of Gold).
 * This code, however, is completely ChaSen-independent and works on any Java-supported
 * byte-oriented encoding.
 * Created 2006-03-01
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @version  $Id: CorrectCharPosHandler.java,v 1.6 2006/06/17 08:49:05 uschaefer Exp $
 */

/* 2006-02-28 : uschaefer: initial version
 * 2006-04-28 : rebecca dridan: fix mapping array length: +1
 * 2006-06-13 : uschaefer: correct cend (-1) for xml_count compatibility
 * 2006-06-15 : uschaefer: insert constant="yes" for punctuation
 */

public class CorrectCharPosHandler extends DefaultHandler
{
    private SAXParserFactory parserFactory = null;
    private StringBuffer output;
    private String cStartAttr;
    private String cEndAttr;
    private int[] posmap;
    private StringBuffer sbWtag = null;
    private boolean bConstant = false;


    /**
     * Correct character positions (instead of byte position as output by ChaSen)
     *
     * @markup original (uncorrected) XML markup as String
     * @rawtext raw text (Unicode String)
     * @encoding name of underlying encoding (e.g. EUC-JP)
     * @cstartattr name of cstart attribute
     * @cendattr name of cend attribute
     * @return modified XML string
     */
    public static String correctCharPos(String markup, String rawtext, String encoding, String cstartattr, String cendattr) 
        throws UnsupportedEncodingException, SAXException, ParserConfigurationException, IOException {
        int[] mapping = new int[(rawtext.getBytes(encoding)).length + 1];
        int i=0;
        byte[] bytes;
        for (int j=0; j<rawtext.length(); j++) {
            bytes=(String.valueOf(rawtext.charAt(j))).getBytes(encoding);
            for (int k=0; k<bytes.length; k++) {
                mapping[i++]=j;
            }
        }
	mapping[i] = mapping[i-1]+1;
        CorrectCharPosHandler handler = new CorrectCharPosHandler();
        return handler.parseAndCorrect(markup, handler, mapping, cstartattr, cendattr);
    }

    public String parseAndCorrect(String markup, CorrectCharPosHandler handler, int[] mapping, String cstartattr, String cendattr) 
        throws SAXException, ParserConfigurationException, IOException {
        if (parserFactory==null) parserFactory = SAXParserFactory.newInstance();
        output = new StringBuffer();
        cStartAttr = cstartattr;
        cEndAttr = cendattr;
        posmap = mapping;
        // Parse the input
        SAXParser saxParser = parserFactory.newSAXParser();
        emit("<?xml version='1.0'?>");
        saxParser.parse(new InputSource(new StringReader(markup)), handler);
        return output.toString();
    }

  
    public void startElement(String namespaceURI,
                             String lName, // local name
                             String qName, // qualified name
                             Attributes attrs)
    throws SAXException
    {
        String eName = lName; // element name
        if ("".equals(eName)) eName = qName; // namespaceAware = false
	
	if ("w".equals(eName)) {
            // sbWtag is used to keep the content of the whole w element
            // including its sub-elements. It is extended with a 
            // constant="yes" attribute in case a punctuation character
            // is indicated by the PoS tag. the whole w is then flushed
            // in endElement if sbWtag
	    sbWtag = new StringBuffer();
	    bConstant = false;
	}
        emit("<"+eName);
        if (attrs != null) {
            for (int i = 0; i < attrs.getLength(); i++) {
                String aName = attrs.getLocalName(i); // Attr name 
                if ("".equals(aName)) aName = attrs.getQName(i);
                emit(" "+aName+"=\"");
		if (aName.equals(cStartAttr))
		    emit(String.valueOf(posmap[Integer.parseInt(attrs.getValue(i))]));
		else if (aName.equals(cEndAttr))
                // subtract one for PIC compatibility (XML_counts)
		emit(String.valueOf(posmap[Integer.parseInt(attrs.getValue(i))]-1));
		else 
		    emit(attrs.getValue(i));
                emit("\"");
                // insert constant="yes" for punctuation characters
		if ((sbWtag!=null) && "tag".equals(aName))
		    if ((attrs.getValue(i)).startsWith("\u8a18\u53f7"))
			bConstant = true;
            }
        }
        emit(">");
    }

    public void endElement(String namespaceURI,
                           String sName, // simple name
                           String qName  // qualified name
                          )
    throws SAXException
    {
        if ("".equals(sName)) sName = qName;
	if ("w".equals(sName)) {
	    if (bConstant) {
		int p = sbWtag.indexOf(">");
		if (p>0) sbWtag.insert(p, " constant=\"yes\"");
	    }
            output.append(sbWtag);
            bConstant = false;
	    sbWtag = null;
	}
        emit("</"+sName+">");
    }

    public void characters(char buf[], int offset, int len)
    throws SAXException
    {
        String s = new String(buf, offset, len);
        if (!s.trim().equals("")) emit(s);
    }


    private void emit(String s)
    {
	if (sbWtag==null)
	    output.append(s);
	else
	    sbWtag.append(s);
    }

}
