/*
 * SproutRmrsTransformerGenerator.java
 *
 * (C) 2004 - 2007  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.util;

import de.dfki.lt.tfs.FeatureStructure;
import de.dfki.lt.tfs.FeatureValuePair;
import de.dfki.lt.tfs.ShUG;
import java.io.*;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.ListIterator;
import java.util.Properties;
import java.util.StringTokenizer;

/*
 * CHANGES 2005-09-21 uschaefer added reading="n" attribute to rmrs (index of
 * reading in matchinfo, starting from 0 for the first reading) 2005-11-30
 * uschaefer renamed "-rel" to "_rel" suffix for type and attribute relation
 * names (added 2 new global constants) 2006-03-14 uschaefer fixed
 * NECSTART/NECEND length bug (array index out of bounds) when filled under
 * non-empty path (OUT; German only) 2007-05-22 uschaefer added ing element
 * (RMRS) 2007-06-06 uschaefer try to reduce stylesheet size because JDK 1.6
 * translet compilation complains 2007-08-02 dcontag template-size reduced
 */

/**
 * @author Ulrich Schaefer Read SProUT TDL type hierarchy from .grm file and
 *         automatically generate a <br>- Sproutput to RMRS XSL transformation
 *         stylesheet <br>- Sproutput to PET XML input chart XSL transformation
 *         stylesheet <br>
 *         for use in SproutModule of the Heart of Gold.<br>
 *         Usage:
 *         <code>SproutRmrsTransformerGenerator hierarchy.grm types.mapping out.path rmrs.xsl pixml.xsl</code>
 *         or (recommended): ant generate_xsl
 * @see de.dfki.lt.hog.modules.SproutModule
 * @see de.dfki.lt.hog.modules.PetModule
 * @version $Date 2005/01/10 23:30:00 $ by $Author: dcontag $
 */
public class SproutRmrsTransformerGenerator {

    /** format for date in comment */
    static SimpleDateFormat datetimeFormat = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z");

    /** additional prefix for type name in RMRS predicate name */
    public static String NE_TYPE_REL_PREFIX = "";

    /** suffix for type name in RMRS predicate name */
    public static String NE_TYPE_REL_SUFFIX = "_rel";

    /** additional prefix for attribute name in RMRS predicate name */
    public static String NE_ATTR_REL_PREFIX = "";

    /** suffix for attribute name in RMRS predicate name */
    public static String NE_ATTR_REL_SUFFIX = "_rel";

    /** name of HPSG default NE type for type mapping */
    public static String NO_HPSG_MAPPING = "NO_HPSG_MAPPING";

    /** default priority for NEs in PET input chart */
    public static String DEFAULT_PRIO = "0.5";

    /**
     * @param args
     */
    public static void main(String[] args) {
        if (args.length != 6) {
            System.err
                    .println("usage: SproutRmrsTransformerGenerator hierarchy.grm types.mapping out.path rmrs.xsl pixml.xsl smaf.xsl");
            System.exit(1);
        }

        ShUG grammar = new ShUG(args[0]);
        String typesfilename = args[1];
        Properties types = new Properties();
        try {
            types.load(new FileInputStream(typesfilename));
        } catch (IOException ioe) {
            System.err.println(ioe.toString());
            System.exit(1);
        }

        String outPrefix = computePrefixPath(args[2]);
        String rmrsfilename = args[3];
        String pixmlfilename = args[4];
        String smaffilename = args[5];
        StringBuffer sbRMRS = new StringBuffer();
        StringBuffer sbPiXml = new StringBuffer();
        StringBuffer sbSmaf = new StringBuffer();
        generateXSL(grammar, types, sbRMRS, sbPiXml, sbSmaf, outPrefix);

        writeStringToFile(rmrsfilename, sbRMRS.toString(), "utf-8");
        writeStringToFile(pixmlfilename, sbPiXml.toString(), "utf-8");
        writeStringToFile(smaffilename, sbSmaf.toString(), "utf-8");
    }

    private static String computePrefixPath(String path) {
        if (path == null || path.equals(""))
            return "";
        else {
            StringTokenizer st = new StringTokenizer(path, " .|:");
            StringBuffer xpath = new StringBuffer();
            while (st.hasMoreTokens()) {
                xpath.append("FS/F[@name=\"" + st.nextToken() + "\"]/");
            }
            return xpath.toString();
        }
    }

    /**
     * @param grammar
     * @param pTypes
     * @param sbRMRS
     * @param sbPiXml
     */
    private static void generateXSL(ShUG grammar, Properties pTypes, StringBuffer sbRMRS, StringBuffer sbPiXml,
            StringBuffer sbSmaf, String outPrefix) {
        generateHeader1(sbRMRS, "translates SProUTput XML to NE RMRS");
        generateHeader1(sbPiXml, "translates SProUTput XML to PET input chart format");
        generateHeader1(sbSmaf, "translates SProUTput XML to SMAF");

        Enumeration eTypes = pTypes.keys();
        while (eTypes.hasMoreElements()) {
            String key = (String) eTypes.nextElement();
            if (grammar.getNumberFromType(key) >= 0) {
                String val = pTypes.getProperty(key);

                sbRMRS.append("<!-- " + key + " -->\n");
                sbPiXml.append("<!-- " + key + " mapped to HPSG type " + (val.equals("") ? NO_HPSG_MAPPING : val)
                        + " -->\n");
                sbSmaf.append("<!-- " + key + " mapped to HPSG type " + (val.equals("") ? NO_HPSG_MAPPING : val)
                        + " -->\n");

            } else
                System.err.println("Type " + key + " is not defined in grammar.");
        }

        generateHeader2(sbRMRS, "xml", "rmrs-list");
        generateHeader2(sbPiXml, "xml", "pet-input-chart");
        generateSmafHeader2(sbSmaf, "xml");

        generateRmrsHeader1(sbRMRS, outPrefix);
        generatePiXmlHeader1(sbPiXml, outPrefix);
        generateSmafHeader1(sbSmaf, outPrefix);

        // as before
        eTypes = pTypes.keys();
        while (eTypes.hasMoreElements()) {
            String type = (String) eTypes.nextElement();
            int tcode = grammar.getNumberFromType(type);
            if (tcode >= 0) {
                FeatureStructure fs = grammar.getPi(tcode);
                generateRmrsDispatch(sbRMRS, type, fs, grammar, outPrefix);
                generatePiXmlDispatch(sbPiXml, type, pTypes.getProperty(type), fs);

                // TODO: testweise dasselbe
                generatePiXmlDispatch(sbSmaf, type, pTypes.getProperty(type), fs);

            }
        }
        generateRmrsHeader2(sbRMRS);

        // generate named templates
        eTypes = pTypes.keys();
        while (eTypes.hasMoreElements()) {
            String type = (String) eTypes.nextElement();
            int tcode = grammar.getNumberFromType(type);
            if (tcode >= 0) {
                FeatureStructure fs = grammar.getPi(tcode);
                generateNamedTemplate(sbRMRS, type, fs, grammar, outPrefix);
            }
        }

        generatePiXmlHeader2(sbPiXml);
        generateSmafHeader2(sbSmaf);

        sbRMRS.append("</xsl:stylesheet>");
        sbPiXml.append("</xsl:stylesheet>");
        sbSmaf.append("</xsl:stylesheet>");
    }

    /**
     * @param sbf
     * @param type
     * @param fs
     */
    private static void showTypeDefinition(StringBuffer sbf, String type, FeatureStructure fs) {
        sbf.append("      <!-- TDL definition (TFS) for type " + type + ": -->\n");
        sbf.append("      <!--\n");
        sbf.append("      ");
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        fs.printFS(pw);
        sbf.append(sw.toString());
        sbf.append("      -->\n");
    }

    /**
     * @param sbf
     * @param comment
     */
    private static void generateHeader1(StringBuffer sbf, String comment) {
        sbf.append("<?xml version='1.0'?>\n");
        sbf.append("<xsl:stylesheet  version='1.0' xmlns:xsl='http://www.w3.org/1999/XSL/Transform'>\n");
        sbf.append("<!-- " + comment + " -->\n");
        sbf.append("<!-- DO NOT EDIT THIS FILE - it has been generated automatically -->\n");
        sbf.append("<!-- by de.dfki.lt.hog.util.SproutRmrsTransformerGenerator -->\n");
        sbf.append("<!-- using the commandline ./ant generate_xsl [-Dproject=...] -->\n");
        sbf.append("<!-- created on " + formatTime() + " by " + System.getProperty("user.name") + " -->\n\n");
        sbf.append("<!-- \n");
        sbf.append("     (C) 2004 - 2007  DFKI Language Technology Lab http://www.dfki.de/lt\n");
        sbf.append("     Author: Ulrich.Schaefer@dfki.de\n\n");
        sbf.append("     This program is free software; you can redistribute it and/or\n");
        sbf.append("     modify it under the terms of the GNU Lesser General Public\n");
        sbf.append("     License as published by the Free Software Foundation; either\n");
        sbf.append("     version 2.1 of the License, or (at your option) any later version.\n\n");
        sbf.append("     This program is distributed in the hope that it will be useful,\n");
        sbf.append("     but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
        sbf.append("     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU\n");
        sbf.append("     Lesser General Public License for more details.\n\n");
        sbf.append("     You should have received a copy of the GNU Lesser General Public\n");
        sbf.append("     License along with this library; if not, write to the Free Software\n");
        sbf.append("     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA\n");
        sbf.append("-->\n\n");
        sbf.append("<!-- recognised types: -->\n");
    }

    /**
     * @param sbf
     * @param method
     * @param root
     */
    private static void generateHeader2(StringBuffer sbf, String method, String root) {
        sbf.append("\n");
        sbf.append("  <xsl:output method='" + method + "' indent='yes'/>\n");
        sbf.append("  <xsl:strip-space elements='*'/>\n");
        sbf.append("  <xsl:param name='sessionID' select='\"unknownsid\"'/>\n");
        sbf.append("  <xsl:param name='collectionID' select='\"unknownacid\"'/>\n");
        sbf.append("  <xsl:param name='annotationID' select='\"unknownaid\"'/>\n");
        sbf
                .append("  <xsl:variable name='uri' select=\"concat('hog://',$sessionID,'/',$collectionID,'/xmltext')\"/>\n");
        sbf.append("  <xsl:variable name='surface' select='document($uri)/text/text()'/>\n\n");
        sbf.append("  <!-- process root element -->\n");
        sbf.append("  <xsl:template match='/SPROUTPUT'>\n");
        sbf.append("    <" + root + ">\n");
        sbf.append("      <xsl:apply-templates/>\n");
        sbf.append("    </" + root + ">\n");
        sbf.append("  </xsl:template>\n\n");
    }
    /**
     * @param sbf
     * @param method
     * @param root
     */
    private static void generateSmafHeader2(StringBuffer sbf, String method) {
        sbf.append("\n");
        sbf.append("  <xsl:output method='" + method + "' indent='yes'/>\n");
        sbf.append("  <xsl:strip-space elements='*'/>\n");
        sbf.append("  <xsl:output doctype-system=\"smaf.dtd\"/>\n");
        sbf.append("  <xsl:param name='sessionID' select='\"unknownsid\"'/>\n");
        sbf.append("  <xsl:param name='collectionID' select='\"unknownacid\"'/>\n");
        sbf.append("  <xsl:param name='annotationID' select='\"unknownaid\"'/>\n");
        sbf.append("  <xsl:variable name='uri' select=\"concat('hog://',$sessionID,'/',$collectionID,'/xmltext')\"/>\n");
        sbf.append("  <xsl:variable name='surface' select='document($uri)/text/text()'/>\n\n");
        sbf.append("  <!-- process root element -->\n");
        sbf.append("  <xsl:template match='/SPROUTPUT'>\n");
        sbf.append("    <smaf document = \"\">\n");
        sbf.append("      <lattice>");
        sbf.append("        <xsl:apply-templates/>\n");
        sbf.append("      </lattice>");
        sbf.append("    </smaf>\n");
        sbf.append("  </xsl:template>\n\n");
    }

    /**
     * @param sbf
     */
    private static void generateRmrsHeader1(StringBuffer sbf, String outPrefix) {
        sbf.append("  <xsl:template match='metadata'>\n");
        sbf.append("    <xsl:copy-of select='.'/>\n");
        sbf.append("  </xsl:template>\n\n");

        sbf.append("  <xsl:template name='listlist'>\n");
        sbf.append("  	<xsl:param name='node'/>\n");
        sbf.append("    <xsl:param name='ctype' select='\"*cons*\"'/>\n");
        sbf.append("    <xsl:variable name='cs' select='$node/F[@name=\"FIRST\"]/FS/@type'/>\n");
        sbf.append("    <xsl:if test=\"$cs and starts-with($cs, '&quot;')\">\n");
        sbf.append("      <xsl:value-of select=\"translate($cs, '&quot;', '')\"/>\n");
        // translate($cs, '&quot;', '')
        sbf.append("      <xsl:if test='$node/F[@name=\"REST\"]/FS/@type = $ctype'>\n");
        sbf.append("        <xsl:text> </xsl:text>\n");
        sbf.append("        <xsl:call-template name='listlist'>\n");
        sbf.append("          <xsl:with-param name='node' select='$node/F[@name=\"REST\"]/FS'/>\n");
        sbf.append("          <xsl:with-param name='ctype' select='$ctype'/>\n");
        sbf.append("        </xsl:call-template>\n");
        sbf.append("      </xsl:if>\n");
        sbf.append("    </xsl:if>\n");
        sbf.append("  </xsl:template>\n\n");

        sbf.append("  <xsl:template match='MATCHINFO'>\n");
        sbf.append("    <xsl:variable name='reading' select='position()-1'/>\n");
        sbf.append("    <xsl:variable name='offset' select='100+100*$reading'/>\n");
        sbf.append("    <xsl:variable name='enamextype' select='" + outPrefix + "FS/@type'/>\n");
        sbf.append("    <xsl:variable name='cfrom'>\n");
        sbf.append("      <xsl:choose>\n");
        sbf.append("        <xsl:when test='" + outPrefix + "FS/F[@name=\"NECSTART\"] and (" + outPrefix
                + "FS/F[@name=\"NECSTART\"]/FS/@type != \"string\")'>\n");
        sbf.append("          <xsl:value-of select='substring(" + outPrefix
                + "FS/F[@name=\"NECSTART\"]/FS/@type, 2, string-length(" + outPrefix
                + "FS/F[@name=\"NECSTART\"]/FS/@type)-2)'/>\n");
        sbf.append("        </xsl:when>\n");
        sbf.append("        <xsl:otherwise>\n");
        sbf.append("          <xsl:value-of select='@cstart'/>\n");
        sbf.append("        </xsl:otherwise>\n");
        sbf.append("      </xsl:choose>\n");
        sbf.append("    </xsl:variable>\n");
        sbf.append("    <xsl:variable name='cto'>\n");
        sbf.append("      <xsl:choose>\n");
        sbf.append("        <xsl:when test='" + outPrefix + "FS/F[@name=\"NECEND\"] and (" + outPrefix
                + "FS/F[@name=\"NECEND\"]/FS/@type != \"string\")'>\n");
        sbf.append("          <xsl:value-of select='substring(" + outPrefix
                + "FS/F[@name=\"NECEND\"]/FS/@type, 2, string-length(" + outPrefix
                + "FS/F[@name=\"NECEND\"]/FS/@type)-2)'/>\n");
        sbf.append("        </xsl:when>\n");
        sbf.append("        <xsl:otherwise>\n");
        sbf.append("          <xsl:value-of select='@cend'/>\n");
        sbf.append("        </xsl:otherwise>\n");
        sbf.append("      </xsl:choose>\n");
        sbf.append("    </xsl:variable>\n");
        sbf.append("    <xsl:variable name='rsurface' select='substring($surface, $cfrom + 2, $cto - $cfrom + 1)'/>\n");
        sbf.append("    <xsl:choose>\n\n");

    }

    /**
     * @param sbf
     */
    private static void generatePiXmlHeader1(StringBuffer sbf, String outPrefix) {
        sbf.append("  <xsl:template match='metadata'>\n");
        sbf.append("    <xsl:copy-of select='.'/>\n");
        sbf.append("  </xsl:template>\n\n");

        sbf.append("  <xsl:template match='DISJ'>\n");
        sbf.append("    <xsl:apply-templates select='MATCHINFO'>\n");
        sbf.append("      <xsl:with-param name='disjindex' select='position()'/>\n");
        sbf.append("    </xsl:apply-templates>\n");
        sbf.append("  </xsl:template>\n\n");

        sbf.append("  <xsl:template match='MATCHINFO'>\n");
        sbf.append("    <xsl:param name='disjindex'/>\n");
        sbf.append("    <xsl:variable name='index' select='concat(string($disjindex),\".\",string(position()))'/>\n");
        sbf.append("    <xsl:variable name='enamextype' select='" + outPrefix + "FS/@type'/>\n");
        sbf.append("    <xsl:variable name='cfrom'>\n");
        sbf.append("      <xsl:choose>\n");
        sbf.append("        <xsl:when test='" + outPrefix + "FS/F[@name=\"NECSTART\"] and (" + outPrefix
                + "FS/F[@name=\"NECSTART\"]/FS/@type != \"string\")'>\n");
        sbf.append("          <xsl:value-of select='substring(" + outPrefix
                + "FS/F[@name=\"NECSTART\"]/FS/@type, 2, string-length(" + outPrefix
                + "FS/F[@name=\"NECSTART\"]/FS/@type)-2)'/>\n");
        sbf.append("        </xsl:when>\n");
        sbf.append("        <xsl:otherwise>\n");
        sbf.append("          <xsl:value-of select='@cstart'/>\n");
        sbf.append("        </xsl:otherwise>\n");
        sbf.append("      </xsl:choose>\n");
        sbf.append("    </xsl:variable>\n");
        sbf.append("    <xsl:variable name='cto'>\n");
        sbf.append("      <xsl:choose>\n");
        sbf.append("        <xsl:when test='" + outPrefix + "FS/F[@name=\"NECEND\"] and (" + outPrefix
                + "FS/F[@name=\"NECEND\"]/FS/@type != \"string\")'>\n");
        sbf.append("          <xsl:value-of select='substring(" + outPrefix
                + "FS/F[@name=\"NECEND\"]/FS/@type, 2, string-length(" + outPrefix
                + "FS/F[@name=\"NECEND\"]/FS/@type)-2)'/>\n");
        sbf.append("        </xsl:when>\n");
        sbf.append("        <xsl:otherwise>\n");
        sbf.append("          <xsl:value-of select='@cend'/>\n");
        sbf.append("        </xsl:otherwise>\n");
        sbf.append("      </xsl:choose>\n");
        sbf.append("    </xsl:variable>\n");
        sbf.append("    <xsl:variable name='rsurface' select='substring($surface, $cfrom + 2, $cto - $cfrom + 1)'/>\n");
        sbf.append("    <xsl:variable name='hpsgtype'>\n");
        sbf.append("      <xsl:choose>\n");
    }

    /**
     * @param sbf
     */
    private static void generateSmafHeader1(StringBuffer sbf, String outPrefix) {
        sbf.append("  <xsl:template match='DISJ'>\n");
        sbf.append("    <xsl:apply-templates select='MATCHINFO'>\n");
        sbf.append("      <xsl:with-param name='disjindex' select='position()'/>\n");
        sbf.append("    </xsl:apply-templates>\n");
        sbf.append("  </xsl:template>\n\n");

        sbf.append("  <xsl:template match='MATCHINFO'>\n");
        sbf.append("    <xsl:param name='disjindex'/>\n");
        sbf.append("    <xsl:variable name='index' select='concat(string($disjindex),\".\",string(position()))'/>\n");
        sbf.append("    <xsl:variable name='enamextype' select='" + outPrefix + "FS/@type'/>\n");
        sbf.append("    <xsl:variable name='cfrom'>\n");
        sbf.append("      <xsl:choose>\n");
        sbf.append("        <xsl:when test='" + outPrefix + "FS/F[@name=\"NECSTART\"] and (" + outPrefix
                + "FS/F[@name=\"NECSTART\"]/FS/@type != \"string\")'>\n");
        sbf.append("          <xsl:value-of select='substring(" + outPrefix
                + "FS/F[@name=\"NECSTART\"]/FS/@type, 2, string-length(" + outPrefix
                + "FS/F[@name=\"NECSTART\"]/FS/@type)-2)'/>\n");
        sbf.append("        </xsl:when>\n");
        sbf.append("        <xsl:otherwise>\n");
        sbf.append("          <xsl:value-of select='@cstart'/>\n");
        sbf.append("        </xsl:otherwise>\n");
        sbf.append("      </xsl:choose>\n");
        sbf.append("    </xsl:variable>\n");
        sbf.append("    <xsl:variable name='cto'>\n");
        sbf.append("      <xsl:choose>\n");
        sbf.append("        <xsl:when test='" + outPrefix + "FS/F[@name=\"NECEND\"] and (" + outPrefix
                + "FS/F[@name=\"NECEND\"]/FS/@type != \"string\")'>\n");
        sbf.append("          <xsl:value-of select='substring(" + outPrefix
                + "FS/F[@name=\"NECEND\"]/FS/@type, 2, string-length(" + outPrefix
                + "FS/F[@name=\"NECEND\"]/FS/@type)-2)'/>\n");
        sbf.append("        </xsl:when>\n");
        sbf.append("        <xsl:otherwise>\n");
        sbf.append("          <xsl:value-of select='@cend'/>\n");
        sbf.append("        </xsl:otherwise>\n");
        sbf.append("      </xsl:choose>\n");
        sbf.append("    </xsl:variable>\n");
        sbf.append("    <xsl:variable name='rsurface' select='substring($surface, $cfrom + 2, $cto - $cfrom + 1)'/>\n");
        sbf.append("    <xsl:variable name='hpsgtype'>\n");
        sbf.append("      <xsl:choose>\n");
    }

    /**
     * @param sbf
     * @param type
     * @param translation
     * @param fs
     */
    private static void generatePiXmlDispatch(StringBuffer sbf, String type, String translation, FeatureStructure fs) {
        sbf.append("        <xsl:when test='$enamextype = \"" + type + "\"'>\n");
        sbf.append("          <xsl:value-of select='\"" + (translation.equals("") ? NO_HPSG_MAPPING : translation)
                + "\"'/>\n");
        sbf.append("        </xsl:when>\n\n");
    }

    /**
     * @param sbf
     */
    private static void generatePiXmlHeader2(StringBuffer sbf) {
        sbf.append("      </xsl:choose>\n");
        sbf.append("    </xsl:variable>\n\n");
        sbf.append("    <!-- generate new PET XML input chart format -->\n");
        sbf.append("    <xsl:if test='$hpsgtype != \"" + NO_HPSG_MAPPING + "\"'>\n");
        sbf.append("      <w id='{concat(\"SPR\", $index)}' cstart='{$cfrom}' cend='{$cto}'");
        sbf.append(" constant='yes' prio='" + DEFAULT_PRIO + "'>\n");
        sbf.append("        <surface><xsl:value-of select='$rsurface'/></surface>\n");
        /*
         * this is redundant (would only apply if a mapping for an HPSG type was
         * not defined which shouldn't be the case) sbf.append(" <pos tag='NNP'
         * prio='1.0'/>\n");
         */
        sbf.append("        <typeinfo id='{concat(\"TIN\", $index)}' baseform='no'>\n");
        sbf.append("          <stem>$<xsl:value-of select='$hpsgtype'/></stem>\n");
        sbf.append("        </typeinfo>\n");
        sbf.append("      </w>\n");
        sbf.append("    </xsl:if>\n\n");
        sbf.append("  </xsl:template>\n");
    }

    /**
     * @param sbf
     */
    private static void generateSmafHeader2(StringBuffer sbf) {
        sbf.append("      </xsl:choose>\n");
        sbf.append("    </xsl:variable>\n\n");
        sbf.append("    <!-- generate new SMAF -->\n");
        sbf.append("    <xsl:if test='$hpsgtype != \"" + NO_HPSG_MAPPING + "\"'>\n");
        
        /* sbf.append("      <edge type=\"token\" id='{concat(\"t\", $disjindex)}'" +
                        " cfrom='{$cfrom}' cto='{$cto + 1}'" +
                        " source='{concat(\"v\", $disjindex - 1)}'" +
                        " target='{concat(\"v\", $disjindex)}' >\n");
        sbf.append("        <slot name='weight'>" + DEFAULT_PRIO + "</slot>");
        sbf.append("      <xsl:value-of select='$rsurface'/></edge>\n");
         */
        
        sbf.append("      <edge type= \"namedEntity\" id='{concat(\"n\", $disjindex)}'" +
                              " cfrom='{$cfrom}' cto='{$cto + 1}'" +
                              " source='{concat(\"v\", $disjindex - 1)}'" +
                              " target='{concat(\"v\", $disjindex)}' >\n");
        sbf.append("        <slot name='weight'>" + DEFAULT_PRIO + "</slot>\n");
        sbf.append("        <slot name=\"surface\"> <xsl:value-of select='$rsurface'/> </slot>\n");
        sbf.append("        <fs type='{$hpsgtype}'></fs>\n");
        sbf.append("      </edge>\n");
        sbf.append("    </xsl:if>\n\n");
        sbf.append("  </xsl:template>\n");
    }

    /**
     * @param sbf
     */
    private static void generateRmrsHeader2(StringBuffer sbf) {
        sbf.append("    </xsl:choose>\n");
        sbf.append("  </xsl:template>\n");
    }

    /**
     * @param sbf
     * @param type
     * @param fs
     * @param tcode
     * @param grammar
     */
    private static void generateRmrsDispatch(StringBuffer sbf, String type, FeatureStructure fs, ShUG grammar,
            String outPrefix) {
        // showTypeDefinition(sbf, type, fs);
        System.out.println(NE_TYPE_REL_PREFIX + type + NE_TYPE_REL_SUFFIX + ":");
        sbf.append("      <xsl:when test='$enamextype = \"" + type + "\"'>\n");

        // append here call to named template.
        sbf.append("        <xsl:call-template name='" + type + "'>\n");
        // needed parameters
        sbf.append("          <xsl:with-param name=\"cfrom\" select=\"$cfrom\" />\n");
        sbf.append("          <xsl:with-param name=\"cto\" select=\"$cto\" />\n");
        sbf.append("          <xsl:with-param name=\"reading\" select=\"$reading\" />\n");
        sbf.append("          <xsl:with-param name=\"rsurface\" select=\"$rsurface\" />\n");
        sbf.append("          <xsl:with-param name=\"offset\" select=\"$offset\" />\n");
        sbf.append("          <xsl:with-param name=\"enamextype\" select=\"$enamextype\" />\n");

        sbf.append("        </xsl:call-template>\n");
        sbf.append("      </xsl:when>\n\n");
    }

    private static void generateNamedTemplate(StringBuffer sbf, String type, FeatureStructure fs, ShUG grammar,
            String outPrefix) {

        // start of named template
        sbf.append("      <xsl:template name =\"" + type + "\">\n");

        // get all parameters
        sbf.append("        <xsl:param name=\"cfrom\" />\n");
        sbf.append("        <xsl:param name=\"cto\" />\n");
        sbf.append("        <xsl:param name=\"reading\" />\n");
        sbf.append("        <xsl:param name=\"rsurface\" />\n");
        sbf.append("        <xsl:param name=\"offset\" />\n");
        sbf.append("        <xsl:param name=\"enamextype\" />\n");
        // RMRS begin
        sbf.append("        <rmrs cfrom='{$cfrom}' cto='{$cto}' reading='{$reading}' surface='{$rsurface}'>\n");
        // vid
        sbf.append("          <label vid='{$offset}'/>\n");
        // main EP
        sbf.append("          <ep cfrom='{$cfrom}' cto='{$cto}' surface='{$rsurface}'>\n");
        sbf.append("            <gpred><xsl:value-of select='concat(\"" + NE_TYPE_REL_PREFIX + "\", $enamextype, \""
                + NE_TYPE_REL_SUFFIX + "\")'/></gpred>\n");
        sbf.append("            <label vid='{$offset}'/>\n");
        sbf.append("            <var sort='x' vid='{$offset}'/>\n");
        sbf.append("          </ep>\n");
        // CARG for main EP
        sbf.append("          <rarg>\n");
        sbf.append("            <label vid='{$offset}'/>\n");
        sbf.append("            <rargname>CARG</rargname>\n");
        sbf.append("            <constant><xsl:value-of select=\"translate($rsurface, '&quot;', '')\"/></constant>\n");
        sbf.append("          </rarg>\n\n");

        // iterate over ne type attributes
        ArrayList attribs = fs.getFeatureValueList();
        ListIterator it = attribs.listIterator();
        String attribute;
        String valuetype;
        FeatureValuePair fvp;
        int i = 0;

        // for span/point, call with outpath appended:
        // FS[@type=\"'span'\"/F[@name=\"TO\"]/
        while (it.hasNext()) {
            fvp = (FeatureValuePair) it.next();
            attribute = grammar.getFeatureFromNumber(fvp.getFeature());
            if (attribute.equals("CSTART") || attribute.equals("CEND") || attribute.equals("NECSTART")
                    || attribute.equals("NECEND") || attribute.equals("ID"))
                continue;
            valuetype = ((FeatureStructure) fvp.getValue()).getTypeName();
            i++; // start with 1 !
            System.out.println("  " + NE_ATTR_REL_PREFIX + attribute.toLowerCase() + NE_ATTR_REL_SUFFIX);
            if (valuetype.equals("*list*") || valuetype.equals("*opencons*") || valuetype.equals("*cons*")) {
                String ctype = "\"*cons*\"";
                if (valuetype.equals("*opencons*"))
                    ctype = "\"*opencons*\"";
                sbf.append("          <xsl:variable name='sprout" + attribute + "'>\n");
                sbf.append("            <xsl:call-template name='listlist'>\n");
                sbf.append("               <xsl:with-param name='node' select='" + outPrefix + "FS[@type=\"" + type);
                sbf.append("\"]/F[@name=\"" + attribute + "\"]/FS'/>\n");
                sbf.append("               <xsl:with-param name='ctype' select='" + ctype + "'/>\n");
                sbf.append("            </xsl:call-template>\n");
                sbf.append("          </xsl:variable>\n");
                sbf.append("          <xsl:if test='$sprout" + attribute + " and ($sprout" + attribute
                        + " != \"\")'>\n");
            } else {
                sbf.append("          <xsl:variable name='sprout" + attribute + "' select='" + outPrefix
                        + "FS[@type=\"" + type);
                sbf.append("\"]/F[@name=\"" + attribute + "\"]/FS/@type'/>\n");
                sbf.append("          <xsl:if test='($sprout" + attribute + " != \"" + valuetype + "\") and ($sprout"
                        + attribute + " != \"\")'>\n");
            }
            sbf.append("            <ep cfrom='{$cfrom}' cto='{$cto}' surface='{$rsurface}'>\n");
            sbf.append("              <gpred>" + NE_ATTR_REL_PREFIX + attribute.toLowerCase() + NE_ATTR_REL_SUFFIX
                    + "</gpred>\n");
            sbf.append("              <label vid='{$offset + " + i + "}'/>\n");

            sbf.append("              <var sort='x' vid='{$offset + " + i + "}'/>\n");
            sbf.append("            </ep>\n");

            sbf.append("            <rarg>\n");
            sbf.append("              <label vid='{$offset + " + i + "}'/>\n");
            sbf.append("              <rargname>CARG</rargname>\n");
            sbf.append("              <constant><xsl:value-of select=\"translate($sprout" + attribute
                    + ", '&quot;', '')\"/></constant>\n");
            sbf.append("            </rarg>\n");

            sbf.append("            <rarg>\n");
            sbf.append("              <label vid='{$offset + " + i + "}'/>\n");
            sbf.append("              <rargname>ARG1</rargname>\n");
            sbf.append("              <var sort='x' vid='{$offset}'/>\n");
            sbf.append("            </rarg>\n");
            sbf.append("            <ing>\n");
            sbf.append("              <ing-a><var sort='h' vid='{$offset}'/></ing-a>\n");
            sbf.append("              <ing-b><var sort='h' vid='{$offset + " + i + "}'/></ing-b>\n");
            sbf.append("            </ing>\n");
            sbf.append("          </xsl:if>\n\n");
        }

        sbf.append("        </rmrs>\n");
        sbf.append("      </xsl:template>\n");
        // end of content of named template
    }

    /**
     * @param sFilename
     * @param sContent
     * @param encoding
     */
    private static void writeStringToFile(String sFilename, String sContent, String encoding) {
        try {
            BufferedWriter file = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(sFilename), encoding));
            file.write(sContent);
            file.newLine();
            file.flush();
            file.close();
        } catch (IOException e) {
            System.err.println(e.toString());
        }
    }

    /**
     * format date and time as ISO/RFC string
     * 
     * @return String containing formatted current time and date
     */
    private static String formatTime() {
        return datetimeFormat.format(new Date());
    }

    /**
     * Read a file completely and return content as String
     * 
     * @param sFilename
     * @param encoding
     * @return text from file as String
     */
    private static String readStringFromFile(String sFilename, String encoding) {
        try {
            byte byTextBuffer[];
            File textFile = new File(sFilename);
            BufferedInputStream in = new BufferedInputStream(new FileInputStream(textFile));
            byTextBuffer = new byte[(int) textFile.length()];
            int iTextLength = in.read(byTextBuffer);
            in.close();
            return new String(byTextBuffer, encoding);
        } catch (IOException e) {
            System.err.println(e.toString());
            return null;
        }
    }
}
