/*
 * Utility.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hog.util;

import java.io.*;
import java.text.*;
import java.util.*;
import java.util.regex.*;
import javax.xml.parsers.*;
import org.apache.log4j.Logger;
import org.apache.xml.serialize.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import de.dfki.lt.hog.MoCoMan;
import de.dfki.lt.hog.container.JToken;

/**
 * Utility class with static methods/constants.
 * Created 2003-11-19
 * @author <a href="mailto:ulrich.schaefer@dfki.de">Ulrich Schaefer</a>
 * @version $Id: Utility.java,v 1.4 2005/08/24 22:45:24 oeze Exp $
 */
public class Utility {
  /** A logger for <code>Module</code>s */
  static protected Logger log = Logger.getLogger(Utility.class);

  public static final String METADATA_ELEMENT = "metadata";
  public static final String SESSIONID_ATTRIBUTE = "sessionid";
  public static final String ACID_ATTRIBUTE = "acid";
  private static final String NEWLINE = System.getProperty("line.separator");

  /** time and date formats for formatXXX methods  */
  static SimpleDateFormat datetimeFormat =
      new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z");
  static SimpleDateFormat durationFormat = new SimpleDateFormat("mm:ss,SS");

  public static String getNewLine() {
    return NEWLINE;
  }

  /** format duration encoded in Date object as String mm:ss,milliseconds
   * @param duration
   * @return String containing formatted duration
   */
  public static String formatDuration(Date duration) {
    return durationFormat.format(duration);
  }

  /**
   * Format a time span given in milliseconds to a String.
   * Example: 2 days 21 hours 11 minutes 32 seconds 382 milliseconds
   * @param totalMillis The time span in milliseconds
   * @return A formatted String
   */
  public static String formatDuration(long totalMillis) {
    StringBuffer sb = new StringBuffer();
    long rest;

    int days = (int) (totalMillis / 86400000);
    rest = totalMillis - (days * 86400000);

    int hours = (int) (rest / 3600000);
    rest = rest - (hours * 3600000);

    int minutes = (int) (rest / 60000);
    rest = rest - (minutes * 60000);

    int seconds = (int) (rest / 1000);
    rest = rest - (seconds * 1000);

    int millis = Math.round(rest);

    if (days > 0) {
      sb.append(days + " days ");
    }

    if (hours > 0) {
      sb.append(hours + " hours ");
    }

    if (minutes > 0) {
      sb.append(minutes + " minutes ");
    }

    if (seconds > 0) {
      sb.append(seconds + " seconds ");
    }

    if (millis > 0) {
      sb.append(millis + " milliseconds");
    }

    return sb.toString();
  } // formatDuration(long)

  /**
  * Format a time span given in milliseconds to a String. Short form.
  * Example: 1T21:11:32
  * @param totalMillis The time span in milliseconds
  * @return A formatted String
  */
 public static String formatDurationShort(long totalMillis) {
   StringBuffer sb = new StringBuffer();
   long rest;

   int days = (int) (totalMillis / 86400000);
   rest = totalMillis - (days * 86400000);

   int hours = (int) (rest / 3600000);
   rest = rest - (hours * 3600000);

   int minutes = (int) (rest / 60000);
   rest = rest - (minutes * 60000);

   int seconds = (int) (rest / 1000);
   rest = rest - (seconds * 1000);

   int millis = Math.round(rest);

   if (days > 0) {
     sb.append(days + "T");
   }

   String h;
   if (hours < 10) {
     h = "0" + hours;
   } else {
     h = "" + hours;
   }
   sb.append(h + ":");

   String m;
   if (minutes < 10) {
     m = "0" + minutes;
   } else {
     m = "" + minutes;
   }
   sb.append(m + ":");

   String s;
   if (seconds < 10) {
     s = "0" + seconds;
   } else {
     s = "" + seconds;
   }
   sb.append(s);

   return sb.toString();
 } // formatDuration(long)


  /** format date and time as ISO/RFC string
   * @return String containing formatted current time and date
   */
  public static String formatTime() {
    return datetimeFormat.format(new Date());
  }

  /** Get computed annotation from MoCoMan, taking sessionID and annotationCollection name from metadata of the input annotation
      The purpose of this method is to provide (from within a Module implementation) access to annotations other than the
      previous one. The implementation via String parsing is an ugly hack.
   * @param mocoman  MoCoMan instance
   * @param inputAnnotation  annotation String that contains the previous annotation with metadata
   * @param seekAnnotationName name of the annotation
   * @return the computed annotation as String
   */
  public static String getComputedAnnotation(MoCoMan mocoman,
                                             String inputAnnotation,
                                             String seekAnnotationName) {

    //TODO true XML parsing
    //TODO write DOM variant

    int pos = inputAnnotation.indexOf(METADATA_ELEMENT);
    if (pos < 0) {
      return null;
    }
    pos += METADATA_ELEMENT.length() + 1;
    String sessionID = getAttributeValue(inputAnnotation, SESSIONID_ATTRIBUTE,
                                         pos);
    if (sessionID == null) {
      return null;
    }
    String acID = getAttributeValue(inputAnnotation, ACID_ATTRIBUTE, pos);
    if (acID == null) {
      return null;
    }
    return mocoman.getAnnotation(sessionID, acID, seekAnnotationName);
  }

  public static String getAttributeValue(String inputString, String attribute,
                                         int startpos) {
    int pos1 = inputString.indexOf(attribute, startpos); // skip attribute name
    if (pos1 < 0) {
      return null;
    }
    int pos2 = inputString.indexOf("\"", pos1 + attribute.length() + 1); // seek starting "
    if (pos2 < 0) {
      return null;
    }
    int pos3 = inputString.indexOf("\"", pos2 + 1); // seek ending "
    if (pos3 < 0) {
      return null;
    }
    return inputString.substring(pos2 + 1, pos3); // return attribute value
  }

  /** read string from file
   * @param sFilename
   * @param encoding
   * @return text from file as String
   */
  public static String readStringFromFile(String sFilename, String encoding) {
    try {
      byte byTextBuffer[];
      File textFile = new File(sFilename);
      BufferedInputStream in =
          new BufferedInputStream(new FileInputStream(textFile));
      byTextBuffer = new byte[ (int) textFile.length()];
      int iTextLength = in.read(byTextBuffer);
      in.close();
      return new String(byTextBuffer, encoding);
    }
    catch (IOException e) {
      log.error(Utility.exception2String(e));
      return null;
    }
  }

  /** write string to file
   * @param sFilename
   * @param sContent
   */
  public static void writeStringToFile(String sFilename, String sContent) {
    try {
      BufferedWriter file = new BufferedWriter(new FileWriter(sFilename));
      file.write(sContent);
      file.newLine();
      file.flush();
      file.close();
    }
    catch (IOException e) {
      log.error(Utility.exception2String(e));
    }
  }

  /**
   * Converts a <code>Map</code> to a <code>String</code>. Maps can be
   * AbstractMap, Attributes, HashMap, Hashtable, IdentityHashMap,
   * RenderingHints, TreeMap or WeakHashMap
   * @param map The Map to be printed.
   * @return The String representing the given Map.
   */
  public static String map2String(Map map) {
    return map2String(map, "", "");
  } // map2String()

  /**
   * Converts a <code>Map</code> to a <code>String</code>. Maps can be
   * AbstractMap, Attributes, HashMap, Hashtable, IdentityHashMap,
   * RenderingHints, TreeMap or WeakHashMap.
   * Each line will be followed by a line separator
   * (e.g. &quot;\n&quot; for Linux).
   * @param map The Map to be printed.
   * @param prefix A String preceding each line.
   * @return The String representing the given Map.
   */
  public static String map2String(Map map, String prefix) {
    return map2String(map, prefix, System.getProperty("line.separator"));
  } // map2String()

  /**
   * Converts a <code>Map</code> to a <code>String</code>. Maps can be
   * AbstractMap, Attributes, HashMap, Hashtable, IdentityHashMap,
   * RenderingHints, TreeMap or WeakHashMap. Use <code>prefix</code>
   * to add a String in front of each line.
   * @param map The Map to be printed.
   * @param prefix A String preceding each line.
   * @param suffix A String following each line.
   * @return The String representing the given Map.
   */
  public static String map2String(Map map, String prefix, String suffix) {
    Iterator iterator = map.keySet().iterator();
    StringBuffer sb = new StringBuffer();
    while (iterator.hasNext()) {
      String key = "";
      String value = "";
      key = iterator.next().toString();
      value = map.get(key).toString();
      sb.append(prefix);
      sb.append(key);
      sb.append(": ");
      sb.append(value);
      sb.append(suffix); //todo: use the OS-dependent line-sep.
    } // while
    return sb.toString();
  } // map2String()

  /**
   * <p>
   * Converts objects of classes implementing <code>Collection</code>
   * to a <code>String</code>. Some known classes implementing interface
   * <code>Collection</code> are
   * <code>LinkedList, ArrayList, Vector, Stack, HashSet, TreeSet, ...</code>
   * </p>
   * Use <code>prefix</code> and <code>suffix</code> to add a String in front
   * and at the end of each line respectively.
   * Note that this method will not care about keys and values, but only
   * about object entries in the collection and their String representaion
   * directly.
   *
   * @param coll The Collection to be printed.
   * @param prefix A String preceding each line.
   * @param suffix A String following each line.
   * @return The String representing the given Map.
   */
  public static String abstractList2String(AbstractList list,
                                           String prefix, String suffix) {
    if (list.isEmpty()) {
      return prefix + suffix;
    }
    StringBuffer sb = new StringBuffer();
    Iterator iterator = list.iterator();
    while (iterator.hasNext()) {
      String value = (String) iterator.next();
      sb.append(prefix);
      sb.append(value);
      sb.append(suffix);
    } // while
    return sb.toString();
  } // map2String()

  /**
   * Cleans an XML <tt>String</tt> by replacing '&amp', '&lt', '&gt' and
   * '&quot;' with '&amp;amp;', '&amp;lt;', '&amp;gt;' and '&amp;quot;',
   * respectively, and by replacing all characters with ASCII code > 127
   * with the numeric representation.
   * @param dirtyXMLString The XML <tt>String</tt> to be cleaned
   * @return The cleaned XML <tt>String</tt>
   */
  public final static String cleanXMLString(String dirtyXMLString) {

    StringBuffer cleanStrBuffer = new StringBuffer();

    // iterate over all chars in string, replace <, >, & or " if encountered
    // also take care of non <127 ASCII chars by replacing them with &#charNum;
    for (int i = 0; i < dirtyXMLString.length(); i++) {
      char c = dirtyXMLString.charAt(i);
      int s = (int) c;

      if (c == '&') {
        cleanStrBuffer.append("&amp;");
      }
      else if (c == '<') {
        cleanStrBuffer.append("&lt;");
      }
      else if (c == '>') {
        cleanStrBuffer.append("&gt;");
      }
      else if (c == '"') {
        cleanStrBuffer.append("&quot;");
      }
      else if (s > 127) {
        cleanStrBuffer.append("&#" + s + ";");
      }
      else {
        cleanStrBuffer.append(c);
      }
    } // for
    return cleanStrBuffer.toString();
  } // cleanXMLString()

  /**
   * Replaces all non-ascii-7 chars with char-entities.
   */
  public static String unicodeTo7bit(String str) {
    if (str == null) {
      str = new String();
    }
    StringBuffer sb = new StringBuffer();
    // replace non <127 ASCII chars with &#charNum;
    for (int i = 0; i < str.length(); i++) {
      char c = str.charAt(i);
      int s = (int) c;
      if (s > 127) {
        sb.append("&#" + s + ";");
      }
      else {
        sb.append(c);
      }
    } // for
    //log.debug("unicodeto7bit(): returning: " + sb.toString());
    return sb.toString();
  }

  /**
   * Converts any given DOM <tt>Node</tt> instance to an XML <tt>String</tt>.
   * @param node The DOM <tt>Node</tt> to be converted.
   * @return The XML <tt>String</tt> representing the given <tt>Node</tt>
   */
  public static String node2String(Node node) {

    StringWriter sWriter = new StringWriter();
    PrintWriter pWriter = new PrintWriter(sWriter);
    OutputFormat format = new OutputFormat();

    String outEncoding = "UTF-8";

    // determine type of node
    if (node.getNodeType() == Node.ELEMENT_NODE) {
      format = new OutputFormat("xml", outEncoding, true);
    }
    else if (node.getNodeType() == Node.DOCUMENT_FRAGMENT_NODE) {
      format = new OutputFormat( ( (DocumentFragment) node).
                                getOwnerDocument(), outEncoding, true);
    }
    else if (node.getNodeType() == Node.DOCUMENT_NODE) {
      format = new OutputFormat( ( (Document) node), outEncoding, true);
    }

    // we don't want any headers (we might change this;
    // maybe using parameters
    format.setOmitDocumentType(true);
    format.setOmitXMLDeclaration(true);

    // generate string
    XMLSerializer out = new XMLSerializer(pWriter, format);
    try {
      if (node.getNodeType() == Node.ELEMENT_NODE) {
        out.serialize( (Element) node);
      }
      else if (node.getNodeType() == Node.DOCUMENT_FRAGMENT_NODE) {
        out.serialize( (DocumentFragment) node);
      }
      else if (node.getNodeType() == Node.DOCUMENT_NODE) {
        out.serialize( (Document) node);
      }
      else if (node.getNodeType() == Node.ATTRIBUTE_NODE) {
        return new String("<attributes>" + ( (Attr) node).getName() + "=" +
                          ( (Attr) node).getValue() + "</attributes>");
      }
    }
    catch (java.io.IOException ioex) {
      log.error(Utility.exception2String(ioex));
    }
    pWriter.flush();
    return sWriter.toString();
  } // node2String()


  /** create DOM element from XML String
   * @param xmlString String to parse
   * @return DOM Element
  */
  public static Element string2Element(String xmlString) {
    if (!xmlString.startsWith("<?xml")) {
      xmlString = "<?xml version=\"1.0\"?>" + xmlString;
    }
    InputSource source = new InputSource(new StringReader(xmlString));
    try {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder = factory.newDocumentBuilder();
      Element el = (Element) builder.parse(source).getDocumentElement();
      return el;
    }
    catch (Exception e) {
      log.error("Could not convert XML string to DOM element " +
                "in string2Element(String xmlString)\n" +
                "Input String: " + xmlString + "\n" +
                Utility.exception2String(e));
      return null;
    }
  }

  /**
   * Creates a CONF XML <tt>String</tt> from the contents of
   * a given <tt>Properties</tt> instance.
   * @param props The <tt>Properties</tt> instance to be converted
   * @return An XML <tt>String</tt> containing all information of <tt>props</tt>
   */
  public static String conf2Xml(Properties props) {

    String e = "        <entry ";
    String n = "name";
    String v = "value";
    String name;
    String value;

    StringBuffer sb = new StringBuffer();
    sb.append("    <conf>\n");

    // iterate over all prop.names and append <entry name=.. value=...> elements
    Enumeration names = props.propertyNames();
    while (names.hasMoreElements()) {
      name = Utility.cleanXMLString(names.nextElement().toString());
      value = Utility.cleanXMLString(props.getProperty(name));
      sb.append(e + n + "=\"" + name + "\" " + v + "=\"" + value + "\"/>\n");
    }

    sb.append("    </conf>\n");
    return sb.toString();
  } // conf2metadataString()

  /**
   * Creates a CONF <tt>Element</tt> from the contents of
   * a given <tt>Properties</tt> instance.
   * @param doc The <tt>Document</tt> the METADATE will be inserted in
   * @param props The <tt>Properties</tt> instance to be converted
   * @return A METADATA <tt>Element</tt> containing all information of
   *         <tt>props</tt>
   */
  public static Element conf2Element(Document doc, Properties props) {

    Element metadata = (Element) doc.createElement(METADATA_ELEMENT);
    Element conf = (Element) doc.createElement("conf");
    metadata.appendChild(conf);

    String name;
    String value;

    // iterate over all prop.names and create <entry ...> elements
    Element tmpEntry = (Element) doc.createElement("entry");
    Enumeration names = props.propertyNames();
    while (names.hasMoreElements()) {
      name = Utility.cleanXMLString(names.nextElement().toString());
      value = Utility.cleanXMLString(props.getProperty(name));
      tmpEntry = (Element) doc.createElement("entry");
      tmpEntry.setAttribute(name, value);
      conf.appendChild(tmpEntry);
    } // while

    return metadata;
  } // conf2metadata()

  /**
   * Split a given string into an ArrayList of lines using "\n" as a separator
   * @param str input String containing newlines
   * @return the split line Strings in an ArrayList
   */
  public static final ArrayList splitLines(String str) {
    ArrayList lines = new ArrayList();
    StringTokenizer st = new StringTokenizer(str, "\n");
    while (st.hasMoreTokens()) {
      lines.add(st.nextToken());
    }
    return lines;
  } // splitLines()

  /**
   * Splits a given string into an ArrayList of lines using "\t" as a separator
   * @param str input String containing tabs
   * @return the split tab Strings in an ArrayList
   */
  public static final ArrayList splitTabs(String str) {
    ArrayList lines = new ArrayList();
    StringTokenizer st = new StringTokenizer(str, "\t");
    while (st.hasMoreTokens()) {
      lines.add(st.nextToken());
    }
    return lines;
  } // splitTabs()

  /**
   * Checks for punctuation, returns <code>true</code> if <code>w</code> is
   * a puncuation character.
   * @param w input String to tesdt for punctuation
   * @return boolean true iff w is a punctuation character
   */
  public static boolean isInterpunct(String w) {
    if (w.length() == 1) {
      String ip = "\"?!.,<>{}[]+*`_'|/%&();:";
      for (int i = 0; i < ip.length(); i++) {
        if (w.charAt(0) == ip.charAt(i))
          return true;
      }
    }
    return false;
  } // isInterpunct()

    /**
     * Write <code>text</code> to string containing
     * <code>&lt;text&gt;&lt;c&gt;t&lt;/c&gt;&lt;c&gt;e&lt;/c&gt;&lt;c&gt;x&lt;/c&gt;&lt;c&gt;t&lt;/c&gt;&lt;/text&gt;</code>
     * (for rmrs2html stylesheet).
     * @param s String to convert
     * @return the converted result as String
    */
    public static String rmrsText(String text) {
        StringBuffer sb = new StringBuffer("\n<text>");
        for (int i=0; i<text.length(); i++) {
            sb.append("<c>");
            char c = text.charAt(i);
            int s = (int) c;
            if (c == '&') {
                sb.append("&amp;");
            }
            else if (c == '<') {
                sb.append("&lt;");
            }
            else if (c == '>') {
                sb.append("&gt;");
            }
            else if (c == '"') {
                sb.append("&quot;");
            }
            else if (s > 127) {
                sb.append("&#" + s + ";");
            }
            else {
                sb.append(c);
            }
            sb.append("</c>");
        }
        sb.append("</text>\n");
        return sb.toString();
    } // rmrsText()

    /**
     * Returns the header part (&lt;HEAD&gt;...&lt;/HEAD&gt;) of a given
     * HTML document
     * @param HtmlDoc
     * @return The header part (&lt;HEAD&gt;...&lt;/HEAD&gt;) of a given
     *         HTML document
     */
    public static String extractHtmlElement(String htmlDoc, String tag) {
      Pattern pattern = Pattern.compile("<" + tag + "[^>]*>(" + ".*)</"
                                        + tag + ">", Pattern.DOTALL
                                        + Pattern.CASE_INSENSITIVE);
      Matcher matcher = pattern.matcher(htmlDoc);
      if (matcher.find())  {
        return matcher.group(1);
      } else {
        return "";
      }
    }

    public static String exception2String(Exception e) {
      StringBuffer sb = new StringBuffer();
      sb.append(e.getClass() + ": " + e.getMessage() + "\n");
      StackTraceElement[] stackTrace = e.getStackTrace();
      for (int i=0; i < stackTrace.length; i++) {
        sb.append("\n" + stackTrace[i]);
      }
      return sb.toString();
    }
    /**
     * Uses the given regular expression to find all matching occurences and
     * returns them as an ArrayList of Strings.
     * @param regex The regex to be matched.
     * @param string The string to be scanned.
     * @return
     */
    public static ArrayList findAllrx(String regex, String string) {
      ArrayList foundOccurences = new ArrayList();
      Pattern pattern = Pattern.compile(regex, Pattern.DOTALL
                                        + Pattern.CASE_INSENSITIVE);
      Matcher matcher = pattern.matcher(string);
      while (matcher.find())  {
        //System.err.println(matcher.group());
        //System.err.println("-------------------------------------------");
        foundOccurences.add(matcher.group());
      }
      return foundOccurences;
    }

   public static String convert(String str, String fromEncoding,
                                String toEncoding) {
     String outStr = null;
     try {
       ByteArrayInputStream streamIn = new ByteArrayInputStream(str.getBytes());
       InputStreamReader reader = new InputStreamReader(streamIn, fromEncoding);
       ByteArrayOutputStream streamOut = new ByteArrayOutputStream();
       OutputStreamWriter writer = new OutputStreamWriter(streamOut, toEncoding);
       outStr = streamOut.toString(toEncoding);
       //int c;
       //while ((c = reader.read()) != -1) {
       //  writer.write(c);
       //}
     } catch(IOException ioex) {
       log.error(ioex);
     }
     return outStr;
  }

    public static byte[] convert(byte[] data, String srcEncoding,
                                 String targetEncoding) {
      // First, decode the data using the source encoding.
      // The String constructor does this (Javadoc).
      String str = null;
      byte[] result = null;

      try {
        str = new String(data, srcEncoding);
      } catch(UnsupportedEncodingException ex) {
        log.error(ex);
      }
      // Next, encode the data using the target encoding.
      // The String.getBytes() method does this.
      try {
        result = str.getBytes(targetEncoding);
      } catch(UnsupportedEncodingException ex) {
        log.error(ex);
      }
      return result;
    }

    /**
     * Marks all sentence boundaries of a given text with a given marker.
     */
    public static String markSentenceBoundaries(String jtokAnno, String marker) {
      marker = " " + marker + " ";
      String tokenizedText;
      StringBuffer tokenBuffer = new StringBuffer();
      ArrayList jtoks = JToken.parseJTokAnnotation(jtokAnno);
      log.debug("text: " + jtokAnno + " list len = " + jtoks.size());
      // split sentences according to JTok
      String sep;
      for (int i = 0; i < jtoks.size(); i++) {
        JToken tok = (JToken) jtoks.get(i);
        log.debug("markSentenceBoundaries: " + tok.getToken());
        if (tok.getType().startsWith("PTERM")) {
          sep = marker;
        }
        else {
          sep = " ";
        }
        tokenBuffer.append(tok.getToken() + sep);
      }
      String tokenizedString = tokenBuffer.toString();
      log.debug("Marked sentence boundaries like this: "
                + tokenizedString);
      return tokenizedString;
    }

    /**
     * Rounds a number to the specified number of decimal places.
     *
     * @param decimals the number of decimal places
     * @param num the number to round
     * @return the value rounded to the specified number of
     *         decimal places
     */
    public static double roundTo(int decimals, double num) {
      if (decimals == 0) {
        num = (int) (num + .5);
      } else {
        double shift = Math.pow(10, decimals);
        num = num * shift;
        num = (int) (num + .5);
        num = num / shift;
      }
      return num;
    }

} // class
