/*
 * HogXmlRpcClient.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hogclient;

import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;
import org.apache.xmlrpc.XmlRpcClientException;
import java.util.Vector;
import java.util.Stack;
import java.lang.ExceptionInInitializerError;

/**
 * This class provides an object oriented XmlRpc-interface to the HOG server.
 *
 * <p>Each instance of this class is connected to a HOG XmlRpc-server and uses
 * its own default values, settings and history of data and exceptions.
 *
 * <p>Instantiate this class by giving it the hostname and the port number
 * of a running HOG server. Use the methods of this instance to create
 * sessions, annotation collections, annotations and to retrieve analyses
 * of the created annotations.
 *
 * <p>Instances of this class keep track of all caught exceptions, the currently
 * used xmlrpc-connection, session-, collection- and annotation IDs, analysis
 * depth, analysis language, session configuration and the default annotation
 * encoding. Each of these values can be changed manually by using the
 * respective setter methods.
 *
 * <p>For a sample of usage, see {@link SampleApplication SampleApplication}.
 *
 * @author <a href=mailto:oeze@dfki.de>Oezguer Demir</a>
 * @version $Id: HogXmlRpcClient.java,v 1.2 2007/08/16 15:24:54 uschaefer Exp $
 */
public class HogXmlRpcClient {

  // Client related variables
  private XmlRpcClient client; // a  connected XmlRpcClient instance
  private boolean connected = false; // connection status of this HogClient
  private Stack exceptionStack = new Stack(); // all catched exceptions

  // Server related variables
  private String lastSID = ""; // last generated by this client session ID
  private String lastCID = ""; // last generated by this client collection ID
  private String lastAID = ""; // last generated by this client annotation ID

  private String configName = "xmlrpcsession.cfg";
  private String language = "en";
  private int depth = 100;
  private String encoding = "utf-8";

  // CONSTRUCTOR
  // -----------

  /**
   * Creates a new <code>HogXmlRpcClient</code> connecting it to a given
   * XML-RPC server.
   *
   * @param   serverName <code>String</code>    the hostname running
   *                                              the XML-RPC server
   * @param   port <code>int</code>             the port number
   *                                              of the XML-RPC server
   * @throws  ExceptionInInitializerError       in case the connection
   *                                              to the given server fails
   */
  public HogXmlRpcClient(String serverName, int port)
      throws
      ExceptionInInitializerError {
    this.connect(serverName, port);
    if (this.isConnected()) {
      this.connected = true;
    }
    else {
      this.connected = false;
      throw new ExceptionInInitializerError("Could not connect to " +
                                            this.client.getURL());
    }
  }

  // PRIVATE CONNECTION AND INVOCATION METHODS (LOW LEVEL)
  // -----------------------------------------------------

  private boolean checkConnection() {
    try {
      this.connected = (this.invoke("mocoman.sayHello").equals("Hello"));
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      this.connected = false;
    }
    return this.connected;
  }

  private String invoke(String method)
      throws XmlRpcClientException {
    Vector param = new Vector();
    return this.invoke(method, param);
  }

  private String invoke(String method, Object arg0)
      throws XmlRpcClientException {
    Vector param = new Vector();
    param.add(arg0);
    return this.invoke(method, param);
  }

  private String invoke(String method, Object arg0, Object arg1)
      throws XmlRpcClientException {
    Vector params = new Vector();
    params.add(arg0);
    params.add(arg1);
    return this.invoke(method, params);
  }

  private String invoke(String method, Object arg0, Object arg1, Object arg2)
      throws XmlRpcClientException {
    Vector params = new Vector();
    params.add(arg0);
    params.add(arg1);
    params.add(arg2);
    return this.invoke(method, params);
  }

  private String invoke(String method,
                        Object arg0, Object arg1, Object arg2, Object arg3)
      throws XmlRpcClientException {
    Vector params = new Vector();
    params.add(arg0);
    params.add(arg1);
    params.add(arg2);
    params.add(arg3);
    return this.invoke(method, params);
  }

  private String invoke(String method,
                        Object arg0, Object arg1, Object arg2, Object arg3,
                        Object arg4)
      throws XmlRpcClientException {
    Vector params = new Vector();
    params.add(arg0);
    params.add(arg1);
    params.add(arg2);
    params.add(arg3);
    params.add(arg4);
    return this.invoke(method, params);
  }

  private String invoke(String method,
                        Object arg0, Object arg1, Object arg2, Object arg3,
                        Object arg4, Object arg5)
      throws XmlRpcClientException {
    Vector params = new Vector();
    params.add(arg0);
    params.add(arg1);
    params.add(arg2);
    params.add(arg3);
    params.add(arg4);
    params.add(arg5);
    return this.invoke(method, params);
  }

  private String invoke(String method, Vector params)
      throws XmlRpcClientException {

    String response = null;

    try {
      Object tmpResponse = this.client.execute(method, params);
      if (tmpResponse.getClass().getSuperclass().toString().endsWith(
          "java.lang.Exception")) {
        response = "";
        throw new XmlRpcClientException("Error invoking " + method + " using " +
                                        client.getURL(), new Exception());
      }
      else {
        response = tmpResponse.toString();
      }
      //System.err.println(
      //    "DEBUG_h205: " + tmp.getClass().getSuperclass().toString());
    }
    catch (XmlRpcException ex0) {
      throw new XmlRpcClientException("Could not invoke " + client.getURL() +
                                      "?" + method + "(" + params + ")",
                                      ex0.getCause());
    }
    catch (java.io.IOException ex1) {
      throw new XmlRpcClientException("Could not invoke " + client.getURL() +
                                      "?" + method + "(" + params + ")",
                                      ex1.getCause());
    }

    if (response != null) {
      return response;
    }
    else {
      return "";
    }
  }

  // PUBLIC HOG CLIENT SPECIFIC METHODS (LOW LEVEL)
  // ----------------------------------------------

  /**
   * Connects this client to a given XML-RPC server.
   *
   * <p>(Disconnecting it from the previous one.)

   * <p><b>Note:</b> ordinarily, this method does not have to be used,
   * as the {@link #HogXmlRpcClient(String, int) constructor} of this class
   * will automatically connect this instance to a given server. Nevertheless,
   * this method can be used to reconnect this instance to the (or possibly
   * another) server.
   *
   * <p><b>Also note</b> that <code>lastSID</code>, <code>lastCID</code> and
   * <code>lastAID</code> are reset to <code>""</code> on using this method.
   *
   * @param   serverName <code>String</code>   the hostname of the machine
   *                                             running XML-RPC server
   * @param   port <code>int</code>            the port number of the
   *                                             XML-RPC server
   */
  public void connect(String serverName, int port) {
    // reset server related member variable
    this.lastSID = "";
    this.lastCID = "";
    this.lastAID = "";
    this.connected = false;
    // connect
    try {
      this.client = new XmlRpcClient(serverName, port);
      this.connected = true;
    }
    catch (java.net.MalformedURLException ex0) {
      this.exceptionStack.push(ex0);
      this.connected = false;
    }
  }

  /**
   * Returns the last caught exception cause and pops it off the internal stack.
   *
   * <p>To be used for error diagnosis. Note that the only public method of
   * this class throwing an exception is the constructor. All other exceptions
   * are handled internally.
   *
   * <p>This method is therefore used to remedy the lack of information arising
   * from this approach. This means, that the application using this API does
   * not have to deal with Java exceptions - error checking can be done by
   * validating the <codeString</code> objects returned by the public methods
   * (rsp. <code>""</code> for failures) - but the caught and recorded
   * exception causes can be retrieved for debugging purposes.
   *
   * <p>For retrieving all caught and recorded exception causes
   * @see #getAllExceptionCauses()
   *
   * @return   <code>Throwable</code>   the last catched exception cause
   */
  public Throwable getLastExceptionCause() {
    if (!this.exceptionStack.isEmpty()) {
      return (Throwable)this.exceptionStack.pop();
    }
    else {
      return new Throwable();
    }
  }

  /**
   * Returns all caught exceptions as a stack.
   *
   * @return   <code>Stack</code>   all caught exceptions as a stack
   */
  public Stack getAllExceptionCauses() {
    if (this.exceptionStack != null) {
      return (Stack)this.exceptionStack.clone();
    }
    else {
      return new Stack();
    }
  }


  /**
   * Return the host name of the server this client is connected to.
   *
   * @return   <code>String</code>   the host name of the server this client
   *                                   is connected to.
   */
  public String getServerHostName() {
    return this.client.getURL().getHost();
  }

  /**
   * Returns the port number of the server this client is connected to.
   *
   * @return   <code>String</code>   the port number of the server this client
   *                                 is connected to
   */
  public int getServerPort() {
    return this.client.getURL().getPort();
  }

  /**
   * Returns the URL of the server this client is connected to.
   *
   * @return   <code>String</code>   the URL of the server this client
   *                                 is connected to
   */
  public String getServerURL() {
    return this.client.getURL().toString();
  }


  /**
   * Returns the <code>String</code> representation of this instance:
   *
   * <p>The name of this class followed by a <code>#</code> and the hashcode
   * of this instance.
   *
   * @return   <code>String</code>   the <code>String</code> representation
   *                                 of this instance
   */
  public String toString() {
    return this.getClass().getName() + "#" + this.hashCode();
    //return this.client.getURL().toString();
  }

  // PUBLIC HOG CLIENT SPECIFIC METHODS
  // ----------------------------------

  /**
   * Sets the name of the session configuation file to be used when creating
   * the next session.
   *
   * @param   conf <code>String</code>   the name of the configuration file
   */
  public void setConfiguration(String conf) {
    this.configName = conf;
  }

  /**
   * Returns the name of the session configuation file to be used when creating
   * the next session.
   *
   * @return   <code>String</code>   the name of the session configuation file
   *                                 to be used when creating the next session
   */
  public String getConfiguration() {
    return this.configName;
  }

  /**
   * Sets the session language to be used when creating the next session.
   * For example: <code>"en"</code>.
   *
   * @param   lang   <code>String</code>   the session language to be used
   *                                         when creating the next session
   */
  public void setLanguage(String lang) {
    this.language = lang;
  }

  /**
   * Returns the session language to be used when creating the next session.
   *
   * @return   <code>String</code>   the session language to be used when
   *                                   creating the next session
   */
  public String getLanguage() {
    return this.language;
  }

  /**
   * Returns the currently set default encoding of annotations for this client.
   *
   * @return   <code>String</code>   the currently set default encoding
   *                                   of annotations for this client
   */
  public String getEncoding() {
    return this.encoding;
  }

  /**
   * Sets the default encoding of annotations for this client.
   *
   * @param   encoding   <code>String</code>   the default encoding of
   *                                             annotations for this client
   */
  public void setEncoding(String encoding) {
    this.encoding = encoding;
  }

  /**
   * Returns the ID of the session most recently used or created.
   *
   * @return   <code>String</code>   the ID of the session most recently
   *                                   used or created
   */
  public String getLastSessionID() {
    return this.lastSID;
  }

  /**
   * Returns the ID of the collection most recently used or created.
   *
   * @return   <code>String</code>   the ID of the collection most recently
   *                                   used or created
   */
  public String getLastCollectionID() {
    return this.lastCID;
  }

  /**
   * Returns the ID of the annotation most recently used or created.
   *
   * @return   <code>String</code>   the ID of the annotation most recently
   *                                   used or created
   */
  public String getLastAnnotationID() {
    return this.lastAID;
  }

  // HOG SERVER SPECIFIC METHODS
  // ---------------------------

  /**
   * Tests the connection between the client and the server.
   *
   * <p>Returns <code>true</code> if connected and ready,
   * <code>false</code> otherwise.
   *
   * @return   <code>boolean</code>   <code>true</code> if connected and ready,
   *                                  <code>false</code> otherwise
   */
  public boolean isConnected() {
    return this.checkConnection();
  }

  /**
   * Requests the server connected to this client to say <code>"Hello"</code>.
   *
   * @return <code>String "Hello"</code> in case the server is running and this
   *         client is connected to it.
   *         <code>String ""</code> in case of failure.
   */
  public String sayHello() {
    try {
      return this.invoke("mocoman.sayHello");
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      return "";
    }
  }

  /**
   * Retrieves and returns a <code>String</code> representation of the current
   * status of the server of this client.
   *
   * @return a <code>String</code> representation of the current
   *         status of the server of this client.
   */
  public String getServerStatusString() {
    try {
      return this.invoke("mocoman.getStatus");
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      return "";
    }

  }

  /**
   * Creates a new Session and return its sessionID.
   *
   * @return   <code>String</code>   the sessionID of the new sentence in case
   *                                 of success,
   *                                 <code>""</code> in case of failure
   */
  public String createSession() {
    String saveLastSID = this.lastSID;
    try {
      this.lastSID = this.invoke("mocoman.createSession",
                                 "conf/" + this.language + "/" +
                                 this.configName);
      return lastSID;
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      // in case the creation fails: preserve old lastSID
      this.lastSID = saveLastSID;
      return "";
    }
  }

  /**
   * Creates a new Session and return its sessionID.
   *
   * <p><b>Note</b> that this method updates <code>language</code> and
   * <code>configuration</code>.
   *
   * @param   language              the session language
   *                                  (e.g. <code>"en"</code>)
   * @param   config                the session configuration file
   *                                  (e.g. <code>"xmlrpcsession.cfg"</code>)
   * @return  <code>String</code>   the sessionID of the new sentence
   *                                  in case of success,
   *                                  <code>""</code> in case of failure
   */
  public String createSession(String language, String config) {
    this.setLanguage(language);
    this.setConfiguration(config);
    return this.createSession();
  }

  /**
   * Creates a new Session and return its sessionID.
   *
   * <p><b>Note</b> that this method updates <code>language</code> and
   * <code>configuration</code>.
   *
   * @param   language              the session language
   *                                  (e.g. <code>"en"</code>)
   * @param   config                the session configuration file
   *                                  (e.g. <code>"xmlrpcsession.cfg"</code>)
   * @param   sessionID                the session ID
   *                                  (e.g. <code>"demosession1"</code>)
   * @return  <code>String</code>   the sessionID of the new sentence
   *                                  in case of success,
   *                                  <code>""</code> in case of failure
   */
  public String createSession(String language, String config, String sessionID) {
    this.setLanguage(language);
    this.setConfiguration(config);
    String saveLastSID = this.lastSID;
    try {
      this.lastSID = this.invoke("mocoman.createSession",
                                 "conf/" + this.language + "/" +
                                 this.configName, sessionID);
      return lastSID;
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      // in case the creation fails: preserve old lastSID
      this.lastSID = saveLastSID;
      return "";
    }
  }


  /**
   * Deletes/closes the most recently created session.
   *
   * <p>Resets <code>lastSID</code> (see {@link #getLastSessionID()})
   * to <code>""</code>.
   *
   * @return   <code>String</code>   <code>"ok"</code> in case of success,
   *                                 <code>""</code> in case of failure
   */
  public String deleteSession() {
    try {
      this.invoke("mocoman.closeSession", this.lastSID);
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      return "";
    }
    this.lastSID = "";
    return "ok";
  }

  /**
   * Deletes/closes session with sessionID.
   *
   * <p><b>Note</b> that <code>lastSID</code> (see {@link #getLastSessionID()})
   * is reset to <code>""</code> if the session to be deleted is the current
   * one.
   *
   * @param   sessionID   <code>String</code>   the ID of the session to be
   *                                              deleted/closed
   * @return  <code>String</code>               <code>"ok"</code> in case of
   *                                              success, <code>""</code> in
   *                                              case of failure
   *
   */
  public String deleteSession(String sessionID) {
    String response = "";
    try {
      response = this.invoke("mocoman.closeSession", sessionID);
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      return "";
    }
    if (sessionID.equals(this.lastSID)) {
      this.lastSID = "";
    }
    if (response.equals("true")) {
      return "ok";
    }
    else {
      this.exceptionStack.push(new Exception(response));
      return "";
    }
  }

  /**
   * Creates a new anntotation collection in the current session
   * and returns its ID. Returns <code>""</code> in case of failure.
   *
   * @return   <code>String</code>   the ID of the new collection,
   *                                 <code>""</code> in case of failure
   */
  public String createCollection() {
    String saveLastCID = this.lastCID;
    try {
      this.lastCID = this.invoke("mocoman.createAnnotationCollection",
                                 this.lastSID);
      return lastCID;
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      this.lastCID = saveLastCID;
      return "";
    }
  }

  /**
   * Creates a new anntotation collection in the given session
   * and returns its ID. Returns <code>""</code> in case of failure.
   *
   * <p><b>Note</b> that the value of lastSID (see {@link #getLastSessionID()})
   * is altered to the given value <code>sessionID</code> in case of success.
   *
   * @param    sessionID   <code>String</code>   the ID of the session in which
   *                                               to create the new collection
   * @return               <code>String</code>   the ID of the new collection,
   *                                               <code>""</code> in case of
   *                                               failure
   */
  public String createCollection(String sessionID) {
    String saveLastSID = this.lastSID;
    String saveLastCID = this.lastCID;
    try {
      this.lastCID = this.invoke("mocoman.createAnnotationCollection",
                                 sessionID);
      this.lastSID = sessionID;
      //System.err.println("DEBUG_h482: " + this.lastCID);
      return lastCID;
    }
    catch (XmlRpcClientException ex0) {
      //System.err.println("DEBUG_h486: CATCHING");
      this.exceptionStack.push(ex0);
      this.lastCID = saveLastCID;
      this.lastSID = saveLastSID;
      return "";
    }
  }

  /**
   * Creates initial annotations in the current session/collection using
   * the currently set default language and encoding.
   *
   * <p>Two annotations will be created when calling this method:
   * <ol>
   * <li>The raw text. ID: <code>rawtext</code></li>
   * <li>The xmlfied text including metadata. ID: <code>xmltext</code></li>
   * </ol>
   *
   * <p>Use this method to initially put text into a new session
   * for later analysis.
   *
   * @see #createAnnotation(String, String, String, String, String)
   * @param    rawtext              the raw input text
   * @return   <code>String</code>   <code>"xmltext"</code> if the operation was
   *                                  successful, <code>""</code> otherwise
   */
  public String createAnnotation(String rawtext) {
    return this.createAnnotation(this.lastSID, this.lastCID, rawtext,
                                 this.language, this.encoding);
  }

  /**
   * Creates initial annotations in the given session/collection using the
   * given settings.
   *
   * <p>Two annotations will be created when calling this method:
   * <ol>
   * <li>The raw text. ID: <code>rawtext</code></li>
   * <li>The xmlfied text including metadata. ID: <code>xmltext</code></li>
   * </ol>
   *
   * <p>Use this method to initially put text into a new session
   * for later analysis.
   *
   * <p><b>Note</b> that this method updates <code>lastSID</code>,
   * <code>lastCID</code> and <code>lastAID</code> in case of success.
   *
   * @see #createAnnotation(String)
   * @param sessionID    <code>String</code>  the ID of the session in which to
   *                                            create the annotation
   * @param collectionID <code>String</code>  the ID of the collection in which
   *                                            create the annotation
   * @param rawtext      <code>String</code>  the raw input text
   * @param language     <code>String</code>  the language of the input text
   * @param encoding     <code>String</code>  the encoding of the input text
   * @return             <code>String</code>  <code>"xmltext"</code> if the
   *                                            operation was successful,
   *                                            <code>""</code> otherwise
   */
  public String createAnnotation(String sessionID, String collectionID,
                                 String rawtext, String language,
                                 String encoding) {
    return this.createAnnotation(sessionID, collectionID, rawtext,
                                 language, "javaXmlRpcClient",
                                 encoding);
  }

  /**
   * Creates initial annotations in the given session/collection using the
   * given settings and a custom name for the creator of the annotation.
   * (The default name when using
   * {@link #createAnnotation(String, String, String, String, String)} or
   * {@link #createAnnotation(String)} is <code>"javaXmlRpcClient"</code>.
   *
   * <p>Two annotations will be created when calling this method:
   * <ol>
   * <li>The raw text. ID: <code>rawtext</code></li>
   * <li>The xmlfied text including metadata. ID: <code>xmltext</code></li>
   * </ol>
   *
   * <p>Use this method to initially put text into a new session
   * for later analysis.
   *
   * <p><b>Note</b> that this method updates <code>lastSID</code>,
   * <code>lastCID</code> and <code>lastAID</code> in case of success.
   *
   * @see #createAnnotation(String, String, String, String, String)
   * @param sessionID    <code>String</code>  the ID of the session in which to
   *                                            create the annotation
   * @param collectionID <code>String</code>  the ID of the collection in which
   *                                            create the annotation
   * @param rawtext      <code>String</code>  the raw input text
   * @param language     <code>String</code>  the language of the input text
   * @param clientName   <code>String</code>  a custom name to be used in the
   *                                            metadata field "component" of
   *                                            annotation to be created
   * @param encoding     <code>String</code>  the encoding of the input text
   * @return             <code>String</code>  <code>"xmltext"</code> if the
   *                                            operation was successful,
   *                                            <code>""</code> otherwise
   */
  public String createAnnotation(String sessionID, String collectionID,
                                 String rawtext, String language,
                                 String clientName, String encoding) {
    String saveLastAID = this.lastAID;
    String saveLastCID = this.lastCID;
    String saveLastSID = this.lastSID;
    try {
      this.lastAID = this.invoke("mocoman.createInitialAnnotation",
                                 sessionID, collectionID, rawtext,
                                 language, clientName, encoding);
      this.lastCID = collectionID;
      this.lastSID = sessionID;
      return lastAID;
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      this.lastAID = saveLastAID;
      this.lastCID = saveLastCID;
      this.lastSID = saveLastSID;
      return "";
    }
  }

  /**
   * Retrieves and returns the annotation with the current session ID,
   * collection ID and annotation ID.
   *
   * <p><b>Note</b>, that this method changes the value of <code>lastAID</code>
   * (see {@link #getLastAnnotationID()}).
   *
   * @return   <code>String</code>   the annotation to be retrieved,
   *                                   <code>""</code> in case of
   *                                   failure
   */
  public String getAnnotation() {
    return this.getAnnotation(this.lastSID, this.lastCID, this.lastAID);
  }

  /**
   * Retrieves and returns the annotation with the given session ID,
   * collection ID and annotation ID.
   *
   * <p><b>Note</b> that this method updates <code>lastSID</code>,
   * <code>lastCID</code> and <code>lastAID</code> in case of success.
   *
   * @param sessionID     <code>String</code>   the ID of the session of the
   *                                              annotation to be retrieved
   * @param collectionID  <code>String</code>   the ID of the collection of the
   *                                              annotation to be retrieved
   * @param annotationID  <code>String</code>   the ID of the annotation to be
   *                                              retrieved
   * @return              <code>String</code>   the annotation to be retrieved,
   *                                              <code>""</code> in case of
   *                                              failure
   */
  public String getAnnotation(String sessionID, String collectionID,
                              String annotationID) {
    String saveLastAID = this.lastAID;
    String saveLastCID = this.lastCID;
    String saveLastSID = this.lastSID;
    String annotation;
    try {
      annotation = this.invoke("mocoman.getAnnotation",
                               sessionID, collectionID, annotationID);
      this.lastAID = annotationID;
      this.lastCID = collectionID;
      this.lastSID = sessionID;
      return annotation;
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      this.lastAID = saveLastAID;
      this.lastCID = saveLastCID;
      this.lastSID = saveLastSID;
      return "";
    }
  }

  /**
   * Retrieves the annotation with the current session ID, collection ID and
   * annotation ID and returns its HTML representation.
   *
   * @return  <code>String</code>  the HTML representation of the annotation
   *                                 with the current session ID,
   *                                 collection ID and annotation ID
   */
  public String getAnnotationAsHTML() {
    return this.getAnnotationAsHTML(this.lastSID, this.lastCID, this.lastAID);
  }

  /**
   * Retrieves the annotation with the given session ID, collection ID and
   * annotation ID and returns its HTML representation.
   *
   * <p><b>Note</b> that this method updates <code>lastSID</code>,
   * <code>lastCID</code> and <code>lastAID</code> in case of success
   *
   * @param sessionID     <code>String</code>   the ID of the session of the
   *                                              annotation to be retrieved
   * @param collectionID  <code>String</code>   the ID of the collection of the
   *                                              annotation to be retrieved
   * @param annotationID  <code>String</code>   the ID of the annotation to be
   *                                              retrieved
   * @return              <code>String</code>   the annotation to be retrieved
   *                                              represented in HTML,
   *                                              <code>""</code> in case of
   *                                              failure
   */
  public String getAnnotationAsHTML(String sessionID, String collectionID,
                                    String annotationID) {
    String saveLastAID = this.lastAID;
    String saveLastCID = this.lastCID;
    String saveLastSID = this.lastSID;
    String annotationHTML;
    try {
      annotationHTML = this.invoke("util.extractHtmlElement",
                                   this.invoke("mocoman.xml2htmlTransformer",
                                               sessionID, collectionID,
                                               annotationID,
                                               "xsl/html/xml2html.xsl", "no"),
                                   "HEAD");
      this.lastAID = annotationID;
      this.lastCID = collectionID;
      this.lastSID = sessionID;
      return annotationHTML;
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      this.lastAID = saveLastAID;
      this.lastCID = saveLastCID;
      this.lastSID = saveLastSID;
      return "";
    }
  }


  /** <p><b>Note</b> that this method updates <code>lastSID</code>,
   * <code>lastCID</code> and <code>lastAID</code>*/

  /**
   * Analyzes the annotation with the current session ID, collection ID and
   * annotation ID and returns the result.
   *
   * @return   <code>String</code>   the result of the analysis,
   *                                 <code>""</code> in case of failure
   */
  public String analyze() {
    return this.analyze(this.lastSID, this.lastCID, this.lastAID);
  }

  /**
   * Analyzes the annotation with given session ID, collection ID and
   * annotation ID and returns the result.
   *
   * @param sessionID     <code>String</code>   the ID of the session of the
   *                                              annotation to be analyzed
   * @param collectionID  <code>String</code>   the ID of the collection of the
   *                                              annotation to be analyzed
   * @param annotationID  <code>String</code>   the ID of the annotation to be
   *                                              analyzed
   * @return              <code>String</code>   the annotation to be analyzed
   *                                              represented in HTML,
   *                                              <code>""</code> in case of
   *                                              failure
   */
  public String analyze(String sessionID, String collectionID,
                        String annotationID) {
    return this.analyze(sessionID, collectionID, annotationID, this.depth);
  }

  /**
   * Analyzes the annotation with given session ID, collection ID and
   * annotation ID at the given depth and returns the result.
   *
   * @param sessionID     <code>String</code>   the ID of the session of the
   *                                              annotation to be analyzed
   * @param collectionID  <code>String</code>   the ID of the collection of the
   *                                              annotation to be analyzed
   * @param annotationID  <code>String</code>   the ID of the annotation to be
   *                                              analyzed
   * @param depth         <code>int</code>      the depth of the analysis
   * @return              <code>String</code>   the annotation to be analyzed
   *                                              represented in HTML,
   *                                              <code>""</code> in case of
   *                                              failure
   */
  public String analyze(String sessionID, String collectionID,
                        String annotationID, int depth) {
    String saveLastAID = this.lastAID;
    String saveLastCID = this.lastCID;
    String saveLastSID = this.lastSID;
    String result;
    try {
      result = this.invoke("mocoman.analyse",
                           sessionID, collectionID, annotationID,
                           new Integer(0), new Integer(0), new Integer(depth));
      this.lastAID = annotationID;
      this.lastCID = collectionID;
      this.lastSID = sessionID;
      return result;
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      this.lastAID = saveLastAID;
      this.lastCID = saveLastCID;
      this.lastSID = saveLastSID;
      return "";
    }
  }

  /**
   * Shuts down the server connected to this client.
   *
   * @return   <code>String</code>    <code>"ok"</code> in case of success,
   *                                  <code>""</code> in case of failure
   */
  public String shutdown() {
    try {
      this.invoke("mocoman.shutdown");
      return "ok";
    }
    catch (XmlRpcClientException ex0) {
      this.exceptionStack.push(ex0);
      return "";
    }
  }

} // class
