/*
 * AnalyzeAll.java
 *
 * (C) 2004 - 2005  DFKI Language Technology Lab http://www.dfki.de/lt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 */

package de.dfki.lt.hogclient.cmdline;

import java.util.*;
import java.io.*;
import java.text.*;
import de.dfki.lt.hog.util.Utility;
import de.dfki.lt.hogclient.*;

/**
 * Scan a given host within a given range of ports
 * for running XmlRpc-HogServers.
 *
 * @author <a href=mailto:oeze@dfki.de>Oezguer Demir</a>
 * @version $Id: AnalyzeAll.java,v 1.8 2006/06/08 15:44:40 uschaefer Exp $
 */

public class AnalyzeAll {

  public AnalyzeAll() {
    // (for static use)
  }

  private static int countLines(String fileName) throws IOException {
    FileReader file = null;
    file = new FileReader(fileName);
    BufferedReader reader = new BufferedReader(file);
    int numLines = 0;
    String line;
    do {
      line = reader.readLine();
      if (line != null) {
        numLines++;
      }
    } while (line != null);
    reader.close();
    file.close();
    return numLines;
  }

  //  M A I N
  public static void main(String args[]) {

    // Initialize variables
    HogXmlRpcClient client;
    if (args.length==0) 
        System.out.println("Usage: AnalyzeAll host port filename [language [encoding]]");
    String hostname = args[0];
    int port = Integer.parseInt(args[1]);
    String fileName = args[2];
    String outFileName = "/tmp" + args[2].substring(args[2].lastIndexOf("/")) +
                         ".analyzed";
    String language = "en";
    if (args.length>3) language=args[3];
    String encoding;
    if (language.equalsIgnoreCase("ja") || language.equalsIgnoreCase("el")) {
      encoding = "utf-8";
    } else {
      encoding = "iso-8859-1";
    }
    if (args.length>4) encoding=args[4];

    long t_now = System.currentTimeMillis();
    double prog = 100.0;
    double rateS = 100.0;
    double rateF = 100.0;
    long eta = 0;

    // Open input file
    System.out.println("Opening input file (read) " + fileName);
    int numLines = 0;
    try {
      numLines = countLines(fileName);
      System.out.println("Contains: " + numLines + " lines.");
    } catch (IOException ex0) {
      ex0.printStackTrace();
      System.exit( -1);
    }
    FileReader freader = null;
    try {
      freader = new FileReader(fileName);
    } catch (FileNotFoundException ex) {
      ex.printStackTrace();
      System.exit( -1);
    }

    // Open output file
    System.out.println("Opening output file (write) " + outFileName);
    FileWriter fwriter = null;
    try {
      fwriter = new FileWriter(outFileName);
    } catch (IOException ex) {
      ex.printStackTrace();
      System.exit( -1);
    }
    BufferedWriter writer = new BufferedWriter(fwriter);

    // Connect to HogServer
    System.out.println("Connecting to " + hostname + ":" + port);
    try {
      client = new HogXmlRpcClient(hostname, port);
    } catch (ExceptionInInitializerError ex0) {
      System.err.println("HogXmlRpcClient.main: "
                         + "Could not initialize client connecting to "
                         + hostname + ":" + port);
      System.err.println("HogXmlRpcClient.main: "
                         + "Server might be down or unreachable.");
      System.err.print("HogXmlRpcClient.main: StackTrace: ");
      ex0.printStackTrace();
      return;
    }

    // Prepare a session and collection
    System.out.println("Creating session and collection (" + language + ")");
    String sid = client.createSession(language, "xmlrpcsession.cfg");
    System.out.println("  Created " + sid);
    String cid = client.createCollection(sid);
    System.out.println("  Created " + cid);

    // Read input file and run the main loop - line per line
    BufferedReader reader = new BufferedReader(freader);
    String line = null;
    int i = 0;
    int successes = 0;
    int failures = 0;
    System.out.println("Running main loop");
    long t_init = System.currentTimeMillis();
    long t_lastProgPrint = System.currentTimeMillis();
    try {
      while ((line = reader.readLine()) != null) {
        i++;

        // Create annotation
        String aid = client.createAnnotation(sid, cid, line, language,
                                             encoding);
        if (!aid.equalsIgnoreCase("xmltext")) {
          System.err.println("ERROR: Could not create Annotation. Text: ");
          System.err.println(line);
        }
        // Analyze annotation
        String result = client.analyze();
        if (result.equals("")) {
          failures++;
          System.err.println("FAILURE: " + line);
        } else {
          successes++;
          // Write result
          writer.write(result + System.getProperty("line.separator"));
        }

        // Print progress information every 5 seconds (if there was progress)
        t_now = System.currentTimeMillis();
        if ((t_now - t_lastProgPrint) >= 5000) {
          eta = ((t_now - t_init) / i) * (numLines - i);
          if (numLines != 0) {
            prog = (double) i / (double) numLines * 100.0;
          }
          if ((successes + failures) != 0) {
            rateS = (double) successes / (double) (successes + failures) *
                    100.0;
          }
          if (rateS != 0) {
            rateF = 100.0 - rateS;
          }
          System.out.println("Done: " + i + " / " + numLines +
                             " (" + Utility.roundTo(2, prog) + "%)" +
//                             " SUCCESS: " + successes +
//                             " (" + Utility.roundTo(3, rateS) + "%)" +
                             " FAIL: " + failures +
                             " (" + Utility.roundTo(3, rateF) + "%)" +
                             " ETA: " + Utility.formatDurationShort(eta) + "h");
          t_lastProgPrint = t_now;
        }
      } // while reader.readline!=null

    } catch (IOException ex0) {
      ex0.printStackTrace();
      System.exit( -1);
    }

    // Print summary
    if (numLines != 0) {
      prog = (double) i / (double) numLines * 100.0;
    }
    if ((successes + failures) != 0) {
      rateS = (double) successes / (double) (successes + failures) *
              100.0;
    }
    if (rateS != 0) {
      rateF = 100.0 - rateS;
    }
    System.out.println("Done: " + i + " / " + numLines +
                       " (" + Utility.roundTo(2, prog) + "%)" +
//                             " SUCCESS: " + successes +
//                             " (" + Utility.roundTo(3, rateS) + "%)" +
                       " FAIL: " + failures +
                       " (" + Utility.roundTo(3, rateF) + "%)" +
                       " Time: " + Utility.formatDurationShort(t_now - t_init) +
                       "h");

    // Close files
    try {
      reader.close();
      freader.close();
      writer.close();
      fwriter.close();
    } catch (IOException ex0) {
      ex0.printStackTrace();
      System.exit( -1);
    }
  }
}
