#!/usr/bin/env python

"""
This script implements a Heart of Gold client for tsdb.
It opens a new session on Heart of Gold,
accepts sentences from stdin linewise, sends them to Heart of Gold,
and returns the analyses for each line immediately after each new
input line to stdout

 ./

USAGE

1. Make sure that

   mocomanserver is running

2. Run

   [python] [./]hogfile.py [hoghost [hogport [inputfile [language [sessioncfg]]]]]

   
"""

__author__ = 'Ulrich Schaefer <ulrich.schaefer@dfki.de>'
__date__ = '$Date: 2007/10/15 15:34:36 $'
__version__ = '$Revision: 1.1 $'
__project__ = 'Quetal/HoGQA'

import xmlrpclib
import string
import sys
import os


srv = None
srvLoc = None

CONF_DIR = 'conf' # config directory (relative to hog dir on server)
CONF_FILE= 'jtok_session.cfg' # config file for session on the server
hogencoding = 'utf-8'
depth = 50
SID = { 'de' : None, 'en' : None }
CID = { 'de' : None, 'en' : None }


def printUsage():
    """Tell the user how to start this client and exit."""
    print 'Usage: [python] [./]hogfile.py [hogserverhost [hogport [inputfile [language [sessioncfg]]]]]'
    print '       default hogqahost: attila'
    print '       default hogqaport: 8411'
    print '       default language: en'
    print '       default sessioncfg: jtok_session.cfg'
    print '       default input file: input.txt'
    print 'Example: hogfile.py attila 7667 input.txt en jtok_session.cfg'
    sys.exit()


def parseCmdLine():
    """Read, interpret and react to the command line arguments."""
    lang = 'en'
    port = 8411
    session = 'jtok_session.cfg'
    host = 'localhost'
    ifile = 'input.txt'
    if len(sys.argv) == 1:
        printUsage()
    elif len(sys.argv) == 2:
        host = sys.argv[1]
    elif len(sys.argv) == 3:
        host = sys.argv[1]
        port = sys.argv[2]
    elif len(sys.argv) == 4:
        host = sys.argv[1]
        port = sys.argv[2]
        lang = sys.argv[3]
    elif len(sys.argv) == 5:
        host = sys.argv[1]
        port = sys.argv[2]
        ifile = sys.argv[3]
        lang = sys.argv[4]
    elif len(sys.argv) == 6:
        host = sys.argv[1]
        port = sys.argv[2]
        ifile = sys.argv[3]
        lang = sys.argv[4]
        session = sys.argv[5]
    else:
        printUsage()
    return (host, port, ifile, lang, session)


def connect(hostname, port):
    """Connect to a given XML-RPC server - return the Server reference."""
    global srv, srvLoc
    srvLoc = 'http://' + hostname + ':' + str(port)
    try:
        srv = xmlrpclib.Server(srvLoc)
        print 'Connected to server %s on port %s'\
              %(hostname, port)
    except:
        print 'Connection failed!'
        print 'Please make sure that hogqaserver is',
        print 'running at', srvLoc
        sys.exit()
    return srv


def checkConnection():
    """Test if the connection to the server is available."""
    global srv, srvLoc
    try:
        srv.mocoman.sayHello()
        return 1
    except:
        print 'Connection failed!'
        print 'Please make sure that hogqaserver is',
        print 'running at', srvLoc
        return 0


def startSession(lang):
    """Initialize by creating new annotation collections in HoG for EN and DE."""
    global SID, CID
    if not checkConnection(): return 'Error: could not connect to HoG server.'
    print "Creating session for " + lang + " ..."
    SID[lang] = createSession(lang)
    print "Creating annotation collection for " + lang + " ..."
    try:
        CID[lang] = createAnnoColl(SID[lang])
    except:
        return 'Error: could not create Annotation Collection for ' + lang + ' in ' + SID[lang]
    print "Annotation collection %s.%s created."%(SID[lang], CID[lang])
    return 'OK'



def createSession(lang):
    """Create a new session and print its id."""
    global srv, CONF_DIR, CONF_FILE
    if not checkConnection(): return
    sid = srv.mocoman.createSession(CONF_DIR + "/" + lang + "/" + CONF_FILE)
    return sid


def deleteSession(sid):
    """Delete/close a session, print whether the operation was successful."""
    global srv
    if not checkConnection(): return
    try:
        srv.mocoman.closeSession(sid)
    except:
        print "Apparently, there is no session with id '%s'"%(sid)


def createAnnoColl(sid):
    """Create a new annotation collection and print its id."""
    global srv
    if not checkConnection(): return
    try:
        cid = srv.mocoman.createAnnotationCollection(sid)
        return cid
    except:
        print "Apparently, there is no session with id '%s'"%(sid)


def analyse(text, lang):
    """Analyse a given annotation."""
    global srv, SID, CID, depth, result, hogencoding
    if not checkConnection(): return
    #text = text.encode(hogencoding)
    try:
        aid = srv.mocoman.createInitialAnnotation(SID[lang], CID[lang], xmlrpclib.Binary(text.encode('utf8')), lang,
                                            "HogQAserver", hogencoding)
    except:
        print "Could not create annotation for sentence " + text
        return ""
    try:
        binary = srv.mocoman.analyseBinary(SID[lang], CID[lang], aid, 0, 0, depth)
        result = unicode(binary.data,"utf-8")
        snip = result.find('<metadata>')
        snap = result.find('</metadata>')
        if (snip >= 0) and (snap >= 0):
            return (result[0:snip] + result[snap+11:])
        else:
            return result
    except:
        print "Could not analyze sentence " + text
        return ""
    

def printStatus(lang):
    """Output MoCoMan's current data status."""
    global srv, SID, CID
    if not checkConnection():
        print "----- Status (not connected) -----"
        print "Currently not connected"
        print "Last connection:"
        print "  %s"%(srv)
        return
    print "----- Status (connected) -----"
    print "Client:"
    print "  Connected to '%s'"%(srv)
    print "  Current session (%s): '%s'"%(lang, SID[lang])
    print "  Current annotation collection (%s): '%s'"%(lang, CID[lang])
    if result: print "  Resultbuffer contains %d bytes ('p' to print)"\
                     %(len(result))
    print "Server:"
    print srv.mocoman.getStatus()


def shutdown():
    """Close HoG client sessions for EN and DE and shutdown this XML-RPC server."""
    global lang
    printStatus(lang)
    deleteSessions()
    sys.exit(0)


def deleteSessions():
    global SID, lang
    deleteSession(SID[lang])
    print "Session %s deleted."%(SID[lang])


def shutdownHoG():
    global srv
    srv.mocoman.shutdown()



def sayHello():
    """Return 'Hello' (for clients to check if server is still alive)."""
    return "Hello"


def init():
    return "OK"
    

def main():
    """Init and run the main loop."""
    global srv, CONF_FILE, lang
    host, port, ifile, lang, CONF_FILE = parseCmdLine()
    fileencoding = 'utf-8'
    try:
        text = unicode(open(ifile, "r").read(),fileencoding)
        print "read file %s with encoding %s" % (ifile, fileencoding)
    except:
        print "Could not open file %s" % (ifile)
        sys.exit()
    user = os.environ['LOGNAME']
    srv = connect(host, port)
    if not checkConnection():
        print "Error connecting to hogserver."
        sys.exit(-1)
    startSession(lang)
    print "Server is ready."
    result = analyse(text, lang)
    print result
    deleteSessions();    
    sys.exit(0)


#
# MAIN 
#
if __name__=='__main__':
    main()


